
#if !defined( commands_h )
#define commands_h

#include <assert.h>

//  Signature for a command object constructor:
//  MyCommand()
//
//  How to register a command in the list of commands:
//  Command<MyCommand> cmd( "My Command", "M" );
//

//  This is not as elegant as "storage-less" auto-registration 
//  using template lists, but, hey, it works!

#define MAX_COMMANDS 64

class CommandBase
{
  public:

    CommandBase( char const * name, char const * defBind )
    {
      assert( numCommands_ < MAX_COMMANDS );
      commands_[numCommands_] = this;
      index_ = numCommands_;
      ++numCommands_;
      isset_ = false;
      name_ = name;
      defBind_ = defBind;
    }

    //  initKeys() returns some combination of these values
    enum {
      DOUBLE_MAPPED_KEY = 1,
      UNMAPPED_COMMAND = 2
    };
    static unsigned int initKeysDefault();
    static unsigned int initKeysText( char const * text, size_t size );
    //  saveKeysText returns how much data is necessary to save, and 
    //  will NOT zero-terminate the output buffer.
    static size_t saveKeysText( char * text, size_t maxSize );

    static CommandBase * keyCommand( unsigned char vKey )
    {
      return keys_[vKey];
    }
    static void setKeyCommand( unsigned char vKey, CommandBase * cmd )
    {
      keys_[vKey] = cmd;
    }

    static int numCommands()
    {
      return numCommands_;
    }
    static CommandBase * commandAt( int ix )
    {
      assert( ix >= 0 && ix < numCommands_ );
      return commands_[ix];
    }
    static unsigned __int64 commandsSet()
    {
      unsigned __int64 todo = 0;
      unsigned __int64 mask = 1;
      for( int i = 0; i < numCommands_; ++i ) {
        if( commands_[i]->isset_ ) {
          todo |= mask;
          mask <<= 1;
        }
      }
      return mask;
    }
    static void doCommands( unsigned __int64 todo )
    {
      for( int i = 0; i < numCommands_; ++i ) {
        if( todo & 1 ) {
          commands_[i]->perform();
        }
        todo >>= 1;
      }
    }

    unsigned char key() const
    {
      return key_;
    }
    bool isset() const
    {
      return isset_;
    }
    void setTo( bool set )
    {
      isset_ = set;
    }
    void set()
    {
      isset_ = true;
    }
    void clear()
    {
      isset_ = false;
    }

  protected:

    virtual void perform() = 0;

  private:

    static CommandBase * commands_[MAX_COMMANDS];
    static CommandBase * keys_[256];
    static int numCommands_;
    int index_;
    char const * name_;
    char const * defBind_;
    unsigned char key_;
    bool isset_;
};

template< class T >
class Command : public CommandBase, public T
{
  public:
    Command( char const * name, char const * defBind )
      : CommandBase( name, defBind )
    {
    }
  protected:
    void perform()
    {
      (*this)();
    }
};

#endif  //  commands_h
