

//  This class will load and render a 3dsmax ASCII Scene Export file.
//  It can issue all triangles in one go, or triangles for a specific 
//  material (it will flatten multi-subs into numbered materials).


//  TODO:
//  - use materials
//  - remove binary search from issue(mtl)
//  - normalize vertex array after load
//  - use VBO
//  - calculate tangents
//  - scale
//  - animation (?)
//  - provide access to individual geomobjects (may be hard when we merge materials) (?)
//  - slice large meshes into smaller chunks for culling/rendering (?)

#include "glut-start.h"
#include <gl/gl.h>

#include "gl-ase.h"
#include "mymath.h"

#include <vector>
#include <cctype>
#include <string>
#include <algorithm>
#include <numeric>

#undef min
#undef max

namespace {
  class Material;

  class ASE : public I_GL_ASE {
    public:
      ASE();
      ~ASE();

      struct Face {
        unsigned int mtl;
        unsigned int vix[3];
        unsigned int nix[3];
        unsigned int tix[3];
        unsigned int cix[3];
      };
      bool load( char const * name, char const * str, int len, float scale );
      void issue( Face const & f );
      void expandMaterials();
      void assignSubMaterials( Material * m );
      void assignSingleMaterial( Material * m );

      virtual void dispose();
      virtual bool normals();
      virtual bool texcoords();
      virtual bool colors();
      virtual unsigned int numMaterials();
      virtual unsigned int numVertices();
      virtual unsigned int numTriangles();
      virtual bool getMesh( float * vertices, uint numVertices, int * triangles, uint numTriangles );
      virtual void getBounds( float * outLow, float* outHigh, float * outRadius );
      virtual void issue( unsigned int mtl );
      virtual void issueAll();
      virtual char const * materialNameAt( unsigned int ix );
      virtual char const * materialDiffuseMapAt( unsigned int ix );

      bool hasNormals_;
      bool hasTexcoords_;
      bool hasColors_;
      size_t vertexOffset_;
      size_t normalOffset_;
      size_t texcoordOffset_;
      size_t colorOffset_;
      size_t faceOffset_;
      GLuint displayList_;

      std::vector< Vector3 > vertices_;
      std::vector< Vector3 > normals_;
      std::vector< Vector3 > texcoords_;
      std::vector< Vector4 > colors_;
      std::vector< Face > faces_;
      std::vector< Material * > materials_;
      std::vector< GLuint > displayLists_;
      Material * curMaterial_;
      std::string name_;
      Bounds bounds_;
      float radius_;
  };


  class Material {
    public:
      static void Delete( Material * m ) {
        delete m;
      }
      Material( Material * parent ) : parent_( parent ) {
        realIndex_ = 0;
      }
      ~Material() {
        //  don't delete materials here, as they will be 
        //  also put at the end of the material array of the ASE
      }
      Material * operator[]( int ix ) {
        if( !subMaterials_.size() ) {
          return this;
        }
        return subMaterials_[ix];
      }
      size_t numChildren() const {
        return subMaterials_.size();
      }
      void setMaterial( size_t ix, Material * m ) {
        size_t curSize = subMaterials_.size();
        subMaterials_.resize( std::max( curSize, ix+1 ), 0 );
        subMaterials_[ix] = m;
      }
      char const * name() const {
        return name_.c_str();
      }
      void setName( char const * str, size_t len ) {
        name_.clear();
        name_.insert( name_.begin(), str, str+len );
      }
      char const * diffuse() const {
        return diffuse_.c_str();
      }
      void setDiffuse( char const * str, size_t len ) {
        diffuse_.clear();
        diffuse_.insert( diffuse_.begin(), str, str+len );
      }
      bool isMulti() {
        return subMaterials_.size() > 0;
      }
      Material * parent() const {
        return parent_;
      }
      void setRealIndex( size_t ix ) {
        realIndex_ = ix;
      }
      size_t realIndex() const {
        return realIndex_;
      }
      std::string name_;
      std::vector< Material * > subMaterials_;
      std::string diffuse_;
      Material * parent_;
      size_t realIndex_;
  };
}

I_GL_ASE * MakeGL_ASE( char const * name, char const * str, size_t len, float scale )
{
  ASE * ase = new ASE();
  if( !ase->load( name, str, (int)len, scale ) ) {
    ase->dispose();
    return 0;
  }
  return ase;
}



ASE::ASE()
{
  hasNormals_ = hasTexcoords_ = hasColors_ = false;
  curMaterial_ = 0;
  displayList_ = 0;
}

ASE::~ASE()
{
  std::for_each( materials_.begin(), materials_.end(), Material::Delete );
}

//  static buffers are SOOO convenient ;-)
static char * NextLine( char const ** str, int * len )
{
  if( *len == 0 ) {
    return 0;
  }
  static char line[ 1024 ];
  char * ptr = line;
  while( *len > 0 && ptr < &line[1023] ) {
    if( **str == '\n' || **str == '\r' ) {
      (*len)--;
      (*str)++;
      break;
    }
    (*len)--;
    *ptr++ = *((*str)++);
  }
  *ptr = 0;
  ptr = line;
  while( *ptr && isspace( *ptr ) ) {
    ++ptr;
  }
  while( *len > 0 && (**str == '\n' || **str == '\r') ) {
    (*len)--;
    (*str)++;
  }
  return ptr;
}

static void GetToken( char ** line, bool * deep, char * token )
{
  assert( **line == '*' );
  ++(*line);
  int len = 0;
  while( **line && !isspace( **line ) && len < 63 ) {
    ++len;
    *token++ = *((*line)++);
  }
  *token = 0;
  while( **line && isspace( **line ) ) {
    ++(*line);
  }
  if( strchr( *line, '{' ) != 0 ) {
    *deep = true;
  }
  else {
    *deep = false;
  }
}

bool MtlLesser( ASE::Face const & a, ASE::Face const & b ) {
  return a.mtl < b.mtl;
}

enum {
  ScanningRoot,
  ScanningGeom,
  ScanningMaterialList,
  ScanningMaterial,
  ScanningDiffuseMap,
  ScanningTM,
  ScanningMesh,
  ScanningVertices,
  ScanningNormals,
  ScanningTexcoords,
  ScanningTFaces,
  ScanningColors,
  ScanningFaces,
};
bool ASE::load( char const * name, char const * str, int len, float scale )
{
  name_ = name;
  assert( vertices_.size() == 0 );
  char * line;
  int depth = 0;
  int state = 0;
  int stateStack[ 32 ];
  bool deep = false;
  char token[ 64 ];
  float nodeTM[3][4];
  int curFace = 0;

  while( line = NextLine( &str, &len ) ) {
    if( !*line ) {
      continue;
    }
    switch( *line ) {
      case '*':
        GetToken( &line, &deep, token );
        if( deep ) {
          if( depth >= 32 ) {
            assert( !"too deep data" );
            return false;
          }
          stateStack[depth] = state;
          ++depth;
        }
        switch( state ) {
          case ScanningRoot:
            if( !strcmp( token, "GEOMOBJECT" ) ) {
              state = ScanningGeom;
              memset( nodeTM, 0, sizeof( nodeTM ) );
              nodeTM[0][0] = 1;
              nodeTM[1][1] = 1;
              nodeTM[2][2] = 1;
              assert( deep );
              if( !deep ) {
                return false;
              }
            }
            else if( !strcmp( token, "MATERIAL_LIST" ) ) {
              state = ScanningMaterialList;
              assert( deep );
              if( !deep ) {
                return false;
              }
            }
            break;
          case ScanningGeom:
            if( !strcmp( token, "MESH" ) ) {
              state = ScanningMesh;
              assert( deep );
              if( !deep ) {
                return false;
              }
              vertexOffset_ = vertices_.size();
              normalOffset_ = normals_.size();
              texcoordOffset_ = texcoords_.size();
              colorOffset_ = colors_.size();
              faceOffset_ = faces_.size();
            }
            else if( !strcmp( token, "NODE_TM" ) ) {
              state = ScanningTM;
              assert( deep );
              if( !deep ) {
                return false;
              }
            }
            else if( !strcmp( token, "MATERIAL_REF" ) ) {
              size_t mtlix = 0;
              sscanf( line, "%ld", &mtlix );
              assert( mtlix < materials_.size() );
              Material * mtl = materials_[mtlix];
              assert( mtl != 0 );
              if( mtl->numChildren() ) {
                assignSubMaterials( mtl );
              }
              else {
                assignSingleMaterial( mtl );
              }
            }
            //  NODE_NAME
            break;
          case ScanningMaterialList:
            if( !strcmp( token, "MATERIAL" ) ) {
              state = ScanningMaterial;
              assert( deep );
              if( !deep ) {
                return false;
              }
              size_t mtlix = 0;
              sscanf( line, "%ld", &mtlix );
              curMaterial_ = new Material( curMaterial_ );
              curMaterial_->setRealIndex( mtlix );
              materials_.resize( std::max( materials_.size(), mtlix+1 ), 0 );
              materials_[mtlix] = curMaterial_;
            }
            break;
          case ScanningMaterial:
            if( !strcmp( token, "MAP_DIFFUSE" ) ) {
              state = ScanningDiffuseMap;
              assert( deep );
              if( !deep ) {
                return false;
              }
            }
            else if( !strcmp( token, "SUBMATERIAL" ) ) {
              state = ScanningMaterial;
              assert( deep );
              if( !deep ) {
                return false;
              }
              size_t mtlix = 0;
              sscanf( line, "%ld", &mtlix );
              curMaterial_ = new Material( curMaterial_ );
              curMaterial_->parent()->setMaterial( mtlix, curMaterial_ );
            }
            else if( !strcmp( token, "MATERIAL_NAME" ) ) {
              assert( *line == '\"' );
              char const * end = strrchr( line+1, '\"' );
              assert( end );
              curMaterial_->setName( line+1, end-(line+1) );
            }
            break;
          case ScanningDiffuseMap:
            if( !strcmp( token, "BITMAP" ) ) {
              assert( *line == '\"' );
              char const * end = strrchr( line+1, '\"' );
              curMaterial_->setDiffuse( line+1, end-(line+1) );
            }
            break;
          case ScanningTM:
            if( !strcmp( token, "TM_ROW0" ) ) {
              sscanf( line, "%f %f %f", &nodeTM[0][0], &nodeTM[1][0], &nodeTM[2][0] );
            }
            else if( !strcmp( token, "TM_ROW1" ) ) {
              sscanf( line, "%f %f %f", &nodeTM[0][1], &nodeTM[1][1], &nodeTM[2][1] );
            }
            else if( !strcmp( token, "TM_ROW2" ) ) {
              sscanf( line, "%f %f %f", &nodeTM[0][2], &nodeTM[1][2], &nodeTM[2][2] );
            }
            else if( !strcmp( token, "TM_ROW3" ) ) {
              sscanf( line, "%f %f %f", &nodeTM[0][3], &nodeTM[1][3], &nodeTM[2][3] );
            }
            break;
          case ScanningMesh:
            if( !strcmp( token, "MESH_NUMVERTEX" ) ) {
              int i = 0;
              sscanf( line, "%d", &i );
              vertices_.resize( vertexOffset_ + size_t( i ) );
            }
            else if( !strcmp( token, "MESH_NUMTVERTEX" ) ) {
              size_t i = 0;
              sscanf( line, "%ld", &i );
              texcoords_.resize( texcoordOffset_ + i );
            }
            else if( !strcmp( token, "MESH_VERTEX_LIST" ) ) {
              state = ScanningVertices;
              assert( deep );
              if( !deep ) {
                return false;
              }
            }
            else if( !strcmp( token, "MESH_TVERTLIST" ) ) {
              state = ScanningTexcoords;
              assert( deep );
              if( !deep ) {
                return false;
              }
            }
            else if( !strcmp( token, "MESH_TFACELIST" ) ) {
              state = ScanningTFaces;
              assert( deep );
              if( !deep ) {
                return false;
              }
            }
            else if( !strcmp( token, "MESH_FACE_LIST" ) ) {
              state = ScanningFaces;
              assert( deep );
              if( !deep ) {
                return false;
              }
            }
            else if( !strcmp( token, "MESH_NORMALS" ) ) {
              state = ScanningNormals;
              assert( deep );
              if( !deep ) {
                return false;
              }
            }
            else {
              break;
            }
            break;
          case ScanningVertices:
            if( !strcmp( token, "MESH_VERTEX" ) ) {
              Vector3 v; int i;
              sscanf( line, "%d %f %f %f", &i, &v.x, &v.y, &v.z );
              assert( size_t(i)+vertexOffset_ < vertices_.size() );
              if( size_t(i)+vertexOffset_ >= vertices_.size() ) {
                return false;
              }

              //  nodeTM is already applied to vertex positions (go figure!)
              float z = v.z;  //  go from Max space to GL space
              v.z = -v.y;
              v.y = z;
              v.x *= scale;
              v.y *= scale;
              v.z *= scale;

              vertices_[ size_t(i)+vertexOffset_ ] = v;
            }
            break;
          case ScanningNormals:
            if( !strcmp( token, "MESH_FACENORMAL" ) ) {
              sscanf( line, "%d", &curFace );
            }
            else if( !strcmp( token, "MESH_VERTEXNORMAL" ) ) {
              Vector3 v; int i;
              sscanf( line, "%d %f %f %f", &i, &v.x, &v.y, &v.z );

              // Apply the nodeTM
              Vector3 u( nodeTM[0][0] * v.x + nodeTM[0][1] * v.y + nodeTM[0][2] * v.z,
                         nodeTM[1][0] * v.x + nodeTM[1][1] * v.y + nodeTM[1][2] * v.z,
                         nodeTM[2][0] * v.x + nodeTM[2][1] * v.y + nodeTM[2][2] * v.z );
              // go to GL space
              v.x = u.x;
              v.y = u.z;
              v.z = -u.y;

              size_t uix = 0;
              while( uix < normals_.size() ) {
                if( normals_[uix] == v ) {
                  break;
                }
                ++uix;
              }
              if( uix == normals_.size() ) {
                normals_.push_back( v );
              }
              assert( curFace + faceOffset_ < faces_.size() );
              bool normal_found = false;
              for( int j = 0; j < 3; ++j ) {
                if( faces_[curFace+faceOffset_].vix[j] == i + vertexOffset_ ) {
                  faces_[curFace+faceOffset_].nix[j] = uint(uix);
                  normal_found = true;
                }
              }
              assert( normal_found || !"normal not found!" );
            }
            break;
          case ScanningTexcoords:
            if( !strcmp( token, "MESH_TVERT" ) ) {
              size_t i; Vector3 v;
              sscanf( line, "%d %f %f %f", &i, &v.x, &v.y, &v.z );
              assert( i < (texcoords_.size() - texcoordOffset_) );
              if( !(i < (texcoords_.size() - texcoordOffset_)) ) {
                return false;
              }
              texcoords_[i+texcoordOffset_] = v;
            }
            break;
          case ScanningTFaces:
            if( !strcmp( token, "MESH_TFACE" ) ) {
              size_t ix; uint i0, i1, i2;
              sscanf( line, "%ld %d %d %d", &ix, &i0, &i1, &i2 );
              assert( ix < (faces_.size() - faceOffset_) );
              if( !( ix < (faces_.size() - faceOffset_) ) ) {
                return false;
              }
              Face & f = faces_[faceOffset_+ix];
              f.tix[0] = i0 + texcoordOffset_;
              f.tix[1] = i1 + texcoordOffset_;
              f.tix[2] = i2 + texcoordOffset_;
            }
            break;
          case ScanningColors:
            break;
          case ScanningFaces:
            if( !strcmp( token, "MESH_FACE" ) ) {
              uint i, i0, i1, i2, _0, _1, _2, sm, mtl;
              sscanf( line, "%d: A: %d B: %d C: %d AB: %d BC: %d CA: %d *MESH_SMOOTHING %d *MESH_MTLID %d",
                  &i, &i0, &i1, &i2, &_0, &_1, &_2, &sm, &mtl );
              size_t s = vertices_.size() - vertexOffset_;
              assert( i0 < s && i1 < s && i2 < s );
              if( !(i0 < s && i1 < s && i2 < s) ) {
                return false;
              }
              s = faces_.size() - faceOffset_;
              if( size_t(i) >= s ) {
                faces_.resize( size_t(i) + faceOffset_ + 1 );
              }
              Face & f = faces_[faceOffset_+size_t(i)];
              f.mtl = mtl;
              f.vix[0] = i0 + vertexOffset_;
              f.vix[1] = i1 + vertexOffset_;
              f.vix[2] = i2 + vertexOffset_;
            }
            break;
          default:
            assert( !"this should not happen" );
            return false;
        }
        break;
      case '}':
        assert( depth > 0 );
        if( depth == 0 ) {
          return false;
        }
        if( state == ScanningMaterial ) {
          curMaterial_ = curMaterial_->parent();
        }
        else if( state == ScanningMaterialList ) {
          expandMaterials();
        }
        --depth;
        state = stateStack[depth];
        break;
      default:
        //  what is this?
        assert( !"unexpected character in file" );
        return false;
        break;
    }
  }
  hasNormals_ = normals_.size() > 0;
  hasTexcoords_ = texcoords_.size() > 0;
  hasColors_ = colors_.size() > 0;
  std::sort( &faces_[0], &faces_[faces_.size()], MtlLesser );
  radius_ = 0;
  if( vertices_.size() > 0 ) {
    Bounds b( vertices_[0] );
    b = std::accumulate( vertices_.begin(), vertices_.end(), b );
    //  fixme: start with the center of the bounding box for radius.
    //  But then it's harder to apply the radius for my user!
    float r = 0;
    for( std::vector<Vector3>::iterator ptr = vertices_.begin();
        ptr != vertices_.end();
        ++ptr ) {
      if( (*ptr).lengthSquared() > r ) {
        r = (*ptr).lengthSquared();
      }
    }
    radius_ = sqrtf( r );
  }
  return vertices_.size() > 0;
}

void ASE::dispose()
{
  delete this;
}

bool ASE::normals()
{
  return hasNormals_;
}

bool ASE::texcoords()
{
  return hasTexcoords_;
}

bool ASE::colors()
{
  return hasColors_;
}

unsigned int ASE::numMaterials()
{
  return (unsigned int)materials_.size();
}

char const * ASE::materialNameAt( unsigned int ix )
{
  if( ix >= materials_.size() ) return 0;
  if( !materials_[ix] ) return 0;
  return materials_[ix]->name();
}

char const * ASE::materialDiffuseMapAt( unsigned int ix )
{
  if( ix >= materials_.size() ) return 0;
  if( !materials_[ix] ) return 0;
  return materials_[ix]->diffuse();
}

unsigned int ASE::numVertices()
{
  return uint(vertices_.size());
}

unsigned int ASE::numTriangles()
{
  return uint(faces_.size());
}

bool ASE::getMesh( float * vertices, uint numVertices, int * triangles, uint numTriangles )
{
  if( numVertices < vertices_.size() || numTriangles < faces_.size() ) {
    return false;
  }
  memcpy( vertices, &vertices_[0], sizeof(float)*3*vertices_.size() );
  for( uint ix = 0; ix < faces_.size(); ++ix ) {
    triangles[0] = faces_[ix].vix[0];
    triangles[1] = faces_[ix].vix[1];
    triangles[2] = faces_[ix].vix[2];
    triangles += 3;
  }
  return true;
}

void ASE::getBounds( float * outLow, float * outHigh, float * outRadius )
{
  *(Vector3 *)outLow = bounds_.low;
  *(Vector3 *)outHigh = bounds_.high;
  *outRadius = radius_;
}

static ASE::Face const * mtl_search( ASE::Face const * a, ASE::Face const * b, unsigned int mtl )
{
  ASE::Face const * mid;
again:
  if( b <= a ) {
    return b;
  }
  mid = a + ((b-a)>>1);
  if( mid->mtl >= mtl ) {
    b = mid;
  }
  else {
    a = mid+1;
  }
  goto again;
}

//  No preparation, no vertex arrays, no sorting, no nothing.
//  Just issue the verts!
void ASE::issue( unsigned int mtl )
{
  //  generate display lists if issue(mtl) hasn't been called previously
  if( displayLists_.size() == 0 ) {
    displayLists_.insert( displayLists_.begin(), materials_.size(), 0 );
  }
  assert( mtl < materials_.size() );
  if( !displayLists_[mtl] ) {
    displayLists_[mtl] = glGenLists( 1 );
    glNewList( displayLists_[mtl], GL_COMPILE );
    //  do a binary search to find the start and end points
    //fixme:  I should store this in each of the materials!
    //todo: perhaps just junk this, as traversal only happens the 
    //  first time something renders.
    Face const * a = mtl_search( &faces_[0], &faces_[faces_.size()], mtl );
    Face const * b = mtl_search( a, &faces_[faces_.size()], mtl+1 );
    glBegin( GL_TRIANGLES );
    while( a < b ) {
      issue( *a );
      ++a;
    }
    glEnd();
    glEndList();
  }
  glCallList( displayLists_[mtl] );
}

void ASE::issueAll()
{
  //  generate display list if issueAll() hasn't been called previously
  if( !displayList_ ) {
    displayList_ = glGenLists( 1 );
    glNewList( displayList_, GL_COMPILE );
    glBegin( GL_TRIANGLES );
    for( size_t ix = 0; ix < faces_.size(); ++ix ) {
      Face & f = faces_[ix];
      issue( f );
    }
    glEnd();
    glEndList();
  }
  glCallList( displayList_ );
}

void ASE::issue( Face const & f )
{
  for( int i = 0; i < 3; ++i ) {
    float * arg;
    if( hasColors_ ) {
      arg = &colors_[f.cix[i]].x;
      glColor4fv( arg );
    }
    if( hasTexcoords_ ) {
      arg = &texcoords_[f.tix[i]].x;
      glTexCoord3fv( arg );
    }
    if( hasNormals_ ) {
      arg = &normals_[f.nix[i]].x;
      glNormal3fv( arg );
    }
    arg = &vertices_[f.vix[i]].x;
    glVertex3fv( arg );
  }
}

void ASE::expandMaterials()
{
  size_t pos = 0;
  size_t next = 0;
  while( pos < materials_.size() ) {
    Material * m = materials_[pos];
    ++pos;
    if( !m ) {
      continue;
    }
    for( size_t i = 0; i < m->numChildren(); ++i ) {
      Material * m2 = (*m)[(int)i];
      if( !m2 ) {
        continue;
      }
      while( next <= materials_.size() ) {
        if( next == materials_.size() ) {
          materials_.push_back( m2 );
          m2->setRealIndex( next );
          break;
        }
        if( !materials_[next] ) {
          materials_[next] = m2;
          m2->setRealIndex( next );
          break;
        }
        ++next;
      }
    }
  }
}

void ASE::assignSubMaterials( Material * m )
{
  for( size_t f = faceOffset_; f < faces_.size(); ++f ) {
    if( faces_[f].mtl >= m->numChildren() ) {
      warning( "Face %d of mesh '%s' uses un-defined submaterial id %d\n", 
          f, name_.c_str(), faces_[f].mtl );
      faces_[f].mtl = 0;
    }
    faces_[f].mtl = (unsigned int)(*m)[faces_[f].mtl]->realIndex();
  }
}

void ASE::assignSingleMaterial( Material * m )
{
  unsigned int rix = (unsigned int)m->realIndex();
  for( size_t f = faceOffset_; f < faces_.size(); ++f ) {
    faces_[f].mtl = rix;
  }
}


void UnittestGL_ASE()
{
  ASE * ase = new ASE();
  assert( !ase->load( "empty", "", 0, 1 ) );
  char * filename = "c:\\3dsmax5\\meshes\\box.ase";
  FILE * f = fopen( filename, "rb" );
  if( !f ) {
    warning( "Could not unit test ASE: %s is missing\n", filename );
    ase->dispose();
    return;
  }
  assert( f );
  fseek( f, 0, 2 );
  long l = ftell( f );
  fseek( f, 0, 0 );
  char * p = new char[ l ];
  fread( p, 1, l, f );
  fclose( f );
  assert( ase->load( "box.ase", p, l, 1 ) );
  assert( ase->normals() );
  assert( ase->texcoords() );
  assert( ase->numMaterials() );
  assert( !ase->colors() );
  delete[] p;
  ase->dispose();
}
