
//  for stock objects
#define WINVER 0x400
#define _WIN32_WINNT 0x500

#include <windows.h>
#include <gl/gl.h>
#include <assert.h>

#include "gl-text.h"
#include "glshaderstate.h"


class GlTextImp {
  public:
    GlTextImp( int width, int height );
    ~GlTextImp();
    void clear();
    void fill( DWORD color, int x, int y, int width, int height );
    void getImage( Image * oImage );
    void drawText( DWORD color, int x, int y, char const * text, size_t len );
    void measureText( char const * text, int * ow, int * oh, int len = -1 );
    void drawLine( DWORD color, int x, int y, int x2, int y2 );
    void updateImage();
    void calculateAlpha();
    TextureRef texture();
    void scroll( DWORD color, int dx, int dy );

    static HFONT font_;
    static HDC utilDC_;
    static HDC screenDC_;
    static unsigned int refCount_;
    static HBITMAP curBitmap_;
    HBITMAP bitmap_;
    void * base_;
    size_t stride_;
    int width_;
    int height_;
    bool dirty_;
    bool textureDirty_;
    TextureRef ref_;
    GLuint tex_;

    void setupDC();
    void teardownDC();
};



HFONT GlTextImp::font_;
HDC GlTextImp::utilDC_;
HDC GlTextImp::screenDC_;
unsigned int GlTextImp::refCount_;
HBITMAP GlTextImp::curBitmap_;

GlTextImp::GlTextImp( int width, int height )
{
  assert( !(width & (width-1)) && !(height & (height-1)) );
  if( !refCount_ ) {
    setupDC();
  }
  ++refCount_;
  width_ = width;
  height_ = height;
  dirty_ = textureDirty_ = true;

  BITMAPINFOHEADER * bmip;
  bmip = (BITMAPINFOHEADER *)malloc( sizeof( BITMAPINFOHEADER ) );
  memset( bmip, 0, sizeof( BITMAPINFOHEADER ) );
  BITMAPINFO & bmi = *(BITMAPINFO *)bmip;
  bmi.bmiHeader.biSize = sizeof( bmi.bmiHeader );
  bmi.bmiHeader.biWidth = width;
  bmi.bmiHeader.biHeight = height;
  bmi.bmiHeader.biPlanes = 1;
  bmi.bmiHeader.biSizeImage = width * height;
  bmi.bmiHeader.biXPelsPerMeter = 4096;
  bmi.bmiHeader.biYPelsPerMeter = 4096;
  bmi.bmiHeader.biBitCount = 32;
  bmi.bmiHeader.biCompression = BI_RGB;
  base_ = 0;
  bitmap_ = ::CreateDIBSection( utilDC_, &bmi, DIB_RGB_COLORS, &base_, 0, 0 );
  stride_ = width*4;
  ::SetBkMode( utilDC_, TRANSPARENT );

  glGenTextures( 1, &tex_ );
  glBindTexture( GL_TEXTURE_2D, tex_ );
  glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 0, GL_BGRA, GL_UNSIGNED_BYTE, 0 );
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
  ref_ = TextureRef( tex_, GL_TEXTURE_2D );
  GLShader::flushState();
}

GlTextImp::~GlTextImp()
{
  assert( refCount_ > 0 );
  --refCount_;
  if( !refCount_ ) {
    teardownDC();
  }
  ::DeleteObject( bitmap_ );
}

void GlTextImp::clear()
{
  if( curBitmap_ != bitmap_ ) {
    DWORD e = ::GetLastError();
    ::SelectObject( utilDC_, bitmap_ );
    DWORD e2 = ::GetLastError();
    curBitmap_ = bitmap_;
  }
  RECT r;
  r.left = 0;
  r.top = 0;
  r.right = width_;
  r.bottom = height_;
  ::FillRect( utilDC_, &r, (HBRUSH)::GetStockObject( BLACK_BRUSH ) );
  dirty_ = true;
}

void GlTextImp::fill( DWORD color, int x, int y, int width, int height )
{
  if( curBitmap_ != bitmap_ ) {
    DWORD e = ::GetLastError();
    ::SelectObject( utilDC_, bitmap_ );
    DWORD e2 = ::GetLastError();
    curBitmap_ = bitmap_;
  }
  RECT r;
  r.left = x;
  r.top = y;
  r.right = x+width;
  r.bottom = y+height;
  HBRUSH br = ::CreateSolidBrush( color );
  ::FillRect( utilDC_, &r, br );
  ::DeleteObject( (HGDIOBJ)br );
  dirty_ = true;
}

void GlTextImp::getImage( Image * oImage )
{
  if( dirty_ ) {
    updateImage();
  }
  //  make an image that won't be auto-freed
  *oImage = Image();
  oImage->base = base_;
  oImage->bits = 32;
  oImage->pitch = width_ * 4;
  oImage->width = width_;
  oImage->height = height_;
}

void GlTextImp::drawText( DWORD color, int x, int y, char const * text, size_t len )
{
  if( curBitmap_ != bitmap_ ) {
    ::SelectObject( utilDC_, bitmap_ );
    curBitmap_ = bitmap_;
  }
  ::SetTextColor( utilDC_, color );
  ::TextOut( utilDC_, x, y, text, (int)len );
  dirty_ = true;
}

void GlTextImp::measureText( char const * text, int * ow, int * oh, int len )
{
  if( curBitmap_ != bitmap_ ) {
    ::SelectObject( utilDC_, bitmap_ );
    curBitmap_ = bitmap_;
  }
  SIZE sz;
  ::GetTextExtentPoint32( utilDC_, text, len < 0 ? (int)strlen(text) : len, &sz );
  *ow = sz.cx;
  *oh = sz.cy;
}

void GlTextImp::drawLine( DWORD color, int x, int y, int x2, int y2 )
{
  if( curBitmap_ != bitmap_ ) {
    ::SelectObject( utilDC_, bitmap_ );
    curBitmap_ = bitmap_;
  }
  ::SelectObject( utilDC_, ::GetStockObject( DC_PEN ) );
  ::SetDCPenColor( utilDC_, color );
  POINT oPoint;
  ::MoveToEx( utilDC_, x, y, &oPoint );
  ::LineTo( utilDC_, x2, y2 );
  dirty_ = true;
}

TextureRef GlTextImp::texture()
{
  if( dirty_ ) {
    updateImage();
  }
  if( textureDirty_ ) {
    calculateAlpha();
    glBindTexture( GL_TEXTURE_2D, tex_ );
    glPixelStorei( GL_UNPACK_ROW_LENGTH, width_ );
    glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, width_, height_, GL_BGRA, GL_UNSIGNED_BYTE, base_ );
    GLShader::flushState();
    textureDirty_ = false;
  }
  return ref_;
}

void GlTextImp::scroll( DWORD color, int dx, int dy )
{
  if( dirty_ ) {
    updateImage();
  }
  //  I scroll first Y, then X.
  //  This may seem wasteful.
  //  That's because it probably is.
  if( dy < 0 ) {
    if( dy <= -height_ ) {
      fill( color, 0, 0, width_, height_ );
    }
    else {
      int nl = height_ + dy;
      memmove( ((unsigned char *)base_) + (-dy)*stride_, (unsigned char *)base_, stride_*nl );
    }
  }
  else if( dy > 0 ) {
    if( dy >= height_ ) {
      fill( color, 0, 0, width_, height_ );
    }
    else {
      int nl = height_ - dy;
      memmove( (unsigned char *)base_, ((unsigned char *)base_) + dy*stride_, stride_*nl );
    }
  }
  if( dx < 0 ) {
    if( dx <= -width_ ) {
      fill( color, 0, 0, width_, height_ );
    }
    else {
      for( int i = 0; i < height_; ++i ) {
        memmove( ((unsigned char *)base_) + i*stride_, ((unsigned char *)base_) + i*stride_ - dx*4, (width_+dx)*4 );
      }
    }
  }
  else if( dx > 0 ) {
    if( dx >= width_ ) {
      fill( color, 0, 0, width_, height_ );
    }
    else {
      for( int i = 0; i < height_; ++i ) {
        memmove( ((unsigned char *)base_) + i*stride_ + dx*4, ((unsigned char *)base_) + i*stride_, (width_-dx)*4 );
      }
    }
  }
  if( dy != 0 ) {
    if( dy < 0 ) {
      fill( color, 0, height_+dy, width_, height_ );
    }
    else {
      fill( color, 0, 0, width_, dy );
    }
  }
  if( dx != 0 ) {
    if( dx < 0 ) {
      fill( color, width_+dx, 0, width_, height_ ) ;
    }
    else {
      fill( color, 0, 0, dx, height_ );
    }
  }
  dirty_ = (dx || dy);
}

void GlTextImp::updateImage()
{
  if( dirty_ ) {
    ::GdiFlush();
    dirty_ = false;
    textureDirty_ = true;
  }
}

void GlTextImp::calculateAlpha()
{
  DWORD * ptr = (DWORD *)base_;
  int bits = 0;
  //todo: text drawing with alpha is annoying, as GDI doesn't support it.
  //  Meanwhile, I take the biggest component, and make alpha be opaque 
  //  where the biggest component is at least 0.5.
  for( int y = 0; y < height_; ++y ) {
    for( int x = 0; x < width_; ++x ) {
      DWORD d = *ptr;
      unsigned char a = unsigned char( d&0xff );
      unsigned char b = unsigned char( (d>>8)&0xff );
      unsigned char c = unsigned char( (d>>16)&0xff );
      if( b > a ) a = b;
      if( c > a ) a = c;
      if( a < 128 ) {
        a <<= 1;
      }
      else {
        a = 255;
      }
      if( d != 0 ) {
        ++bits;
      }
      d = d | ((int)a << 24);
      *ptr = d;
      ++ptr;
    }
  }
}

void GlTextImp::setupDC()
{
  screenDC_ = ::CreateDC( "DISPLAY", 0, 0, 0 );
  utilDC_ = ::CreateCompatibleDC( ::wglGetCurrentDC() );
  LOGFONT lf;
  memset( &lf, 0, sizeof( lf ) );
  lf.lfHeight = 15;
  lf.lfWeight = FW_NORMAL;
  lf.lfCharSet = ANSI_CHARSET;
  lf.lfOutPrecision = OUT_DEFAULT_PRECIS;
  lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
  lf.lfQuality = ANTIALIASED_QUALITY;
  lf.lfPitchAndFamily = FIXED_PITCH | FF_MODERN;
  strcpy( lf.lfFaceName, "Courier New" );
  font_ = ::CreateFontIndirect( &lf );
  // select into screen to turn on anti-aliasing
  HGDIOBJ f = ::SelectObject( screenDC_, font_ );
  ::SelectObject( screenDC_, f );
  ::SelectObject( utilDC_, (HGDIOBJ)font_ );
}

void GlTextImp::teardownDC()
{
}




GlText::GlText( int width, int height )
{
  imp_ = new GlTextImp( width, height );
}

void GlText::dispose()
{
  delete imp_;
  delete this;
}

void GlText::clear()
{
  imp_->clear();
}

void GlText::fill( GWColor color, int x, int y, int width, int height )
{
  imp_->fill( RGB( color.red, color.green, color.blue ), x, y, width, height );
}

void GlText::getImage( Image * oImage )
{
  imp_->getImage( oImage );
}

void GlText::drawText( GWColor color, int x, int y, char const * text, int len )
{
  imp_->drawText( RGB( color.red, color.green, color.blue ), x, y, text, len < 0 ? strlen( text ) : len );
}

void GlText::measureText( char const * text, int * ow, int * oh, int len )
{
  imp_->measureText( text, ow, oh, len );
}

void GlText::drawLine( GWColor color, int x, int y, int x2, int y2 )
{
  imp_->drawLine( RGB(color.red,color.green,color.blue), x, y, x2, y2 );
}

TextureRef GlText::texture()
{
  return imp_->texture();
}

void GlText::dirty()
{
  imp_->dirty_ = true;
}

void GlText::scroll( GWColor color, int dx, int dy )
{
  imp_->scroll( color.color, dx, dy );
}

GlText * MakeGlText( int width, int height )
{
  //  fixme: GlTextImp should derive from GlText, and point 
  //  to the same object. But it's unlikely to matter.
  return new GlText( width, height );
}
