
#include "glprograms.h"
#include <assert.h>
#include <algorithm>

////////////////////////
//  GLProgramValue
////////////////////////


namespace {
  enum {
    MAX_TEXTURES = 16  //  fixme: up this value
  };
  struct ProgramState {
    ProgramState() {
      assert( !glGetError() );
      magic = 0x69696969;
      program = 0;
      maxTexture = 4;
      glGetIntegerv( GL_MAX_TEXTURE_IMAGE_UNITS_ARB, &maxTexture );
      if( maxTexture > MAX_TEXTURES ) {
        maxTexture = MAX_TEXTURES;
      }
      for( int i = 0; i < maxTexture; ++i ) {
        texture[i] = 0;
      }
      glActiveTexture( GL_TEXTURE0_ARB );
      //  this assert will fire on NVIDIA hardware contrary to spec
      assert( !glGetError() );
    }
    ~ProgramState() {
      assert( magic == 0x69696969 );
      magic = 0xefefefef;
    }
    int magic;
    GLint maxTexture;
    GLhandleARB program;
    GLenum texture[MAX_TEXTURES];
  };
  ProgramState * gCurState;
};


GLProgram GLProgram::Null;
GLProgramInstance GLProgramInstance::Null( &GLProgram::Null );

void GLSwitchToProgramInstanceState( GLProgramInstanceState & ustate )
{
  ProgramState * ps = reinterpret_cast< ProgramState * >( ustate );
  if( !ps ) {
    //  First usage of this state instance.
    //  This may blow away previous gCurState, but that should be 
    //  remembered by some previous call to GLSwitchToProgramInstanceState(), 
    //  unless it was allocated by GLProgramInstance::apply() (which is a 
    //  usage error, that I allow with an assertion).
    assert( GL_ARB_shader_objects_Flag );
    gCurState = new ProgramState;
    ustate = reinterpret_cast< GLProgramInstanceState >( gCurState );
  }
  else {
    assert( ps->magic == 0x69696969 );
    gCurState = ps;
  }
}

void GLFreeProgramInstanceState( GLProgramInstanceState & ustate )
{
  ProgramState * ps = reinterpret_cast< ProgramState * >( ustate );
  ustate = 0;
  delete ps;
  if( ps == gCurState ) {
    gCurState = 0;
  }
}





GLProgramValue::GLProgramValue()
{
  pave();
}

GLProgramValue::GLProgramValue( float f )
{
  pave();
  type_ = FloatN;
  n_ = 1;
  m_ = 1;
  floats_ = new float[1];
  floats_[0] = f;
}

GLProgramValue::GLProgramValue( float const * p, size_t n )
{
  assert( n <= 4 );
  pave();
  type_ = FloatN;
  n_ = n;
  m_ = 1;
  floats_ = new float[n_];
  memcpy( floats_, p, sizeof(float)*n );
}

GLProgramValue::GLProgramValue( float const * p, size_t n, bool matrix, size_t m )
{
  assert( m >= 1 );
  assert( n <= 4 );
  pave();
  size_t n2 = n_;
  n_ = n;
  m_ = m;
  if( matrix ) {
    type_ = FloatNxM;
    n2 = n_*n_;
  }
  else {
    type_ = FloatN;
  }
  floats_ = new float[n2*m_];
  memcpy( floats_, p, sizeof(float)*n2*m_ );
}

GLProgramValue::GLProgramValue( TextureRef const & t )
{
  pave();
  switch( t.textureTargetHack() ) {
  case GL_TEXTURE_2D:
    type_ = Sampler2D;
    break;
  case GL_TEXTURE_CUBE_MAP_ARB:
    type_ = SamplerCube;
    break;
  default:
    assert( !"what kind of texture is this?" );
  }
  texture_ = t;
}

GLProgramValue::GLProgramValue( GLProgramValue const & o )
{
  pave();
  copy( o );
}

GLProgramValue & GLProgramValue::operator=( GLProgramValue const & o )
{
  nuke();
  copy( o );
  return *this;
}

void GLProgramValue::get( float * p, size_t n, size_t m ) const
{
  assert( (type_ == FloatN) && "not a float vector in GLProgramValue::get()" );
  assert( n_ == n && m_ == m );
  memcpy( p, floats_, sizeof(float)*n*m );
}

void GLProgramValue::getMatrix( float * p, size_t n, size_t m ) const
{
  assert( (type_ == FloatNxM) && "not a float matrix in GLProgramValue::get()" );
  assert( n_ == n && m_ == m );
  memcpy( p, floats_, sizeof(float)*n*n*m );
}

void GLProgramValue::get( TextureRef * oRef ) const
{
  assert( (type_ == Sampler2D || type_ == SamplerCube) && "not a texture in GLProgramValue::get()" );
  *oRef = texture_;
}

void GLProgramValue::set( float const * p, size_t n, size_t m )
{
  assert( (type_ == FloatN) && "not a float vector in GLProgramValue::set()" );
  assert( n_ == n && m_ == m );
  memcpy( floats_, p, sizeof(float)*n*m );
}

void GLProgramValue::setMatrix( float const * p, size_t n, size_t m )
{
  assert( (type_ == FloatNxM) && "not a float matrix in GLProgramValue::set()" );
  assert( n_ == n && m_ == m );
  memcpy( floats_, p, sizeof(float)*n*n*m );
}

void GLProgramValue::set( TextureRef const & oRef )
{
  assert( (type_ == Sampler2D || type_ == SamplerCube) && "not a texture in GLProgramValue::set()" );
  texture_ = oRef;
}

static void glUniformNfv( int location, size_t n, size_t m, float const * f )
{
  switch( n ) {
  case 1:
    glUniform1fvARB( location, GLsizei(m) , f );
    break;
  case 2:
    glUniform2fvARB( location, GLsizei(m), f );
    break;
  case 3:
    glUniform3fvARB( location, GLsizei(m), f );
    break;
  case 4:
    glUniform4fvARB( location, GLsizei(m), f );
    break;
  default:
    assert( !"bad n in glUniformNfv()" );
  }
}

static void glUniformMatrixNfv( int location, size_t n, size_t m, float const * f )
{
  switch( n ) {
  case 2:
    glUniformMatrix2fvARB( location, GLsizei(m), false, f );
    break;
  case 3:
    glUniformMatrix3fvARB( location, GLsizei(m), false, f );
    break;
  case 4:
    glUniformMatrix4fvARB( location, GLsizei(m), false, f );
    break;
  default:
    assert( !"bad n in glUniformMatrixNfv()" );
  }
}

void GLProgramValue::apply( int location ) const
{
  assert( !glGetError() );
  int tu = -1;
  switch( type_ ) {
  case GLProgramValue::Sampler2D:
  case GLProgramValue::SamplerCube:
    for( int i = 0; i < gCurState->maxTexture; ++i ) {
      if( gCurState->texture[i] == 0 ) {
        tu = i;
        break;
      }
    }
    assert( tu >= 0 );
    if( tu < 0 ) {
      //should I warn? How could this happen?
      return;
    }
    glActiveTexture( GL_TEXTURE0_ARB + tu );
    assert( !glGetError() );
    gCurState->texture[tu] = texture_.textureTargetHack();
    glBindTexture( gCurState->texture[tu], texture_.textureIdHack() );
    assert( !glGetError() );
    glUniform1iARB( location, tu );
    break;
  case GLProgramValue::FloatN:
    glUniformNfv( location, n_, m_, floats_ );
    break;
  case GLProgramValue::FloatNxM:
    glUniformMatrixNfv( location, n_, m_, floats_ );
    break;
  default:
    ;
  }
  assert( !glGetError() );
}

void GLProgramValue::nuke()
{
  texture_ = TextureRef();
  delete[] floats_;
}

void GLProgramValue::pave()
{
  type_ = Null;
  n_ = 0;
  m_ = 0;
  floats_ = 0;
}

void GLProgramValue::copy( GLProgramValue const & o )
{
  type_ = o.type_;
  switch( type_ ) {
  default:
    assert( !"bad value in GLProgramValue::copy()" );
  case Null:
    floats_ = 0;
    break;
  case FloatN:
    assert( o.m_ > 0 );
    assert( o.n_ > 0 );
    n_ = o.n_; m_ = o.m_;
    floats_ = new float[n_*m_];
    memcpy( floats_, o.floats_, sizeof(float)*n_*m_ );
    break;
  case FloatNxM:
    assert( o.m_ > 0 );
    assert( o.n_ > 0 );
    n_ = o.n_; m_ = o.m_;
    floats_ = new float[n_*n_*m_];
    memcpy( floats_, o.floats_, sizeof(float)*n_*n_*m_ );
    break;
  case Sampler2D:
    floats_ = 0;
    texture_ = o.texture_;
    break;
  case SamplerCube:
    floats_ = 0;
    texture_ = o.texture_;
    break;
  }
}




////////////////////////
//  GLProgramFragment
////////////////////////


namespace {
  char * const shaderError_ = "(0): GLProgramFragment is not compiled yet";
  bool sFirstInit;
  bool sIsATI;
}

GLProgramFragment::GLProgramFragment()
{
  refCount_ = 1;
  if( !sFirstInit ) {
    //  if we get the generic context, try again later
    if( strstr( (char const *)glGetString( GL_VENDOR ), "Microsoft" ) == 0 ) {
      sIsATI = strstr( (char const *)glGetString( GL_VENDOR ), "ATI " ) != 0;
      sFirstInit = true;
    }
  }
  pave();
}

GLProgramFragment::GLProgramFragment( GLuint type )
{
  refCount_ = 1;
  if( !sFirstInit ) {
    if( strstr( (char const *)glGetString( GL_VENDOR ), "Microsoft" ) == 0 ) {
      sIsATI = strstr( (char const *)glGetString( GL_VENDOR ), "ATI " ) != 0;
      sFirstInit = true;
    }
  }
  assert( !glGetError() );
  pave();
  type_ = type;
  shaderHandle_ = GLRefCountedHandle( glCreateShaderObjectARB( type ) );
  assert( !glGetError() );
}

GLProgramFragment::GLProgramFragment( GLProgramFragment const & o )
{
  refCount_ = 1;
  pave();
  copy( o );
}

GLProgramFragment & GLProgramFragment::operator=( GLProgramFragment const & o )
{
  nuke();
  copy( o );
  return *this;
}

GLProgramFragment::~GLProgramFragment()
{
  nuke();
}

bool GLProgramFragment::compile( char const * text, int size, char const * fileName )
{
  int lineNo = 1;
  return compile( 1, &text, &size, &fileName, &lineNo );
}

bool GLProgramFragment::compile( size_t n, char const ** text, int const * size, char const ** fileName, int const * lineNo )
{
  assert( n <= 10 );
  if( n > 10 ) {
    log_ = new char[100];
    _snprintf( log_, 100, 
        "GLProgramFragment::compile() can only paste 10 code fragments; requested %d.\n",
        n );
    error( "%s", log_ );
    return false;
  }

  char const * lineText[20];
  char lineLine[10][100];
  int lineSize[20];
  for( size_t i = 0; i < n; ++i ) {
    char const * fn = fileName[i];
    if( strlen( fn ) > 70 ) {
      fn = fn + strlen( fn ) - 70;
    }
    lineText[i*2] = lineLine[i];
    sprintf( lineLine[i], "%sline %d \"%s\"\n", sIsATI ? "// " : "#", lineNo[i], fn );
    lineSize[i*2] = int( strlen( lineLine[i] ) );
    lineText[i*2+1] = text[i];
    lineSize[i*2+1] = size[i];
  }

  debug( "Compiling %d fragments from '%s'\n", n, fileName[0] );

  assert( (type_ != 0) && "You cannot compile an uninitialized GLProgramFragment" );
  // walk the shader data to make sure it's valid
#if !defined( NDEBUG )
  for( int i = 0; i < (int)n*2; ++i ) {
    for( int j = 0; j < lineSize[i]; ++j ) {
      *(volatile char *)&lineText[i][j];
    }
  }
#endif
  glShaderSourceARB( shaderHandle_, (GLsizei)n*2, lineText, lineSize );
  if( glGetError() ) {
    log_ = new char[100];
    strcpy( log_, "(0): ShaderSourceARB() failed" );
    return false;
  }
  glCompileShaderARB( shaderHandle_ );
  assert( !glGetError() );
  GLint i = 0;
  if( log_ != shaderError_ ) {
    delete[] log_;
  }
  log_ = 0;
  glGetObjectParameterivARB( shaderHandle_, GL_OBJECT_COMPILE_STATUS_ARB, &i );
  char log[2000];
  GLsizei len = sizeof( log );
  glGetInfoLogARB( shaderHandle_, len, &len, log );
  assert( !glGetError() );
  if( len >= sizeof( log ) ) {
    len = sizeof( log )-1;
  }
  log[len] = 0;
  log_ = new char[len+1];
  memcpy( log_, log, len+1 );
  if( i == GL_TRUE ) {
    compiled_ = true;
    return true;
  }
  return false;
}

void GLProgramFragment::nuke()
{
  if( log_ != shaderError_ ) {
    delete[] log_;
  }
  shaderHandle_ = GLRefCountedHandle( 0 );
}

void GLProgramFragment::pave()
{
  type_ = 0;
  compiled_ = false;
  log_ = shaderError_;
}

void GLProgramFragment::copy( GLProgramFragment const & o )
{
  type_ = o.type_;
  shaderHandle_ = o.shaderHandle_;
  compiled_ = o.compiled_;
  log_ = o.log_;
}



////////////////////////
//  GLProgramParameter
////////////////////////


GLProgramParameter::GLProgramParameter()
{
  pave();
}

GLProgramParameter::GLProgramParameter( GLProgramParameter const & o )
{
  copy( o );
}

GLProgramParameter & GLProgramParameter::operator=( GLProgramParameter const & o )
{
  nuke();
  copy( o );
  return *this;
}

GLProgramParameter::~GLProgramParameter()
{
  nuke();
}

void GLProgramParameter::nuke()
{
}

void GLProgramParameter::pave()
{
  type_ = GLProgramValue::Null;
  n_ = 0;
  m_ = 0;
  location_ = -1;
  name_[0] = 0;
}

void GLProgramParameter::copy( GLProgramParameter const & o )
{
  type_ = o.type_;
  n_ = o.n_;
  m_ = o.m_;
  location_ = o.location_;
  memcpy( name_, o.name_, sizeof( name_ ) );
}





////////////////////////
//  GLProgram
////////////////////////

static char * programError_ = "(0): GLProgram is not linked yet.";

GLProgram::GLProgram()
{
  refCount_ = 1;
  pave();
  if( this == &Null ) {
    linked_ = true;
  }
}

GLProgram::GLProgram( GLProgram const & o )
{
  refCount_ = 1;
  pave();
  copy( o );
}

GLProgram & GLProgram::operator=( GLProgram const & o )
{
  refCount_ = 1;
  nuke();
  copy( o );
  return *this;
}

GLProgram::~GLProgram()
{
  nuke();
}

bool GLProgram::link( GLProgramFragment * vertex, GLProgramFragment * fragment )
{
  nuke();
  pave();

  programHandle_ = GLRefCountedHandle( glCreateProgramObjectARB() );
  vertexShader_ = vertex;
  if( vertexShader_ ) {
    vertexShader_->addRef();
  }
  fragmentShader_ = fragment;
  if( fragmentShader_ ) {
    fragmentShader_->addRef();
  }
  log_ = 0;

  //  Don't use std::string, because not everyone might want to pull 
  //  in STL in their material handling.
  bool ok = true;
  char const * vsLog = 0;
  if( vertexShader_ ) {
    if( !vertexShader_->compiled() ) {
      ok = false;
    }
    vsLog = vertexShader_->log();
  }
  if( !vsLog ) {
    vsLog = "";
  }
  char const * fsLog = 0;
  if( fragmentShader_ ) {
    if( !fragmentShader_->compiled() ) {
      ok = false;
    }
    fsLog = fragmentShader_->log();
  }
  if( !fsLog ) {
    fsLog = "";
  }
  char log[2000];
  if( ok ) {
    if( vertexShader_ ) {
      glAttachObjectARB( programHandle_, vertexShader_->handle() );
      assert( !glGetError() );
    }
    if( fragmentShader_ ) {
      glAttachObjectARB( programHandle_, fragmentShader_->handle() );
      assert( !glGetError() );
    }
    glLinkProgramARB( programHandle_ );
    assert( !glGetError() );
    GLint linked = 0;
    glGetObjectParameterivARB( programHandle_, GL_OBJECT_LINK_STATUS_ARB, &linked );
    ok = ok && (linked != 0);
    GLsizei len = sizeof( log );
    glGetInfoLogARB( programHandle_, len, &len, log );
    assert( !glGetError() );
    if( len >= sizeof( log ) ) {
      len = sizeof( log )-1;
    }
    log[len] = 0;
    //  catch degenerate cases
    assert( !strstr( log, "software" ) );
    if( strstr( log, "software" ) ) {
      warning( "Shader program seems to use emulation.\n" );
    }
  }
  else {
    strcpy( log, "Could not link program." );
  }
  size_t l = strlen( vsLog ) + 1 + strlen( fsLog ) + 1 + strlen( log ) + 1 + 1;
  log_ = new char[ l ];
  sprintf( log_, "%s\n%s\n%s\n", vsLog, fsLog, log );
  if( ok ) {
    linked_ = true;
    //  extract parameters
    GLint np = 0;
    glGetObjectParameterivARB( programHandle_, GL_OBJECT_ACTIVE_UNIFORMS_ARB, &np );
    parameterCount_ = np;
    parameters_ = new GLProgramParameter[ np ];
    unsigned int ix = 0;
    for( int i = 0; i < np; ++i ) {
      GLsizei l = 0;
      GLint s = 0;
      GLenum t = 0;
      glGetActiveUniformARB( programHandle_, ix, MAX_PARAMETER_NAME, &l, &s, &t, parameters_[i].name_ );
      if( l >= MAX_PARAMETER_NAME ) l = MAX_PARAMETER_NAME-1;
      parameters_[i].name_[l] = 0;
      parameters_[i].location_ = glGetUniformLocationARB( programHandle_, parameters_[i].name_ );
      parameters_[i].m_ = s;
      if( !strncmp( parameters_[i].name_, "gl_", 3 ) || (parameters_[i].location_ < 0) ) {
        //  built-in state, or ignored state
        --np;
        --i;
      }
      else {
        switch( t ) {
        case GL_FLOAT:
          parameters_[i].type_ = GLProgramValue::FloatN;
          parameters_[i].n_ = 1;
          break;
        case GL_FLOAT_VEC2_ARB:
          parameters_[i].type_ = GLProgramValue::FloatN;
          parameters_[i].n_ = 2;
          break;
        case GL_FLOAT_VEC3_ARB:
          parameters_[i].type_ = GLProgramValue::FloatN;
          parameters_[i].n_ = 3;
          break;
        case GL_FLOAT_VEC4_ARB:
          parameters_[i].type_ = GLProgramValue::FloatN;
          parameters_[i].n_ = 4;
          break;
        case GL_FLOAT_MAT2_ARB:
          parameters_[i].type_ = GLProgramValue::FloatNxM;
          parameters_[i].n_ = 2;
          parameters_[i].m_ = 2 * s;
          break;
        case GL_FLOAT_MAT3_ARB:
          parameters_[i].type_ = GLProgramValue::FloatNxM;
          parameters_[i].n_ = 3;
          parameters_[i].m_ = 3 * s;
          break;
        case GL_FLOAT_MAT4_ARB:
          parameters_[i].type_ = GLProgramValue::FloatNxM;
          parameters_[i].n_ = 4;
          parameters_[i].m_ = 4 * s;
          break;
        case GL_SAMPLER_2D_ARB:
          parameters_[i].type_ = GLProgramValue::Sampler2D;
          parameters_[i].n_ = s;
          break;
        case GL_SAMPLER_CUBE_ARB:
          parameters_[i].type_ = GLProgramValue::SamplerCube;
          parameters_[i].n_ = s;
          break;
        default:
          assert( !"Wooah! I don't support this type!" );
          parameters_[i].type_ = GLProgramValue::Null;
        }
      }
      assert( !glGetError() );
      ++ix;
    }
    parameterCount_ = np; //  re-set the count once we've filtered out the GL internal states
  }

  return ok;
}

size_t GLProgram::parameterCount() const
{
  return parameterCount_;
}

GLProgramParameter * GLProgram::parameterAt( size_t index ) const
{
  assert( index < parameterCount_ );
  return &parameters_[index];
}

int GLProgram::parameterIndex( char const * name ) const
{
  size_t ub = parameterCount_;
  for( size_t i = 0; i < ub; ++i ) {
    if( !stricmp( name, parameters_[i].name() ) ) {
      return int(i);
    }
  }
  return -1;
}

void GLProgram::nuke()
{
  if( vertexShader_ ) vertexShader_->release();
  if( fragmentShader_ ) fragmentShader_->release();
  delete[] parameters_;
  if( log_ != programError_ ) {
    delete[] log_;
  }
}

void GLProgram::pave()
{
  vertexShader_ = 0;
  fragmentShader_ = 0;
  parameters_ = 0;
  parameterCount_ = 0;
  linked_ = false;
  log_ = programError_;
}

void GLProgram::copy( GLProgram const & o )
{
  pave();
  if( o.linked_ ) {
    (void)link( o.vertexShader_, o.fragmentShader_ );
  }
}

void GLProgram::flushState()
{
  glUseProgramObjectARB( 0 );
  gCurState->program = 0;
}



////////////////////////
//  GLProgramInstance
////////////////////////

GLProgramInstance::GLProgramInstance()
{
  pave();
}

GLProgramInstance::GLProgramInstance( GLProgram * program )
{
  pave();
  assign( program );
}

GLProgramInstance::GLProgramInstance( GLProgramInstance const & o )
{
  pave();
  copy( o );
}

GLProgramInstance & GLProgramInstance::operator=( GLProgramInstance const & o )
{
  nuke();
  copy( o );
  return *this;
}

GLProgramInstance::~GLProgramInstance()
{
  nuke();
}

GLProgramParameter * GLProgramInstance::parameterAt( size_t index ) const
{
  assert( index < numValues_ );
  return values_[index].param;
}

int GLProgramInstance::parameterIndex( char const * name ) const
{
  return program_->parameterIndex( name );
}

GLProgramValue & GLProgramInstance::valueAt( size_t index )
{
  assert( index < numValues_ );
  return values_[index].value;
}

GLProgramValue const & GLProgramInstance::valueAt( size_t index ) const
{
  assert( index < numValues_ );
  return values_[index].value;
}

bool GLProgramInstance::setParameterValueAt( size_t index, GLProgramValue const & v ) const
{
  assert( index < numValues_ );
  values_[index].value = v;
  return true;
}

void GLProgramInstance::apply()
{
  if( !gCurState ) {
    assert( gCurState && "you need to GLSwitchToProgramInstanceState() once per context" );
    gCurState = new ProgramState;
  }
  assert( !glGetError() );
  if( !program_->linked_ ) {
    assert( !"attempt to use un-linked program" );
    return;
  }
  GLhandleARB p = program_->programHandle_;
  if( gCurState->program != p ) {
    glUseProgramObjectARB( p );
    assert( !glGetError()) ;
    gCurState->program = p;
  }
  int mt = gCurState->maxTexture;
  memset( gCurState->texture, 0, sizeof( gCurState->texture ) );
  for( size_t i = 0; i < numValues_; ++i ) {
    values_[i].value.apply( values_[i].param->location() );
  }
  glActiveTexture( GL_TEXTURE0_ARB );
}

void GLProgramInstance::nuke()
{
  if( program_ ) {
    program_->release();
  }
  delete[] values_;
}

void GLProgramInstance::pave()
{
  program_ = 0;
  values_ = 0;
  numValues_ = 0;
}

void GLProgramInstance::copy( GLProgramInstance const & o )
{
  program_ = o.program_;
  if( program_ ) {
    program_->addRef();
  }
  numValues_ = o.numValues_;
  if( numValues_ ) {
    values_ = new Value[numValues_];
    std::copy( o.values_, &o.values_[numValues_], values_ );
  }
  else {
    values_ = 0;
  }
}

static GLProgramValue floatValue( size_t n, size_t m )
{
  float * f = new float[n*m];
  memset( f, 0, sizeof(float)*n*m );
  GLProgramValue ret = GLProgramValue( f, n, false, m );
  delete[] f;
  return ret;
}

static GLProgramValue matrixValue( size_t n, size_t m )
{
  float * f = new float[n*n*m];
  memset( f, 0, sizeof(float)*n*n*m );
  GLProgramValue ret = GLProgramValue( f, n, true, m );
  delete[] f;
  return ret;
}

void GLProgramInstance::assign( GLProgram * p )
{
  program_ = p;
  if( !p ) {
    numValues_ = 0;
    values_ = 0;
    return;
  }
  p->addRef();
  numValues_ = p->parameterCount();
  values_ = new Value[numValues_];
  for( size_t i = 0; i < numValues_; ++i ) {
    values_[i].param = p->parameterAt(i);
    //todo: do I want to set up default parameter values?
    switch( values_[i].param->type() ) {
    default:
    case GLProgramValue::Null:
      break;
    case GLProgramValue::FloatN:
      values_[i].value = floatValue( values_[i].param->n(), values_[i].param->m() );
      break;
    case GLProgramValue::FloatNxM:
      values_[i].value = matrixValue( values_[i].param->n(), values_[i].param->m() );
      break;
    case GLProgramValue::Sampler2D:
      values_[i].value = GLProgramValue( TextureRef::null );
      break;
    case GLProgramValue::SamplerCube:
      values_[i].value = GLProgramValue( TextureRef::nullCube );
      break;
    }
  }
}


