
#if !defined( glprograms_h )
#define glprograms_h

#include <windows.h>
#include <GL/GL.h>
#include <assert.h>
#include "extsetup.h"
#include "textures.h"



//  Context management: because render states are not shared between 
//  contexts, you need to save the program instance specific state 
//  after switching to another context. Use it like so:
//
//  HDC dc1, dc2 = ...;
//  HGLRC rc1, rc2 = ...;
//  GLProgramInstanceState is1 = 0, is2 = 0;
//
//  wglMakeCurrent( dc1, rc1 );
//  GLSwitchToProgramInstanceState( is1 );
//  ...
//  wglMakeCurrent( dc2, rc2 );
//  GLSwitchToProgramInstanceState( is2 );
//  ...
//  wglMakeCurrent( dc1, rc1 );
//  GLSwitchToProgramInstanceState( is1 );
//
//  However, it is assumes that you will wglShareLists() between the 
//  contexts so that the program objects and texture objects will 
//  translate between them; this is only to keep multiple caches of 
//  non-shared context render state.
//
typedef struct GLProgramInstanceState_t * GLProgramInstanceState;
void GLSwitchToProgramInstanceState( GLProgramInstanceState & );
void GLFreeProgramInstanceState( GLProgramInstanceState & );


//todo: maybe use OBJECT_ACTIVE_UNIFORM_MAX_LENGTH_ARB instead?
#define MAX_PARAMETER_NAME 32

class GLProgramValue {
  public:

    GLProgramValue();
    explicit GLProgramValue( float f );
    //  Maximum value of n is 4.
    GLProgramValue( float const * p, size_t n );
    GLProgramValue( float const * p, size_t n, bool matrix, size_t m );
    explicit GLProgramValue( TextureRef const & t );
    GLProgramValue( GLProgramValue const & o );
    GLProgramValue & operator=( GLProgramValue const & o );
    ~GLProgramValue() {}

    //  fixme: "Type" should probably contain N and M, 
    //  so make it a three-element struct (which makes 
    //  the size of this object even bigger).
    enum Type {
      Null,       //  no storage
      FloatN,     //  (scalar/vector) array
      FloatNxM,   //  (matrix) array
      Sampler2D,  //  stored in texture_
      SamplerCube,//  stored in texture_
    };
    Type type() const { return type_; }

    //  fixme: I should support partial get and set
    void get( float * p, size_t n, size_t m ) const;
    void getMatrix( float * p, size_t n, size_t m ) const;
    void get( TextureRef * oRef ) const;
    // You can't set() a parameter to a different type than it already is
    void set( float const * p, size_t n, size_t m );
    void setMatrix( float const * p, size_t n, size_t m );
    void set( TextureRef const & oRef );
    size_t n() const { return n_; }
    size_t m() const { return m_; }
    void apply( int location ) const;

  protected:
    //  todo: this is grody -- fix it up some
    //  - storing N and M as size_t each? does it matter?
    //  - N and M should probably be part of type_
    //  - matrix or array should be cleaner separated
    //  - floats_ should point to local memory when small
    //  - union of texture and local floats?
    Type type_;
    size_t n_;
    size_t m_;
    TextureRef texture_;
    float * floats_;

    void nuke();
    void pave();
    void copy( GLProgramValue const & o );
};

//  The theory is that I pass along a pointer to a ref count 
//  whenever I pass along the actual handle. This means that 
//  I can correctly keep a ref count between multiple copies 
//  of GLRefCountedHandle. Ownership for the allocated ref 
//  count transfers with the handle. When the ref count reaches 
//  zero, the count, and the handle, are destroyed.
//  You must not use the explicit constructor from handle on 
//  a GL handle more than once, or you will get double delete 
//  problems. You have been warned!
class GLRefCountedHandle {
  public:
    GLRefCountedHandle() {
      refCount_ = new unsigned int( 1 );
      handle_ = 0;
    }
    explicit GLRefCountedHandle( GLhandleARB h ) {
      refCount_ = new unsigned int( 1 );
      handle_ = h;
    }
    GLRefCountedHandle( GLRefCountedHandle const & o ) {
      refCount_ = o.refCount_;
      assert( *refCount_ > 0 );
      ++*refCount_;
      handle_ = o.handle_;
    }
    GLRefCountedHandle & operator=( GLRefCountedHandle const & o ) {
      assert( *refCount_ > 0 );
      if( !--*refCount_ ) {
        if( handle_ ) {
          glDeleteObjectARB( handle_ );
        }
        delete refCount_;
      }
      refCount_ = o.refCount_;
      assert( *refCount_ > 0 );
      ++*refCount_;
      handle_ = o.handle_;
      return *this;
    }
    ~GLRefCountedHandle() {
      assert( *refCount_ > 0 );
      if( !--*refCount_ ) {
        if( handle_ ) {
          glDeleteObjectARB( handle_ );
        }
        delete refCount_;
      }
    }

    operator GLhandleARB() const {
      return handle_;
    }
  protected:
    GLhandleARB handle_;
    unsigned int * refCount_;
};

template< class T >
class GLPtr {
  public:
    GLPtr() : t_( 0 ) {}
    explicit GLPtr( T * t ) : t_( t ) {}
    GLPtr( GLPtr const & o ) {
      t_ = o.t_;
      if( t_ ) t_->addRef();
    }
    GLPtr & operator=( GLPtr const & o ) {
      if( &o == this ) return *this;
      if( t_ ) t_->release();
      t_ = o.t_;
      if( t_ ) t_->addRef();
      return *this;
    }
    ~GLPtr() { if( t_ ) t_->release(); }
    T * operator&() {
      assert( t_ == 0 );
    }
    GLPtr & operator=( T * t ) {
      if( t_ ) t_->release();
      t_ = t;
      if( t_ ) t->addRef();
      return *this;
    }
    T * operator->() {
      return t_;
    }
    T & operator*() {
      return *t_;
    }
    bool operator!() {
      return !t_;
    }
    operator T *() {
      return t_;
    }
  private:
    T * t_;
};

class GLProgramFragment {
  public:
    GLProgramFragment();
    explicit GLProgramFragment( GLuint type );
    GLProgramFragment( GLProgramFragment const & o );
    GLProgramFragment & operator=( GLProgramFragment const & o );

    void addRef() { ++refCount_; }
    void release() { if( !--refCount_ ) delete this; }

    bool compile( char const * text, int size, char const * fileName );
    bool compile( size_t n, char const ** text, int const * size, char const ** fileNames, int const * lineNos );
    bool compiled() { return compiled_; }
    char const * log() const { return log_; }
    GLhandleARB handle() const { return shaderHandle_; }

  protected:

    ~GLProgramFragment();

    int             refCount_;
    GLuint          type_;    //  vertex or fragment
    GLRefCountedHandle shaderHandle_;
    bool            compiled_;
    char *          log_;

    void nuke();
    void pave();
    void copy( GLProgramFragment const & o );
    
};

class GLProgramParameter {
  public:
    GLProgramParameter();
    GLProgramParameter( GLProgramParameter const & o );
    GLProgramParameter & operator=( GLProgramParameter const & o );
    ~GLProgramParameter();

    GLProgramValue::Type type() const { return type_; }
    size_t n() const { return n_; }
    size_t m() const { return m_; }
    char const * name() const { return name_; }
    int location() const { return location_; }

  protected:

    friend class GLProgram;
    GLProgramValue::Type type_;
    size_t          n_;
    size_t          m_;
    int             location_;
    char            name_[MAX_PARAMETER_NAME];

    void nuke();
    void pave();
    void copy( GLProgramParameter const & o );
};

class GLProgram {
  public:
    GLProgram();
    GLProgram( GLProgram const & o );
    GLProgram & operator=( GLProgram const & o );

    void addRef() { ++refCount_; }
    void release() { if( !--refCount_ ) delete this; }

    bool link( GLProgramFragment * vertex, GLProgramFragment * fragment );
    bool linked() const { return linked_; }
    char const * log() const { return log_; }

    size_t parameterCount() const;
    GLProgramParameter * parameterAt( size_t index ) const;
    int parameterIndex( char const * name ) const;
    GLhandleARB handle() const { return programHandle_; }

    static GLProgram Null;

    static void flushState();

  private:
    friend class GLProgramInstance;
    ~GLProgram();

    int             refCount_;
    GLRefCountedHandle programHandle_;
    GLProgramFragment *      vertexShader_;
    GLProgramFragment *      fragmentShader_;
    //  todo: caching program uniform values, so unnecessary re-loads are not needed
    GLProgramParameter *  parameters_;
    size_t          parameterCount_;
    bool            linked_;
    char *          log_;

    void nuke();
    void pave();
    void copy( GLProgram const & o );
};

class GLProgramInstance {
  public:
    GLProgramInstance();
    explicit GLProgramInstance( GLProgram * program );
    GLProgramInstance( GLProgramInstance const & o );
    GLProgramInstance & operator=( GLProgramInstance const & o );
    ~GLProgramInstance();

    GLProgram * program() const { return program_; }
    size_t parameterCount() const { return numValues_; }
    GLProgramParameter * parameterAt( size_t index ) const;
    int parameterIndex( char const * name ) const;
    GLProgramValue & valueAt( size_t index );
    GLProgramValue const & valueAt( size_t index ) const;
    bool setParameterValueAt( size_t index, GLProgramValue const & v ) const;
    void apply();

    static GLProgramInstance Null;

  protected:
    GLProgram * program_;
    struct Value {
      GLProgramParameter * param;
      GLProgramValue value;
    };
    Value * values_;
    size_t numValues_;

    void nuke();
    void pave();
    void copy( GLProgramInstance const & o );
    void assign( GLProgram * p );
};


#endif  //  glprograms_h
