

#include "glshaderstate.h"
#include "myfilesystem.h"
#include "xmlscanner.h"

#undef min

#include <algorithm>
#include <hash_map>
#include <string>
#include <cctype>
#include <cstdlib>


namespace {
  struct ShaderState {
    ShaderState() {
      programState_ = 0;
    }
    ~ShaderState() {
      GLFreeProgramInstanceState( programState_ );
    }
    GLShaderContext context_;
    GLProgramInstanceState programState_;
  };
  ShaderState * gCurState;

  class Shader : public GLShader {
    public:
      friend class ShaderFactory;

      Shader( ShaderFactory * sf );
      ~Shader();

      virtual void dispose();
      virtual void apply();
      virtual void setTexture( char const * name, TextureRef const & ref );
      virtual void setFloat( char const * name, float f );
      virtual void setFloatN( char const * name, float const * f, int n, int m );
      virtual void setMatrix( char const * name, float const * f, int n, int m );

      ShaderFactory * factory_;
  };

  class ShaderFactory : public GLShaderFactory {
    public:
      ShaderFactory();
      ~ShaderFactory();

      virtual void dispose();
      virtual GLShader * newShader( char const * handle );
      virtual void clear();
      virtual void setSharedTexture( char const * name, TextureRef const & ref );
      virtual void setSharedFloat( char const * name, float f );
      virtual void setSharedFloatN( char const * name, float const * f, int n, int m = 1 );
      virtual void setSharedMatrix( char const * name, float const * f, int n, int m = 1 );

      struct ProgramData {
        GLPtr< GLProgram > program_;
        GLShaderContext context_;
        Data defaults_;
      };
      bool readProgramData( Data & d, ProgramData & pd, char const * fileName );
      bool readContext( char const * base, char const * end, GLShaderContext & out );
      bool readParameters( char const * base, char const * end, GLProgramInstance * pi );

      stdext::hash_map< std::string, ProgramData > programs_;
      stdext::hash_map< std::string, GLProgramValue > sharedParameters_;
  };
}



void GLSwitchToShaderInstanceState( GLShaderInstanceState & ustate )
{
  ShaderState * ss = reinterpret_cast< ShaderState * >( ustate );
  if( !ss ) {
    //  First usage of this state instance.
    //  This may blow away previous gCurState, but that should be 
    //  remembered by some previous call to GLSwitchToShaderInstanceState()
    gCurState = new ShaderState;
    ustate = reinterpret_cast< GLShaderInstanceState >( gCurState );
  }
  else {
    gCurState = ss;
  }
  GLSwitchToProgramInstanceState( gCurState->programState_ );
  //  This is a little redundant, but won't hurt much compared 
  //  to the HGLRC switch we presumably just did. Note that we 
  //  don't force the program state in GLSwitchToProgramInstanceState().
  GLForceContext( gCurState->context_ );
}

void GLFreeShaderInstanceState( GLShaderInstanceState & ustate )
{
  ShaderState * ss = reinterpret_cast< ShaderState * >( ustate );
  ustate = 0;
  delete ss;
  if( ss == gCurState ) {
    gCurState = 0;
  }
}



void GLForceContext( GLShaderContext const & c )
{
  if( c.alphaTest ) {
    glEnable( GL_ALPHA_TEST );
  }
  else {
    glDisable( GL_ALPHA_TEST );
  }
  glAlphaFunc( c.alphaTestFunc, c.alphaTestLevel );

  if( c.blend ) {
    glEnable( GL_BLEND );
  }
  else {
    glDisable( GL_BLEND );
  }
  glBlendFunc( c.blendSrc, c.blendDst );

  if( c.depthTest ) {
    glEnable( GL_DEPTH_TEST );
  }
  else {
    glDisable( GL_DEPTH_TEST );
  }
  glDepthFunc( c.depthTestFunc );

  glDepthMask( c.depthWrite ? GL_TRUE : GL_FALSE );

  glColor4ub( c.color.red_, c.color.green_, c.color.blue_, c.color.alpha_ );

  if( c.cullFace == GL_NONE ) {
    glDisable( GL_CULL_FACE );
  }
  else {
    glEnable( GL_CULL_FACE );
    glCullFace( c.cullFace );
  }

  assert( !glGetError() );

  gCurState->context_ = c;
}


void GLContextDelta( GLShaderContext const & nu, GLShaderContext & old )
{
  if( nu.alphaTest ) {
    if( !old.alphaTest ) {
      glEnable( GL_ALPHA_TEST );
      old.alphaTest = true;
    }
    if( old.alphaTestFunc != nu.alphaTestFunc ||
        old.alphaTestLevel != nu.alphaTestLevel ) {
      glAlphaFunc( nu.alphaTestFunc, nu.alphaTestLevel );
      old.alphaTestFunc = nu.alphaTestFunc;
      old.alphaTestLevel = nu.alphaTestLevel;
    }
  }
  else if( old.alphaTest ) {
    glDisable( GL_ALPHA_TEST );
    old.alphaTest = false;
  }

  if( nu.blend ) {
    if( !old.blend ) {
      glEnable( GL_BLEND );
      old.blend = true;
    }
    if( old.blendSrc != nu.blendSrc ||
        old.blendDst != nu.blendDst ) {
      glBlendFunc( nu.blendSrc, nu.blendDst );
      old.blendSrc = nu.blendSrc;
      old.blendDst = nu.blendDst;
    }
  }
  else if( old.blend ) {
    glDisable( GL_BLEND );
    old.blend = false;
  }

  if( nu.depthTest ) {
    if( !old.depthTest ) {
      glEnable( GL_DEPTH_TEST );
      old.depthTest = true;
    }
    if( old.depthWrite != nu.depthWrite ) {
      glDepthMask( nu.depthWrite ? GL_TRUE : GL_FALSE );
      old.depthWrite = nu.depthWrite;
    }
    if( old.depthTestFunc != nu.depthTestFunc ) {
      glDepthFunc( nu.depthTestFunc );
      old.depthTestFunc = nu.depthTestFunc;
    }
  }
  else if( old.depthTest ) {
    glDisable( GL_DEPTH_TEST );
    old.depthTest = false;
  }

  //  Colors are changed by vertex arrays and whatnot, so I 
  //  might as well slam it for each draw.
  glColor4ub( nu.color.red_, nu.color.green_, nu.color.blue_, nu.color.alpha_ );
  old.color = nu.color;

  if( old.cullFace != nu.cullFace ) {
    if( nu.cullFace == GL_NONE ) {
      glDisable( GL_CULL_FACE );
    }
    else {
      glEnable( GL_CULL_FACE );
      glCullFace( nu.cullFace );
    }
    old.cullFace = nu.cullFace;
  }
  assert( !glGetError() );
}




/////////////////////////////
//  Shader
/////////////////////////////

Shader::Shader( ShaderFactory * sf )
{
  instance_ = 0;
  factory_ = sf;
}

Shader::~Shader()
{
  delete instance_;
}

void Shader::dispose()
{
  delete this;
}

void Shader::apply()
{
  //  If this triggers, you forgot to set up the instance state 
  //  with GLSwitchToInstanceState() after you created your context.
  assert( gCurState );

  //  apply the appropriate instance
  if( instance_ ) {
    //  Map shared parameters to the instance. It's pretty bad that I do 
    //  it here, but the GLSL language doesn't have shared uniforms anyway.
    size_t m = instance_->parameterCount();
    for( size_t i = 0; i < m; ++i ) {
      GLProgramParameter * pp = instance_->parameterAt( i );
      stdext::hash_map< std::string, GLProgramValue >::iterator sp = factory_->sharedParameters_.find( pp->name() );
      if( sp != factory_->sharedParameters_.end() ) {
        instance_->setParameterValueAt( i, (*sp).second );
      }
    }
    instance_->apply();
  }
  else {
    GLProgramInstance::Null.apply();
  }
  //  do a state delta for the context
  GLContextDelta( context_, gCurState->context_ );
}

void Shader::setTexture( char const * name, TextureRef const & ref )
{
  if( !instance_ ) return;
  int i = instance_->parameterIndex( name );
  if( i < 0 ) return;
  GLProgramValue pv( ref );
  instance_->setParameterValueAt( i, pv );
}

void Shader::setFloat( char const * name, float f )
{
  if( !instance_ ) return;
  int i = instance_->parameterIndex( name );
  if( i < 0 ) return;
  GLProgramValue pv( f );
  instance_->setParameterValueAt( i, pv );
}

void Shader::setFloatN( char const * name, float const * f, int n, int m = 1 )
{
  if( !instance_ ) return;
  int i = instance_->parameterIndex( name );
  if( i < 0 ) return;
  GLProgramValue pv( f, n, false, m );
  instance_->setParameterValueAt( i, pv );
}

void Shader::setMatrix( char const * name, float const * f, int n, int m = 1 )
{
  if( !instance_ ) return;
  int i = instance_->parameterIndex( name );
  if( i < 0 ) return;
  GLProgramValue pv( f, n, true, m );
  instance_->setParameterValueAt( i, pv );
}



/////////////////////////////
//  ShaderFactory
/////////////////////////////


ShaderFactory::ShaderFactory()
{
}

ShaderFactory::~ShaderFactory()
{
}

void ShaderFactory::dispose()
{
  delete this;
}

static bool glEnumGet( char const * str, GLenum * val )
{
  static struct {
    char const * str;
    GLenum val;
  }
  vals[] = {
    //  blends
    { "one", GL_ONE },
    { "zero", GL_ZERO },
    { "src_alpha", GL_SRC_ALPHA },
    { "one_minus_src_alpha", GL_ONE_MINUS_SRC_ALPHA },
    { "dst_alpha", GL_DST_ALPHA },
    { "one_minus_dst_alpha", GL_ONE_MINUS_DST_ALPHA },
    { "src_color", GL_SRC_COLOR },
    { "dst_color", GL_DST_COLOR },

    //  tests
    { "always", GL_ALWAYS },
    { "never", GL_NEVER },
    { "less", GL_LESS },
    { "lequal", GL_LEQUAL },
    { "equals", GL_EQUAL },
    { "gequal", GL_GEQUAL },
    { "greater", GL_GREATER },
    { "notequal", GL_NOTEQUAL },

    //  cull modes
    { "none", GL_NONE },
    { "back", GL_BACK },
    { "front", GL_FRONT },
  };
  for( int i = 0; i < sizeof(vals)/sizeof(vals[0]); ++i ) {
    if( !stricmp( vals[i].str, str ) ) {
      *val = vals[i].val;
      return true;
    }
  }
  return false;
}

static bool token( char const ** base, char const * end, char * tok, size_t s )
{
  char const * ptr = *base;
  while( isspace( *ptr ) && ptr < end ) {
    ++ptr;
  }
  if( ptr == end ) return false;
  bool inQuote = false;
  if( *ptr == '\"' ) {
    inQuote = true;
    ++ptr;
  }
  bool lastQuote = true;
  while( (!isspace( *ptr ) || inQuote) && ptr < end && s > 0 ) {
    if( *ptr == '\"' && !lastQuote ) {
      inQuote = !inQuote;
      ++ptr;
      lastQuote = true;
    }
    else {
      lastQuote = false;
      *tok++ = *ptr++;
      --s;
    }
  }
  *base = ptr;
  if( s == 0 ) {
    tok[-1] = 0;
  }
  else {
    *tok = 0;
  }
  return true;
}

static bool token_flt( char const ** base, char const * end, float * f, char * tok, size_t s )
{
  tok[0] = 0;
  if( !token( base, end, tok, s ) ) {
    return false;
  }
  char * term = 0;
  *f = (float)strtod( tok, &term );
  if( term == tok || !term ) {
    return false;
  }
  return true;
}

char const * next_line( char const * base, char const * end )
{
  bool inNl = false;
  while( base < end ) {
    if( *base == '\n' ) {
      inNl = true;
    }
    else if( !isspace( *base ) && inNl ) {
      return base;
    }
    ++base;
  }
  return base;
}

inline static float clampf( float f, float lo, float hi ) {
  return (f < lo) ? lo : (f > hi) ? hi : f;
}

bool ShaderFactory::readContext( char const * base, char const * end, GLShaderContext & out )
{
  char tok[128];
  while( token( &base, end, tok, sizeof(tok) ) ) {
    if( tok[0] == '/' && tok[1] == '/' ) {
      base = next_line( base, end );
    }
    else if( !strcmp( tok, "alphaTest" ) ) {
      if( !token( &base, end, tok, sizeof(tok) ) ) {
        error( "'alphaTest' needs arguments in context description.\n" );
        return false;
      }
      else if( !strcmp( tok, "disable" ) ) {
        out.alphaTest = false;
      }
      else if( !glEnumGet( tok, &out.alphaTestFunc ) ) {
        error( "'alphaTest' function '%s' is not recognized.\n", tok );
        return false;
      }
      else if( !token_flt( &base, end, &out.alphaTestLevel, tok, sizeof(tok) ) ) {
        error( "'alphaTest' needs reference value as third argument.\n" );
        return false;
      }
      else {
        out.alphaTest = true;
      }
    }
    else if( !strcmp( tok, "depthTest" ) ) {
      if( !token( &base, end, tok, sizeof(tok) ) ) {
        error( "'depthWrite' needs arguments in context description.\n" );
        return false;
      }
      else if( !strcmp( tok, "disable" ) ) {
        out.depthTest = false;
      }
      else if( !glEnumGet( tok, &out.depthTestFunc ) ) {
        error( "'depthTest' function '%s' is not recognized.\n", tok );
        return false;
      }
    }
    else if( !strcmp( tok, "depthWrite" ) ) {
      if( !token( &base, end, tok, sizeof(tok) ) ) {
        error( "'depthWrite' needs arguments in context description.\n" );
        return false;
      }
      else if( !strcmp( tok, "disable" ) ) {
        out.depthWrite = false;
      }
      else if( !strcmp( tok, "enable" ) ) {
        out.depthWrite = true;
      }
      else {
        error( "'depthWrite' needs enable or disable argument.\n" );
        return false;
      }
    }
    else if( !strcmp( tok, "blend" ) ) {
      if( !token( &base, end, tok, sizeof(tok) ) ) {
        error( "'blend' needs 2 arguments in context description.\n" );
        return false;
      }
      else if( !strcmp( tok, "disable" ) ) {
        out.blend = false;
      }
      else if( !glEnumGet( tok, &out.blendSrc ) ) {
        error( "'blendSrc' function '%s' is not recognized.\n", tok );
        return false;
      }
      else if( !token( &base, end, tok, sizeof(tok) ) ) {
        error( "'blend' needs 2 arguments in context description.\n" );
        return false;
      }
      else if( !glEnumGet( tok, &out.blendDst ) ) {
        error( "'blendDst' function '%s' is not recognized.\n", tok );
        return false;
      }
      else {
        out.blend = true;
      }
    }
    else if( !strcmp( tok, "color" ) ) {
      float r, g, b, a;
      if( !token_flt( &base, end, &r, tok, sizeof(tok) ) ||
          !token_flt( &base, end, &g, tok, sizeof(tok) ) ||
          !token_flt( &base, end, &b, tok, sizeof(tok) ) ||
          !token_flt( &base, end, &a, tok, sizeof(tok) ) ) {
        error( "'color' requires four float values near '%s'.\n", tok );
        return false;
      }
      out.color = GLColor( 
          (unsigned char )(clampf( r, 0, 1 ) * 255),
          (unsigned char )(clampf( g, 0, 1 ) * 255),
          (unsigned char )(clampf( b, 0, 1 ) * 255),
          (unsigned char )(clampf( a, 0, 1 ) * 255) );
    }
    else if( !strcmp( tok, "cullFace" ) ) {
      if( !token( &base, end, tok, sizeof(tok) ) ) {
        error( "'cullFace' needs 1 argument in the context description.\n" );
        return false;
      }
      else if( !glEnumGet( tok, &out.cullFace ) ) {
        error( "'cullFace' needs NONE, FRONT or BACK in the context description.\n" );
        return false;
      }
      else if( out.cullFace != GL_NONE && out.cullFace != GL_FRONT && out.cullFace != GL_BACK ) {
        error( "'cullFace' needs NONE, FRONT or BACK in the context description.\n" );
        out.cullFace = GL_NONE;
        return false;
      }
    }
    else {
      error( "Unexpected token '%s' in context description.\n", tok );
      return false;
    }
  }
  return true;
}

char const * find_str( char const * haystack, char const * end, char const * needle )
{
  size_t l = strlen( needle );
  char root = *needle;    //  this means that finding the empty string never succeeds
  while( haystack <= end-l ) {
    if( (*haystack == root) && !strncmp( haystack, needle, l ) ) {
      return haystack;
    }
    ++haystack;
  }
  return 0;
}

char const * string_param( char const * base, char const * end )
{
  static char string_param_ret[128];
  while( base < end ) {
    if( *base == '\"' ) {
      if( token( &base, end, string_param_ret, sizeof(string_param_ret) ) ) {
        return string_param_ret;
      }
    }
    ++base;
  }
  return 0;
}

bool ShaderFactory::readProgramData( Data & d, ProgramData & pd, char const * fileName )
{
  //  format of a program file:
  //
  //  shared shader declarations (varyings)
  //  #section context
  //  <context directives>
  //  #section default
  //  <default values>
  //  #section vertex-shader
  //  <vertex shader source>
  //  #section fragment-shader
  //  <fragment shader source>
  //
  //  if there is a string after the section name, that 
  //  is taken as a file name that's included, instead 
  //  of using the text after the section directive

  Data vData;
  GLProgramFragment * vShader = 0;
  GLProgramFragment * fShader = 0;
  GLProgram * program = 0;

  char const * base = d.readBase();
  char const * end = base + d.size();
  char const * sharedTop = base;
  char const * error_pos = base;
  char const * looking_for = "#section context";
  char const * contextTop = find_str( error_pos, end, looking_for );
  if( !contextTop ) {
    goto syntax_error;
  }
  char const * sharedEnd = contextTop;
  error_pos = next_line( sharedEnd, end );
  looking_for = "#section default";
  char const * defaultsTop = find_str( error_pos, end, looking_for );
  if( !defaultsTop ) {
    goto syntax_error;
  }
  char const * contextEnd = defaultsTop;
  error_pos = next_line( contextEnd, end );
  looking_for = "#section vertex-shader";
  char const * vertexTop = find_str( error_pos, end, looking_for );
  if( !vertexTop ) {
    goto syntax_error;
  }
  char const * defaultsEnd = vertexTop;
  error_pos = next_line( contextEnd, end );
  looking_for = "#section fragment-shader";
  char const * fragmentTop = find_str( error_pos, end, looking_for );
  if( !fragmentTop ) {
    goto syntax_error;
  }
  char const * vertexEnd = fragmentTop;
  char const * fragmentEnd = end;

  //  create vertex shader
  char const * nl = next_line( vertexTop, end );
  char const * sparam = string_param( vertexTop, nl );
  vertexTop = nl;
  error_pos = nl;
  char vFileName[128];
  strncpy( vFileName, fileName, 128 );
  vFileName[127] = 0;
  char const * vFileNames[2] = { fileName, vFileName };
  int vFileLines[2] = { 1, 1 };
  for( char const * ptr = d.readBase(); ptr < vertexTop; ++ptr ) {
    if( *ptr == '\n' ) {
      ++vFileLines[1];
    }
  }
  int fileVLines = vFileLines[1];
  for( char const * ptr = vertexTop; ptr < fragmentTop; ++ptr ) {
    if( *ptr == '\n' ) {
      ++vFileLines[1];
    }
  }
  int fileFLines = vFileLines[1];
  vFileLines[1] = fileVLines;
  //  get it from a separate file if indicated
  if( sparam ) {
    //  todo: cache and re-use the program fragment
    vFileLines[1] = 0;
    strncpy( vFileName, sparam, 128 );
    vFileName[127] = 0;
    if( vertexEnd > nl ) {
      warning( "Ignoring extra data when reading vertex shader from file '%s'\n", sparam );
    }
    if( !FS::readFileByPath( sparam, &vData ) && !FS::readFileByHandle( sparam, &vData ) ) {
      looking_for = sparam;
      goto syntax_error;
    }
    vertexTop = vData.readBase();
    vertexEnd = vertexTop + vData.size();
  }
  vShader = new GLProgramFragment( GL_VERTEX_SHADER_ARB );
  char const * text[2];
  int size[2];
  text[0] = sharedTop;
  text[1] = vertexTop;
  size[0] = int( sharedEnd - sharedTop );
  size[1] = int( vertexEnd - vertexTop );
  if( !vShader->compile( 2, text, size, vFileNames, vFileLines ) ) {
    error( "Vertex shader compile error:\n%s\n", vShader->log() );
    looking_for = "vertex shader";
    goto syntax_error;
  }

  //  create fragment shader
  nl = next_line( fragmentTop, end );
  sparam = string_param( fragmentTop, nl );
  fragmentTop = nl;
  error_pos = nl;
  vFileLines[1] = fileFLines;
  strncpy( vFileName, fileName, 128 );
  vFileName[127] = 0;
  //  get it from a separate file if indicated
  if( sparam ) {
    //  todo: cache and re-use the program fragment
    vFileLines[1] = 0;
    strncpy( vFileName, sparam, 128 );
    vFileName[127] = 0;
    if( fragmentEnd > nl ) {
      warning( "Ignoring extra data when reading fragment shader from file '%s'\n", sparam );
    }
    if( !FS::readFileByPath( sparam, &vData ) && !FS::readFileByHandle( sparam, &vData ) ) {
      looking_for = sparam;
      goto syntax_error;
    }
    fragmentTop = vData.readBase();
    fragmentEnd = fragmentTop + vData.size();
  }
  fShader = new GLProgramFragment( GL_FRAGMENT_SHADER_ARB );
  text[0] = sharedTop;
  text[1] = fragmentTop;
  size[0] = int( sharedEnd - sharedTop );
  size[1] = int( fragmentEnd - fragmentTop );
  if( !fShader->compile( 2, text, size, vFileNames, vFileLines ) ) {
    error( "Fragment shader compile error:\n%s\n", fShader->log() );
    looking_for = "fragment shader";
    goto syntax_error;
  }

  //  link the program
  program = new GLProgram();
  if( !program->link( vShader, fShader ) ) {
    error( "Shader link error:\n%s\n", program->log() );
    looking_for = "linking program";
    goto syntax_error;
  }
  info( "Shader linked: %s\n", program->log() );

  //  program now has ref counts for shaders
  vShader->release();
  fShader->release();

  //  extract context
  nl = next_line( contextTop, end );
  sparam = string_param( contextTop, nl );
  contextTop = nl;
  error_pos = contextTop;
  if( sparam ) {
    if( contextEnd > nl ) {
      warning( "Ignoring extra data when reading context data from file '%s'\n", sparam );
    }
    if( !FS::readFileByPath( sparam, &vData ) && !FS::readFileByHandle( sparam, &vData ) ) {
      looking_for = sparam;
      goto syntax_error;
    }
    contextTop = vData.readBase();
    contextEnd = contextTop + vData.size();
  }
  if( !readContext( contextTop, contextEnd, pd.context_ ) ) {
    looking_for = "shader context";
    goto syntax_error;
  }

  //  Whew! We're done!
  pd.program_ = program;
  //  This won't copy the data. That's OK, because we're passing 
  //  a value in here that will be copied when put into the map.
  pd.defaults_ = Data( defaultsTop, defaultsEnd-defaultsTop );

  debug( "Successfully loaded shader data.\n" );
  return true;

syntax_error:
  error( "Error compiling program; looking for '%s'; near '%.*s'\n",
      looking_for, std::min( end-error_pos, 20 ), error_pos );
  if( program ) program->release();
  if( vShader ) vShader->release();
  if( fShader ) fShader->release();
  return false;
}

void GLShader::flushState()
{
  GLForceContext( gCurState->context_ );
  GLProgram::flushState();
}

static void getFloatN( char const * base, char const * end, GLProgramInstance * pi, size_t ix )
{
  GLProgramParameter * pp = pi->parameterAt( ix );
  size_t n = pp->n();
  size_t m = pp->m();
  float * temp = new float[n*m];
  float * cur = temp;
  float * top = &temp[n*m];
  char * nu = 0;
  while( base < end && isspace( *base ) ) {
    ++base;
  }
  while( (cur < top) && (base < end) ) {
    *cur = float( strtod( base, &nu ) );
    if( nu == base ) {
      warning( "Expected float but got '%.*s' for %s.\n", 
        int( end - base ), base, pp->name() );
      break;
    }
    base = nu;
    if( *base == ',' ) ++base;
    ++cur;
    while( base < end && isspace( *base ) ) {
      ++base;
    }
  }
  if( cur < top ) {
    warning( "Parameter '%s' is missing %d values.\n", pp->name(), top-cur );
    while( cur < top ) {
      *cur++ = 0;
    }
  }
  pi->valueAt( ix ).set( temp, n, m );
  delete[] temp;
}

static void getFloatNxM( char const * base, char const * end, GLProgramInstance * pi, size_t ix )
{
  GLProgramParameter * pp = pi->parameterAt( ix );
  size_t n = pp->n();
  size_t m = pp->m();
  float * temp = new float[n*n*m];
  float * cur = temp;
  float * top = &temp[n*n*m];
  char * nu = 0;
  while( base < end && isspace( *base ) ) {
    ++base;
  }
  while( (cur < top) && (base < end) ) {
    *cur = float( strtod( base, &nu ) );
    if( nu == base ) {
      warning( "Expected matrix float but got '%.*s' for %s.\n", 
        int( end - base ), base, pp->name() );
      break;
    }
    base = nu;
    if( *base == ',' ) ++base;
    ++cur;
    while( base < end && isspace( *base ) ) {
      ++base;
    }
  }
  if( cur < top ) {
    warning( "Parameter '%s' is missing %d values.\n", pp->name(), top-cur );
    while( cur < top ) {
      *cur++ = 0;
    }
  }
  pi->valueAt( ix ).setMatrix( temp, n, m );
  delete[] temp;
}

static void getSampler2D( char const * base, char const * end, GLProgramInstance * pi, size_t ix )
{
  GLProgramParameter * pp = pi->parameterAt( ix );
  char name[128];
  if( !token( &base, end, name, sizeof(name) ) ) {
    warning( "Parameter '%s' is missing a texture name.\n", pp->name() );
    return;
  }
  if( !strchr( name, '.' ) ) {
    char path[128];
    if( FS::translateHandle( name, path, sizeof(path) ) ) {
      strcpy( name, path );
    }
  }
  TextureRef tr( name );
  if( tr.textureTargetHack() != GL_TEXTURE_2D ) {
    warning( "Parameter '%s' should be 2D texture.\n", pp->name() );
    return;
  }
  pi->valueAt( ix ).set( tr );
}

static void getSamplerCube( char const * base, char const * end, GLProgramInstance * pi, size_t ix )
{
  GLProgramParameter * pp = pi->parameterAt( ix );
  char name[128];
  if( !token( &base, end, name, sizeof(name) ) ) {
    warning( "Parameter '%s' is missing a texture name.\n", pp->name() );
    return;
  }
  TextureRef tr( name );
  if( tr.textureTargetHack() != GL_TEXTURE_CUBE_MAP_ARB ) {
    warning( "Parameter '%s' should be cube map.\n", pp->name() );
    return;
  }
  pi->valueAt( ix ).set( tr );
}

bool ShaderFactory::readParameters( char const * base, char const * end, GLProgramInstance * pi )
{
  while( *base && isspace( *base ) && base < end ) {
    ++base;
  }
  while( base < end ) {
    char const * le = next_line( base, end );
    char const * nl = le;
    while( nl > base && isspace( nl[-1] ) ) {
      --nl;
    }
    if( *base != '#' && *base != '/' ) {
      char name[128];
      if( token( &base, nl, name, sizeof(name) ) ) {
        while( base < le && isspace( *base ) ) {
          ++base;
        }
        //  the value is at base through nl
        int pix = pi->parameterIndex( name );
        if( pix < 0 ) {
          warning( "Default parameter '%s' was not found in program.\n", name );
        }
        else {
          GLProgramParameter * pp = pi->parameterAt( pix );
          switch( pp->type() ) {
          default:
          case GLProgramValue::Null:
            warning( "Default parameter '%s' references a bad type (%d).\n", 
                name, pp->type() );
            break;
          case GLProgramValue::FloatN:
            getFloatN( base, nl, pi, pix );
            break;
          case GLProgramValue::FloatNxM:
            getFloatNxM( base, nl, pi, pix );
            break;
          case GLProgramValue::Sampler2D:
            getSampler2D( base, nl, pi, pix );
            break;
          case GLProgramValue::SamplerCube:
            getSamplerCube( base, nl, pi, pix );
            break;
          }
        }
      }
    }
    base = le;
  }
  return true;
}

GLShader * ShaderFactory::newShader( char const * handle )
{
  Data d;
  char path[ 256 ];
  //  I support both path names and handle names
  if( !strchr( handle, '.' ) && !strchr( handle, '/' ) ) {
    if( FS::translateHandle( handle, path, sizeof( path ) ) ) {
      handle = path;
    }
    else {
      warning( "newShader( %s ): handle not found in filemap\n", handle );
    }
  }
  std::string n( handle );
  std::transform( n.begin(), n.end(), n.begin(), tolower );
  info( "ShaderFactory::newShader( %s )\n", n.c_str() );
  stdext::hash_map< std::string, ProgramData >::iterator pptr = programs_.find( n );
  if( pptr == programs_.end() ) {
    //  load shader
    if( !FS::readFileByPath( handle, &d ) ) {
      error( "newShader( %s ): file not found\n", handle );
      return new Shader( this );
    }
    ProgramData pd;
    if( !readProgramData( d, pd, handle ) ) {
      return new Shader( this );
    }
    programs_[n] = pd;
    pptr = programs_.find( n );
    assert( pptr != programs_.end() );
  }
  // make program instance
  GLProgramInstance * pi = new GLProgramInstance( (*pptr).second.program_ );
  readParameters( (*pptr).second.defaults_.readBase(),
      (*pptr).second.defaults_.readBase()+(*pptr).second.defaults_.size(), pi );

  //  return instance
  GLShader * s = new Shader( this );
  setShaderInstance( s, pi );
  s->context() = (*pptr).second.context_;
  return s;
}

void ShaderFactory::clear()
{
  //  this releases all ref counts
  programs_.clear();
  sharedParameters_.clear();
}

void ShaderFactory::setSharedTexture( char const * name, TextureRef const & ref )
{
  sharedParameters_[name] = GLProgramValue( ref );
}

void ShaderFactory::setSharedFloat( char const * name, float f )
{
  sharedParameters_[name] = GLProgramValue( f );
}

void ShaderFactory::setSharedFloatN( char const * name, float const * f, int n, int m )
{
  sharedParameters_[name] = GLProgramValue( f, n, false, m );
}

void ShaderFactory::setSharedMatrix( char const * name, float const * f, int n, int m )
{
  sharedParameters_[name] = GLProgramValue( f, n, true, m );
}



GLShaderFactory * NewShaderFactory()
{
  return new ShaderFactory();
}
