
#if !defined( glshaderstate_h )
#define glshaderstate_h

//  GLShaderState encapsulates GLProgramInstance and 
//  pulls in auxiliary support such as XML parsing to 
//  build more of an "effect" system.

#include "glprograms.h"

//  GLShaderInstanceState serves the same purpose as GLProgramInstanceState, 
//  and in fact wraps the latter. If you use GLShaderState, then you should 
//  switch your contexts with GLSwitchToShaderInstanceState(), not 
//  GLSwitchToProgramInstanceState().
typedef struct GLShaderInstanceState_t * GLShaderInstanceState;
void GLSwitchToShaderInstanceState( GLShaderInstanceState & );
void GLFreeShaderInstanceState( GLShaderInstanceState & );

struct GLColor {
  GLColor() {
    color_ = 0xffffffff;
  }
  GLColor( unsigned int c ) {
    color_ = c;
  }
  GLColor( unsigned char r, unsigned char g, unsigned char b, unsigned char a ) {
    blue_ = b; green_ = g; red_ = r; alpha_= a;
  }
  operator unsigned int() const {
    return color_;
  }
  bool operator!=( GLColor const & o ) const {
    return color_ != o.color_;
  }
  bool operator==( GLColor const & o ) const {
    return color_ == o.color_;
  }
  union {
    unsigned int color_;
    struct {
      unsigned char blue_;
      unsigned char green_;
      unsigned char red_;
      unsigned char alpha_;
    };
  };
};

struct GLShaderContext {
  GLShaderContext() {
    blend = false;
    blendSrc = GL_ONE;
    blendDst = GL_ONE;
    alphaTest = false;
    alphaTestFunc = GL_GREATER;
    alphaTestLevel = 0.5f;
    depthTest = true;
    depthWrite = true;
    depthTestFunc = GL_LEQUAL;
    cullFace = GL_BACK;
    //  todo: add stencil
  }

  bool blend;
  bool alphaTest;
  bool depthTest;
  bool depthWrite;

  GLenum blendSrc;
  GLenum blendDst;
  GLenum alphaTestFunc;
  float alphaTestLevel;
  GLenum depthTestFunc;
  GLColor color;
  GLenum cullFace;
};

void GLForceContext( GLShaderContext const & c );
void GLContextDelta( GLShaderContext const & target, GLShaderContext & source );

class GLShader {
  public:
    virtual void dispose() = 0;
    virtual void apply() = 0;
    virtual void setTexture( char const * name, TextureRef const & ref ) = 0;
    virtual void setFloat( char const * name, float f ) = 0;
    virtual void setFloatN( char const * name, float const * f, int n, int m = 1 ) = 0;
    virtual void setMatrix( char const * name, float const * f, int n, int m = 1 ) = 0;
    GLProgramInstance * instance() { return instance_; }
    GLProgramInstance const * instance() const { return instance_; }
    GLShaderContext & context() { return context_; }
    GLShaderContext const & context() const { return context_; }
    static void flushState();
  protected:
    friend class GLShaderFactory;
    virtual ~GLShader() {}
    GLProgramInstance * instance_;
    GLShaderContext context_;
};

class GLShaderFactory {
  public:
    virtual void dispose() = 0;
    virtual GLShader * newShader( char const * handle ) = 0;
    virtual void clear() = 0;
    virtual void setSharedTexture( char const * name, TextureRef const & ref ) = 0;
    virtual void setSharedFloat( char const * name, float f ) = 0;
    virtual void setSharedFloatN( char const * name, float const * f, int n, int m = 1 ) = 0;
    virtual void setSharedMatrix( char const * name, float const * f, int n, int m = 1 ) = 0;
  protected:
    virtual ~GLShaderFactory() {}
    void setShaderInstance( GLShader * s, GLProgramInstance * pi ) const {
      s->instance_ = pi;
    }
};

GLShaderFactory * NewShaderFactory();


#endif  //  glshaderstate_h

