

#include "image.h"
#include "mytga.h"
#include "mydxt.h"
#include "mydds.h"

#include <new>


bool loadImage( char const * path, Image * oImage )
{
  if( strstr( path, ".dds" ) ) {
    return loadDdsImage( path, oImage );
  }
  if( strstr( path, ".tga" ) ) {
    return loadTgaImage( path, oImage );
  }
  warning( "loadImage(): path '%s' is not .tga or .dds\n", path );
  return false;
}

//  TODO: factor TGA loading (and class Image) into separate file

static bool getTgaInfo( char * data, long size, int * width, int * height, int * bits )
{
  TGA_header const * hdr = (TGA_header *)data;
  if( hdr->imageIDLengthInBytes != 0 ) {
    warning( "TGA image ID is not supported (length %d, or not a TGA)\n", hdr->imageIDLengthInBytes );
    return false;
  }
  if( hdr->colorMapType != TGA_colorMapType_direct ) {
    warning( "colorMapType %d is not TGA direct\n", hdr->colorMapType );
    return false;
  }
  switch( hdr->imageTypeCode ) {
  case TGA_imageTypeCode_direct:
    if( hdr->imagePixelSize != 32 && hdr->imagePixelSize != 24 ) {
      warning( "imagePixelSize %d is not TGA 24/32 bits\n", hdr->imagePixelSize );
      return false;
    }
    if( (hdr->imageDescriptor & TGA_imageDescriptor_alphaBitCount) != (hdr->imagePixelSize - 24) ) {
      warning( "imageDescriptor %d is not TGA RGB possibly with 8 bits alpha\n", hdr->imageDescriptor );
      return false;
    }
    break;
  case TGA_imageTypeCode_grayscale:
    if( hdr->imagePixelSize != 8 ) {
      warning( "imagePixelSize %d is not grayscale TGA 8 bits\n", hdr->imagePixelSize );
      return false;
    }
    if( ((hdr->imageDescriptor & TGA_imageDescriptor_alphaBitCount) & 7) != 0 ) {
      warning( "imageDescriptor %d is not TGA grayscale possibly as 8 bits alpha\n", hdr->imageDescriptor );
      return false;
    }
    break;
  default:
    warning( "imageTypeCode %d is not TGA direct or grayscale\n", hdr->imageTypeCode );
    return false;
  }
  *width = hdr->widthOfImage;
  *height = hdr->heightOfImage;
  *bits = hdr->imagePixelSize;
  if( *width < 1 || *width > 1024 || *height < 1 || *height > 1024 ) {
    warning( "image dimensions %dx%d are not within acceptable TGA limits\n", *width, *height );
    return false;
  }
  return true;
}

bool loadTgaImage( char const * path, Image * oImage )
{
  FILE * f = fopen( path, "rb" );
  if( !f ) {
    if( strlen( path ) > 250 ) {
      warning( "TGA image '%s' not found; path too long to try alternate\n", path );
    }
    else {
      char path2[ 256 ] = "";
      char const * s = strrchr( path, '/' );
      if( s ) {
        strncpy( path2, path, s-path+1 );
        path2[s-path+1] = 0;
      }
      else {
        s = path;
      }
      strcat( path2, "alt-" );
      strcat( path2, s+1 );
      warning( "TGA image '%s' not found; trying alternate '%s'\n", path, path2 );
      f = fopen( path2, "rb" );
    }
  }
  if( !f ) {
    error( "TGA image '%s' not found\n", path );
    return false;
  }
  fseek( f, 0, 2 );
  long l = ftell( f );
  fseek( f, 0, 0 );
  int rlen = int(strlen(path)+1);
  rlen += (8 - (rlen&7)&7);
  char * toFree = (char *)malloc( l+sizeof(int)+rlen );
  if( !toFree ) {
    warning( "TGA image '%s' too large: %ld bytes failed\n", path, l );
    return false;
  }
  strcpy( toFree + sizeof(int), path );
  strlwr( toFree + sizeof(int) );
  fread( toFree+sizeof(int)+rlen, 1, l, f );
  fclose( f );
  oImage->~Image();
  new( oImage ) Image();
  if( !getTgaInfo( toFree+sizeof(int)+rlen, l, &oImage->width, &oImage->height, &oImage->bits ) ) {
    warning( "TGA image '%s' mal-formed\n", path );
    free( toFree );
    return false;
  }
  oImage->pitch = oImage->width * (oImage->bits/8);
  oImage->base = (char *)toFree + 18 + sizeof(int) + rlen;
  oImage->toFree = toFree;
  *(int *)toFree = 1;
  return true;
}

bool loadDdsImage( char const * path, Image * oImage )
{
  FILE * f = fopen( path, "rb" );
  if( !f ) {
    if( strlen( path ) > 250 ) {
      warning( "DDS image '%s' not found; path too long to try alternate\n", path );
    }
    else {
      char path2[ 256 ] = "";
      char const * s = strrchr( path, '/' );
      if( s ) {
        strncpy( path2, path, s-path+1 );
        path2[s-path+1] = 0;
      }
      else {
        s = path;
      }
      strcat( path2, "alt-" );
      strcat( path2, s+1 );
      warning( "DDS image '%s' not found; trying alternate '%s'\n", path, path2 );
      f = fopen( path2, "rb" );
    }
  }
  if( !f ) {
    error( "DDS image '%s' not found\n", path );
    return false;
  }
  fseek( f, 0, 0 );
  DDS_header hdr;
  memset( &hdr, 0, sizeof( hdr ) );
  fread( &hdr, sizeof( hdr ), 1, f );
  if( hdr.dwMagic != DDS_MAGIC ) {
    warning( "DDS file '%s' is not a DDS image.\n", path );
    fclose( f );
    return false;
  }
  bool dxt1 = false;
  if( PF_IS_DXT1( hdr.sPixelFormat ) ) {
    dxt1 = true;
  }
  else if( PF_IS_DXT5( hdr.sPixelFormat ) ) {
    dxt1 = false;
  }
  else {
    warning( "DDS file '%s' is not a supported DDS image.\n", path );
    fclose( f );
    return false;
  }
  int rlen = int(strlen(path)+1);
  rlen += (8 - (rlen&7)&7);
  int w = hdr.dwWidth; if( w < 4 ) w = 4;
  int h = hdr.dwHeight; if( h < 4 ) h = 4;
  size_t l = w * h * 4;
  size_t toRead = w * h / (dxt1 ? 2 : 1);
  unsigned char * buffer = (unsigned char *)malloc( toRead );
  if( !buffer ) {
    warning( "DDS image '%s' is too large: %ld bytes failed\n", path, toRead );
    fclose( f );
    return false;
  }
  fread( buffer, toRead, 1, f );
  fclose( f );
  char * toFree = (char *)malloc( l+sizeof(int)+rlen );
  if( !toFree ) {
    warning( "DDS image '%s' too large: %ld bytes failed\n", path, l );
    free( buffer );
    return false;
  }
  strcpy( toFree + sizeof(int), path );
  strlwr( toFree + sizeof(int) );
  oImage->~Image();
  new( oImage ) Image();
  oImage->width = w;
  oImage->height = h;
  oImage->bits = 32;
  oImage->pitch = oImage->width * (oImage->bits/8);
  oImage->base = (char *)toFree + sizeof(int) + rlen;
  oImage->toFree = toFree;
  *(int *)toFree = 1;
  dxt_flip_data( buffer, w, h, dxt1 );
  dxt_decompress( buffer, w, h, (unsigned char *)oImage->base, dxt1 );
  free( buffer );
  return true;
}

bool saveTgaImage( char const * path, Image const * iImage )
{
  Image const & image = *iImage;
  if( image.bits != 24 && image.bits != 32 && image.bits != 8 ) {
    warning( "attempt to save tga with %d bits per pixel as '%s' failed\n", image.bits, path );
  }
  FILE * f = fopen( path, "wb" );
  if( !f ) {
    warning( "cannot open '%s' for writing TGA\n", path );
    return false;
  }

  TGA_header h;
  memset( &h, 0, sizeof( h ) );
  h.imageTypeCode = (image.bits == 8) ? TGA_imageTypeCode_grayscale : TGA_imageTypeCode_direct;
  h.widthOfImage = image.width;
  h.heightOfImage = image.height;
  h.imagePixelSize = image.bits;
  h.imageDescriptor = (image.bits == 32) ? 8 : 0;

  //fixme: what if fwrite fails?
  fwrite( &h, sizeof( h ), 1, f );
  for( int row = 0; row < image.height; ++row ) {
    fwrite( ((char const *)image.base)+(row*image.pitch), image.width, image.bits/8, f );
  }
  fclose( f );

  return true;
}

Image::~Image()
{
  if( toFree ) {
    assert( *(int *)toFree > 0 );
    *(int *)toFree -= 1;
    if( *(int *)toFree == 0 ) {
      free( toFree );
    }
  }
}

Image::Image( Image const & o )
{
  width = o.width;
  height = o.height;
  pitch = o.pitch;
  bits = o.bits;
  base = o.base;
  toFree = o.toFree;
  if( toFree ) {
    assert( *(int *)toFree > 0 );
    *(int *)toFree += 1;
  }
}

void Image::allocate( int bi, int wi, int he )
{
  assert( wi < 8200 && he < 8200 );
  assert( wi >= 0 && he >= 0 );
  assert( bi == 8 || bi == 24 || bi == 32 );
  this->~Image();
  new( this ) Image();
  bits = bi;
  width = wi;
  height = he;
  pitch = wi * (bits/8);
  toFree = (char *)malloc( 32 + pitch * height );
  *(int*)toFree = 1;
  sprintf( (char*)toFree + sizeof(int), "%dx%dx%d", wi, he, bi );
  assert( strlen( (char*)toFree+sizeof(int) ) < 28 );
  base = (char*)toFree + 32;
}

Image & Image::operator=( Image const & o )
{
  if( &o == this ) {
    return *this;
  }
  this->~Image();
  new( this ) Image( o );
  return *this;
}

char const * Image::name() const
{
  return toFree ? ((char const *)toFree + sizeof(int)) : 0;
}


