
#if !defined( lowpassfilter1D_h )
#define lowpassfilter1D_h

#include <assert.h>

//  12 dB/octave rejection
//
//Type : biquad, tweaked butterworth
//References : Posted by Patrice Tarrabia
//Code :
//r  = rez amount, from sqrt(2) to ~ 0.1
//f  = cutoff frequency
//(from ~0 Hz to SampleRate/2 - though many
//synths seem to filter only  up to SampleRate/4)
//
//The filter algo:
//out(n) = a1 * in + a2 * in(n-1) + a3 * in(n-2) - b1*out(n-1) - b2*out(n-2)
//
//Lowpass:
//      c = 1.0 / tan(pi * f / sample_rate);
//
//      a1 = 1.0 / ( 1.0 + r * c + c * c);
//      a2 = 2* a1;
//      a3 = a1;
//      b1 = 2.0 * ( 1.0 - c*c) * a1;
//      b2 = ( 1.0 - r * c + c * c) * a1;
//
//Hipass:
//      c = tan(pi * f / sample_rate);
//
//      a1 = 1.0 / ( 1.0 + r * c + c * c);
//      a2 = -2*a1;
//      a3 = a1;
//      b1 = 2.0 * ( c*c - 1.0) * a1;
//      b2 = ( 1.0 - r * c + c * c) * a1;

class LowpassFilter {
    float a1_, b1_, b2_;
    float in_1, in_2, out_1, out_2;
  public:
    LowpassFilter( float cutoff )   //  0 through 0.5f
      {
        assert( !"test me!" );
        in_1 = 0;
        in_2 = 0;
        out_1 = 0;
        out_2 = 0;
        setCutoff( cutoff );
      }
    void setCutoff( float cutoff )
      {
        assert( cutoff > 0 && cutoff < 0.5f );
        out_1 = out_2 = 0;
        float r = sqrtf( 2.0f );
        float c = 1.0f / tanf( 3.1415926f * cutoff );
        a1_ = 1.0f / (1.0f + r * c + c * c);
        //  There's some chance the filter becomes unstable 
        //  when changing the cut-off, so re-scale out_1 and 
        //  out_2 based on the difference. This is to avoid 
        //  the filter suddently introducing energy and 
        //  oscillating out to infinity.
        float ob1 = b1_;
        float ob2 = b2_;
        b1_ = 2.0f * (1.0f - c*c) * a1_;
        b2_ = (1.0f - r*c + c*c) * a1_;
        if( b1_ > ob1 || b2_ > ob2 ) {
          out_1 *= ob1 / b1_;
          out_2 *= ob2 / b2_;
        }
      }
    void process( float const * iData, float * oData, unsigned int cnt )
      {
        //  load
        float in1 = in_1;
        float in2 = in_2;
        float out1 = out_1;
        float out2 = out_2;

        float const a1 = a1_;
        float const b1 = b1_;
        float const b2 = b2_;

        //  process
        while( cnt-- > 0 ) {
          float in = *iData++;
          //  use a2 == 2 * a1
          //  use a3 == a1
          float res = (a1 * in) + (a1 * 2 * in1) + (a1 * in2) -
              (b1 * out1) - (b2 * out2);
          out2 = out1;
          out1 = res;
          in2 = in1;
          in1 = in;
          *oData++ = res;
        }

        //  store
        in_1 = in1;
        in_2 = in2;
        out_1 = out1;
        out_2 = out2;
      }
};


#endif  //  lowpassfilter1D_h
