

#include "mycamera.h"
#include "mytweakables.h"
#include "misc.h"

#include <assert.h>


static double specialLookAtTime;
static Vector3 specialLookAt;
static Vector3 specialLookFrom;

//fixme: special looks should be per camera (so that I can use 
//  a camera for setting up shadow mapping, for example).
void CameraSpecialLook( double time, Vector3 const & at, Vector3 const & from )
{
  specialLookAtTime = time;
  specialLookAt = at;
  specialLookFrom = from;
}

namespace {
  struct CameraVisParams {
    dGeomID geom;
    MySystem * focus;
  };
  extern "C" {
    void CameraVisibilityCallback( void * d, dGeomID g1, dGeomID g2 ) {
      assert( dGeomGetClass( g2 ) == dRayClass );
      MySystem * ms = (MySystem *)dGeomGetData( g1 );
      if( ms == ((CameraVisParams *)d)->focus ) {
        return;
      }
      if( dGeomIsSpace( g1 ) ) {
        return dSpaceCollide2( g1, g2, d, CameraVisibilityCallback );
      }
      dContactGeom cg;
      int nc = dCollide( g1, g2, 1, &cg, sizeof( cg ) );
      if( nc ) {
        ((CameraVisParams *)d)->geom = g1;
      }
    }
  }
}

MyCamera::MyCamera()
{
  dBodyID tempBody = dBodyCreate( world() );
  dBodySetPosition( tempBody, -20, 2, 0 );
  dBodySetGravityMode( tempBody, 0 );
  dBodySetAutoDisableFlag( tempBody, 0 );
  dMass m;
  dMassSetSphereTotal( &m, 0.2f, 1 );
  dBodySetMass( tempBody, &m );
  addBody( tempBody, 0, 0 );
  dGeomID g = dCreateSphere( worldSpace(), 1 );
  addGeom( g, tempBody );
  lastVel = Vector3::Null;
  velChange = 0;
  lastCut = 0;
  lastCutTime = 0;
  desiredShot_ = ChaseCam;
  focus_ = 0;
  preventCutTime_ = 0;
  aim_ = 0;
}

MyCamera::CameraShot MyCamera::desiredShot() const
{
  if( specialLookAtTime > 0 ) {
    return SpecialLook;
  }
  return desiredShot_;
}

void MyCamera::setDesiredShot( MyCamera::CameraShot shot )
{
  assert( shot != SpecialLook );
  desiredShot_ = shot;
  if( desiredShot_ == FirstPersonIsh ) {
    desiredFirstPersonCamera( &firstPersonPos_, &firstPersonLookDir_, &firstPersonUp_ );
  }
  lastCutTime = curTime();
}

MySystem * MyCamera::focus() const
{
  return focus_;
}

void MyCamera::setFocus( MySystem * focus )
{
  focus_ = focus;
}

void MyCamera::setAim( float aim, float elevation )
{
  aim_ = aim;
  elevation_ = elevation;
}

void MyCamera::parameters( Vector3 * lookFrom, Vector3 * lookAt, Vector3 * up ) const {
  if( specialLookAtTime > 0 ) {
    *lookAt = specialLookAt;
    *lookFrom = specialLookFrom;
    *up = Vector3::Y;
    return;
  }
  switch( desiredShot_ ) {
  default:
    assert( !"bad camera data" );
  case ChaseCam:
    {
      if( focus_ ) {
        Vector3 left, cup, front;
        focus_->basis( &left, &cup, &front );
        *lookAt = focus_->P() + front + cup;
      }
      else {
        *lookAt = P() + Vector3( 0, 0, -1 );
      }
      *up = Vector3::Y;
      *lookFrom = P(); //rotateLookFromAround( P(), aim_, *lookAt, *up );
    }
    break;
  case SideCam:
    if( focus_ ) {
      *lookAt = focus_->P();
    }
    else {
      *lookAt = P() + Vector3( 1, 0, 0 );
    }
    *lookFrom = P();
    *up = Vector3::Y;
    break;
  case FirstPersonIsh:
    *up = firstPersonUp_;
    *lookFrom = firstPersonPos_;
    {
      Vector3 right = firstPersonLookDir_ * firstPersonUp_;
      right *= 1.0f / right.lengthSquaredNotZero();
      *lookAt = firstPersonPos_ + 
          rotateLookFromAround( 
              rotateLookFromAround( firstPersonLookDir_, elevation_, Vector3::Null, right ),
              aim_,
              Vector3::Null,
              *up );
    }
    break;
  }
}

bool MyCamera::physical()
{
  return false;
}

bool MyCamera::wantsContact( dContact & c, MySystemBase * other, dGeomID you, dGeomID him, bool firstTest )
{
  if( firstTest && (!other || other->wantsContact( c, this, him, you, false )) ) {
    // camera
    c.surface.mu = c.surface.mu2 = 0;
    c.surface.mode = dContactFDir1;
    dJointID tempJoint = dJointCreateContact( world(), tempJointGroup(), &c );
    dJointAttach( tempJoint, bodies_[0].body, 0 );
  }
  return false;
}

void MyCamera::preStep() {
  //  camera
  dReal const * cR = focus_ ? focus_->R() : &Matrix3::Identity.xx;
  dReal const * cP = focus_ ? &(focus_->P().x) : &Vector3::Null.x;
  Vector3 front( cR[2], 0, cR[10] );    //  project to XZ plane
  front *= 1.0f / sqrtf( front.lengthSquaredNotZero() );
  Vector3 left( cR[0], 0, cR[8] );      //  project to XZ plane
  left *= 1.0f / sqrtf( left.lengthSquaredNotZero() );
  Vector3 pos( cP[0], cP[1], cP[2] );
  Vector3 desired;
  switch( desiredShot_ ) {
  default:
    assert( !"bad camera state" );
  case FirstPersonIsh:  //  first person doesn't really care about the other stuff, 
                        //  as the camera is placed relative to target
    {
      Vector3 pos, lookAt, up;
      desiredFirstPersonCamera( &pos, &lookAt, &up );
      firstPersonPos_ *= 0.5f; firstPersonPos_ += (pos * 0.5f);
      firstPersonLookDir_ *= 0.9f; firstPersonLookDir_ += (lookAt * 0.1f);
      firstPersonUp_ *= 0.5f; firstPersonUp_ += (up * 0.5f);
      firstPersonUp_ *= 1.0f / sqrtf( firstPersonUp_.lengthSquaredNotZero() );
    }
    //  fall through to chase cam
  case ChaseCam:
      //  hacky that the camera relies on SINK_AMT!
    desired = rotateLookFromAround(
        rotateLookFromAround(
            pos - (front * CAMERA_BACK) + (Vector3::Y * (CAMERA_UP + SINK_AMT)),
            -elevation_,
            pos,
            left ),
        aim_,
        pos,
        Vector3::Y );
    break;
  case SideCam:
    desired = pos + left * CAMERA_BACK * 2;
    break;
  }
  Vector3 vel = L();
  Vector3 actual = P();
  actual += vel * CAMERA_LOOKAHEAD;
  desired -= actual;
  velChange += (vel - lastVel).lengthSquared();
  velChange *= 0.9f;
  lastVel = vel;
  if( preventCutTime_ > 0 ) {
    preventCutTime_ -= TIME_STEP;
    goto no_cut;
  }
  if( ((desired.lengthSquared() > MAX_CAMERA_DISTANCE_SQUARED) || (velChange > MAX_CAMERA_VEL_CHANGE))
      && (curTime() - lastCutTime > CAMERA_CUT_INTERVAL) ) {
    info( "desired distance: %.1f  velChange: %.1f  -> cutting camera\n", desired.lengthSquared(), velChange );
    debug( "desired: %.1f,%.1f,%.1f  actual: %.1f,%.1f,%.1f  vel: %.1f,%.1f,%.1f\n",
        desired.x, desired.y, desired.z, actual.x, actual.y, actual.z, vel.x, vel.y, vel.z );
    cutCamera();
    lastCutTime = curTime();
  }
  else {
no_cut:
    desired *= CAMERA_DISTANCE_FORCE;
    dBodyAddForce( bodies_[0].body, desired.x, desired.y, desired.z );
  }
}

void MyCamera::cutCamera()
{
  if( preventCutTime_ > 0 ) {
    return;
  }
  if( desiredShot_ != FirstPersonIsh ) {
    desiredShot_ = ChaseCam;
  }
  //  cut to a location that can see the vehicle
  Matrix3 rot = ((Matrix3 *)R())->transpose();
  Vector3 deltaRelative[] = {
    Vector3( 0, 4, 5 ),
    Vector3( 2, 6, 3 ),
    Vector3( -2, 6, 3 ),
    Vector3( 3, 8, -4 ),
  };
  Vector3 deltaAbsolute[] = {
    Vector3( -5, 4, -6 ),
    Vector3( 5, 4, -6 ),
    Vector3( 6, 4, 5 ),
    Vector3( -6, 4, 5 ),
  };
  lastVel = Vector3::Null;
  velChange = 0;
  ClearForceAndTorque( bodies_[0].body );
  dBodySetRotation( bodies_[0].body, &Matrix3::Identity.xx );
  Vector3 pos = focus_ ? focus_->P() : Vector3::Null;
  for( size_t i = 0; i < 4; ++i ) {
    Vector3 dir = rot * deltaRelative[(i+lastCut)&3];
    float l = sqrtf( dir.lengthSquared() );
    dir *= (1.0f / l);
    dGeomRaySet( utilRay(), pos.x, pos.y, pos.z, dir.x, dir.y, dir.z );
    dGeomRaySetLength( utilRay(), l );
    CameraVisParams cvp;
    cvp.focus = focus_;
    cvp.geom = 0;
    dSpaceCollide2( (dGeomID)worldSpace(), utilRay(), &cvp, CameraVisibilityCallback );
    if( cvp.geom == 0 ) {
      dBodySetPosition( bodies_[0].body, pos.x+dir.x*l, pos.y+dir.y*l, pos.z+dir.z*l );
      debug( "using relative cut %d\n", (i + lastCut)&3 );
      debugVar( 45, 1, "relative %d ", (i+lastCut)&3 );
      lastCut = (1 + i + lastCut)&3;
      return;
    }
  }
  for( size_t i = 0; i < 4; ++i ) {
    Vector3 dir = deltaAbsolute[(i+lastCut)&3];
    float l = sqrtf( dir.lengthSquared() );
    dir *= (1.0f / l);
    dGeomRaySet( utilRay(), pos.x, pos.y, pos.z, dir.x, dir.y, dir.z );
    dGeomRaySetLength( utilRay(), l );
    CameraVisParams cvp;
    cvp.focus = focus_;
    cvp.geom = 0;
    dSpaceCollide2( (dGeomID)worldSpace(), utilRay(), &cvp, CameraVisibilityCallback );
    if( cvp.geom == 0 ) {
      dBodySetPosition( bodies_[0].body, pos.x+dir.x*l, pos.y+dir.y*l, pos.z+dir.z*l );
      debug( "using absolute cut %d\n", (i + lastCut)&3 );
      debugVar( 45, 1, "absolute %d ", (i+lastCut)&3 );
      lastCut = (1 + i + lastCut)&3;
      return;
    }
  }
  debug( "using last-resort cut\n" );
  debugVar( 45, 1, "last resort" );
  dBodySetPosition( bodies_[0].body, pos.x+4, pos.y+4, pos.z+4 );
}

void MyCamera::postStep()
{
  //  MyCamera
  MySystem::postStep();
  each_body( DoDampen( -0.08f, -0.1f ) );
}

void MyCamera::desiredFirstPersonCamera( Vector3 * pos, Vector3 * look, Vector3 * u ) const
{
  if( focus_ ) {
    Vector3 left, up, front;
    Vector3 p = focus_->P();
    focus_->basis( &left, &up, &front );
    p += up * 1.15f;
    p += front * -0.4f;
    *pos = p;
    *look = front;
    *u = up;
  }
  else {
    *pos = P();
    *look = Vector3( 0, 0, -1 );
    *u = Vector3( 0, 1, 0 );
  }
}

void MyCamera::step( float deltaTime )
{
  if( specialLookAtTime > 0 ) {
    specialLookAtTime -= deltaTime;
  }
}

Vector3 MyCamera::rotateLookFromAround( Vector3 const & from, float amt, Vector3 const & point, Vector3 const & up ) const
{
  if( amt == 0 ) return from;
  Matrix4 r;
  RotationMatrixAroundPoint( &r, up, amt, point );
  return from * r;
}
