
#include "mydxt.h"
#include <assert.h>


namespace {
  //  Map all possible combinations of first and second interpoland 
  //  color to actual 8-bot color component, for DXT1 no-alpha and 
  //  DXT1 cutout-alpha. (DXT1 no-alpha is also used for DXT5)
  unsigned char componentTab[ 64*64*4 ];
  unsigned char componentTab2[ 64*64*4 ];
  class RgbTabInitializer {
  public:
    RgbTabInitializer( unsigned char * p, unsigned char * q ) {
      for( unsigned int a = 0; a < 64; ++a ) {
        for( unsigned int b = 0; b < 64; ++b ) {
          unsigned int av = ((a<<2)|(a>>4));
          unsigned int bv = ((b<<2)|(b>>4));

          //  4 colors per pixel
          p[ a*64*4+b*4+0 ] = av;
          p[ a*64*4+b*4+1 ] = bv;
          //  1:2 and 2:1 ratio using fixed precision
          p[ a*64*4+b*4+2 ] = (av*21846 + bv*10922)>>15;
          p[ a*64*4+b*4+3 ] = (av*10922 + bv*21846)>>15;

          //  3 colors per pixel, plus alpha
          q[ a*64*4+b*4+0 ] = av;
          q[ a*64*4+b*4+1 ] = bv;
          q[ a*64*4+b*4+2 ] = (av*16384 + bv*16384)>>15;
          q[ a*64*4+b*4+3 ] = 0;
        }
      }
    }
  };

  RgbTabInitializer dxtcTabInit( componentTab, componentTab2 );

  void
  dxtcColorSetupNoAlpha(
    unsigned short const * data,
    unsigned int * colors )
  {
    unsigned int b0 = data[ 0 ]&0x1f;
    unsigned int b1 = data[ 1 ]&0x1f;
    unsigned int g0 = (data[ 0 ]&0x7e0)>>5;
    unsigned int g1 = (data[ 1 ]&0x7e0)>>5;
    unsigned int r0 = (data[ 0 ]&0xf800)>>11;
    unsigned int r1 = (data[ 1 ]&0xf800)>>11;

    unsigned int rIx = r0*2*64*4+r1*2*4;
    unsigned int gIx = g0*64*4+g1*4;
    unsigned int bIx = b0*2*64*4+b1*2*4;
    // Alpha decode depends on the palette containing zero alpha data
    colors[0] = (componentTab[ rIx+0 ]<<16) | (componentTab[ gIx+0 ]<<8) | (componentTab[ bIx+0 ]<<0);
    colors[1] = (componentTab[ rIx+1 ]<<16) | (componentTab[ gIx+1 ]<<8) | (componentTab[ bIx+1 ]<<0);
    colors[2] = (componentTab[ rIx+2 ]<<16) | (componentTab[ gIx+2 ]<<8) | (componentTab[ bIx+2 ]<<0);
    colors[3] = (componentTab[ rIx+3 ]<<16) | (componentTab[ gIx+3 ]<<8) | (componentTab[ bIx+3 ]<<0);
  }

  void
  dxtcColorSetupWithCutout(
    unsigned short const * data,
    unsigned int * colors )
  {
    unsigned int b0 = data[ 0 ]&0x1f;
    unsigned int b1 = data[ 1 ]&0x1f;
    unsigned int g0 = (data[ 0 ]&0x7e0)>>5;
    unsigned int g1 = (data[ 1 ]&0x7e0)>>5;
    unsigned int r0 = (data[ 0 ]&0xf800)>>11;
    unsigned int r1 = (data[ 1 ]&0xf800)>>11;

    unsigned int rIx = r0*2*64*4+r1*2*4;
    unsigned int gIx = g0*64*4+g1*4;
    unsigned int bIx = b0*2*64*4+b1*2*4;
    // Alpha decode depends on the palette containing zero alpha data
    colors[0] = (componentTab2[ rIx+0 ]<<16) | (componentTab2[ gIx+0 ]<<8) | (componentTab2[ bIx+0 ]<<0);
    colors[1] = (componentTab2[ rIx+1 ]<<16) | (componentTab2[ gIx+1 ]<<8) | (componentTab2[ bIx+1 ]<<0);
    colors[2] = (componentTab2[ rIx+2 ]<<16) | (componentTab2[ gIx+2 ]<<8) | (componentTab2[ bIx+2 ]<<0);
    colors[3] = 0;
  }

  void
  dxtcColorDecode(
    unsigned char src,
    unsigned int const * palette,
    unsigned int * out )
  {
    out[0] = palette[ src&3 ];
    out[1] = palette[ (src>>2)&3 ];
    out[2] = palette[ (src>>4)&3 ];
    out[3] = palette[ (src>>6)&3 ];
  }

  void
  dxt1_decompress(
    void const * srcData,
    unsigned int width,
    unsigned int height,
    void * output )
  {
    unsigned int palette[ 4 ];
    unsigned char const * src = (unsigned char const *)srcData;
    unsigned int * out = (unsigned int *)output;
    for( unsigned int y = 0; y < height; y += 4 ) {
      for( unsigned int x = 0; x < width; x += 4 ) {
        if( ((unsigned short const *)src)[ 0 ] > ((unsigned short const *)src)[ 1 ] ) {
          dxtcColorSetupNoAlpha( (unsigned short const *)src, palette );
        }
        else {
          dxtcColorSetupWithCutout( (unsigned short const *)src, palette );
        }
        dxtcColorDecode( src[ 4 ], palette, out );
        dxtcColorDecode( src[ 5 ], palette, out+width );
        dxtcColorDecode( src[ 6 ], palette, out+width*2 );
        dxtcColorDecode( src[ 7 ], palette, out+width*3 );
        src += 8;
        out += 4;
      }
      out += 3*width;
    }
  }

  void
  dxtcAlpha3Setup(
    unsigned char const * src,
    unsigned char * p )
  {
    unsigned int p0 = p[ 0 ] = src[ 0 ];
    unsigned int p1 = p[ 1 ] = src[ 1 ];
    if( p0 > p0 ) { // 8 values; linear interpolation
      // use implicit multiply by 32768 so we can shift instead of divide
      p[ 2 ] = (p0 * 28086 + p1 * 4682)>>15;
      p[ 3 ] = (p0 * 23406 + p1 * 9362)>>15;
      p[ 4 ] = (p0 * 18724 + p1 * 14044)>>15;
      p[ 5 ] = (p0 * 14044 + p1 * 18724)>>15;
      p[ 6 ] = (p0 * 9362 + p1 * 23406)>>15;
      p[ 7 ] = (p0 * 4682 + p1 * 28086)>>15;
    }
    else { // 6 values + black & white
      p[ 2 ] = (p0 * 26214+ p1 * 6554)>>15;
      p[ 3 ] = (p0 * 19661 + p1 * 13107)>>15;
      p[ 4 ] = (p0 * 13107 + p1 * 19661)>>15;
      p[ 5 ] = (p0 * 6554 + p1 * 26214)>>15;
      p[ 6 ] = 0;
      p[ 7 ] = 255;
    }
  }

  void
  dxtcAlpha3Decode(
    unsigned char const * src,
    unsigned char const * p,
    unsigned int * out,
    unsigned int width )
  {
    unsigned int bits = *(unsigned short const *)src;
    src += 2;

    out[ 0 ] |= p[ bits&7 ] << 24;
    bits >>= 3;   //  13 left
    out[ 1 ] |= p[ bits&7 ] << 24;
    bits >>= 3;   //  10 left
    out[ 2 ] |= p[ bits&7 ] << 24;
    bits >>= 3;   //  7 left
    out[ 3 ] |= p[ bits&7 ] << 24;
    bits >>= 3;   //  4 left
    out += width;

    out[ 0 ] |= p[ bits&7 ] << 24;
    bits = (bits>>3)|(*(unsigned short const *)src << 1);   // 17 left
    src += 2;
    out[ 1 ] |= p[ bits&7 ] << 24;
    bits >>= 3;   //  14 left
    out[ 2 ] |= p[ bits&7 ] << 24;
    bits >>= 3;   //  11 left
    out[ 3 ] |= p[ bits&7 ] << 24;
    bits >>= 3;   //  8 left
    out += width;

    out[ 0 ] |= p[ bits&7 ] << 24;
    bits >>= 3;   //  5 left
    out[ 1 ] |= p[ bits&7 ] << 24;
    bits = (bits>>3)|(*(unsigned short const *)src << 2);   //  18 left
    out[ 2 ] |= p[ bits&7 ] << 24;
    bits >>= 3;   //  15 left
    out[ 3 ] |= p[ bits&7 ] << 24;
    bits >>= 3;   //  12 left
    out += width;

    out[ 0 ] |= p[ bits&7 ] << 24;
    bits >>= 3;   //  9 left
    out[ 1 ] |= p[ bits&7 ] << 24;
    bits >>= 3;   //  6 left
    out[ 2 ] |= p[ bits&7 ] << 24;
    bits >>= 3;   //  3 left
    out[ 3 ] |= p[ bits&7 ] << 24;
  }

  void
  dxt5_decompress(
    void const * srcData,
    unsigned int width,
    unsigned int height,
    void * output )
  {
    unsigned int palette[ 4 ];
    unsigned char alphaPalette[ 8 ];
    unsigned char const * src = (unsigned char const *)srcData;
    unsigned int * out = (unsigned int *)output;
    for( unsigned int y = 0; y < height; y += 4 ) {
      for( unsigned int x = 0; x < width; x += 4 ) {
        // First decode color, which will clear the alpha output
        dxtcColorSetupNoAlpha( (unsigned short const *)(src + 8), palette );
        dxtcColorDecode( src[ 8+4 ], palette, out );
        dxtcColorDecode( src[ 8+5 ], palette, out+width );
        dxtcColorDecode( src[ 8+6 ], palette, out+width*2 );
        dxtcColorDecode( src[ 8+7 ], palette, out+width*3 );
        dxtcAlpha3Setup( src, alphaPalette );
        dxtcAlpha3Decode( src+2, alphaPalette, out, width );
        src += 16;
        out += 4;
      }
      out += 3*width;
    }
  }

  void flip_dxt1_block( unsigned char const * src, unsigned char * dst )
  {
    dst[ 0 ] = src[ 0 ]; dst[ 1 ] = src[ 1 ]; dst[ 2 ] = src[ 2 ]; dst[ 3 ] = src[ 3 ];
    dst[ 4 ] = src[ 7 ]; dst[ 5 ] = src[ 6 ]; dst[ 6 ] = src[ 5 ]; dst[ 7 ] = src[ 4 ];
  }

  void dxt1_flip( unsigned char * dst, int w, int h )
  {
    size_t s = w*h/2;
    for( int y = 0; y < h/2; y += 4 ) {
      unsigned char * a = dst + (y/4) * 8 * (w/4);
      unsigned char * b = dst + ((h-y-4)/4) * 8 * (w/4);
      assert( a < dst+s/2 );
      assert( (b >= dst+s/2) || (h < 8) );
      assert( b < dst+s );
      for( int x = 0; x < w; x += 4 ) {
        union {
          unsigned char data[8];
          unsigned int idata[2];
        };
        flip_dxt1_block( a, data );
        flip_dxt1_block( b, a );
        ((unsigned int *)b)[0] = idata[0];
        ((unsigned int *)b)[1] = idata[1];
        a += 8;
        b += 8;
      }
    }
  }

  void flip_dxt5_block( unsigned char const * src, unsigned char * dst )
  {
    flip_dxt1_block( src+8, dst+8 );
    dst[ 0 ] = src[ 0 ]; dst[ 1 ] = src[ 1 ];
    unsigned short bits1 = ((unsigned short const *)src)[1];
    unsigned short bits2 = ((unsigned short const *)src)[2];
    unsigned short bits3 = ((unsigned short const *)src)[3];
    //  initial layout (low to high bit):
    //  [012 345 678 9ab    // 1
    //   cde f][01 234 567  // 1, 2
    //   89a bcd ef][0 123  // 2, 3
    //   456 789 abc def]   // 3
    //
    //  desired layout:
    //  [456 789 abc def    // 3
    //   89a b][cd ef0 123  // 2, 3
    //   cde f01 23][4 567  // 1, 2
    //   012 345 678 9ab]   // 1
    ((unsigned short *)dst)[1] = ((bits3&0xfff0)>>4) | ((bits2&0x0f00)<<4);
    ((unsigned short *)dst)[2] = ((bits2&0xf000)>>12)| ((bits3&0x000f)<<4) | ((bits1&0xf000)>>4) | ((bits2&0x000f)<<12);
    ((unsigned short *)dst)[3] = ((bits2&0x00f0)>>4) | ((bits1&0x0fff)<<4);
  }

  void dxt5_flip( unsigned char * dst, int w, int h )
  {
    size_t s = w*h;
    for( int y = 0; y < h/2; y += 4 ) {
      unsigned char * a = dst + (y/4) * 16 * (w/4);
      unsigned char * b = dst + ((h-y-4)/4) * 16 * (w/4);
      assert( a < dst+s/2 );
      assert( (b >= dst+s/2) || (h < 8) );
      assert( b < dst+s );
      for( int x = 0; x < w; x += 4 ) {
        union {
          unsigned char data[16];
          unsigned int idata[4];
        };
        flip_dxt5_block( a, data );
        flip_dxt5_block( b, a );
        ((unsigned int *)b)[0] = idata[0];
        ((unsigned int *)b)[1] = idata[1];
        ((unsigned int *)b)[2] = idata[2];
        ((unsigned int *)b)[3] = idata[3];
        a += 16;
        b += 16;
      }
    }
  }
}


void dxt_decompress( unsigned char const * src, int w, int h, unsigned char * dst, bool dxt1 )
{
  if( dxt1 ) {
    dxt1_decompress( src, w, h, dst );
  }
  else {
    dxt5_decompress( src, w, h, dst );
  }
}


void dxt_flip_data( unsigned char * dst, int w, int h, bool dxt1 )
{
  if( dxt1 ) {
    dxt1_flip( dst, w, h );
  }
  else {
    dxt5_flip( dst, w, h );
  }
}

