
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <new>
#include <assert.h>

#include "log.h"
#include "myfilesystem.h"



Data::Data( void * ptr, size_t size )
{
  ptr_ = ptr;
  size_ = size;
  pos_ = 0;
  ownsPtr_ = false;
  canWrite_ = true;
}

Data::Data( void const * ptr, size_t size )
{
  //  this is safe, because we set canWrite_ to false!
  ptr_ = const_cast< void * >( ptr );
  size_ = size;
  pos_ = 0;
  ownsPtr_ = false;
  canWrite_ = false;
}

Data::Data( void const * ptr, size_t size, bool copy )
{
  if( !copy ) {
    new( this ) Data( ptr, size );
    return;
  }
  new( this ) Data( size );
  memcpy( ptr_, ptr, size );
}

Data::Data( size_t size )
{
  ptr_ = ::calloc( size, 1 );
  size_ = ptr_ ? size : 0;
  pos_ = 0;
  ownsPtr_ = true;
  canWrite_ = ptr_ != 0;
}

Data::Data()
{
  ptr_ = 0;
  size_ = 0;
  pos_ = 0;
  ownsPtr_ = true;
  canWrite_ = false;
}

Data::~Data()
{
  if( ownsPtr_ ) {
    memset( ptr_, 0xfd, size_ );
    free( ptr_ );
  }
}

Data::Data( Data const & other )
{
  ptr_ = ::calloc( other.size_, 1 );
  size_ = ptr_ ? other.size_ : 0;
  pos_ = 0;
  ownsPtr_ = true;
  canWrite_ = true;
  ::memcpy( ptr_, other.ptr_, size_ );
}

Data & Data::operator=( Data const & other )
{
  this->~Data();
  new( this ) Data( other );
  return *this;
}

size_t Data::size() const
{
  return size_;
}

char * Data::writeBase()
{
  assert( canWrite_ );
  if( !canWrite_ ) return 0;
  return reinterpret_cast< char * >( ptr_ );
}

char const * Data::readBase() const
{
  return reinterpret_cast< const char * >( ptr_ );
}

size_t Data::pos() const
{
  return pos_;
}

bool Data::setPos( size_t pos )
{
  assert( pos <= size_ );
  if( pos > size_ ) {
    return false;
  }
  else {
    pos_ = pos;
  }
  return true;
}

size_t Data::read( void * buf, size_t amount )
{
  size_t toRead = size_ - pos_;
  if( toRead > amount ) {
    toRead = amount;
  }
  memcpy( buf, ((char *)ptr_)+pos_, toRead );
  pos_ = pos_ + toRead;
  return toRead;
}

size_t Data::write( void const * buf, size_t amount )
{
  assert( canWrite_ );
  if( !canWrite_ ) {
    return 0;
  }
  size_t toWrite = size_-pos_;
  if( toWrite > amount ) {
    toWrite = amount;
  }
  memcpy( ptr_, buf, toWrite );
  pos_ = pos_ + toWrite;
  return toWrite;
}

bool Data::trim( size_t newSmallerSize )
{
  if( newSmallerSize > size_ ) {
    return false;
  }
  size_ = newSmallerSize;
  return true;
}

bool Data::ownsPtr() const
{
  return ownsPtr_;
}

bool Data::canWrite() const
{
  return canWrite_;
}

void * Data::detach()
{
  assert( ownsPtr_ );
  if( !ownsPtr_ ) {
    return 0;
  }
  void * ret = ptr_;
  pos_ = 0;
  size_ = 0;
  ptr_ = 0;
  canWrite_ = true;
  return ret;
}

Data & Data::allocate( size_t size )
{
  this->~Data();
  new( this ) Data( size );
  return *this;
}

Data & Data::attach( void * ptr, size_t size )
{
  this->~Data();
  new( this ) Data( ptr, size );
  return *this;
}

Data const & Data::attach( void const * ptr, size_t size )
{
  this->~Data();
  new( this ) Data( ptr, size );
  return *this;
}


#include <map>
#include <string>
#include <algorithm>
#include <cctype>


namespace {
  std::map< std::string, std::string > gHandles;
  std::string & lowercase( std::string & s ) {
    std::transform( s.begin(), s.end(), s.begin(), (int(*)(int))tolower );
    return s;
  }
  bool gDumpedFileMap = false;
}

using namespace FS;

namespace FS {

//  Handles is an XML document which can map virtual file names 
//  to physical paths. All handles loaded are merged, until such 
//  time as you call clearHandles(). Everything is lowercased 
//  internally (handles and paths).
void clearHandles()
{
  gHandles.clear();
  gDumpedFileMap = false;
}

bool addHandles( char const * path )
{
  gDumpedFileMap = false;
  Data handleXml;
  if( !readFileByPath( path, &handleXml ) ) {
    warning( "Could not addHandles() from '%s'.\n", path );
    return false;
  }
  return addHandles( handleXml );
}

bool addHandles( Data const & iData )
{
  XMLDom dom( iData.readBase(), iData.size() );
  if( dom.error() ) {
    error( "XML parse error in data: %s\n", dom.error() );
    return false;
  }
  XMLNode * root = dom.root();
  bool b = false;
  while( root ) {
    if( (root->token().kind == XMLScanner::Token::Tag) && !strnicmp( root->token().key, "filemap", root->token().keySize ) ) {
      if( addHandles( root ) ) {
        b = true;
      }
    }
    root = root->next();
  }
  return b;
}

bool addHandles( XMLNode * filemap )
{
  XMLNode * cld = filemap->child( 0 );
  char const * curTag = 0;
  size_t curTagSize = 0;
  while( cld ) {
    if( cld->token().kind == XMLScanner::Token::Tag ) {
      curTag = cld->token().key;
      curTagSize = cld->token().keySize;
      XMLNode * txt = cld->text( 0 );
      while( txt ) {
        if( txt->token().kind == XMLScanner::Token::CDATA ) {
          std::string key(curTag, curTagSize);
          std::string value( txt->token().value, txt->token().valueSize );
          std::transform( key.begin(), key.end(), key.begin(), tolower );
          if( gHandles.find( key ) != gHandles.end() ) {
            warning( "File map handle '%s' is re-defined, from '%s' to '%s'.\n",
                key.c_str(), gHandles[key].c_str(), value.c_str() );
          }
          gHandles[lowercase(key)] = value;
          curTag = 0;
          break;
        }
        txt = txt->next();
      }
      if( curTag != 0 ) {
        warning( "Filemap contains XML tag '%*s' which has no data.\n", curTag, curTagSize );
      }
    }
    cld = cld->next();
  }
  return true;
}

bool translateHandle( char const * handle, char * path, size_t outSize )
{
  std::string key( lowercase( std::string( handle ) ) );
  std::map< std::string, std::string >::iterator found;
  if( (found = gHandles.find( key )) == gHandles.end() ) {
    error( "Could not translate handle '%s'.\n", handle );
if( !gDumpedFileMap ) {
  for( found = gHandles.begin(); found != gHandles.end(); ++found ) {
    debug( "%s:%s\n", (*found).first.c_str(), (*found).second.c_str() );
  }
  gDumpedFileMap = true;
}
    return false;
  }
  size_t len = (*found).second.length();
  if( len > outSize-1 ) {
    warning( "Buffer of %ld bytes is too small for %ld bytes path '%s' (key '%s').\n", 
        outSize, len, (*found).second.c_str(), key.c_str() );
    return false;
  }
  assert( len < outSize );    //  thus strncpy() WILL terminate the string!
  strncpy( path, (*found).second.c_str(), outSize );
  return true;
}

bool readFileByHandle( char const * token, Data * oData )
{
  char path[ 256 ];
  if( !translateHandle( token, path, 256 ) ) {
    return false;
  }
  return readFileByPath( path, oData );
}

bool readFileByPath( char const * file, Data * oData )
{
  std::string filePath( file );
  FILE * f = fopen( lowercase( filePath ).c_str(), "rb" );
  if( !f ) {
    warning( "File '%s' not found.\n", filePath.c_str() );
    return false;
  }
  fseek( f, 0, 2 );
  fpos_t size = ftell( f );
  if( size > 0x10000000UL ) {
    warning( "File '%s' is bigger than %ld bytes!\n", 0x10000000UL );
    fclose( f );
    return false;
  }
  size_t bytes = static_cast< size_t >( size );
  fseek( f, 0, 0 );
  oData->allocate( bytes );
  if( !oData->writeBase() ) {
    error( "Could not allocate data for %ld bytes for '%s'.\n", bytes, filePath.c_str() );
    fclose( f );
    return false;
  }
  fread( oData->writeBase(), 1, bytes, f );
  fclose( f );
  debug( "File '%s' is %ld bytes.\n", filePath.c_str(), bytes );
  return true;
}

bool writeFileByPath( char const * file, Data const & iData )
{
  std::string bakPath;
  if( strlen( file ) > 250 ) {
    error( "Output path is too long: '%s'\n", file );
    return false;
  }
  bakPath = lowercase( std::string( file ) + std::string( ".bak" ) );
  unlink( bakPath.c_str() );
  rename( file, bakPath.c_str() );
  std::string filePath( file );
  FILE * f = fopen( lowercase( filePath ).c_str(), "wb" );
  if( !f ) {
    error( "Could not create: '%s'\n", filePath.c_str() );
    return false;
  }
  fwrite( iData.readBase(), 1, iData.size(), f );
  fclose( f );
  info( "Wrote '%s' with backup in '%s'.\n", filePath.c_str(), bakPath.c_str() );
  return true;
}

}
