
#include "mygrenade.h"
#include "mytweakables.h"
#include "myparticles.h"
#include "myweapon.h"
#include "glshaderstate.h"
#include "misc.h"
#include "gl-ase.h"

#include <assert.h>



//todo: should this be a set<> for efficient remove?
std::vector< MySystem * > MyGrenade::grenades_;


namespace {

  class GrenadeGun : public I_Weapon {
    public:
      GrenadeGun();
      ~GrenadeGun();

      void renderSight( float screenWidth, float screenHeight, Matrix4 const & camMtx, Matrix4 const & projMtx );
      void setAim( Matrix3 const & aim, Vector3 const & source, Vector3 const & ptVel );
      bool readyToFire( double );
      bool fire( double );
      void renderIndicators( float phi, float alpha );

      GLShader * reticleShader_;
      GLShader * indicatorShader_;
      GLShader * ammoShader_;
      I_GL_ASE * ammoModel_;
      Matrix3 ori_;
      Vector3 pos_;
      Vector3 vel_;
      double lastFired_;
  };
}

I_Weapon * NewGrenadeWeapon()
{
  return new GrenadeGun();
}


GrenadeGun::GrenadeGun()
{
  reticleShader_ = gShaderFactory->newShader( "grenadeReticle" );
  indicatorShader_ = gShaderFactory->newShader( "ammoIndicator" );
  ammoShader_ = gShaderFactory->newShader( "grenadeShader" );
  ammoModel_ = ASEFromFile( asePath( "grenadeAmmo" ) );
  ori_ = Matrix3::Identity;
  pos_ = Vector3::Null;
  lastFired_ = 0;
}

GrenadeGun::~GrenadeGun()
{
  reticleShader_->dispose();
  indicatorShader_->dispose();
  ammoShader_->dispose();
  ammoModel_->dispose();
}

void GrenadeGun::renderSight( float screenWidth, float screenHeight, Matrix4 const & camMtx, Matrix4 const & projMtx )
{
  int w = 32;
  if( screenWidth >= 1000 && screenHeight >= 750 ) {
    w = 32;
  }
  else {
    w = 16;
  }

  //  figure out where we'd be 20 meters out from the gun
  Vector4 pos = Vector4(pos_ + ori_.front() * 20) * camMtx * projMtx;
  pos *= 1.0f / pos.w;
  float midX = (pos.x + 1) * screenWidth/2;
  float midY = (pos.y + 1) * screenHeight/2;

  reticleShader_->apply();
  glBegin( GL_QUADS );
  glTexCoord2f( 0, 0 ); glVertex2f( midX-w*2, midY-w );
  glTexCoord2f( 1, 0 ); glVertex2f( midX+w*2, midY-w );
  glTexCoord2f( 1, 1 ); glVertex2f( midX+w*2, midY+w );
  glTexCoord2f( 0, 1 ); glVertex2f( midX-w*2, midY+w );
  glEnd();
}

void GrenadeGun::setAim( Matrix3 const & aim, Vector3 const & source, Vector3 const & vel )
{
  ori_ = aim;
  pos_ = source;
  vel_ = vel;
}

bool GrenadeGun::readyToFire( double time )
{
  return time > (lastFired_ + GRENADE_REFIRE_TIME);
}

bool GrenadeGun::fire( double time )
{
  if( !readyToFire( time ) ) {
    return false;
  }
  lastFired_ = time;
  MyGrenade * g = new MyGrenade( pos_, ori_, vel_, ammoShader_, ammoModel_ );
  return true;
}

void GrenadeGun::renderIndicators( float phi, float alpha )
{
  assert( alpha >= 0 );
  assert( phi > -3.1416f && phi < 6.2832f );
  glRotatef( 120 * phi, 0, 1, 0 );
  glTranslatef( 0, 0, -0.2f );
  indicatorShader_->context().color.alpha_ = (alpha > 1) ? 0x80 : (uchar)(0x80 * alpha);
  indicatorShader_->apply();
  ammoModel_->issueAll();
}


extern "C" {
  static void GrenadeRayCallback( void * gr, dGeomID g1, dGeomID g2 )
  {
    assert( dGeomGetClass( g2 ) == dRayClass );
    if( dGeomIsSpace( g1 ) ) {
      dSpaceCollide2( g1, g2, gr, GrenadeRayCallback );
    }
    else {
      dContactGeom cg;
      MyGrenade * theG = (MyGrenade *)gr;
      if( !theG->hasGeom( g1 ) && dCollide( g2, g1, 1, &cg, sizeof( cg ) ) ) {
        MySystem * s1 = (MySystem *)dGeomGetData( g2 );
        if( !s1 || s1->physical() ) {
          theG->exploding_ = true;
        }
      }
    }
  }
}


//  Todo: deal with recoil

MyGrenade::MyGrenade( Vector3 const & pos, Matrix3 const & ori, Vector3 const & ptVel, GLShader * shader, I_GL_ASE * mesh )
{
  explodeParams_.radius_ = GRENADE_EXPLODE_RADIUS;
  explodeParams_.force_ = GRENADE_EXPLODE_FORCE;
  exploding_ = false;
  exploding_ = false;
  dBodyID tempBody = dBodyCreate( world() );
  dMass m;
  dMassSetBoxTotal( &m, 1, GRENADE_RADIUS, GRENADE_RADIUS, GRENADE_RADIUS*2 );
  dBodySetMass( tempBody, &m );
  Matrix3 m3b = ori.transpose();
  dBodySetRotation( tempBody, &m3b.xx );
  dBodySetPosition( tempBody, pos.x, pos.y, pos.z );
  oldPos_ = pos;
  dBodySetLinearVel( tempBody, ptVel.x, ptVel.y, ptVel.z );
  Vector3 left, up, front;
  ori.leftUpFront( &left, &up, &front );
  Vector3 gForce = front * GRENADE_FORCE + up * GRENADE_UP_FORCE;
  dBodyAddForce( tempBody, gForce.x, gForce.y, gForce.z );
  addBody( tempBody, shader, mesh );
  dGeomID g = dCreateSphere( worldSpace(), GRENADE_RADIUS );
  addGeom( g, tempBody );
  grenades_.push_back( this );

  MyParticles * smoke = MakeParticleSystem( "smokepuff", worldSpace() );
  Vector3 basis[3] = { front, up, left };
  smoke->setEmitterPos( pos, basis );
}

MyGrenade::~MyGrenade()
{
  grenades_.erase( std::find( grenades_.begin(), grenades_.end(), this ) );
}

void MyGrenade::preStep()
{
  //  MyGrenade
  MySystem::preStep();
}

bool MyGrenade::wantsContact( dContact & c, MySystemBase * other, dGeomID you, dGeomID him, bool firstTest )
{
  if( !other || other->physical() ) {
    exploding_ = true;
    return false;
  }
  return true;
}

void MyGrenade::postStep()
{
  //  MyGrenade
  MySystem::postStep();
  Vector3 pos = P();
  Vector3 vel = pos - oldPos_;
  float l = sqrtf( vel.lengthSquared() );
  if( l > 1e-3 ) {
    float li = 1.0f / l;
    Vector3 norm( vel.x*li, vel.y*li, vel.z*li );
    dGeomRaySetLength( utilRay(), l );
    dGeomRaySetParams( utilRay(), 0, 1 );
    dGeomRaySet( utilRay(), pos.x, pos.y, pos.z, norm.x, norm.y, norm.z );
    dSpaceCollide2( utilRay(), (dGeomID)worldSpace(), this, &GrenadeRayCallback );
  }
  oldPos_ = pos;
  if( exploding_ ) {
    explodeParams_.pos_ = pos;
    explode();
  }
}

void MyGrenade::preStepAll()
{
  std::for_each( grenades_.begin(), grenades_.end(), MySystem::preStepOne );
}

void MyGrenade::postStepAll()
{
  std::for_each( grenades_.begin(), grenades_.end(), MySystem::postStepOne );
}

void MyGrenade::renderAll()
{
  std::for_each( grenades_.begin(), grenades_.end(), MySystem::renderOne );
}
