

#include "mygui.h"
#include "gl-text.h"
#include "glshaderstate.h"
#include "misc.h"


#include <assert.h>
#include <algorithm>




void GuiContained::hold()
{
  //  would you ever want to hold something more than 3 times?
  assert( refCount_ > 0 && refCount_ < 3 );
  ++refCount_;
}

void GuiContained::release()
{
  assert( refCount_ > 0 );
  --refCount_;
  if( !refCount_ ) {
    delete this;
  }
}

void GuiContained::addedToContainer( GuiRoot * top, GuiContainer * parent, WidgetSize const & inSize )
{
  assert( !top_ );
  assert( !parent_ );
  top_ = top;
  parent_ = parent;
  size_ = inSize;
}

void GuiContained::removedFromContainer( GuiRoot * top, GuiContainer * parent )
{
  assert( top == top_ );
  assert( parent == parent_ );
  top_ = 0;
  parent_ = 0;
}

void GuiContained::mouseEnter( unsigned int x, unsigned int y )
{
}

void GuiContained::mouseOver( unsigned int x, unsigned int y )
{
}

void GuiContained::mouseLeave()
{
}

GuiContained::GuiContained()
{
  refCount_ = 1;
  parent_ = 0;
  top_ = 0;
  size_.left = size_.top = size_.right = size_.bottom;
}

void GuiContained::releaseContained( GuiContained * c )
{
  c->release();
}





GuiWidget::GuiWidget()
{
}

void GuiWidget::invalidate()
{
  top_->invalidate();
}

bool GuiWidget::beginClick( unsigned int x, unsigned int y, unsigned int mb )
{
  return true;
}

void GuiWidget::endClick( bool over )
{
  //  OK, we're done!
  if( over ) {
    top_->setFocus( this );
  }
}

bool GuiWidget::setData( void const * data, size_t size )
{
  return false;
}

size_t GuiWidget::getData( void * data, size_t size )
{
  return 0;
}





GuiContainer::GuiContainer()
{
  lastOverTarget_ = 0;
}

GuiContainer::~GuiContainer()
{
  std::for_each( children_.begin(), children_.end(), GuiContained::releaseContained );
}

void GuiContainer::addChild( GuiContained * child, WidgetSize const & size, GuiContained * before )
{
  assert( child->parent() == 0 );
  child->hold();
  if( !before ) {
    children_.push_back( child );
  }
  else {
    children_.insert( std::find( children_.begin(), children_.end(), before ), child );
  }
  child->addedToContainer( top_, this, size );
}

void GuiContainer::removeChild( GuiContained * child )
{
  std::vector< GuiContained * >::iterator ptr = std::find( children_.begin(), children_.end(), child );
  assert( ptr != children_.end() );
  child->removedFromContainer( top_, this );
  children_.erase( ptr );
  childRemoved( child );
  child->release();
}

unsigned int GuiContainer::numChildren()
{
  return (unsigned int)children_.size();
}

GuiContained * GuiContainer::childAt( unsigned int ix )
{
  if( ix >= children_.size() ) {
    return 0;
  }
  return children_[ ix ];
}

namespace {
  class ContainedCoordMatch {
    public:
      unsigned int x_, y_;
      ContainedCoordMatch( unsigned int x, unsigned int y ) : x_( x ), y_( y ) {}
      bool operator()( GuiContained * c ) const {
        WidgetSize ws;
        c->getPos( &ws );
        return (ws.left <= x_ && ws.right > x_ && ws.top <= y_ && ws.bottom > y_);
      }
  };
}

GuiContained * GuiContainer::childAt( unsigned int xPos, unsigned int yPos )
{
  std::vector< GuiContained * >::iterator i = std::find_if( 
      children_.begin(), children_.end(), ContainedCoordMatch( xPos, yPos ) );
  if( i == children_.end() ) {
    return 0;
  }
  return (*i);
}

void GuiContainer::childRemoved( GuiContained * cld )
{
  if( parent_ ) {
    parent_->childRemoved( cld );
  }
}


namespace {
  class UpdateChild {
    public:
      UpdateChild() {}
      void operator()( GuiContained * gc ) {
        GuiWidget * gw = gc->widget();
        if( gw ) {
          gw->render();
        }
        GuiContainer * cc = gc->container();
        if( cc ) {
          cc->renderChildren();
        }
      }
  };
}

void GuiContainer::renderChildren()
{
  std::for_each( children_.begin(), children_.end(), UpdateChild() );
}

void GuiContainer::mouseEnter( unsigned int x, unsigned int y )
{
  mouseOver( x, y );
}

void GuiContainer::mouseOver( unsigned int x, unsigned int y )
{
  GuiContained * gc = childAt( x, y );
  if( gc != lastOverTarget_ ) {
    if( lastOverTarget_ ) {
      lastOverTarget_->mouseLeave();
    }
    lastOverTarget_ = gc;
    if( gc ) {
      gc->mouseEnter( x, y );
    }
  }
  else if( gc ) {
    gc->mouseOver( x, y );
  }
}

void GuiContainer::mouseLeave()
{
  if( lastOverTarget_ ) {
    lastOverTarget_->mouseLeave();
    lastOverTarget_ = 0;
  }
}



class C_Gui;


class C_GuiRenderer : public GuiRenderer {
  public:
    C_GuiRenderer( C_Gui * gui, GuiWidget * w, GlText * t ) {
      refCount_ = 1;
      gui_ = gui;
      widget_ = w;
      text_ = t;
    }
    void hold() {
      assert( refCount_ > 0 );
      ++refCount_;
    }
    void release() {
      assert( refCount_ > 0 );
      --refCount_;
      if( !refCount_ ) {
        delete this;
      }
    }
    void fillRect( WidgetSize const & rect, unsigned int color ) {
      text_->fill( color, rect.left, rect.top, rect.right-rect.left, rect.bottom-rect.top );
    }
    void frameRect( WidgetSize const & rect, unsigned int color ) {
      text_->drawLine( color, rect.left, rect.top, rect.left, rect.bottom );
      text_->drawLine( color, rect.left, rect.bottom, rect.right, rect.bottom );
      text_->drawLine( color, rect.right, rect.bottom, rect.right, rect.top );
      text_->drawLine( color, rect.right, rect.top, rect.left, rect.top );
    }
    void drawLine( unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2, unsigned int color ) {
      text_->drawLine( color, x1, y1, x2, y2 );
    }
    void drawText( char const * text, unsigned int x, unsigned int y, unsigned int color ) {
      text_->drawText( color, x, y, text );
    }
    void measureText( char const * text, unsigned int * width, unsigned int * height ) {
      int w = 0, h = 0;
      text_->measureText( text, &w, &h );
      *width = (uint)w;
      *height = (uint)h;
    }
    //  Input image is bottom-to-top orientation.
    //  Output image is ... bottom-to-top?
    void drawImage( Image const & i, unsigned int x, unsigned int y, bool alphaTest );
    void drawImage8A( Image const & i, unsigned int x, unsigned int y, unsigned int w, unsigned int h ) {
      assert( !"not implemented image format" );
    }
    void drawImage8( Image const & i, unsigned int x, unsigned int y, unsigned int w, unsigned int h ) {
      assert( !"not implemented image format" );
    }
    void drawImage24A( Image const & i, unsigned int x, unsigned int y, unsigned int w, unsigned int h ) {
      assert( !"not implemented image format" );
    }
    void drawImage24( Image const & i, unsigned int x, unsigned int y, unsigned int w, unsigned int h ) {
      assert( !"not implemented image format" );
    }
    void drawImage32A( Image const & i, unsigned int x, unsigned int y, unsigned int w, unsigned int h );
    void drawImage32( Image const & i, unsigned int x, unsigned int y, unsigned int w, unsigned int h );

    unsigned int refCount_;
    C_Gui * gui_;
    GuiWidget * widget_;
    GlText * text_;
};



class C_Gui : public GuiRoot, public GuiContainer {
  public:
    C_Gui( WidgetSize const & size );
    ~C_Gui();

    //  GuiContained
    GuiContainer * container() { return this; }

    //  GuiContainer
    void childRemoved( GuiContained * cld );

    void dispose() { unlock(); }

    void setFocus( GuiWidget * focus );
    GuiWidget * focus() { return focus_; }
    GuiRenderer * renderer( GuiWidget * forWidget );
    void invalidate();

    void render();
    void beginClick( unsigned int x, unsigned int y, unsigned int mb );
    void endClick( unsigned int x, unsigned int y );
    void mouseMove( unsigned int x, unsigned int y );
    //  add() will recurse, finding the right container (based on top-left)
    void add( GuiContained * contained, WidgetSize const & size );

    bool captureKey( unsigned int x, unsigned int y, unsigned char key );
    void setCallback( MyGuiCallback * callback );

    bool recurseClick( GuiContainer * c, unsigned int x, unsigned int y, unsigned int mb );
    void lock() {
      ++lockCount_;
    }
    bool unlock() {
      assert( lockCount_ > 0 );
      if( --lockCount_ == 0 ) {
        if( callback_ ) {
          callback_->guiClosing( this );
        }
        delete this;
        return true;
      }
      return false;
    }

    GuiWidget * focus_;
    GuiWidget * clicked_;
    GlText * text_;
    unsigned int width_;
    unsigned int height_;
    bool dirty_;
    Image image_;
    unsigned int lockCount_;
    MyGuiCallback * callback_;
    GLShader * shader_;
};

C_Gui::C_Gui( WidgetSize const & size )
{
  top_ = this;
  parent_ = 0;
  size_ = size;
  unsigned int w = 1;
  while( w < (size.right-size.left) ) {
    w <<= 1;
  }
  width_ = w;
  unsigned int h = 1;
  while( h < (size.bottom-size.top) ) {
    h <<= 1;
  }
  height_ = h;
  text_ = MakeGlText( w, h );
  focus_ = 0;
  clicked_ = 0;
  dirty_ = true;
  lockCount_ = 1;
  callback_ = 0;
  shader_ = gShaderFactory->newShader( "guiShader" );
  GLProgramInstance * pi = shader_->instance();
  shader_->setTexture( "diffuseMap", text_->texture() );
}

C_Gui::~C_Gui()
{
  text_->dispose();
}

void C_Gui::setFocus( GuiWidget * focus )
{
  lock();
  if( focus != focus_ ) {
    if( focus_ ) {
      focus_->invalidate();
    }
    focus_ = focus;
    if( focus ) {
      focus->invalidate();
    }
  }
  unlock();
}

void C_Gui::childRemoved( GuiContained * cld )
{
  lock();
  if( focus_ && cld == focus_ ) {
    focus_ = 0;
  }
  if( clicked_ && cld == clicked_ ) {
    clicked_ = 0;
  }
  unlock();
}

GuiRenderer * C_Gui::renderer( GuiWidget * forWidget )
{
//todo: count number of renderers, assert if there are any left on update.
  return new C_GuiRenderer( this, forWidget, text_ );
}

void C_Gui::invalidate()
{
  dirty_ = true;
}

void C_Gui::render()
{
  lock();
  if( dirty_ ) {
    //  clear to a gray background -- this will also dirty the bitmap
    text_->fill( GWColor( 0x00404040 ), 0, 0, width_, height_ );
    text_->getImage( &image_ );
    renderChildren();
    dirty_ = false;
    (void)text_->texture(); //  updates texture image
  }

  shader_->apply();

  glBegin( GL_QUADS );
    glTexCoord2f( 0, 1 );
    glVertex2i( size_.left, size_.top );
    glTexCoord2f( 0, 1-(size_.bottom-size_.top)/float(height_) );
    glVertex2i( size_.left, size_.bottom );
    glTexCoord2f( (size_.right-size_.left)/float(width_), 1-(size_.bottom-size_.top)/float(height_) );
    glVertex2i( size_.right, size_.bottom );
    glTexCoord2f( (size_.right-size_.left)/float(width_), 1 );
    glVertex2i( size_.right, size_.top );
  glEnd();

  assert( !glGetError() );
  unlock();
}


void C_Gui::add( GuiContained * contained, WidgetSize const & size )
{
  lock();
  GuiContainer * prev = this;
  while( true ) {
    GuiContained * gc = prev->childAt( size.left, size.top );
    if( !gc || !gc->container() ) {
      break;
    }
    prev = gc->container();
  }
  prev->addChild( contained, size, 0 );
  //  The container incremented the ref count, but I take the 
  //  ownership of one ref count away from my caller.
  contained->release();
  unlock();
}

void C_Gui::beginClick( unsigned int x, unsigned int y, unsigned int mb )
{
  lock();
  assert( !clicked_ );
  recurseClick( this, x - size_.left, y - size_.top, mb );
  unlock();
}

bool C_Gui::recurseClick( GuiContainer * c, unsigned int x, unsigned int y, unsigned int mb )
{
  GuiContained * gc = c->childAt( x, y );
  if( !gc ) {
    return false;
  }
  GuiContainer * cc = gc->container();
  if( cc != 0 ) {
    if( recurseClick( cc, x, y, mb ) ) {
      return true;
    }
  }
  GuiWidget * gw = gc->widget();
  if( gw != 0 && gw->beginClick( x, y, mb ) ) {
    if( !clicked_ ) {
      //  Allow for nasty hacks by only setting it if it's not already set here.
      //  Note that beginClick() asserts that it's not set at the top level.
      clicked_ = gw;
    }
    return true;
  }
  return false;
}

void C_Gui::endClick( unsigned int x, unsigned int y )
{
  lock();
  x -= size_.left;
  y -= size_.top;
  if( clicked_ ) {
    WidgetSize size;
    clicked_->getPos( &size );
    //  Allow for nasty hacks by clearing clicked_ before dispatching.
    GuiWidget * c = clicked_;
    clicked_ = 0;
    c->endClick( size.left <= x && size.right > x &&
        size.top <= y && size.bottom > y );
    if( callback_ ) {
      callback_->clickDone( this, c );
    }
  }
  unlock();
}

void C_Gui::mouseMove( unsigned int x, unsigned int y )
{
  lock();
  bool in = (size_.left <= x && size_.right > x && size_.top <= y && size_.bottom > y);
  if( !in ) {
    if( lastOverTarget_ ) {
      lastOverTarget_->mouseLeave();
      lastOverTarget_ = 0;
    }
  }
  else {
    mouseOver( x - size_.left, y - size_.top );
  }
  unlock();
}

bool C_Gui::captureKey( unsigned int x, unsigned int y, unsigned char key )
{
  if( !callback_ ) {
    return false;
  }
  return callback_->captureKey( this, key );
}

void C_Gui::setCallback( MyGuiCallback * callback )
{
  callback_ = callback;
}




void C_GuiRenderer::drawImage( Image const & i, unsigned int x, unsigned int y, bool alphaTest )
{
  unsigned int w = i.width;
  unsigned int h = i.height;
  WidgetSize pos;
  gui_->getPos( &pos );
  unsigned int tw = pos.right - pos.left;
  unsigned int th = pos.bottom - pos.top;
  if( x >= tw || y >= th ) {
    return;
  }
  if( x + w > tw ) {
    w = tw-x;
  }
  if ( y + h > th ) {
    h = th-y;
  }
  //  Now, adjust for the fact that destination is bottom-to-top, 
  //  and "0" means "top" -- the drawImageX functions will use this 
  //  Y value correctly.
  y = th-h-y;
  switch( i.bits ) {
    default:
      assert( !"unsupported image format" );
      break;
    case 8:
      if( alphaTest ) {
        drawImage8A( i, x, y, w, h );
      }
      else {
        drawImage8( i, x, y, w, h );
      }
      break;
    case 24:
      if( alphaTest ) {
        drawImage24A( i, x, y, w, h );
      }
      else {
        drawImage24( i, x, y, w, h );
      }
      break;
    case 32:
      if( alphaTest ) {
        drawImage32A( i, x, y, w, h );
      }
      else {
        drawImage32( i, x, y, w, h );
      }
      break;
  }
}

void C_GuiRenderer::drawImage32A( Image const & i, unsigned int x, unsigned int y, unsigned int w, unsigned int h ) {
  Image d;
  text_->getImage( &d );
  WidgetSize pos;
  gui_->getPos( &pos );
  unsigned int th = pos.bottom-pos.top;
  //  flush to top
  y += d.height - th;
  unsigned char * dst = (unsigned char *)d.base;
  unsigned char const * src = (unsigned char const *)i.base;
  for( unsigned int r = 0; r < h; ++r ) {
    unsigned int * dp = (unsigned int *)(dst + d.pitch * (r+y) + x * 4);
    unsigned int const * sp = (unsigned int *)(src + i.pitch * r);
    for( unsigned int c = 0; c < w; ++c ) {
      unsigned int pixel = sp[c];
      if( (pixel & 0xff000000) >= 0x40000000 ) {
        dp[c] = pixel;
      }
    }
  }
}

void C_GuiRenderer::drawImage32( Image const & i, unsigned int x, unsigned int y, unsigned int w, unsigned int h ) {
  Image d;
  text_->getImage( &d );
  WidgetSize pos;
  gui_->getPos( &pos );
  unsigned int th = pos.bottom-pos.top;
  //  flush to top
  y += d.height - th;
  unsigned char * dst = (unsigned char *)d.base;
  unsigned char const * src = (unsigned char const *)i.base;
  for( unsigned int r = 0; r < h; ++r ) {
    memcpy( dst + d.pitch * (r+y) + x * 4, src + i.pitch * r, w * 4 );
  }
}




GuiRoot * NewGui( WidgetSize const & size )
{
  return new C_Gui( size );
}




class C_StaticTextWidget : public GuiWidget {
  public:
    C_StaticTextWidget( char const * str, int align, GWColor color ) {
      str_ = new char[strlen(str)+1];
      strcpy( str_, str );
      align_ = align;
      color_ = color;
    }
    ~C_StaticTextWidget() {
      delete[] str_;
    }
    void render() {
      GuiRenderer * r = top_->renderer( this );
      unsigned int tw, th;
      r->measureText( str_, &tw, &th );
      r->drawText( str_, size_.left + ((size_.right-size_.left-tw)*(1+align_))/2, 
          size_.top + (size_.bottom-size_.top-th)/2, color_.color );
      r->release();
    }
    bool setData( void const * data, size_t size ) {
      delete[] str_;
      size_t l = strlen( (char const *)data );
      str_ = new char[l+1];
      strcpy( str_, (char const *)data );
      invalidate();
      return true;
    }
    size_t getData( void * data, size_t size ) {
      size_t l = strlen( str_ );
      if( size > l+1 ) {
        size = l+1;
      }
      if( size > 0 ) {
        strncpy( (char *)data, str_, size );
        ((char *)data)[size-1] = 0;
      }
      return l;
    }
    char * str_;
    int align_;
    GWColor color_;
};


GuiWidget * NewStaticTextWidget( char const * text, int align, GWColor color )
{
  return new C_StaticTextWidget( text, align, color );
}


class C_GuiButton : public GuiWidget {
  public:
    C_GuiButton( char const * text, MyButtonCallback * callback, 
        GWColor colorNormal, GWColor colorHilite ) {
      str_ = new char[ strlen( text )+1 ];
      strcpy( str_, text );
      callback_ = callback;
      normal_ = colorNormal;
      hilite_ = colorHilite;
      clicking_ = false;
      clickMBs_ = 0;
    }
    ~C_GuiButton() {
      delete[] str_;
    }
    void render() {
      GuiRenderer * r = top_->renderer( this );
      GWColor color = clicking_ ? hilite_ : normal_;
      r->frameRect( size_, color.color );
      unsigned int tw, th;
      r->measureText( str_, &tw, &th );
      r->drawText( str_, 
          size_.left + (size_.right-size_.left-tw)/2 + (clicking_ ? 1 : 0), 
          size_.top + (size_.bottom-size_.top-th)/2 + (clicking_ ? 1 : 0), color.color );
      r->release();
    }
    bool beginClick( unsigned int x, unsigned int y, unsigned int mb ) {
      clickMBs_ = mb;
      clicking_ = true;
      invalidate();
      return true;
    }
    void mouseEnter( unsigned int x, unsigned int y ) {
      if( clickMBs_ ) {
        clicking_ = true;
        invalidate();
      }
    }
    void mouseLeave() {
      if( clicking_ ) {
        clicking_ = false;
        invalidate();
      }
    }
    void endClick( bool over ) {
      if( over ) {
        callback_->onButtonClicked( this, clickMBs_ );
      }
      clicking_ = false;
      clickMBs_ = 0;
      invalidate();
    }
    

    char * str_;
    MyButtonCallback * callback_;
    bool clicking_;
    uint clickMBs_;
    GWColor normal_;
    GWColor hilite_;
};

GuiWidget * NewButton( char const * text, MyButtonCallback * callback, 
    GWColor colorNormal, GWColor colorHilite )
{
  return new C_GuiButton( text, callback, colorNormal, colorHilite );
}

class C_GuiCheckbox : public GuiWidget {
  public:
    C_GuiCheckbox( char const * text, MyButtonCallback * callback, bool checked ) {
      str_ = new char[strlen(text)+1];
      strcpy( str_, text );
      callback_ = callback;
      checked_ = checked;
      hilite_ = GWColor(0x00ff8080);
      normal_ = GWColor(0x00ffffc0);
      clickMBs_ = 0;
      clicking_ = 0;
    }
    ~C_GuiCheckbox() {
      delete[] str_;
    }

    void render() {
      GuiRenderer * r = top_->renderer( this );
      GWColor color = clicking_ ? hilite_ : normal_;
      WidgetSize s = size_;
      s.right = s.left+s.bottom-s.top;
      r->frameRect( s, color.color );
      if( clicking_ != checked_ ) {
        r->drawLine( s.left, s.top, s.right, s.bottom, color.color );
        r->drawLine( s.left, s.bottom, s.right, s.top, color.color );
      }
      unsigned int width, height;
      r->measureText( str_, &width, &height );
      r->drawText( str_, 
          s.right+4,
          s.top+(s.bottom-s.top-height)/2,
          color.color );
      r->release();
    }
    bool beginClick( unsigned int x, unsigned int y, unsigned int mb ) {
      clickMBs_ = mb;
      clicking_ = true;
      invalidate();
      return true;
    }
    void mouseEnter( unsigned int x, unsigned int y ) {
      if( clickMBs_ ) {
        clicking_ = true;
        invalidate();
      }
    }
    void mouseLeave() {
      if( clicking_ ) {
        clicking_ = false;
        invalidate();
      }
    }
    void endClick( bool over ) {
      if( over ) {
        checked_ = !checked_;
        if( callback_ ) {
          callback_->onButtonClicked( this, clickMBs_ );
        }
      }
      clicking_ = false;
      clickMBs_ = 0;
      invalidate();
    }

    bool setData( void const * data, size_t size ) {
      if( size != sizeof( checked_ ) ) {
        assert( !"please use the right setData() protocol" );
        return false;
      }
      memcpy( &checked_, data, sizeof( checked_ ) );
      invalidate();
      return true;
    }
    size_t getData( void * data, size_t size ) {
      if( size != sizeof( checked_ ) ) {
        assert( !"please use the right getData() protocol" );
        return 0;
      }
      memcpy( data, &checked_, size );
      return sizeof( checked_ );
    }

    char * str_;
    MyButtonCallback * callback_;
    unsigned int clickMBs_;
    bool clicking_;
    bool checked_;
    GWColor hilite_;
    GWColor normal_;
};

GuiWidget * NewCheckboxWidget( char const * str, MyButtonCallback * callback, bool checked )
{
  return new C_GuiCheckbox( str, callback, checked );
}

class C_GuiSelector : public GuiWidget {
  public:
    C_GuiSelector( char const * text, char const * const * selections, size_t count, 
        MyButtonCallback * callback, int align, GWColor normal, GWColor hilite ) {
      str_ = new char[strlen(text)+1];
      strcpy( str_, text );
      selections_ = new char *[count];
      cur_ = 0;
      count_ = count;
      for( size_t i = 0; i < count; ++i ) {
        selections_[i] = new char[strlen(selections[i])+1];
        strcpy(selections_[i],selections[i]);
      }
      align_ = align;
      normal_ = normal;
      hilite_ = hilite;
      clickMBs_ = 0;
      clicking_ = false;
      curClick_ = 0;
      callback_ = callback;
    }
    ~C_GuiSelector() {
      delete[] str_;
      for( size_t i = 0; i < count_; ++i ) {
        delete[] selections_[i];
      }
      delete[] selections_;
    }
    bool clicking_;
    unsigned int clickMBs_;
    char * str_;
    char * * selections_;
    int align_;
    size_t cur_;
    size_t count_;
    GWColor hilite_;
    GWColor normal_;
    size_t curClick_;
    MyButtonCallback * callback_;

    void render() {
      GuiRenderer * r = top_->renderer( this );
      size_t epl = (size_.bottom-size_.top)/(count_+1);
      unsigned int sw, sh;
      r->measureText( str_, &sw, &sh );
      r->drawText( str_, size_.left + (size_.right-size_.left-sw)*(1+align_)/2, 
          size_.top + (epl-sh)/2, normal_.color );
      size_t hCur = (clicking_ ? curClick_ : cur_);
      for( size_t i = 0; i < count_; ++i ) {
        r->measureText( selections_[i], &sw, &sh );
        GWColor color = (hCur == i) ? hilite_ : normal_;
        r->drawText( selections_[i], size_.left + (size_.right-size_.left-sw)*(1+align_)/2,
          size_.top + (i+1)*epl + (epl-sh)/2, color.color );
      }
      WidgetSize size = { size_.left, size_.top+epl*(hCur+1), size_.right, size_.top+epl*(hCur+2)-1 };
      r->frameRect( size, hilite_.color );
      r->release();
    }
    bool beginClick( unsigned int x, unsigned int y, unsigned int mb ) {
      size_t epl = (size_.bottom-size_.top)/(count_+1);
      size_t e = (y - size_.top) / epl;
      if( e == 0 || e > count_ ) {
        return false;
      }
      curClick_ = e-1;
      clickMBs_ = mb;
      clicking_ = true;
      invalidate();
      return true;
    }
    void mouseEnter( unsigned int x, unsigned int y ) {
      if( clickMBs_ ) {
        clicking_ = true;
        invalidate();
      }
    }
    void mouseLeave() {
      if( clicking_ ) {
        clicking_ = false;
        invalidate();
      }
    }
    void endClick( bool over ) {
      if( over ) {
        cur_ = curClick_;
        if( callback_ ) {
          callback_->onButtonClicked( this, clickMBs_ );
        }
      }
      clicking_ = false;
      clickMBs_ = 0;
      invalidate();
    }
    bool setData( void const * data, size_t size ) {
      if( size != sizeof( cur_ ) ) {
        assert( !"please use the right setData() protocol" );
        return false;
      }
      memcpy( &cur_, data, sizeof( cur_ ) );
      if( cur_ >= count_ ) cur_ = count_-1;
      if( count_ == 0 ) cur_ = 0;
      invalidate();
      return true;
    }
    size_t getData( void * data, size_t size ) {
      if( size != sizeof( cur_ ) ) {
        assert( !"please use the right getData() protocol" );
        return 0;
      }
      memcpy( data, &cur_, size );
      return sizeof( cur_ );
    }
};

GuiWidget * NewSelector( char const * text, char const * const * selections, size_t count, 
    MyButtonCallback * callback, int align, 
    GWColor colorNormal, GWColor colorHilite )
{
  return new C_GuiSelector( text, selections, count, callback, align, colorNormal, colorHilite );
}


class C_GuiImage : public GuiWidget {
  public:
    C_GuiImage( Image const & i ) : image_( i ) {
    }
    void render() {
      GuiRenderer * r = top_->renderer( this );
      r->drawImage( image_, size_.left, size_.top, true );
    }
    Image image_;
    bool setData( void const * data, size_t size ) {
      if( size != sizeof( Image ) ) {
        assert( !"please use the right setData() protocol" );
        return false;
      }
      image_ = *(Image const *)data;
      return true;
    }
    size_t getData( void * data, size_t size ) {
      if( size != sizeof( Image ) ) {
        assert( !"please use the right getData() protocol" );
        return 0;
      }
      *(Image *)data = image_;
      return sizeof( Image );
    }
};


GuiWidget * NewImageWidget( Image const & image )
{
  return new C_GuiImage( image );
}



class C_GuiImageSelector : public GuiWidget {
  public:
    C_GuiImageSelector( char const * text, Image const ** selections, size_t count, 
        MyButtonCallback * callback, GWColor normal, GWColor hilite ) {
      str_ = new char[strlen(text)+1];
      strcpy( str_, text );
      selections_ = new Image[count];
      cur_ = 0;
      count_ = count;
      totalWidth_ = 0;
      maxHeight_ = 0;
      for( size_t i = 0; i < count; ++i ) {
        assert( selections[i] );
        assert( selections[i]->height && selections[i]->width );
        Image const * im = selections[i];
        selections_[i] = *im;
        totalWidth_ += im->width;
        if( size_t(im->height) > maxHeight_ ) {
          maxHeight_ = im->height;
        }
      }
      normal_ = normal;
      hilite_ = hilite;
      clickMBs_ = 0;
      clicking_ = false;
      curClick_ = 0;
      callback_ = callback;
    }
    ~C_GuiImageSelector() {
      delete[] str_;
      delete[] selections_;
    }
    bool clicking_;
    unsigned int clickMBs_;
    char * str_;
    Image * selections_;
    int align_;
    size_t cur_;
    size_t count_;
    size_t totalWidth_;
    size_t maxHeight_;
    GWColor hilite_;
    GWColor normal_;
    size_t curClick_;
    MyButtonCallback * callback_;

    void render() {
      GuiRenderer * r = top_->renderer( this );
      size_t width = size_.right-size_.left;
      size_t left = 0;
      size_t pad = 0;
      size_t down = 16;
      if( maxHeight_ < size_.bottom-size_.top-16 ) {
        down = 16 + (size_.bottom-size_.top-16-maxHeight_)/2;
      }
      if( totalWidth_ >= width || count_ == 0 ) {
      }
      else {
        pad = (width-totalWidth_)/count_;
        left = (width-totalWidth_-(pad*(count_-1)))/2;
      }
      r->drawText( str_, size_.left, size_.top, normal_.color );
      for( size_t i = 0; i < count_; ++i ) {
        size_t top = (maxHeight_-selections_[i].height)/2+down;
        WidgetSize ws = { size_.left+left, size_.top+top, 
            size_.left+left+selections_[i].width, 
            size_.top+top+selections_[i].height };
        if( clicking_ && (curClick_ == i) ) {
          r->fillRect( ws, hilite_.color );
        }
        else if( cur_ == i ) {
          r->fillRect( ws, normal_.color );
        }
        r->drawImage( selections_[i], ws.left, ws.top, true );
        left += pad + selections_[i].width;
      }
    }
    bool beginClick( unsigned int x, unsigned int y, unsigned int mb ) {
      size_t width = size_.right-size_.left;
      size_t left = 0;
      size_t pad = 0;
      size_t down = 16;
      if( maxHeight_ < size_.bottom-size_.top-16 ) {
        down = 16 + (size_.bottom-size_.top-16-maxHeight_)/2;
      }
      if( totalWidth_ >= width || count_ == 0 ) {
      }
      else {
        pad = (width-totalWidth_)/count_;
        left = (width-totalWidth_-(pad*(count_-1)))/2;
      }
      for( size_t i = 0; i < count_; ++i ) {
        size_t top = (maxHeight_-selections_[i].height)/2+down;
        WidgetSize ws = { size_.left+left, size_.top+top, 
            size_.left+left+selections_[i].width, 
            size_.top+top+selections_[i].height };
        if( x >= ws.left && x < ws.right && y >= ws.top && y < ws.bottom ) {
          curClick_ = i;
          clickMBs_ = mb;
          clicking_ = true;
          invalidate();
          return true;
        }
        left += pad + selections_[i].width;
      }
      return false;
    }
    void mouseEnter( unsigned int x, unsigned int y ) {
      if( clickMBs_ ) {
        clicking_ = true;
        invalidate();
      }
    }
    void mouseLeave() {
      if( clicking_ ) {
        clicking_ = false;
        invalidate();
      }
    }
    void endClick( bool over ) {
      if( over ) {
        cur_ = curClick_;
        if( callback_ ) {
          callback_->onButtonClicked( this, clickMBs_ );
        }
      }
      clicking_ = false;
      clickMBs_ = 0;
      invalidate();
    }
    bool setData( void const * data, size_t size ) {
      if( size != sizeof( cur_ ) ) {
        assert( !"please use the right setData() protocol" );
        return false;
      }
      memcpy( &cur_, data, sizeof( cur_ ) );
      if( cur_ >= count_ ) cur_ = count_-1;
      if( count_ == 0 ) cur_ = 0;
      invalidate();
      return true;
    }
    size_t getData( void * data, size_t size ) {
      if( size != sizeof( cur_ ) ) {
        assert( !"please use the right getData() protocol" );
        return 0;
      }
      memcpy( data, &cur_, size );
      return sizeof( cur_ );
    }
};

GuiWidget * NewImageSelectorWidget( char const * text, Image const ** selections, size_t count,
    MyButtonCallback * callback, GWColor colorNormal, GWColor colorHilite )
{
  return new C_GuiImageSelector( text, selections, count, callback, colorNormal, colorHilite );
}
