
#if !defined( mygui_h )
#define mygui_h

#include "image.h"
#include "glwrap.h" //  for GWColor

#include <stdlib.h>
#include <vector>

struct WidgetSize {
  unsigned int left, top, right, bottom;
};

class GuiContainer;
class GuiWidget;
class GuiRoot;

class GuiContained {
  public:
    virtual void hold();
    virtual void release();
    //  Note that your location (found in size_) is global to the top-level view, 
    //  so you could render outside yourself if you're not careful.
    virtual void addedToContainer( GuiRoot * top, GuiContainer * parent, WidgetSize const & inSize );
    virtual void removedFromContainer( GuiRoot * top, GuiContainer * parent );
    virtual void getPos( WidgetSize * outSize ) { *outSize = size_; }
    virtual void mouseEnter( unsigned int x, unsigned int y );
    virtual void mouseOver( unsigned int x, unsigned int y );
    virtual void mouseLeave();
    virtual GuiContainer * parent() { return parent_; }
    virtual GuiContainer * container() { return 0; }
    virtual GuiWidget * widget() { return 0; }
  protected:
    GuiContained();
    virtual ~GuiContained() {}
    static void releaseContained( GuiContained * c );

    unsigned int refCount_;
    GuiContainer * parent_;
    GuiRoot * top_;
    WidgetSize size_;
};

class GuiWidget : public GuiContained {
  public:
    virtual void invalidate();
    //  A GuiWidget can assume that its background has been cleared to 
    //  something sensible, and should only draw its actual text and 
    //  decorations, not background.
    virtual void render() = 0;
    virtual bool beginClick( unsigned int x, unsigned int y, unsigned int mb );
    virtual void endClick( bool over );
    virtual GuiWidget * widget() { return this; }

    virtual bool setData( void const * data, size_t size );
    virtual size_t getData( void * data, size_t size );

  protected:
    GuiWidget();
    virtual ~GuiWidget() {}
};


class GuiContainer : public GuiContained {
  public:

    virtual void addChild( GuiContained * child, WidgetSize const & size, GuiContained * before );
    virtual void removeChild( GuiContained * child );
    void renderChildren();
    virtual unsigned int numChildren();
    virtual GuiContained * childAt( unsigned int ix );
    virtual GuiContained * childAt( unsigned int xPos, unsigned int yPos );
    virtual void mouseEnter( unsigned int x, unsigned int y );
    virtual void mouseOver( unsigned int x, unsigned int y );
    virtual void mouseLeave();
    virtual GuiContainer * container() { return this; }

  protected:
    GuiContainer();
    virtual ~GuiContainer();
    virtual void childRemoved( GuiContained * cld );

    std::vector< GuiContained * > children_;
    GuiContained * lastOverTarget_;
};

class GuiRenderer {
  public:
    virtual void hold() = 0;
    virtual void release() = 0;

    virtual void fillRect( WidgetSize const & rect, unsigned int color ) = 0;
    virtual void frameRect( WidgetSize const & rect, unsigned int color ) = 0;
    virtual void drawLine( unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2, unsigned int color ) = 0;
    virtual void drawText( char const * text, unsigned int x, unsigned int y, unsigned int color ) = 0;
    virtual void measureText( char const * text, unsigned int * width, unsigned int * height ) = 0;
    virtual void drawImage( Image const & i, unsigned int x, unsigned int y, bool alphaTest = true ) = 0;
  protected:
    virtual ~GuiRenderer() {}
};

class MyGuiCallback {
  public:
    virtual bool captureKey( GuiRoot * root, unsigned char key ) = 0;
    virtual void guiClosing( GuiRoot * root ) = 0;
    virtual void clickDone( GuiRoot * root, GuiWidget * target ) = 0;
};

//  a "gui" is like a dialog
class GuiRoot {
  public:

    virtual void dispose() = 0;

    virtual void setFocus( GuiWidget * widget ) = 0;
    virtual GuiWidget * focus() = 0;
    //  renderer() really is ONLY for the child widgets
    virtual GuiRenderer * renderer( GuiWidget * forWidget ) = 0;
    virtual void invalidate() = 0;

    //  GuiRoot::render() actually draws to the renderer.
    //  GuiWidget and GuiContainer don't do that; instead, they render 
    //  to the texture of GuiRoot; they are only called when the Gui is 
    //  invalidated.
    virtual void render() = 0;
    //  beginClick for the root is global-relative
    virtual void beginClick( unsigned int x, unsigned int y, unsigned int mb ) = 0;
    //  endClick for the root is global-relative
    virtual void endClick( unsigned int x, unsigned int y ) = 0;
    //  mouseMove for the root is global-relative
    virtual void mouseMove( unsigned int x, unsigned int y ) = 0;

    //  GUI may choose to capture a key
    virtual bool captureKey( unsigned int x, unsigned int y, unsigned char key ) = 0;
    virtual void setCallback( MyGuiCallback * callback ) = 0;

    //  add() will recurse, finding the right container (based on top-left).
    //  add() will take ownership of the contained (not increment ref count) -- 
    //  this is different from pretty much all other functions in this interface!
    virtual void add( GuiContained * contained, WidgetSize const & size ) = 0;

  protected:

    GuiRoot() {}
    ~GuiRoot() {}
};

class MyButtonCallback {
  public:
    virtual void onButtonClicked( GuiWidget * widget, uint mbMask ) = 0;
};


GuiRoot * NewGui( WidgetSize const & size );
GuiWidget * NewStaticTextWidget( char const * text, int align = -1, GWColor color = GWColor(0x00ffffc0) );
GuiWidget * NewImageWidget( Image const & image );
GuiWidget * NewButton( char const * text, MyButtonCallback * callback, 
    GWColor colorNormal = GWColor(0x00ffffc0), GWColor colorHilite = GWColor(0x00ff8080) );
GuiWidget * NewSelector( char const * text, char const * const * selections, size_t count, 
    MyButtonCallback * callback = 0, int align = -1, 
    GWColor colorNormal = GWColor(0x00ffffc0), GWColor colorHilite = GWColor(0x00ff8080) );
GuiWidget * NewImageSelectorWidget( char const * text, Image const ** selections, size_t count,
    MyButtonCallback * callback = 0, 
    GWColor colorNormal = GWColor(0x00ffffc0), GWColor colorHilite = GWColor(0x00ff8080) );
GuiWidget * NewCheckboxWidget( char const * text, MyButtonCallback * callback = 0, bool checked = true );

#endif  //  mygui_h
