
#include "mymachinegun.h"
#include "myweapon.h"
#include "misc.h"
#include "glshaderstate.h"
#include "gl-ase.h"
#include "myparticles.h"
#include "mysystem.h"
#include "mytweakables.h"


//  CMachinegunBulletManager is shared between all machine guns.
class CMachinegunBulletManager {
  public:
    CMachinegunBulletManager();
    ~CMachinegunBulletManager();
    //  addBullet() has immediate physical effect
    void addBullet( Vector3 const & src, Vector3 const & dir, float distance );
    void renderBullets();
    enum {
      MAX_BULLETS = 50
    };

    struct MGVertex {
      Vector3 pos;
      uint color;
    };
    unsigned int numBullets_;
    MGVertex vertices_[MAX_BULLETS*2];
    GLShader * bulletShader_;
    MyParticles * impactEffect_;
 
    //  these are used when hit testing
    float nearest_;
    MySystem * hit_;
    dGeomID hitGeom_;
};

extern "C" {
  static void BulletRay( void * data, dGeomID g1, dGeomID g2 )
  {
    assert( dGeomGetClass( g2 ) == dRayClass );
    if( dGeomIsSpace( g1 ) ) {
      dSpaceCollide2( g1, g2, data, &BulletRay );
    }
    else {
      dContactGeom cg;
      CMachinegunBulletManager * ptr = (CMachinegunBulletManager *)data;
      MySystem * s1 = (MySystem *)dGeomGetData( g1 );
      if( (!s1 || s1->physical()) && dCollide( g2, g1, 1, &cg, sizeof( cg ) ) ) {
        if( ptr->nearest_ > cg.depth ) {
          ptr->nearest_ = cg.depth;
          ptr->hit_ = s1;
          ptr->hitGeom_ = g1;
        }
      }
    }
  }
}

CMachinegunBulletManager::CMachinegunBulletManager()
{
  numBullets_ = 0;
  bulletShader_ = gShaderFactory->newShader( "bulletShader" );
  impactEffect_ = MakeParticleSystem( "bulletImpact", MySystemBase::worldSpace() );
}

CMachinegunBulletManager::~CMachinegunBulletManager()
{
  bulletShader_->dispose();
  delete impactEffect_;
}

void CMachinegunBulletManager::addBullet( Vector3 const & src, Vector3 const & dir, float distance )
{
  nearest_ = distance;
  hit_ = 0;
  hitGeom_ = 0;
  dGeomID g = MySystemBase::utilRay();
  dGeomRaySet( g, src.x, src.y, src.z, dir.x, dir.y, dir.z );
  dGeomRaySetLength( g, distance );
  //  figure out whether we hit anything
  dSpaceCollide2( (dGeomID)MySystemBase::worldSpace(), g, this, &BulletRay );

  Vector3 hitPos = src + dir * nearest_;

  if( numBullets_ < MAX_BULLETS ) {
    MGVertex * v = &vertices_[numBullets_*2];
    v[0].pos = src;
    v[0].color = 0x40404040;
    v[1].pos = hitPos;
    v[1].color = 0xc0c0c0c0;
    ++numBullets_;
  }

  if( nearest_ < distance ) {   //  emit a hit

    //  particle effect
    Vector3 vel = Vector3::Null;
    if( hit_ != 0 ) {
      vel = hit_->L();
    }
    vel += dir * -0.3f;
    vel += Vector3::Y * 0.3f;
    hitPos += dir * -0.02f;
    impactEffect_->emit( hitPos, vel );

    //  physical effect
    if( hit_ ) {
      hit_->affectByImpact( hitGeom_, hitPos, dir, MACHINEGUN_BULLET_FORCE );
    }
  }
}

void CMachinegunBulletManager::renderBullets()
{
  if( numBullets_ ) {
    glEnableClientState( GL_VERTEX_ARRAY );
    glVertexPointer( 3, GL_FLOAT, sizeof( MGVertex ), &vertices_[0].pos.x );
    glEnableClientState( GL_COLOR_ARRAY );
    glColorPointer( 4, GL_UNSIGNED_BYTE, sizeof( MGVertex ), &vertices_[0].color );
    glDrawArrays( GL_LINES, 0, numBullets_*2 );
    glDisableClientState( GL_VERTEX_ARRAY );
    glDisableClientState( GL_COLOR_ARRAY );
    numBullets_ = 0;
  }
}


class CMachinegunWeapon : public I_Weapon {
  public:
    CMachinegunWeapon();
    ~CMachinegunWeapon();

    void renderSight( float screenWidth, float screenHeight, Matrix4 const & camMtx, Matrix4 const & projMtx );
    void setAim( Matrix3 const & aim, Vector3 const & source, Vector3 const & ptVel );
    bool readyToFire( double );
    bool fire( double );
    void renderIndicators( float phi, float alpha );
    virtual bool autoFire() { return true; }

    GLShader * reticleShader_;
    GLShader * indicatorShader_;
    GLShader * ammoShader_;
    I_GL_ASE * machinegunModel_;
    Matrix3 ori_;
    Vector3 pos_;
    Vector3 vel_;
    double lastFired_;
    static CMachinegunBulletManager * bullets_;
};

CMachinegunBulletManager * CMachinegunWeapon::bullets_;

I_Weapon * NewMachinegunWeapon()
{
  if( !CMachinegunWeapon::bullets_ ) {
    CMachinegunWeapon::bullets_ = new CMachinegunBulletManager();
  }
  return new CMachinegunWeapon();
}

void RenderMachinegunBullets()
{
  if( CMachinegunWeapon::bullets_ ) {
    CMachinegunWeapon::bullets_->renderBullets();
  }
}


CMachinegunWeapon::CMachinegunWeapon()
{
  reticleShader_ = gShaderFactory->newShader( "machinegunReticle" );
  indicatorShader_ = gShaderFactory->newShader( "ammoIndicator" );
  machinegunModel_ = ASEFromFile( asePath( "machinegun" ) );
  ori_ = Matrix3::Identity;
  pos_ = Vector3::Null;
  lastFired_ = 0;
}

CMachinegunWeapon::~CMachinegunWeapon()
{
  reticleShader_->dispose();
  indicatorShader_->dispose();
  machinegunModel_->dispose();
}

void CMachinegunWeapon::renderSight( float screenWidth, float screenHeight, Matrix4 const & camMtx, Matrix4 const & projMtx )
{
  int w = 32;
  if( screenWidth >= 1000 && screenHeight >= 750 ) {
    w = 32;
  }
  else {
    w = 16;
  }

  //  figure out where we'd be 20 meters out from the gun
  Vector4 pos = Vector4(pos_ + ori_.front() * 20) * camMtx * projMtx;
  pos *= 1.0f / pos.w;
  float midX = (pos.x + 1) * screenWidth/2;
  float midY = (pos.y + 1) * screenHeight/2;

  reticleShader_->apply();
  glBegin( GL_QUADS );
  glTexCoord2f( 0, 0 ); glVertex2f( midX-w, midY-w );
  glTexCoord2f( 1, 0 ); glVertex2f( midX+w, midY-w );
  glTexCoord2f( 1, 1 ); glVertex2f( midX+w, midY+w );
  glTexCoord2f( 0, 1 ); glVertex2f( midX-w, midY+w );
  glEnd();
}

void CMachinegunWeapon::setAim( Matrix3 const & aim, Vector3 const & source, Vector3 const & ptVel )
{
  ori_ = aim;
  pos_ = source;
  vel_ = ptVel;
}

bool CMachinegunWeapon::readyToFire( double time )
{
  return time > (lastFired_ + MACHINEGUN_REFIRE_TIME);
}

bool CMachinegunWeapon::fire( double time )
{
  if( !readyToFire( time ) ) {
    return false;
  }
  lastFired_ = time;
  bullets_->addBullet( pos_, ori_.front(), 150 );
  return true;
}

void CMachinegunWeapon::renderIndicators( float phi, float alpha )
{
  assert( alpha >= 0 );
  assert( phi > -3.1416f && phi < 6.2832f );
  glRotatef( 120 * phi, 0, 1, 0 );
  glTranslatef( 0, 0, -0.2f );
  glScalef( 0.5f, 0.5f, 0.5f );
  indicatorShader_->context().color.alpha_ = (alpha > 1) ? 0x80 : (uchar)(0x80 * alpha);
  indicatorShader_->apply();
  machinegunModel_->issueAll();
}
