
#include "mymath.h"
#include "ode/ode.h"
#include <assert.h>

Vector3 Vector3::X( 1, 0, 0 );
Vector3 Vector3::Y( 0, 1, 0 );
Vector3 Vector3::Z( 0, 0, 1 );
Vector3 Vector3::Null( 0, 0, 0 );
Vector4 Vector4::W( 0, 0, 0, 1 );
Vector4 Vector4::Null( 0, 0, 0, 0 );
Matrix3 Matrix3::Identity( Vector3::X, Vector3::Y, Vector3::Z );
Matrix3 Matrix3::Null( Vector3::Null, Vector3::Null, Vector3::Null );
Matrix4 Matrix4::Identity( Matrix3::Identity, Vector4::W );
Matrix4 Matrix4::Null( Matrix3::Null, Vector4::Null );



Matrix3::Matrix3( Vector3 const & axis, float angle )
{
  //  ODE uses column-major convention
  dRFromAxisAndAngle( &xx, axis.x, axis.y, axis.z, -angle );
  wx = wy = wz = 0;
}


void RotationMatrixAroundPoint( Matrix4 * out, Vector3 const & up, float amt, Vector3 const & point )
{
  Matrix4 a = Matrix4::Identity;
  a.xw = -point.x;
  a.yw = -point.y;
  a.zw = -point.z;
  Matrix4 b( up, amt );
  *out = a * b;
  out->xw += point.x;
  out->yw += point.y;
  out->zw += point.z;
}

namespace {
  class UnitTestMatrixRotation {
    public:
      UnitTestMatrixRotation() {
        Matrix4 a;
        RotationMatrixAroundPoint( &a, Vector3::Y, 0, Vector3::Null );
        assert( fabsf( a.xx-1 ) < 0.001f );
        assert( fabsf( a.yx ) < 0.001f );
        assert( fabsf( a.zx ) < 0.001f );
        assert( fabsf( a.wx ) < 0.001f );
        assert( fabsf( a.xy ) < 0.001f );
        assert( fabsf( a.yy-1 ) < 0.001f );
        assert( fabsf( a.zy ) < 0.001f );
        assert( fabsf( a.wy ) < 0.001f );
        assert( fabsf( a.xz ) < 0.001f );
        assert( fabsf( a.yz ) < 0.001f );
        assert( fabsf( a.zz-1 ) < 0.001f );
        assert( fabsf( a.wz ) < 0.001f );
        assert( fabsf( a.xw ) < 0.001f );
        assert( fabsf( a.yw ) < 0.001f );
        assert( fabsf( a.zw ) < 0.001f );
        assert( fabsf( a.ww-1 ) < 0.001f );
        Vector3 p = Vector3( 1, 3, 4 ) * a;
        assert( fabsf( p.x-1 ) < 0.001f );
        assert( fabsf( p.y-3 ) < 0.001f );
        assert( fabsf( p.z-4 ) < 0.001f );
        RotationMatrixAroundPoint( &a, Vector3::Y, 3.1415926f/2, Vector3::Z );
        p = Vector3( 1, 0, 0 ) * a;
        assert( fabsf( p.x+1 ) < 0.001f );
        assert( fabsf( p.y ) < 0.001f );
        assert( fabsf( p.z ) < 0.001f );
      }
  };
  UnitTestMatrixRotation utmr;
}

//  Return true if the sphere described has non-zero intersection with the AABB
bool PointRadiusIntersectsAABB( Vector3 const & point, float radius, Vector3 const & low, Vector3 const & high )
{
  return
    point.x - radius < high.x &&
    point.x + radius > low.x &&
    point.y - radius < high.y &&
    point.y + radius > low.y &&
    point.z - radius < high.z &&
    point.z + radius > low.z;
}

//  Return true if the AABB described has non-zero intersection with the AABB
bool AABBIntersectsAABB( Vector3 const & aLow, Vector3 const & aHigh, Vector3 const & bLow, Vector3 const & bHigh, bool * outFullyVisible )
{
  //  check for a completely visible within b
  if( outFullyVisible != 0 ) {
    *outFullyVisible = false;
    if( aLow.x >= bLow.x && 
        aHigh.x <= bHigh.x && 
        aLow.y >= bLow.y &&
        aHigh.y <= bHigh.y &&
        aLow.z >= bLow.z &&
        aLow.z <= bHigh.z ) {
      *outFullyVisible = true;
      return true;
    }
  }
  return
    aLow.x < bHigh.x &&
    aHigh.x > bLow.x &&
    aLow.y < bHigh.y &&
    aHigh.y > bLow.y &&
    aLow.z < bHigh.z &&
    aHigh.z > bLow.z;
}

//  Return true if the sphere described may have non-zero intersection with the planes.
//  This implementation is conservative, it may return TRUE for some non-intersecting cases.
bool PointRadiusIntersectsPlanes( Vector3 const & point, float radius, Vector4 const * planes, uint numPlanes )
{
  //  This is conservative, because no single plane may be able to 
  //  fully cull away the sphere, yet it may not intersect the polyhedron
  //  described by all the planes together.
  for( uint p = 0; p < numPlanes; ++p ) {
    float d = planes[p].vector3() / point;
    d += planes[p].w;
    d += radius;
    if( d < 0 ) {
      return false;
    }
  }
  return true;
}

//  Return true if the point lies on or within all the planes.
bool PointIntersectsPlanes( Vector3 const & point, Vector4 const * planes, uint numPlanes )
{
  for( uint i = 0; i < numPlanes; ++i ) {
    float d = point / planes[i].vector3();
    d += planes[i].w;
    if( d < 0 ) {
      return false;
    }
  }
  return true;
}

//  Return true if any points intersects the plane
bool AnyPointIntersectsPlane( Vector3 const * points, uint numPoints, Vector4 const & plane, bool * outAllIntersect )
{
  if( outAllIntersect ) {
    bool anyIntersect = false;
    *outAllIntersect = true;
    for( uint i = 0; i < numPoints; ++i ) {
      float d = points[i] / plane.vector3();
      d += plane.w;
      if( d >= 0 ) {
        anyIntersect = true;
      }
      else {
        *outAllIntersect = false;
      }
    }
    return anyIntersect;
  }
  for( uint i = 0; i < numPoints; ++i ) {
    float d = points[i] / plane.vector3();
    d += plane.w;
    if( d >= 0 ) {
      return true;
    }
  }
  return false;
}

//  Return true if the AABB described may have non-zero intersection with the planes.
//  This implementation is conservative, it may return TRUE for some non-intersecting cases.
bool AABBIntersectsPlanes( Vector3 const & low, Vector3 const & high, Vector4 const * planes, uint numPlanes, bool * outFullyVisible )
{
  //fixme: implement this better
#if 0
  //  Find the bounding sphere of the AABB, and test against that
  Vector3 c( (low.x+high.x)*0.5f, (low.y+high.y)*0.5f, (low.z+high.z)*0.5f );
  float r = sqrtf( (c.x-low.x)*(c.x-low.x) + (c.y-low.y)*(c.y-low.y) + (c.z-low.z)*(c.z-low.z) );
  return PointRadiusIntersectsPlanes( c, r, planes, numPlanes );
#endif

  Vector3 points[8] = {
    Vector3( low.x, low.y, low.z ), Vector3( high.x, low.y, low.z ), 
    Vector3( low.x, high.y, low.z ), Vector3( high.x, high.y, low.z ), 
    Vector3( low.x, low.y, high.z ), Vector3( high.x, low.y, high.z ), 
    Vector3( low.x, high.y, high.z ), Vector3( high.x, high.y, high.z ), 
  };
  for( uint i = 0; i < numPlanes; ++i ) {
    if( !AnyPointIntersectsPlane( points, 8, planes[i], outFullyVisible ) ) {
      //  we found a plane that culls the entire point set
      return false;
    }
  }
  return true;
}

bool OBBIntersectsAABB( OBB const & obb, Vector3 const & low, Vector3 const & high )
{
return true;    //  implement me!
  assert( 0 );  //  do this tomorrow
}

bool OBBIntersectsPlanes( OBB const & obb, Vector4 const * planes, uint numPlanes )
{
return true;    //  implement me!
  assert( 0 );  //  do this tomorrow
}


void ExtractFrustumPlanes( Matrix4 const & modelView, Matrix4 const & projection, Vector4 planes[6] )
{
  //  for good derivation, see http://www2.ravensoft.com/users/ggribb/plane%20extraction.pdf
  Matrix4 const & m = modelView * projection;
  planes[0] = m.col(3) + m.col(0);
  planes[1] = m.col(3) - m.col(0);
  planes[2] = m.col(3) + m.col(1);
  planes[3] = m.col(3) - m.col(1);
  planes[4] = m.col(3) + m.col(2);
  planes[5] = m.col(3) - m.col(2);

  for( uint i = 0; i < 6; ++i ) {
    float l = sqrtf( planes[i].vector3().lengthSquared() );
    planes[i] *= 1.0f / l;
  }
}


namespace {
  class UnittestMathFunctions {
    public:
      UnittestMathFunctions() {

//  Return true if the sphere described has non-zero intersection with the AABB
//  bool PointRadiusIntersectsAABB( Vector3 const & point, float radius, Vector3 const & low, Vector3 const & high );
        assert( PointRadiusIntersectsAABB( Vector3( 1, 1, 1 ), 2, Vector3( 0, 0, 0 ), Vector3( 2, 2, 0 ) ) );
        assert( !PointRadiusIntersectsAABB( Vector3( 0, 0, 3 ), 2, Vector3( 0, 0, 0 ), Vector3( 2, 2, 0 ) ) );

//  Return true if the AABB described has non-zero intersection with the AABB
//  bool AABBIntersectsAABB( Vector3 const & aLow, Vector3 const & aHigh, Vector3 const & bLow, Vector3 const & bHigh );
        assert( AABBIntersectsAABB( Vector3( 1, 1, 1 ), Vector3( 2, 3, 4 ), Vector3( 1.5f, 0, 0 ), Vector3( 1.6f, 1.1f, 5 ) ) );
        assert( !AABBIntersectsAABB( Vector3( 1, 1, 1 ), Vector3( 2, 3, 4 ), Vector3( 0.5f, 0, 0 ), Vector3( 0.6f, 1.1f, 5 ) ) );

//  Return true if the point lies on or within all the planes.
//  bool PointIntersectsPlanes( Vector3 const & point, Vector4 const * planes, uint numPlanes );
        Vector4 planes[2] = { Vector4( 0, 1, 0, 0 ), Vector4( 1, 0, 0, 2 ) };
        assert( PointIntersectsPlanes( Vector3( 0, 0, 0 ), planes, 2 ) );
        assert( !PointIntersectsPlanes( Vector3( -3, 1, 1 ), planes, 2 ) );

//  Return true if any points intersects the plane
//  bool AnyPointIntersectsPlane( Vector3 const * points, uint numPoints, Vector4 const & plane );
        Vector3 points[3] = { Vector3( -2, -1, -1 ), Vector3( -2, 1, -1 ), Vector3( 1, 0, 0 ) };
        assert( AnyPointIntersectsPlane( points, 3, Vector4( 1, 0, 0, 1 ) ) );
        assert( !AnyPointIntersectsPlane( points, 2, Vector4( 1, 0, 0, 1 ) ) );
        assert( AnyPointIntersectsPlane( points, 3, Vector4( 1, 0, 0, -1 ) ) );
        assert( !AnyPointIntersectsPlane( points, 2, Vector4( 0, 1, 0, -2 ) ) );

//  Return true if the sphere described may have non-zero intersection with the planes.
//  This implementation is conservative, it may return TRUE for some non-intersecting cases.
//  bool PointRadiusIntersectsPlanes( Vector3 const & point, float radius, Vector4 const * planes, uint numPlanes );
        assert( PointRadiusIntersectsPlanes( Vector3( 0, 0, 0 ), 0, planes, 2 ) );
        assert( !PointRadiusIntersectsPlanes( Vector3( -3, 1, 1 ), 0.9f, planes, 2 ) );
        assert( PointRadiusIntersectsPlanes( Vector3( -3, 1, 1 ), 2, planes, 2 ) );

//  Return true if the AABB described may have non-zero intersection with the planes.
//  This implementation is conservative, it may return TRUE for some non-intersecting cases.
//  bool AABBIntersectsPlanes( Vector3 const & low, Vector3 const & high, Vector4 const * planes, uint numPlanes );
        assert( AABBIntersectsPlanes( Vector3( -1, -1, -1 ), Vector3( 1, 1, 1 ), planes, 2 ) );
        assert( !AABBIntersectsPlanes( Vector3( -5, -5, -5 ), Vector3( -4, -2, -2 ), planes, 2 ) );

//  Extract the left, right, bottom, top, near and far clipping planes for a view frustum
//  defined by a modelview and a projection matrix.
//  void ExtractFrustumPlanes( Matrix4 const & modelView, Matrix4 const & projection, Vector4 planes[6] );
        Matrix4 mvm( 1, 0, 0, 0, 0, 0, 1, 0, 0, -1, 0, 0, 10, 0, 0, 0 );
        Matrix4 pm( 0.75, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0 );
        Vector4 planes2[7];
        planes2[6] = Vector4( 100, 100, 100, 100 );
        ExtractFrustumPlanes( mvm, pm, planes2 );  //  at least it shouldn't crash!
        assert( planes2[6] == Vector4( 100, 100, 100, 100 ) );
      }
  };
  UnittestMathFunctions utmf;
}

