
#if !defined( mymath_h )
#define mymath_h

#include "glut-start.h"
#undef min
#undef max
#include <algorithm>

//  The conventions are:

//  row vectors
//  multiplied from the left into matrices
//  matrices are stored row major
//  translation lives at the bottom of the matrix
//  (i e elements 12, 13 and 14 when numbering from 0 to 15)

class Vector3 {
  public:
    Vector3( float a, float b, float c ) : x(a), y(b), z(c) {}
    Vector3() {}
    float x, y, z;
    static Vector3 X;
    static Vector3 Y;
    static Vector3 Z;
    static Vector3 Null;

    bool operator==( Vector3 const & o ) const {
      return (x == o.x) && (y == o.y) && (z == o.z);
    }

    //  modify by vector
    Vector3 & operator+=( Vector3 const & o ) {
      x += o.x; y += o.y; z += o.z; return *this;
    }
    Vector3 & operator-=( Vector3 const & o ) {
      x -= o.x; y -= o.y; z -= o.z; return *this;
    }

    //  return by vector
    Vector3 operator+( Vector3 const & o ) const {
      return Vector3( x+o.x, y+o.y, z+o.z );
    }
    Vector3 operator-( Vector3 const & o ) const {
      return Vector3( x-o.x, y-o.y, z-o.z );
    }

    //  dot product as division
    float operator/( Vector3 const & o ) const {
      return x*o.x + y*o.y + z*o.z;
    }

    //  cross product as multiplication; X cross Y is Z
    Vector3 operator*( Vector3 const & o ) const {
      return Vector3( y*o.z - z*o.y, z*o.x - x*o.z, x*o.y - y*o.x );
    }

    //  modify by scalar
    Vector3 & operator*=( float f ) {
      x *= f; y *= f; z *= f; return *this;
    }

    //  return by scalar
    Vector3 operator*( float f ) const {
      return Vector3( x*f, y*f, z*f );
    }
    Vector3 operator-() const {
      return Vector3( -x, -y, -z );
    }

    float lengthSquared() const {
      return x*x + y*y + z*z;
    }
    float lengthSquaredNotZero() const {
      float l = x*x + y*y + z*z;
      if ( l <= 1e-10f ) {
        return 1e-10f;
      }
      return l;
    }
};
class Vector4 : public Vector3 {
  public:
    explicit Vector4( Vector3 const & v3 ) : Vector3( v3 ), w( 1 ) {}
    Vector4() {}
    Vector4( float a, float b, float c, float d ) : w(d) {
      x = a; y = b; z = c;
    }
    bool operator==( Vector4 const & o ) const {
      return (x == o.x) && (y == o.y) && (z == o.z) && (w == o.w);
    }
    Vector3 & vector3() {
      return *this;
    }
    Vector3 const & vector3() const {
      return *this;
    }
    float w;
    static Vector4 W;
    static Vector4 Null;

    Vector4 & operator*=( float f ) {
      x *= f; y *= f; z *= f; w *= f; return *this;
    }
    Vector4 & operator+=( Vector4 const & o ) {
      x += o.x; y += o.y; z += o.z; w += o.w; return *this;
    }
    Vector4 & operator-=( Vector4 const & o ) {
      x -= o.x; y -= o.y; z -= o.z; w -= o.w; return *this;
    }
    Vector4 operator+( Vector4 const & o ) const {
      return Vector4( x+o.x, y+o.y, z+o.z, w+o.w );
    }
    Vector4 operator-( Vector4 const & o ) const {
      return Vector4( x-o.x, y-o.y, z-o.z, w-o.w );
    }

  private:
    operator Vector3&() {
      return *this;
    }
    operator Vector3 const &() const {
      return *this;
    }
};
class Matrix3 {
  public:

    //  Storage is row major.
    //  Translation does not live in this matrix; 
    //  it would go in a fourth row below these three.
    //  Naming is column-row !
    union {
      struct {
        float xx, yx, zx, wx;
        float xy, yy, zy, wy;
        float xz, yz, zz, wz;
      };
      struct {
        float e00, e01, e02, e03;
        float e10, e11, e12, e13;
        float e20, e21, e22, e23;
      };
    };

    Matrix3() {}
    Matrix3( Vector3 const & a, Vector3 const & b, Vector3 const & c, bool basis = false ) {
      wx = wy = wz = 0;
      if( basis ) {
        xx = a.x; xy = a.y; xz = a.z;
        yx = b.y; yy = b.y; yz = b.z;
        zx = c.z; zy = c.y; zz = c.z;
      }
      else {
        xx = a.x; yx = a.y; zx = a.z;
        xy = b.x; yy = b.y; zy = b.z;
        xz = c.x; yz = c.y; zz = c.z;
      }
    }
    Matrix3( 
        float _00, float _01, float _02, float _03,
        float _10, float _11, float _12, float _13,
        float _20, float _21, float _22, float _23 ) :
        xx(_00), yx(_01), zx(_02), wx(_03),
        xy(_10), yy(_11), zy(_12), wy(_13),
        xz(_20), yz(_21), zz(_22), wz(_23) {}
    Matrix3( Vector4 const & a, Vector4 const & b, Vector4 const & c ) {
      (Vector4 &)xx = a;
      (Vector4 &)xy = b;
      (Vector4 &)xz = c;
    }
    Matrix3( Vector3 const & axis, float angle );
    Matrix3 & ori() { return (Matrix3 &)*this; }
    Matrix3 const & ori() const { return (Matrix3 const &)*this; }
    Matrix3 transpose() const {
      return Matrix3( Vector3( xx, xy, xz ), Vector3( yx, yy, yz ), Vector3( zx, zy, zz ) );
    }
    void leftUpFront( Vector3 * left, Vector3 * up, Vector3 * front ) const {
      left->x = xx; left->y = yx; left->z = zx;
      up->x = xy; up->y = yy; up->z = zy;
      front->x = xz; front->y = yz; front->z = zz;
    }
    Vector3 left() const {
      return Vector3( xx, yx, zx );
    }
    Vector3 up() const {
      return Vector3( xy, yy, zy );
    }
    Vector3 front() const {
      return Vector3( xz, yz, zz );
    }
    Matrix3 operator*( Matrix3 const & m ) const {
      return Matrix3( 
          xx*m.xx + yx*m.xy + zx*m.xz, xx*m.yx + yx*m.yy + zx*m.yz, xx*m.zx + yx*m.zy + zx*m.zz, 0,
          xy*m.xx + yy*m.xy + zy*m.xz, xy*m.yx + yy*m.yy + zy*m.yz, xy*m.zx + yy*m.zy + zy*m.zz, 0,
          xz*m.xx + yz*m.xy + zz*m.xz, xz*m.yx + yz*m.yy + zz*m.yz, xz*m.zx + yz*m.zy + zz*m.zz, 0 );
    }
    Vector3 operator*( Vector3 const & v ) const {
      return Vector3( v.x * xx + v.y * xy + v.z * xz, v.x * yx + v.y * yy + v.z * yz, v.x * zx + v.y * zy + v.z * zz );
    }
    static Matrix3 Identity;
    static Matrix3 Null;
};
class Matrix4 : public Matrix3 {
  public:
    Matrix4( Matrix3 const & m3 ) : Matrix3( m3 ), xw(0), yw(0), zw(0), ww(1) {}
    Matrix4() {}
    Matrix4( Matrix3 const & m3, Vector4 const & v4 ) : Matrix3( m3 ), 
      xw(v4.x), yw(v4.y), zw(v4.z), ww(v4.w) {}
    Matrix4( Vector3 const & axis, float angle ) : Matrix3( axis, angle ), xw(0), yw(0), zw(0), ww(1) {}
    Matrix4( 
        float _00, float _01, float _02, float _03,
        float _10, float _11, float _12, float _13,
        float _20, float _21, float _22, float _23,
        float _30, float _31, float _32, float _33 ) :
        Matrix3( _00, _01, _02, _03, _10, _11, _12, _13, _20, _21, _22, _23 ),
        xw(_30), yw(_31), zw(_32), ww(_33) {}

    union {
      struct {
        float xw, yw, zw, ww;
      };
      struct {
        float e30, e31, e32, e33;
      };
    };

    Vector3 & pos() { return (Vector3 &)xw; }
    Vector3 const & pos() const { return (Vector3 const &)xw; }
    Vector3 operator*( Vector3 const & v ) {
      return Vector3( v.x * xx + v.y * xy + v.z * xz + xw, v.x * yx + v.y * yy + v.z * yz + yw, v.x * zx + v.y * zy + v.z * zz + zw );
    }
    Vector4 col( unsigned int ix ) const {
      float const * p = &e00;
      assert( ix < 4 );
      return Vector4( p[ix], p[ix+4], p[ix+8], p[ix+12] );
    }
    Vector4 const & row( unsigned int ix ) const {
      float const * p = &e00;
      assert( ix < 4 );
      return (Vector4 const &)(p[4*ix]);
    }
    Matrix4 operator*( Matrix4 const & o ) const {
      Matrix4 ret;
      float * rp = &ret.xx;
      float const * ap = &xx;
      float const * bp = &o.xx;
      //  this is too much code for me to write out, so I'll do the loop :-)
      for( int r = 0; r < 4; ++r ) {
        for( int c = 0; c < 4; ++c ) {
          float v = 0;
          for( int i = 0; i < 4; ++i ) {
            v += ap[i]*bp[i*4];
          }
          *rp++ = v;
          bp += 1;
        }
        bp -= 4;
        ap += 4;
      }
      return ret;
    }
    static Matrix4 Identity;
    static Matrix4 Null;
};

inline Vector3 operator*( Vector3 const & v, Matrix3 const & m ) {
  return Vector3( 
    v.x*m.xx + v.y*m.xy + v.z*m.xz,
    v.x*m.yx + v.y*m.yy + v.z*m.yz,
    v.x*m.zx + v.y*m.zy + v.z*m.zz );
}
inline Vector3 operator*( Vector3 const & v, Matrix4 const & m ) {
  return Vector3( 
    v.x*m.xx + v.y*m.xy + v.z*m.xz + m.xw,
    v.x*m.yx + v.y*m.yy + v.z*m.yz + m.yw,
    v.x*m.zx + v.y*m.zy + v.z*m.zz + m.zw );
}

inline Vector4 operator*( Vector4 const & v, Matrix4 const & m ) {
  return Vector4( 
    v.x*m.xx + v.y*m.xy + v.z*m.xz + v.w*m.xw,
    v.x*m.yx + v.y*m.yy + v.z*m.yz + v.w*m.yw,
    v.x*m.zx + v.y*m.zy + v.z*m.zz + v.w*m.zw,
    v.x*m.wx + v.y*m.wy + v.z*m.wz + v.w*m.ww );
}

class Bounds {
  public:
    Vector3 low;
    Vector3 high;
    Bounds() : low( Vector3::Null ), high( Vector3::Null ) {}
    Bounds( Vector3 const & v ) : low( v ), high( v ) {}
    Bounds( Vector3 const & v, Vector3 const & w ) : low( v ), high( w ) {}

    void set( Vector3 const & a, Vector3 const & b ) {
      low = a;
      high = b;
    }

    //  Grr... this is really inefficient, unless the compiler is ultra-super-uber-smart!
    //  But it's needed for std::accumulate().
    Bounds operator+( Vector3 const & v ) const {
      return Bounds( Vector3( std::min( v.x, low.x ), std::min( v.y, low.y ), std::min( v.z, low.z ) ),
                     Vector3( std::max( v.x, high.x ), std::max( v.y, high.y ), std::max( v.z, high.z ) ) );
    }
    Bounds & operator+=( Vector3 const & v ) {
      if( v.x < low.x ) low.x = v.x;
      if( v.y < low.y ) low.y = v.y;
      if( v.z < low.z ) low.z = v.z;
      if( v.x > high.x ) high.x = v.x;
      if( v.y > high.y ) high.y = v.y;
      if( v.z > high.z ) high.z = v.z;
      return *this;
    }
};

class OBB {
  public:
    OBB() {}
    OBB( Bounds const & b ) {
      set( b.low, b.high );
    }
    OBB( Bounds const & b, Matrix4 const & m4 ) {
      set( b.low, b.high, m4 );
    }
    void set( Bounds const & b ) {
      set( b.low, b.high );
    }
    void set( Bounds const & b, Matrix4 const & m4 ) {
      set( b.low, b.high, m4 );
    }
    void set( Vector3 const & low, Vector3 const & high ) {
      base = low;
      x = Vector3( high.x-low.x, 0, 0 );
      y = Vector3( 0, high.y-low.y, 0 );
      z = Vector3( 0, 0, high.z-low.z );
    }
    void set( Vector3 const & low, Vector3 const & high, Matrix4 const & m4 ) {
      set( low, high );
      base = base * m4;
      x = x * (Matrix3 const &)m4;
      y = y * (Matrix3 const &)m4;
      z = z * (Matrix3 const &)m4;
    }

    Vector3 base;
    Vector3 x;
    Vector3 y;
    Vector3 z;
};

void RotationMatrixAroundPoint( Matrix4 * out, Vector3 const & up, float amt, Vector3 const & point );

//  Return true if the sphere described has non-zero intersection with the AABB (conservative)
bool PointRadiusIntersectsAABB( Vector3 const & point, float radius, Vector3 const & low, Vector3 const & high );
//  Return true if the oriented bounding box described has non-zero intersection with the AABB (conservative)
bool OBBIntersectsAABB( OBB const & obb, Vector3 const & low, Vector3 const & high );
//  Return true if the AABB described has non-zero intersection with the AABB (conservative)
bool AABBIntersectsAABB( Vector3 const & aLow, Vector3 const & aHigh, Vector3 const & bLow, Vector3 const & bHigh, bool * outFullyWithin = 0 );
//  Return true if the point lies on or within all the planes.
bool PointIntersectsPlanes( Vector3 const & point, Vector4 const * planes, uint numPlanes );
//  Return true if any points intersects the plane
bool AnyPointIntersectsPlane( Vector3 const * points, uint numPoints, Vector4 const & plane, bool * outAllIntersect = 0 );
//  Return true if the sphere described may have non-zero intersection with the planes (conservative).
bool PointRadiusIntersectsPlanes( Vector3 const & point, float radius, Vector4 const * planes, uint numPlanes );
//  Return true if the oriented bounding box described may have non-zero intersection with the planes (conservative).
bool OBBIntersectsPlanes( OBB const & obb, Vector4 const * planes, uint numPlanes );
//  Return true if the AABB described may have non-zero intersection with the planes (conservative).
bool AABBIntersectsPlanes( Vector3 const & low, Vector3 const & high, Vector4 const * planes, uint numPlanes, bool * outFullyWithin = 0 );
//  Extract the left, right, bottom, top, near and far clipping planes for a view frustum
//  defined by a modelview and a projection matrix.
void ExtractFrustumPlanes( Matrix4 const & modelView, Matrix4 const & projection, Vector4 planes[6] );

#endif  //  mymath_h
