

#include "mymine.h"
#include "mytweakables.h"
#include "mycamera.h"
#include "myweapon.h"
#include "misc.h"
#include "glshaderstate.h"
#include "gl-ase.h"


std::vector< MySystem * > MyMine::activeMines_;


namespace {

  class MineDropper : public I_Weapon {
    public:
      MineDropper();
      ~MineDropper();

      void renderSight( float screenWidth, float screenHeight, Matrix4 const & camMtx, Matrix4 const & projMtx );
      void setAim( Matrix3 const & aim, Vector3 const & source, Vector3 const & ptVel );
      bool readyToFire( double );
      bool fire( double );
      void renderIndicators( float phi, float alpha );

      GLShader * reticleShader_;
      GLShader * indicatorShader_;
      GLShader * ammoShader_;
      I_GL_ASE * ammoModel_;
      Matrix3 ori_;
      Vector3 pos_;
      Vector3 vel_;
      double lastFired_;
  };
}


I_Weapon * NewMineWeapon()
{
  return new MineDropper();
}

MineDropper::MineDropper()
{
  reticleShader_ = gShaderFactory->newShader( "mineReticle" );
  indicatorShader_ = gShaderFactory->newShader( "ammoIndicator" );
  ammoShader_ = gShaderFactory->newShader( "mineShader" );
  ammoModel_ = ASEFromFile( asePath( "mineAmmo" ) );
  ori_ = Matrix3::Identity;
  pos_ = Vector3::Null;
  lastFired_ = 0;
}

MineDropper::~MineDropper()
{
  reticleShader_->dispose();
  indicatorShader_->dispose();
  ammoShader_->dispose();
  ammoModel_->dispose();
}

void MineDropper::renderSight( float screenWidth, float screenHeight, Matrix4 const & camMtx, Matrix4 const & projMtx )
{
  int w = 32;
  if( screenWidth >= 1000 && screenHeight >= 750 ) {
    w = 32;
  }
  else {
    w = 16;
  }

  //  figure out where we'd be 1 meters down from the dropper
  Vector4 pos = Vector4(pos_ - ori_.up() * 1.0f) * camMtx * projMtx;
  pos *= 1.0f / pos.w;
  float midX = (pos.x + 1) * screenWidth/2;
  float midY = (pos.y + 1) * screenHeight/2;

  reticleShader_->apply();
  glBegin( GL_QUADS );
  glTexCoord2f( 0, 0 ); glVertex2f( midX-w, midY-w );
  glTexCoord2f( 1, 0 ); glVertex2f( midX+w, midY-w );
  glTexCoord2f( 1, 1 ); glVertex2f( midX+w, midY+w );
  glTexCoord2f( 0, 1 ); glVertex2f( midX-w, midY+w );
  glEnd();
}

void MineDropper::setAim( Matrix3 const & aim, Vector3 const & source, Vector3 const & vel )
{
  ori_ = aim;
  pos_ = source;
  vel_ = vel;
}

bool MineDropper::readyToFire( double time )
{
  return time > (lastFired_ + MINE_REFIRE_TIME);
}

bool MineDropper::fire( double time )
{
  if( !readyToFire( time ) ) {
    return false;
  }
  lastFired_ = time;
  MyMine * g = new MyMine( pos_, ammoShader_, ammoModel_ );
  return true;
}

void MineDropper::renderIndicators( float phi, float alpha )
{
  assert( alpha >= 0 );
  assert( phi > -3.1416f && phi < 6.2832f );
  glRotatef( 120 * phi, 0, 1, 0 );
  glTranslatef( 0, 0, -0.2f );
  indicatorShader_->context().color.alpha_ = (alpha > 1) ? 0x80 : (uchar)(0x80 * alpha);
  indicatorShader_->apply();
  ammoModel_->issueAll();
}







void MyMine::preStepAll()
{
  std::for_each( activeMines_.begin(), activeMines_.end(), MySystem::preStepOne );
}

void MyMine::postStepAll()
{
  std::for_each( activeMines_.begin(), activeMines_.end(), MySystem::postStepOne );
}

void MyMine::renderAll()
{
  std::for_each( activeMines_.begin(), activeMines_.end(), MySystem::renderOne );
}

MyMine::MyMine( Vector3 const & pos, GLShader * shader, I_GL_ASE * mesh )
{
  explodeParams_.radius_ = MINE_RADIUS;
  explodeParams_.force_ = MINE_EXPLODE_FORCE;
  dBodyID tempBody = dBodyCreate( world() );
  dMass m;
  dMassSetBoxTotal( &m, 0.1f, 0.3f, 0.1f, 0.3f );
  dBodySetMass( tempBody, &m );
  dBodySetPosition( tempBody, pos.x, pos.y, pos.z );
  addBody( tempBody, shader, mesh );
  dGeomID tempGeom = dCreateBox( worldSpace(), 0.3f, 0.1f, 0.3f );
  addGeom( tempGeom, tempBody );
  time_ = MINE_FUSE_TIME;
  activeMines_.push_back( this );
  cameraCut_ = false;
}

MyMine::~MyMine()
{
  activeMines_.erase( std::find( activeMines_.begin(), activeMines_.end(), this ) );
}

void MyMine::preStep()
{
  //  mine
  time_ -= TIME_STEP;
  if( time_ < 0.7f && !cameraCut_ ) {
    cameraCut_ = true;
    CameraSpecialLook( 2.5f, P(), P() + Vector3( 10, 20, -20 ) );
  }
  if( time_ <= 0 ) {
    explodeParams_.pos_ = P();
    explode();
  }
}

bool MyMine::wantsContact( dContact & c, MySystemBase * other, dGeomID you, dGeomID him, bool firstTest )
{
  if( other && other->physical() && dGeomGetBody( him ) ) { //  mines only stick to static environment, but not the camera
    return false;
  }
  return true;
}
