

#include "myparticles.h"
#include "myfilesystem.h"
#include "xmlscanner.h"
#include "xmlutil.h"
#include "misc.h"

#include <assert.h>
#include <algorithm>
#include <new>
#include <string>
#include <map>
#include <vector>



dGeomID MyParticles::particleRay_;

std::vector< MyParticles * > MyParticles::allParticleSystems_;
std::vector< MyParticles * > MyParticles::toRemove_;


MyParticles::MyParticles( Vector3 const & pos, ParticleSettings const * ps, dSpaceID space )
{
  emitting_ = true;
  assert( (ps->particleCount != 0) == ps->recycleParticles );
  *(ParticleSettings *)this = *ps;
  basePos_ = pos;
  baseParticle_ = 0;
  liveParticles_ = 0;
  particlesEmitted_ = 0;
  particleCount_ = maxParticlesToEmit;
  if( particleCount ) {
    particleCount_ = particleCount;
  }
  particles_ = new Particle[ particleCount_ ];
  memset( particles_, 0, sizeof( Particle ) * particleCount_ );
  particleBuffer_ = 0;
  particleFlux_ = 0;

  if( !particleRay_ ) {
    particleRay_ = dCreateRay( space, 1 );
  }
  allParticleSystems_.push_back( this );
}

MyParticles::~MyParticles()
{
  std::vector< MyParticles * >::iterator me = 
      std::find( allParticleSystems_.begin(), allParticleSystems_.end(), this );
  assert( me != allParticleSystems_.end() );
  allParticleSystems_.erase( me );
  if( particleBuffer_ ) {
    glDeleteBuffersARB( 1, &particleBuffer_ );
  }
  delete[] particles_;
}

void MyParticles::setEmitterPos( Vector3 const & pos, Vector3 const (& basis)[3] )
{
  basePos_ = pos;
  emitterVectors[0] = basis[0];
  emitterVectors[1] = basis[1];
  emitterVectors[2] = basis[2];
}

void MyParticles::setEmitterAlpha( float nuAlpha )
{
  if( nuAlpha < 0 ) {
    emitterAlpha = 0;
  }
  else if( nuAlpha > 1 ) {
    emitterAlpha = 1;
  }
  else {
    emitterAlpha = nuAlpha;
  }
}

void MyParticles::setEmitterRate( float nuRate )
{
  if( nuRate < 0 ) {
    emitterRate = 0;
  }
  else {
    assert( nuRate < 1000 );
    if( !(nuRate < 1000) ) {
      nuRate = 1000;
    }
    emitterRate = nuRate;
  }
}

namespace {
  class StepParticleSystem {
    public:
      float dt_;
      StepParticleSystem( float dt ) : dt_( dt ) {}
      void operator()( MyParticles * mp ) {
        mp->step( dt_ );
      }
  };
  class RenderParticleSystem {
    public:
      void operator()( MyParticles * mp ) {
        mp->render();
      }
  };
  class DeleteParticleSystem {
    public:
      void operator()( MyParticles * mp ) {
        delete mp;
      }
  };
}

void MyParticles::stepAll( float delta )
{
  std::for_each( allParticleSystems_.begin(), allParticleSystems_.end(), StepParticleSystem( delta ) );
  std::for_each( toRemove_.begin(), toRemove_.end(), DeleteParticleSystem() );
  toRemove_.clear();
}

void MyParticles::renderAll()
{
  //todo:  sort systems from far to near
  //todo:  merge with other transparency sorting
  glDepthMask( GL_FALSE );
  std::for_each( allParticleSystems_.begin(), allParticleSystems_.end(), RenderParticleSystem() );
  glDepthMask( GL_TRUE );
}

bool MyParticles::emit( Vector3 const & pos, Vector3 const & vel )
{
  size_t nu = (baseParticle_ + liveParticles_) % particleCount_;
  if( liveParticles_ == particleCount_ ) {
    if( !recycleParticles ) {
      return false;   //  sorry, list full!
    }
    killParticle( particles_[nu] );
    --liveParticles_;
    baseParticle_ = (baseParticle_+1) % particleCount_;
  }
  initParticle( particles_[nu], pos, vel );
  ++liveParticles_;
  ++particlesEmitted_;
  return true;
}

bool MyParticles::step( float delta )
{
  //  emit, if rate says we should
  if( delta > 0.1f ) {
    delta = 0.1f;
  }
  if( emitting_ ) {
    particleFlux_ += delta * emitterRate;
    while( (particleFlux_ >= 1) && (!maxParticlesToEmit || (particlesEmitted_ < maxParticlesToEmit)) ) {
      float v1 = (rand() & 0x7fff)/32767.0f;
      float v2 = (rand() & 0x7fff)/32767.0f;
      float v3 = (rand() & 0x7fff)/32767.0f;
      Vector3 vel = 
        emitterVectors[0] * (emitterMinMax[0][0] +  v1 * (emitterMinMax[0][1] - emitterMinMax[0][0])) +
        emitterVectors[1] * (emitterMinMax[1][0] +  v2 * (emitterMinMax[1][1] - emitterMinMax[1][0])) +
        emitterVectors[2] * (emitterMinMax[2][0] +  v3 * (emitterMinMax[2][1] - emitterMinMax[2][0]));
      emit( basePos_, vel );
      particleFlux_ -= 1;
    }
  }

  //  powf() is very expensive, so only calculate dampening once per op
  //  rather than once per particle. This forces us to use un-clean 
  //  storage in the description of the particle ops.
  //  Todo: remove un-cleanliness by copying particleOps when initializing 
  //  from data -- when initializing from XML, we could fill in expanded 
  //  struct directly.
  for( size_t i = 0; i < numParticleOps; ++i ) {
    (const_cast<ParticleOp &>(particleOps[i])).temp = powf( particleOps[i].speedDampening, delta );
    (const_cast<ParticleOp &>(particleOps[i])).spinS = sinf( delta * particleOps[i].spin );
    (const_cast<ParticleOp &>(particleOps[i])).spinC = cosf( delta * particleOps[i].spin );
  }

  //  evolve particles, possibly killing them
  size_t killed = 0;
  for( size_t i = 0; i < liveParticles_; ++i ) {

    Particle & p = particles_[(baseParticle_+i) % particleCount_];

    //  This evolution loop is a morass of branches.
    //  At least I don't run a ray collision per particle!
    //  No, wait...

    //  if killing particle, swap with particles_[killed+...] and increment base
    if( p.curOpDuration <= 0 ) {
      p.curOp++;
      if( p.curOp >= particleOps+numParticleOps ) {
        //  Kill particle; replace with some other particle that may be live, 
        //  but has already been iterated in that case.
        p = particles_[(baseParticle_+killed) % particleCount_];
        ++killed;
        continue;
      }
      p.curOpDuration = p.curOp->maxDuration;
    }

    //  duration
    float d = delta;
    if( d > p.curOpDuration ) {
      d = p.curOpDuration;
    }
    p.curOpDuration -= d;

    //  alpha
    float a = p.curOp->alphaDelta * d;
    p.alpha += a;
    if( (a > 0) ? (p.alpha > p.curOp->alphaTarget) : (p.alpha < p.curOp->alphaTarget) ) {
      p.alpha = p.curOp->alphaTarget;
      p.curOpDuration = 0;  //  will advance next time
    }

    //  size
    float s = (p.curOp->sizeDelta + sqrtf(p.vel.lengthSquared())*p.curOp->sizeDeltaSpeedWeight) * d;
    p.size += s;
    if( (s > 0) ? (p.size > p.curOp->sizeTarget) : (p.size < p.curOp->sizeTarget) ) {
      p.size = p.curOp->sizeTarget;
      p.curOpDuration = 0;  //  will advance next time
    }

    //  spin
    float x = p.xPhase, y = p.yPhase;
    p.xPhase = (x * p.curOp->spinC - y * p.curOp->spinS);
    p.yPhase = (x * p.curOp->spinS + y * p.curOp->spinC);

    //  todo: run intersection test

    //  position & velocity
    p.pos += p.vel * d;
    p.vel *= p.curOp->temp;
    p.vel += p.curOp->deltaVperSec * d;
  }

  baseParticle_ = (baseParticle_ + killed) % particleCount_;
  assert( liveParticles_ >= killed );
  liveParticles_ -= killed;

  if( autoDeath && liveParticles_ == 0 && particlesEmitted_ > 0 ) {
    toRemove_.push_back( this );
    return false;
  }
  return liveParticles_ > 0;
}

void MyParticles::render()
{
  //todo: it would be rad if I could sort non-additive particles far to near
  //      -- it wouldn't be THAT slow!
  //todo: when the emitter moves around, I should calculate a new center 
  //      of the particles for sorting
  assert( !glGetError() );

  float mvm[16];
  glGetFloatv( GL_MODELVIEW_MATRIX, mvm );

  void * mem = 0;
  float * v = 0;
  static bool useMapping = true;
  if( useMapping ) {
    if( particleBuffer_ == 0 ) {
      glGenBuffersARB( 1, &particleBuffer_ );
    }
    glBindBufferARB( GL_ARRAY_BUFFER_ARB, particleBuffer_ );
    glBufferDataARB( GL_ARRAY_BUFFER_ARB, liveParticles_*sizeof(float)*6*4, 0, GL_DYNAMIC_DRAW_ARB );
    v = (float *)glMapBufferARB( GL_ARRAY_BUFFER_ARB, GL_WRITE_ONLY_ARB );
  }

  assert( !glGetError() );

  if( !v ) {
    mem = malloc( sizeof( float )*6*4*liveParticles_ );
    v = (float *)mem;
  }
  Vector3 r = Vector3( mvm[0], mvm[4], mvm[8] );   //  fixme: take this from camera right
  Vector3 u = Vector3( mvm[1], mvm[5], mvm[9] );      //  fixme: take this from camera up
  Vector3 a = r - u;
  Vector3 b = r + u;
  for( size_t i = 0; i < liveParticles_; ++i ) {
    Particle & p = particles_[(baseParticle_+i) % particleCount_];
    uint32 color;
    if( p.alpha < 0 ) {
      color = 0x00ffffff;
    }
    else if( p.alpha > 1 ) {
      color = 0xffffffff;
    }
    else {
      color = 0xffffff | ((uint32)(0xff * p.alpha) << 24);
    }
#if 1
    *(Vector3 *)v = p.pos + (a * p.yPhase - b * p.xPhase) * p.size;
    v += 3; *v++ = 0; *v++ = 0;   //  u, v
    *(uint32*)(v++) = color;
    *(Vector3 *)v = p.pos + (b * p.yPhase + a * p.xPhase) * p.size;
    v += 3; *v++ = 1; *v++ = 0;   //  u, v
    *(uint32*)(v++) = color;
    *(Vector3 *)v = p.pos + (a * (-p.yPhase) + b * p.xPhase) * p.size;
    v += 3; *v++ = 1; *v++ = 1;   //  u, v
    *(uint32*)(v++) = color;
    *(Vector3 *)v = p.pos + (b * (-p.yPhase) - a * p.xPhase) * p.size;
    v += 3; *v++ = 0; *v++ = 1;   //  u, v
    *(uint32*)(v++) = color;
#else
    *(Vector3 *)v = p.pos + right * -p.size + up * -p.size;
    v += 3; *v++ = 0; *v++ = 0;   //  u, v
    *(uint32*)(v++) = color;
    *(Vector3 *)v = p.pos + right * p.size + up * -p.size;
    v += 3; *v++ = 1; *v++ = 0;   //  u, v
    *(uint32*)(v++) = color;
    *(Vector3 *)v = p.pos + right * p.size + up * p.size;
    v += 3; *v++ = 1; *v++ = 1;   //  u, v
    *(uint32*)(v++) = color;
    *(Vector3 *)v = p.pos + right * -p.size + up * p.size;
    v += 3; *v++ = 0; *v++ = 1;   //  u, v
    *(uint32*)(v++) = color;
#endif
  }
  if( !mem ) {
    glUnmapBufferARB( GL_ARRAY_BUFFER_ARB );
  }

  assert( !glGetError() );

  glEnableClientState( GL_VERTEX_ARRAY );
  glEnableClientState( GL_TEXTURE_COORD_ARRAY );
  glDisableClientState( GL_NORMAL_ARRAY );
  glEnableClientState( GL_COLOR_ARRAY );

  assert( !glGetError() );

  glVertexPointer( 3, GL_FLOAT, 6*sizeof(float), (char *)mem + 0 );
  glTexCoordPointer( 2, GL_FLOAT, 6*sizeof(float), (char *)mem + 3*sizeof(float) );
  glColorPointer( 4, GL_UNSIGNED_BYTE, 6*sizeof(float), (char *)mem + 5*sizeof(float) );
  material->apply();
  glDrawArrays( GL_QUADS, 0, static_cast< GLsizei >( liveParticles_*4 ) );

  assert( !glGetError() );

  if( useMapping ) {
    glBindBufferARB( GL_ARRAY_BUFFER_ARB, 0 );
  }
  if( mem ) {
    free( mem );
  }
  glDisableClientState( GL_COLOR_ARRAY );

  assert( !glGetError() );
}

void MyParticles::killParticle( Particle & p )
{
  p.curOp = 0;
}

void MyParticles::initParticle( Particle & p, Vector3 const & pos, Vector3 const & vel )
{
  p.curOp = particleOps;
  p.curOpDuration = p.curOp->maxDuration;
  p.pos = pos;
  p.vel = vel;
  p.alpha = emitterAlpha + emitterAlphaVariation * (rand() & 0x7fff) / 32767.0f;
  p.size = emitterSize;
  float phi = (rand() & 0x7fff)/32767.0f * (emitterSpinMax-emitterSpinMin) + emitterSpinMin;
  p.xPhase = sinf( phi );
  p.yPhase = cosf( phi );
}


namespace {
  struct LoadedParticles {
    enum {
      MAX_OPS = 8,
    };
    ParticleSettings settings_;
    ParticleOp ops_[MAX_OPS];
    LoadedParticles() {
      settings_.particleOps = ops_;
      settings_.material = 0;
    }
    ~LoadedParticles() {}
    LoadedParticles( LoadedParticles const & o ) {
      settings_ = o.settings_;
      std::copy( o.ops_, &o.ops_[MAX_OPS], ops_ );
      settings_.particleOps = ops_;
    }
    LoadedParticles & operator=( LoadedParticles const & o ) {
      if( &o != this ) {
        this->~LoadedParticles();
        new( this ) LoadedParticles( o );
      }
      return *this;
    }
  };
  
  std::map< std::string, LoadedParticles * > loadedParticles_;

  bool XMLReadMaterial( XMLNode * n, GLShader * & material, char const * name )
  {
    size_t ix = 0;
    XMLNode * c;
    std::string shader = "particleShader";
    TextureRef tr( "media/missing.dds" );
    GLenum blendSrc = GL_ONE;
    GLenum blendDst = GL_ZERO;
    while( (c = n->child( ix )) != 0 ) {
      if( c->token().kind == XMLScanner::Token::Attribute ) {
        if( c->token().keyIs( "shader" ) ) {
          shader = std::string( c->token().value, c->token().valueSize );
        }
        else if( c->token().keyIs( "texture" ) ) {
          std::string tName( c->token().value, c->token().valueSize );
          if( !strchr( tName.c_str(), '.' ) ) {
            char path[128];
            if( FS::translateHandle( tName.c_str(), path, sizeof(path) ) ) {
              tName = path;
            }
          }
          tr = TextureRef( tName.c_str() );
        }
        else if( c->token().keyIs( "blend" ) ) {
          std::string bName( c->token().value, c->token().valueSize );
          if( bName == "modulate" ) {
            blendSrc = GL_ZERO;
            blendDst = GL_SRC_COLOR;
          }
          else if( bName == "add" ) {
            blendSrc = GL_ONE;
            blendDst = GL_ONE;
          }
          else if( bName == "blend" ) {
            blendSrc = GL_SRC_ALPHA;
            blendDst = GL_ONE_MINUS_SRC_ALPHA;
          }
          else if( bName == "premult" ) {
            blendSrc = GL_ONE;
            blendDst = GL_ONE_MINUS_SRC_ALPHA;
          }
          else if( bName == "replace" ) {
            blendSrc = GL_ONE;
            blendDst = GL_ZERO;
          }
          else {
            warning( "Particle material blend '%.*s' is not recognized; assuming 'replace'.\n" );
          }
        }
        else {
          warning( "Particle material attribute '%.*s' ignored in '%s'\n", 
              c->token().keySize, c->token().key, name );
        }
      }
      else if( c->token().kind == XMLScanner::Token::Tag ) {
        warning( "Particle material sub-tag '%.*s' ignored in '%s'\n",
            c->token().keySize, c->token().key, name );
      }
      ++ix;
    }
    material = gShaderFactory->newShader( shader.c_str() );
    material->setTexture( "diffuseMap", tr );
    if( blendDst != GL_ZERO ) {
      material->context().blendSrc = blendSrc;
      material->context().blendDst = blendDst;
      material->context().blend = true;
    }
    return true;
  }

  bool XMLReadVectorParams( XMLNode * n, float * minMax, Vector3 & vec )
  {
    if( !XMLUtil::ToVector( n, vec ) ) return false;
    XMLNode * c;
    size_t ix = 0;
    while( (c = n->child( ix )) != 0 ) {
      if( c->token().kind == XMLScanner::Token::Attribute ) {
        if( c->token().keyIs( "min" ) ) {
          XMLUtil::ToNumber( c, minMax[0] );
        }
        else if( c->token().keyIs( "max" ) ) {
          XMLUtil::ToNumber( c, minMax[1] );
        }
      }
      ++ix;
    }
    return true;
  }

/*

 <particles 
    emitterRate="" emitterSize="" emitterAlpha="" emitterAlphaVariation="" 
    emitterSpinMin="" emitterSpinMax="" 
    maxParticlesToEmit="" particleCount="" 
    dieOnContact="" recycleParticles="" autoDeath="">
   <phase alphaDelta="" alphaTarget="" sizeDelta="" sizeTarget="" sizeDeltaSpeedWeight="" speedDampening="" maxDuration="">
     <deltaVperSec> x,y,z </deltaVperSec>
   </phase>
   <material texture="" blend="" />
   <emitterVectors0 min="" max=""> x,y,z </emitterVectors0>
   <emitterVectors1 min="" max=""> x,y,z </emitterVectors1>
   <emitterVectors2 min="" max=""> x,y,z </emitterVectors2>
 </particles>
 */

  LoadedParticles * ExtractLoadedParticlesFromText( Data const & d, char const * name )
  {
    XMLDom dom( d.readBase(), d.size() );
    if( dom.error() ) {
      error( "ExtractLoadedParticlesFromText(%s): %s\n", name, dom.error() );
      return 0;       
    }
    XMLNode * n = dom.root();
    if( !n ) {
      error( "Error extracting XML data from '%s'\n", name );
      return 0;
    }
    //  I only pay attention to the first top-level element in the file
    XMLNode * c;
    size_t ix = 0;
    size_t curPhase = 0;
    LoadedParticles * lp = new LoadedParticles;
    while( (c = n->child( ix )) != 0 ) {
      if( c->token().kind == XMLScanner::Token::Attribute ) {
        //  attribute of particle system

#define MAYBEGET(c,lp,n) \
  if( c->token().keyIs(#n) ) { \
    XMLUtil::ToNumber(c,lp->settings_.n); \
    goto done_attr; \
  }
#define MAYBEGETB(c,lp,n) \
  if( c->token().keyIs(#n) ) { \
    XMLUtil::ToBool(c,lp->settings_.n); \
    goto done_attr; \
  }
        MAYBEGET(c,lp,maxParticlesToEmit);
        MAYBEGET(c,lp,emitterRate);
        MAYBEGET(c,lp,emitterSize);
        MAYBEGET(c,lp,emitterAlpha);
        MAYBEGET(c,lp,emitterAlphaVariation);
        MAYBEGET(c,lp,emitterSpinMin);
        MAYBEGET(c,lp,emitterSpinMax);
        MAYBEGET(c,lp,particleCount);
        MAYBEGETB(c,lp,dieOnContact);
        MAYBEGETB(c,lp,recycleParticles);
        MAYBEGETB(c,lp,autoDeath);
#undef MAYBEGET
        warning( "Unknown attribute '%.*s' in XML particle description '%s'.\n", c->token().keySize, c->token().key, name );
      }
      else if( c->token().kind == XMLScanner::Token::Tag ) {
        if( c->token().keyIs( "phase" ) ) {
          if( lp->settings_.numParticleOps == LoadedParticles::MAX_OPS ) {
            warning( "Too many phases in XML particle description '%s' (max is %d)\n", 
                name, LoadedParticles::MAX_OPS );
            goto done_phase;
          }
          ParticleOp * op = &lp->ops_[lp->settings_.numParticleOps];
          XMLNode * p;
          size_t cix = 0;
          while( (p = c->child( cix )) != 0 ) {
            if( p->token().kind == XMLScanner::Token::Attribute ) {
              //  attribute of phase
#define MAYBEGET(p,op,n) \
  if( p->token().keyIs(#n) ) { \
    XMLUtil::ToNumber(p,op->n); \
    goto done_phase_attr; \
  }
              MAYBEGET(p,op,alphaDelta);
              MAYBEGET(p,op,alphaTarget);
              MAYBEGET(p,op,sizeDelta);
              MAYBEGET(p,op,sizeTarget);
              MAYBEGET(p,op,spin);
              MAYBEGET(p,op,maxDuration);
              MAYBEGET(p,op,speedDampening);
              MAYBEGET(p,op,sizeDeltaSpeedWeight);
#undef MAYBEGET
            }
            else if( p->token().kind == XMLScanner::Token::Tag ) {
              if( p->token().keyIs( "deltaVperSec" ) ) {
                XMLUtil::ToVector(p,op->deltaVperSec);
              }
              else {
                warning( "XML Particles phase %d tag '%.*s' is unknown in '%s'.\n",
                    lp->settings_.numParticleOps, p->token().keySize, p->token().key, name );
              }
            }
            else {
              debug( "XML Particles '%s': Ignoring token type %d (%.*s:%.*s)\n", name, c->token().kind, 
                  c->token().keySize, c->token().key, c->token().valueSize, c->token().value );
            }
done_phase_attr:;
            ++cix;
          }
          ++lp->settings_.numParticleOps;
done_phase:;
        }
        else if( c->token().keyIs( "material" ) ) {
          XMLReadMaterial( c, lp->settings_.material, name );
        }
        else if( c->token().keyIs( "emitterVectors0" ) ) {
          XMLReadVectorParams( c, lp->settings_.emitterMinMax[0], lp->settings_.emitterVectors[0] );
        }
        else if( c->token().keyIs( "emitterVectors1" ) ) {
          XMLReadVectorParams( c, lp->settings_.emitterMinMax[1], lp->settings_.emitterVectors[1] );
        }
        else if( c->token().keyIs( "emitterVectors2" ) ) {
          XMLReadVectorParams( c, lp->settings_.emitterMinMax[2], lp->settings_.emitterVectors[2] );
        }
        else {
          warning( "Unknown tag in XML particle data '%s': %.*s\n", c->token().keySize, c->token().key, name );
        }
      }
      else {
        debug( "XML Particles '%s': Ignoring token type %d (%.*s:%.*s)\n", name, c->token().kind, 
            c->token().keySize, c->token().key, c->token().valueSize, c->token().value );
        //  ignore this
      }
done_attr:
      ++ix;
    }
    return lp;
  }
  //todo: if root has siblings, we should load one particle system per each, 
  //  and return an amalgamation of all of them. Think fire + smoke + kick-up.

  MyParticles * MakeParticleSystem( LoadedParticles const & lp, dSpaceID space )
  {
    MyParticles * ret = new MyParticles( Vector3::Null, &lp.settings_, space );
    return ret;
  }
}

MyParticles * MakeParticleSystem( char const * handleName, dSpaceID space )
{
  std::string n( handleName );
  std::map< std::string, LoadedParticles * >::iterator iter = loadedParticles_.find( handleName );
  if( iter != loadedParticles_.end() ) {
    return MakeParticleSystem( *(*iter).second, space );
  }
  Data d;
  if( !FS::readFileByHandle( handleName, &d ) ) {
    warning( "Could not find particle system handle '%s'\n", handleName );
    return 0;
  }
  LoadedParticles * lp;
  if( !(lp = ExtractLoadedParticlesFromText( d, handleName )) ) {
    warning( "Could not extract particle information from file for handle '%s'\n", handleName );
    return 0;
  }
  loadedParticles_[n] = lp;
  return MakeParticleSystem( *lp, space );
}

