
#include "myrocket.h"
#include "mytweakables.h"
#include "myweapon.h"
#include "gl-ase.h"


std::vector< MySystem * > MyRocket::rockets_;


namespace {
  //  todo: factor the commonalities between weapons
  class RocketLauncher : public I_Weapon {
    public:
      RocketLauncher();
      ~RocketLauncher();

      void renderSight( float screenWidth, float screenHeight, Matrix4 const & camMtx, Matrix4 const & projMtx );
      void setAim( Matrix3 const & aim, Vector3 const & source, Vector3 const & vel );
      bool readyToFire( double );
      bool fire( double );
      void renderIndicators( float phi, float alpha );

      GLShader * reticleShader_;
      GLShader * indicatorShader_;
      GLShader * ammoShader_;
      I_GL_ASE * ammoModel_;
      Matrix3 ori_;
      Vector3 pos_;
      Vector3 vel_;
      double lastFired_;
  };
}


I_Weapon * NewRocketWeapon()
{
  return new RocketLauncher;
}



RocketLauncher::RocketLauncher()
{
  reticleShader_ = gShaderFactory->newShader( "rocketReticle" );
  indicatorShader_ = gShaderFactory->newShader( "ammoIndicator" );
  ammoShader_ = gShaderFactory->newShader( "rocketShader" );
  ammoModel_ = ASEFromFile( asePath( "rocketAmmo" ) );
  ori_ = Matrix3::Identity;
  pos_ = Vector3::Null;
  lastFired_ = 0;
}

RocketLauncher::~RocketLauncher()
{
  reticleShader_->dispose();
  indicatorShader_->dispose();
  ammoShader_->dispose();
  ammoModel_->dispose();
}

void RocketLauncher::renderSight( float screenWidth, float screenHeight, Matrix4 const & camMtx, Matrix4 const & projMtx )
{
  int w = 32;
  if( screenWidth >= 1000 && screenHeight >= 750 ) {
    w = 32;
  }
  else {
    w = 16;
  }

  //  figure out where we'd be 100 meters from the launcher
  Vector4 pos = Vector4(pos_ + ori_.front() * 100) * camMtx * projMtx;
  pos *= 1.0f / pos.w;
  float midX = (pos.x + 1) * screenWidth/2;
  float midY = (pos.y + 1) * screenHeight/2;

  reticleShader_->apply();
  glBegin( GL_QUADS );
  glTexCoord2f( 0, 0 ); glVertex2f( midX-w, midY-w );
  glTexCoord2f( 1, 0 ); glVertex2f( midX+w, midY-w );
  glTexCoord2f( 1, 1 ); glVertex2f( midX+w, midY+w );
  glTexCoord2f( 0, 1 ); glVertex2f( midX-w, midY+w );
  glEnd();
}

void RocketLauncher::setAim( Matrix3 const & aim, Vector3 const & source, Vector3 const & vel )
{
  ori_ = aim;
  pos_ = source;
  vel_ = vel;
}

bool RocketLauncher::readyToFire( double time )
{
  return time > (lastFired_ + ROCKET_REFIRE_TIME);
}

bool RocketLauncher::fire( double time )
{
  if( !readyToFire( time ) ) {
    return false;
  }
  lastFired_ = time;
  MyRocket * g = new MyRocket( pos_, ori_, vel_, ammoShader_, ammoModel_ );
  return true;
}

void RocketLauncher::renderIndicators( float phi, float alpha )
{
  assert( alpha >= 0 );
  assert( phi > -3.1416f && phi < 6.2832f );
  glRotatef( 120 * phi, 0, 1, 0 );
  glTranslatef( 0, 0, -0.2f );
  indicatorShader_->context().color.alpha_ = (alpha > 1) ? 0x80 : (uchar)(0x80 * alpha);
  indicatorShader_->apply();
  ammoModel_->issueAll();
}



extern "C" {
  static void RocketRayCallback( void * gr, dGeomID g1, dGeomID g2 )
  {
    assert( dGeomGetClass( g2 ) == dRayClass );
    if( dGeomIsSpace( g1 ) ) {
      dSpaceCollide2( g1, g2, gr, RocketRayCallback );
    }
    else {
      dContactGeom cg;
      MyRocket * theR = (MyRocket *)gr;
      if( !theR->hasGeom( g1 ) && dCollide( g2, g1, 1, &cg, sizeof( cg ) ) ) {
        MySystem * s1 = (MySystem *)dGeomGetData( g2 );
        if( !s1 || s1->physical() ) {
          theR->exploding_ = true;
        }
      }
    }
  }
}

static SoundHandle launchSound_;

MyRocket::MyRocket( Vector3 const & pos, Matrix3 const & ori, Vector3 const & ptVel, GLShader * shader, I_GL_ASE * mesh )
{
  if( launchSound_ == 0 ) {
    launchSound_ = gSoundManager->loadSound( wavPath( "rocket" ) );
  }
  if( launchSound_ ) {
    gSoundManager->playInstance( launchSound_, ROCKET_VOLUME );
  }
  age_ = 0;
  explodeParams_.radius_ = ROCKET_EXPLODE_RADIUS;
  explodeParams_.force_ = ROCKET_EXPLODE_FORCE;
  exploding_ = false;
  dBodyID tempBody = dBodyCreate( world() );
  dMass m;
  dMassSetSphereTotal( &m, 1, 0.1f );
  dBodySetMass( tempBody, &m );
  Matrix3 const m3b = ori.transpose();
  dBodySetRotation( tempBody, &m3b.xx );
  Vector3 left, up, front;
  ori.leftUpFront( &left, &up, &front );
  dBodySetPosition( tempBody, pos.x, pos.y, pos.z );
  oldPos_ = pos;
//  Vector3 const & vel = (Vector3 const &)*dBodyGetLinearVel( chas ) + front * ROCKET_SPEED;
//  dBodySetLinearVel( tempBody, vel.x, vel.y, vel.z );
  Vector3 vel = ptVel + front*2;
  dBodySetLinearVel( tempBody, vel.x, vel.y, vel.z );
  dBodySetGravityMode( tempBody, 0 );
  addBody( tempBody, shader, mesh );
  dGeomID g = dCreateCCylinder( worldSpace(), ROCKET_RADIUS, ROCKET_LENGTH );
  addGeom( g, tempBody );
  rockets_.push_back( this );

  smokeTrail_ = MakeParticleSystem( "smoketrail", worldSpace() );
  Vector3 basis[3] = { front, up, left };
  smokeTrail_->setEmitterPos( pos, basis );
  smokeTrail_->setEmitting( true );
}

MyRocket::~MyRocket()
{
  rockets_.erase( std::find( rockets_.begin(), rockets_.end(), this ) );
  smokeTrail_->setEmitting( false );
}

void MyRocket::preStep()
{
  dReal const * rot = R();
  Vector3 front( rot[2], rot[6], rot[10] );
  Vector3 up( rot[1], rot[5], rot[9] );
  Vector3 left( rot[0], rot[4], rot[8] );
  Vector3 nuPos = P();
  Vector3 basis[3] = { front, up, left };
  smokeTrail_->setEmitterPos( nuPos, basis );
  age_ += TIME_STEP;
  if( age_ > ROCKET_MAX_AGE ) {
    exploding_ = true;
  }
  else {
    dBodyAddForceAtRelPos( bodyAt(0), front.x * ROCKET_FORCE, front.y * ROCKET_FORCE, front.z * ROCKET_FORCE,
        0, 0, 0.2f );
    //  stabilize
    Vector3 vel = *(Vector3 const *)dBodyGetLinearVel( bodyAt(0) );
    vel -= front * (vel / front);
    dBodyAddForce( bodyAt(0), vel.x * -ROCKET_STABILIZE, vel.y * -ROCKET_STABILIZE, vel.z * -ROCKET_STABILIZE );
  }

  //  MyRocket
  MySystem::preStep();
}

bool MyRocket::wantsContact( dContact & c, MySystemBase * other, dGeomID you, dGeomID him, bool firstTest )
{
  if( !other || other->physical() ) {
    exploding_ = true;
    return false;
  }
  return true;
}

void MyRocket::postStep()
{
  //  MyRocket
  MySystem::postStep();
  Vector3 pos = P();
  Vector3 vel = pos - oldPos_;
  float l = sqrtf( vel.lengthSquared() );
  if( l > 1e-3 ) {
    float li = 1.0f / l;
    Vector3 norm( vel.x*li, vel.y*li, vel.z*li );
    dGeomRaySetLength( utilRay(), l );
    dGeomRaySetParams( utilRay(), 0, 1 );
    dGeomRaySet( utilRay(), pos.x, pos.y, pos.z, norm.x, norm.y, norm.z );
    dSpaceCollide2( utilRay(), (dGeomID)worldSpace(), this, &RocketRayCallback );
  }
  oldPos_ = pos;
  if( exploding_ ) {
    explodeParams_.pos_ = pos;
    explode();
  }
}

void MyRocket::preStepAll()
{
  std::for_each( rockets_.begin(), rockets_.end(), MySystem::preStepOne );
}

void MyRocket::postStepAll()
{
  std::for_each( rockets_.begin(), rockets_.end(), MySystem::postStepOne );
}

void MyRocket::renderAll()
{
  std::for_each( rockets_.begin(), rockets_.end(), MySystem::renderOne );
}
