
#include "mysystem.h"
#include "tweakable.h"
#include "gl-ase.h"
#include "glshaderstate.h"

#include <assert.h>
#include <algorithm>
#include <windows.h>
#include <GL/GL.h>


TW_FLOAT(simulation, TIME_STEP, 0.01f);



double MySystemBase::curTime_;
dGeomID MySystemBase::utilRay_;
dGeomID MySystemBase::utilSphere_;
dWorldID MySystemBase::world_;
dSpaceID MySystemBase::worldSpace_;
dJointGroupID MySystemBase::tempGroup_;


void DampenBody( dBodyID body, float vScale, float aScale )
{
  assert( vScale <= 0 && aScale <= 0 );
  if( !dBodyIsEnabled( body ) ) {
    return;
  }
  dReal const * V = dBodyGetLinearVel( body );
  dBodyAddForce( body, vScale*V[0], vScale*V[1], vScale*V[2] );
  dReal const * A = dBodyGetAngularVel( body );
  dBodyAddTorque( body, aScale*A[0], aScale*A[1], aScale*A[2] );
}


MySystem::MySystem() : space_( 0 )
{
}

MySystem::~MySystem()
{
  std::for_each( geoms_.begin(), geoms_.end(), destroyGeom );
  std::for_each( joints_.begin(), joints_.end(), destroyJoint );
  std::for_each( bodies_.begin(), bodies_.end(), destroyBody );
  if( space_ ) dSpaceDestroy( space_ );
}

// preStep() is where you accumulate forces prior to integration
void MySystem::preStep()
{
  //  MySystem
}

bool MySystem::affectByExplosion( Vector3 const & center, float radius, float unitForce )
{
  std::for_each( bodies_.begin(), bodies_.end(), ApplyExplosionForce( unitForce / TIME_STEP, center, radius ) );
  return true;
}

bool MySystem::affectByImpact( dGeomID geom, Vector3 const & pos, Vector3 const & dir, float force )
{
  dBodyID b = 0;
  if( geom ) {
    b = dGeomGetBody( geom );
  }
  if( !b ) {
    b = bodyAt( 0 );
  }
  Vector3 f = dir * force;
  dBodyEnable( b );
  dBodyAddForceAtPos( b, f.x, f.y, f.z, pos.x, pos.y, pos.z );
  return true;
}


void MySystem::postStep()
{
  //todo: copy position/rotation here
  //todo: copy time stamp here
  //todo: keep two pos/rot/timestamps
  doDampening();
}

void MySystem::doDampening()
{
  each_body( DoDampen( -0.01f, -0.01f ) );
}

Vector3 MySystem::P( size_t ix ) const
{
  assert( ix < bodies_.size() );
  return (Vector3 const &)*dBodyGetPosition( bodies_[ix].body );
}

dReal const * MySystem::R( size_t ix ) const
{
  assert( ix < bodies_.size() );
  return dBodyGetRotation( bodies_.front().body );
}

Vector3 MySystem::L( size_t ix ) const
{
  assert( ix < bodies_.size() );
  return (Vector3 const &)*dBodyGetLinearVel( bodies_[ix].body );
}

Vector3 MySystem::A( size_t ix ) const
{
  assert( ix < bodies_.size() );
  return (Vector3 const &)*dBodyGetAngularVel( bodies_[ix].body );
}

void MySystem::basis( Vector3 * left, Vector3 * up, Vector3 * front) const
{
  dReal const * r = R();
  *left = Vector3( r[0], r[4], r[8] );
  *up = Vector3( r[1], r[5], r[9] );
  *front = Vector3( r[2], r[6], r[10] );
}

void MySystem::moveBy( float dx, float dy, float dz )
{
  each_body( DoMove( dx, dy, dz ) );
}

void MySystem::addBody( dBodyID b, GLShader * shader, I_GL_ASE * mesh, Vector3 const & meshDisplacement, void * data )
{
  BodyInfo bi;
  bi.body = b;
  bi.system = this;
  bi.systemData = data;
  bodies_.push_back( bi );
  meshes_.push_back( MeshInfo( shader, mesh, meshDisplacement ) );
}

void MySystem::addGeom( dGeomID g, dBodyID b )
{
  geoms_.push_back( g );
  assert( std::find( bodies_.begin(), bodies_.end(), b ) != bodies_.end() );
  dGeomSetBody( g, b );
  dGeomSetData( g, (void *)this );
}

void MySystem::addJoint( dJointID joint, dBodyID a, dBodyID b )
{
  joints_.push_back( joint );
  assert( std::find( bodies_.begin(), bodies_.end(), a ) != bodies_.end() );
  assert( std::find( bodies_.begin(), bodies_.end(), b ) != bodies_.end() );
  dJointAttach( joint, a, b );
}

size_t MySystem::bodyCount() const
{
  return bodies_.size();
}

dBodyID MySystem::bodyAt( size_t n ) const
{
  assert( n < bodies_.size() );
  return bodies_[n].body;
}

dMass MySystem::massAt( size_t n ) const
{
  assert( n < bodies_.size() );
  dMass m;
  dBodyGetMass( bodies_[n].body, &m );
  return m;
}

void MySystem::render()
{
  //todo: use the copied positions/rotations to render an interpolated frame
  size_t cnt = bodies_.size();
  for( size_t ix = 0; ix < cnt; ++ix ) {
    renderAt( ix );
  }
}

MySystem::ApplyExplosionForce::ApplyExplosionForce( float unitForce, Vector3 const & center, float radius )
{
  f_ = unitForce;
  c_ = center;
  r_ = radius;
}

void MySystem::ApplyExplosionForce::operator()( BodyInfo const & bi )
{
  dBodyID b = bi.body;
  //  I'm using mass as a weak substitute for surface area.
  //  The benefit is that all bodies in a system will be 
  //  accelerated equally, leading to better stability.
  dMass m;
  dBodyGetMass( b, &m );
  Vector3 pos = *(Vector3 const *)dBodyGetPosition( b );
  pos -= c_;
  float l = sqrtf( pos.lengthSquared() );
  if( l >= r_ ) return;
  if( l <= 0 ) {
    dBodyAddForce( b, 0, m.mass * f_, 0 );
    debug( "Explode force: 0.00,%.2f,0.00\n", m.mass*f_ );
  }
  else {
    l = (1 - l/r_) * m.mass * f_ / l;
    dBodyAddForce( b, pos.x*l, pos.y*l, pos.z*l );
    debug( "Explode force: %.2f,%.2f,%.2f\n", pos.x*l, pos.y*l, pos.z*l );
  }
  dBodyEnable( b );
}

MySystem::DoDampen::DoDampen( float l, float a ) : linear_( l ), angular_( a )
{
  assert( l < 0 && a < 0 );
}

void MySystem::DoDampen::operator()( BodyInfo const & bi )
{
  dBodyID b = bi.body;
  DampenBody( b, linear_, angular_ );
}

MySystem::DoMove::DoMove( float dx, float dy, float dz ) : dx_( dx ), dy_( dy ), dz_( dz )
{
}

void MySystem::DoMove::operator()( BodyInfo const & bi )
{
  dBodyID b = bi.body;
  dReal const * P = dBodyGetPosition( b );
  dBodySetPosition( b, dx_+P[0], dy_+P[1], dz_+P[2] );
}

void MySystem::renderAt( size_t ix )
{
  assert( ix < bodies_.size() );
  dBodyID b = bodies_[ix].body;
  dReal const * r = dBodyGetRotation( b );
  Matrix4 m4 = ((Matrix3 *)r)->transpose();
  m4.pos() = *(Vector3 const *)dBodyGetPosition( b );
  //  adjust according to rendering displacement
  glPushMatrix();
  glMultMatrixf( &m4.xx );
  Vector3 const & d = meshes_[ix].displacement;
  glTranslatef( d.x, d.y, d.z );
  meshes_[ix].shader->apply();
  meshes_[ix].mesh->issueAll();
  glPopMatrix();
  assert( !glGetError() );
}

void MySystem::renderOne( MySystem * ms )
{
  ms->render();
}

void MySystem::preStepOne( MySystem * ms )
{
  ms->preStep();
}

void MySystem::postStepOne( MySystem * ms )
{
  ms->postStep();
}

void MySystem::destroy( MySystem * ms )
{
  delete ms;
}

void MySystem::destroyGeom( dGeomID g )
{
  dGeomDestroy( g );
}

void MySystem::destroyJoint( dJointID j )
{
  dJointDestroy( j );
}

void MySystem::destroyBody( BodyInfo const & b )
{
  dBodyDestroy( b.body );
}

void MySystemBase::setCurTime( double time )
{
  curTime_ = time;
}

double MySystemBase::curTime()
{
  return curTime_;
}

void MySystemBase::setUtilRay( dGeomID ray )
{
  utilRay_ = ray;
}

dGeomID MySystemBase::utilRay()
{
  return utilRay_;
}

void MySystemBase::setUtilSphere( dGeomID sphere )
{
  utilSphere_ = sphere;
}

dGeomID MySystemBase::utilSphere()
{
  return utilSphere_;
}

void MySystemBase::setWorld( dWorldID world )
{
  world_ = world;
}

dWorldID MySystemBase::world()
{
  return world_;
}

void MySystemBase::setWorldSpace( dSpaceID worldSpace )
{
  worldSpace_ = worldSpace;
}

dSpaceID MySystemBase::worldSpace()
{
  return worldSpace_;
}

void MySystemBase::setTempJointGroup( dJointGroupID tempGroup )
{
  tempGroup_ = tempGroup;
}

dJointGroupID MySystemBase::tempJointGroup()
{
  return tempGroup_;
}




