
#if !defined( mysystem_h )
#define mysystem_h

#include <ode/ode.h>
#include <vector>
#include <algorithm>

#include "mymath.h"
class I_GL_ASE;

class MySystem;
class GLShader;

//  Trees and rocks really don't need much stepping...
class MySystemBase {
  public:
    virtual bool physical() { return true; }
    virtual bool wantsContact( dContact & c, MySystemBase * other, dGeomID you, dGeomID him, bool firstTest ) { return true; }
    virtual bool canBePickedUp() { return true; }
    virtual bool affectByExplosion( Vector3 const & center, float radius, float unitForce ) { return false; }
    virtual bool affectByImpact( dGeomID geom, Vector3 const & pos, Vector3 const & dir, float force ) { return false; }
    virtual void render() = 0;
    virtual MySystem * richSystem() { return 0; }

    static void setCurTime( double time );
    static double curTime();
    static void setUtilRay( dGeomID ray );
    static dGeomID utilRay();
    static void setUtilSphere( dGeomID sphere );
    static dGeomID utilSphere();
    static void setWorld( dWorldID world );
    static dWorldID world();
    static void setWorldSpace( dSpaceID worldSpace );
    static dSpaceID worldSpace();
    static void setTempJointGroup( dJointGroupID tempGroup );
    static dJointGroupID tempJointGroup();


  protected:
    virtual ~MySystemBase() {}

  private:
    static double curTime_;
    static dGeomID utilRay_;
    static dGeomID utilSphere_;
    static dWorldID world_;
    static dSpaceID worldSpace_;
    static dJointGroupID tempGroup_;
};

//  A MySystem is a collection of geoms, bodies, joints and 
//  optionally a space that they're in.
class MySystem : public MySystemBase {
  public:
    //todo: Should have a struct of body ID, system pointer and 
    //  system-specific data (pointer/int) for each body, and 
    //  set as body data.
    MySystem();
    virtual ~MySystem();
    virtual MySystem * richSystem() { return this; }

    // preStep() is where you accumulate forces prior to integration
    virtual void preStep();
    virtual bool affectByExplosion( Vector3 const & center, float radius, float unitForce );
    virtual bool affectByImpact( dGeomID geom, Vector3 const & pos, Vector3 const & dir, float force );
    virtual void postStep();
    virtual void doDampening();

    // note: these are not virtual!
    Vector3 P( size_t ix = 0 ) const;
    dReal const * R( size_t ix = 0 ) const;
    Vector3 L( size_t ix = 0 ) const;
    Vector3 A( size_t ix = 0 ) const;
    void basis( Vector3 * left, Vector3 * up, Vector3 * front) const;
    void moveBy( float dx, float dy, float dz );

    template< class T > each_body( T & func )
      {
        std::for_each( bodies_.begin(), bodies_.end(), func );
      }

    struct BodyInfo {
      dBodyID body;
      MySystem * system;
      void * systemData;
      bool operator==( dBodyID const & o ) const {
        return body == o;
      }
      bool operator==( BodyInfo const & o ) const {
        return body == o.body;
      }
    };
    struct MeshInfo {
      GLShader * shader;
      I_GL_ASE * mesh;
      Vector3 displacement;
      MeshInfo() {}
      MeshInfo( GLShader * s, I_GL_ASE * m, Vector3 const & v ) :
        shader( s ), mesh( m ), displacement( v ) {}
    };

    //  these are mostly for use by subclasses, but public anyway
    void addBody( dBodyID b, GLShader * s, I_GL_ASE * mesh, Vector3 const & meshDisplacement = Vector3::Null, void * data = 0 );
    void addGeom( dGeomID g, dBodyID b );
    void addJoint( dJointID joint, dBodyID a, dBodyID b );
    size_t bodyCount() const;
    dBodyID bodyAt( size_t n ) const;
    dMass massAt( size_t n ) const;

    virtual void render();

    static void renderOne( MySystem * ms );
    static void preStepOne( MySystem * ms );
    static void postStepOne( MySystem * ms );
    static void destroy( MySystem * ms );
    static void destroyGeom( dGeomID g );
    static void destroyJoint( dJointID j );
    static void destroyBody( BodyInfo const & b );

  protected:

    dSpaceID space_;
    std::vector< BodyInfo > bodies_;
    std::vector< dGeomID > geoms_;
    std::vector< dJointID > joints_;
    std::vector< MeshInfo > meshes_;
    dSpaceID customSpace_;    //  if null, it's world space

    class ApplyExplosionForce {
      public:
        ApplyExplosionForce( float unitForce, Vector3 const & center, float radius );
        void operator()( BodyInfo const & bi );

        float f_;
        Vector3 c_;
        float r_;
    };
    class DoDampen {
      public:
        DoDampen( float l, float a );
        void operator()( BodyInfo const & bi );

        float linear_;
        float angular_;
    };
    class DoMove {
      public:
        DoMove( float dx, float dy, float dz );
        void operator()( BodyInfo const & bi );

        float dx_, dy_, dz_;
    };
    
    void renderAt( size_t ix );
};


void DampenBody( dBodyID, float lScale, float aScale );
void ClearForceAndTorque( dBodyID body );
void AddDeferredDelete( MySystem * );

//  These return pointers to statics!
char const * imagePath( char const * handle );
char const * asePath( char const * handle );
char const * wavPath( char const * handle );


#endif  //  mysystem_h
