
#include <windows.h>

#include "owavefile.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>

#pragma warning( push, 1 )
#include <vector>
#pragma warning( pop )

using namespace std;



class WaveFile1
  : public OWaveFile
{
public:
  WaveFile1();
  ~WaveFile1();

  void setPath( char const * path );
  void setOpenMode( WaveOpenMode mode );
  bool open();
  void destroy();

  //  read only functions
  unsigned int countChunks();
  unsigned int chunkType( unsigned int ix );
  unsigned int chunkSize( unsigned int ix );
  bool readChunkData( unsigned int ix, void * buffer, unsigned int offset, unsigned int * ioSize );
  bool fmtHeader( WAVFmtHeader * outHdr, void * extraData, unsigned int * extraDataSize );
  unsigned int countFrames();
  unsigned int readFrames( float ** channels, unsigned int frameOffset, unsigned int frameCount );

  bool writeChunkData( unsigned int ix, void const * buffer, unsigned int offset, unsigned int * ioSize );

  //  write only functions
  bool newChunk( unsigned int type );
  bool writeNewData( void const * data, unsigned int size );
  bool endChunk();
  bool setFmtHeader( WAVFmtHeader const & hdr, void const * extraData, unsigned int extraSize );
  bool beginData();
  bool writeFrames( float const * const * channels, unsigned int frameCount );
  bool endData();

  //  implementation
  bool slurpChunks();
  bool writeRiff();
  unsigned int calcFileSize( unsigned int frameCount );
  unsigned int calcFilePos( unsigned int frameOffset );
  void * tempBlock( unsigned int size );

  bool waveHeaderIsSane( WAVFmtHeader const & hdr );
  bool parseFmtHeader();
  void translateDataIn( void const * data, unsigned int frameCount, float ** buffers );
  void translateDataOut( float const * const * buffers, unsigned int frameCount, void * data );

  FILE * file_;
  vector< pair< unsigned int, WAVChunkHeader > > headers_;
  unsigned int dataOffset_;
  unsigned int frameCount_;
  WAVFmtHeader fmtHeader_;
  void * fmtExtraData_;
  unsigned int fmtExtraSize_;
  char path_[ 300 ];
  WaveOpenMode mode_;
  unsigned int curChunkOffset_;
  void * tempBlock_;
  unsigned int tempBlockSize_;

  void (WaveFile1::*packFunc_)( float const * const *, unsigned int, void * );
  void (WaveFile1::*unpackFunc_)( void const *, unsigned int, float ** );

  //  for packing
  float dither1_;
  float dither2_;
  float max1_;
  float max2_;
  float min1_;
  float min2_;

  //  pack functions
  void sep_flt_to_int_16_1( float const *const*, unsigned int, void * );
  void sep_flt_to_int_16_2( float const *const*, unsigned int, void * );
  void sep_flt_to_int_24_1( float const *const*, unsigned int, void * );
  void sep_flt_to_int_24_2( float const *const*, unsigned int, void * );
  void sep_flt_to_int_32_1( float const *const*, unsigned int, void * );
  void sep_flt_to_int_32_2( float const *const*, unsigned int, void * );
  void sep_flt_to_int_flt_1( float const *const*, unsigned int, void * );
  void sep_flt_to_int_flt_2( float const *const*, unsigned int, void * );

  //  unpack functions
  void int_16_to_sep_flt_1( void const *, unsigned int, float ** );
  void int_16_to_sep_flt_2( void const *, unsigned int, float ** );
  void int_24_to_sep_flt_1( void const *, unsigned int, float ** );
  void int_24_to_sep_flt_2( void const *, unsigned int, float ** );
  void int_32_to_sep_flt_1( void const *, unsigned int, float ** );
  void int_32_to_sep_flt_2( void const *, unsigned int, float ** );
  void int_flt_to_sep_flt_1( void const *, unsigned int, float ** );
  void int_flt_to_sep_flt_2( void const *, unsigned int, float ** );
};



WaveFile1::WaveFile1()
{
  file_ = 0;
  dataOffset_ = 0;
  frameCount_ = 0;
  curChunkOffset_ = 0;
  memset( &fmtHeader_, 0, sizeof( fmtHeader_ ) );
  fmtExtraData_ = 0;
  fmtExtraSize_ = 0;
  path_[ 0 ] = 0;
  mode_ = WaveOpenNull;
  tempBlock_ = 0;
  tempBlockSize_ = 0;
  dither1_ = dither2_ = 0;
  min1_ = min2_ = 0;
  max1_ = max2_ = 0;
}

WaveFile1::~WaveFile1()
{
  assert( file_ == 0 );
  if( file_ ) {
    fclose( file_ );
  }
}


void WaveFile1::setPath( char const * path )
{
  strncpy( path_, path, sizeof( path_ ) );
  path_[ sizeof( path_ )-1 ] = 0;
}

void WaveFile1::setOpenMode( WaveOpenMode mode )
{
  mode_ = mode;
}

bool WaveFile1::open()
{
  assert( file_ == 0 ); //  can't open a file that's already open
  assert( mode_ != WaveOpenNull );
  assert( path_[ 0 ] != 0 );
  if( mode_ == WaveOpenNull || path_[ 0 ] == 0 ) {
    return false;
  }
  char const * mode = 0;
  switch( mode_ ) {
  default:
    assert( !"mode_ is wrong" );
  case WaveOpenRead:
    mode = "rb";
    break;
  case WaveOpenCreate:
    mode = "w+b";
    break;
  case WaveOpenUpdate:
    mode = "r+b";
    break;
  }
  file_ = fopen( path_, mode );
  if( !file_ ) {
    return false;
  }
  bool ok = false;
  if( (mode_ == WaveOpenRead) || (mode_ == WaveOpenUpdate) ) {
    ok = slurpChunks();
  }
  else if( mode_ == WaveOpenCreate ) {
    ok = writeRiff();
  }
  if( !ok ) {
    fclose( file_ );
    file_ = 0;
    return false;
  }
  return true;
}

void WaveFile1::destroy()
{
  if( file_ ) {
    if( mode_ == WaveOpenCreate ) {
      fseek( file_, 0, 2 );
      unsigned int sz = ftell( file_ )-8;
      fseek( file_, 4, 0 );
      fwrite( &sz, 1, 4, file_ );
    }
    fclose( file_ );
    file_ = 0;
  }
  delete this;
}


//  read only functions
unsigned int WaveFile1::countChunks()
{
  assert( file_ );
  assert( (mode_ == WaveOpenUpdate) || (mode_ == WaveOpenRead) );
  return unsigned int( headers_.size() );
}

unsigned int WaveFile1::chunkType( unsigned int ix )
{
  assert( file_ );
  assert( ix < headers_.size() );
  assert( (mode_ == WaveOpenUpdate) || (mode_ == WaveOpenRead) );
  return headers_[ ix ].second.tag;
}

unsigned int WaveFile1::chunkSize( unsigned int ix )
{
  assert( file_ );
  assert( ix < headers_.size() );
  assert( (mode_ == WaveOpenUpdate) || (mode_ == WaveOpenRead) );
  return headers_[ ix ].second.size;
}

bool WaveFile1::readChunkData( unsigned int ix, void * buffer, unsigned int offset, unsigned int * ioSize )
{
  assert( file_ );
  assert( ix < headers_.size() );
  assert( (mode_ == WaveOpenUpdate) || (mode_ == WaveOpenRead) );
  if( offset > headers_[ ix ].second.size ) {
    *ioSize = 0;
    return false;
  }
  unsigned int maxRead = headers_[ ix ].second.size - offset;
  if( *ioSize > maxRead ) {
    *ioSize = maxRead;
  }
  fseek( file_, headers_[ ix ].first + offset, 0 );
  maxRead = unsigned int( fread( buffer, 1, *ioSize, file_ ) );
  *ioSize = maxRead;
  return true;
}

bool WaveFile1::fmtHeader( WAVFmtHeader * outHdr, void * extraData, unsigned int * extraDataSize )
{
  assert( file_ );
  assert( (mode_ == WaveOpenUpdate) || (mode_ == WaveOpenRead) );
  *outHdr = fmtHeader_;
  bool ok = true;
  if( *extraDataSize > fmtExtraSize_ ) {
    *extraDataSize = fmtExtraSize_;
  }
  else if( *extraDataSize < fmtExtraSize_ ) {
    ok = false;
  }
  memcpy( extraData, fmtExtraData_, *extraDataSize );
  return ok;
}

unsigned int WaveFile1::countFrames()
{
  assert( file_ );
  assert( (mode_ == WaveOpenUpdate) || (mode_ == WaveOpenRead) );
  return frameCount_;
}

unsigned int WaveFile1::readFrames( float ** channels, unsigned int frameOffset, unsigned int frameCount )
{
  assert( !curChunkOffset_ );
  assert( file_ );
  assert( (mode_ == WaveOpenUpdate) || (mode_ == WaveOpenRead) );
  if( frameOffset >= frameCount_ ) {
    return 0;
  }
  if( frameCount > frameCount_ - frameOffset ) {
    frameCount = frameCount_ - frameOffset;
  }
  unsigned int size = calcFileSize( frameCount );
  void * block = tempBlock( size );
  if( !block ) {
    return 0;
  }
  unsigned int pos = calcFilePos( frameOffset );
  fseek( file_, pos, 0 );
  if( size != fread( block, 1, size, file_ ) ) {
    assert( !"file read error" );
    return 0;
  }
  translateDataIn( block, frameCount, channels );
  return frameCount;
}

bool WaveFile1::writeChunkData( unsigned int ix, void const * buffer, unsigned int offset, unsigned int * ioSize )
{
  assert( !curChunkOffset_ ); //  can't be writing a new chunk
  assert( file_ );
  assert( ix < headers_.size() );
  assert( (mode_ == WaveOpenUpdate) || (mode_ == WaveOpenRead) );
  if( offset >= headers_[ ix ].second.size ) {
    *ioSize = 0;
    return false;
  }
  if( *ioSize > headers_[ ix ].second.size - offset ) {
    *ioSize = headers_[ ix ].second.size - offset;
  }
  fseek( file_, offset+headers_[ ix ].first + offset, 0 );
  *ioSize = unsigned int( fwrite( buffer, 1, *ioSize, file_ ) );
  return true;
}


//  write only functions
bool WaveFile1::newChunk( unsigned int type )
{
  assert( !curChunkOffset_ ); //  can't be writing a new chunk
  assert( file_ );
  assert( mode_ == WaveOpenCreate );
  WAVChunkHeader hdr;
  hdr.tag = type;
  hdr.size = 0;
  fseek( file_, 0, 2 );
  if( 8 != fwrite( &hdr, 1, sizeof( hdr ), file_ ) ) {
    assert( !"write error" );
    return false;
  }
  curChunkOffset_ = ftell( file_ );
  headers_.push_back( pair< unsigned int, WAVChunkHeader >( curChunkOffset_, hdr ) );
  return true;
}

bool WaveFile1::writeNewData( void const * data, unsigned int size )
{
  assert( curChunkOffset_ );
  fseek( file_, 0, 2 );
  if( size != fwrite( data, 1, size, file_ ) ) {
    return false;
  }
  return true;
}

bool WaveFile1::endChunk()
{
  assert( curChunkOffset_ );
  fseek( file_, 0, 2 );
  WAVChunkHeader hdr = (*headers_.rbegin()).second;
  long fPos = ftell( file_ );
  hdr.size = fPos - curChunkOffset_;
  if( fPos & 3 ) {
    //  pad chunk out to 4 bytes
    fwrite( "\0\0\0\0", 1, 4-(fPos & 3), file_ );
  }
  fseek( file_, curChunkOffset_ - sizeof( hdr ), 0 );
  curChunkOffset_ = 0;
  if( 8 != fwrite( &hdr, 1, sizeof( hdr ), file_ ) ) {
    return false;
  }
  (*headers_.rbegin()).second = hdr;
  return true;
}

bool WaveFile1::setFmtHeader( WAVFmtHeader const & hdr, void const * extraData, unsigned int extraSize )
{
  assert( !curChunkOffset_ ); //  can't be writing a new chunk
  assert( file_ );
  assert( !fmtExtraData_ );
  assert( mode_ == WaveOpenCreate );
  assert( fmtHeader_.format == 0 );
  assert( waveHeaderIsSane( hdr ) );
  fmtHeader_ = hdr;
  if( !parseFmtHeader() ) {
    assert( !"unsupported FMT header" );
    return false;
  }
  fmtExtraData_ = malloc( extraSize );
  memcpy( fmtExtraData_, extraData, extraSize );
  fmtExtraSize_ = extraSize;
  bool b = newChunk( FMT__TAG );
  b = b && writeNewData( &hdr, sizeof( hdr ) );
  b = b && (!extraSize || writeNewData( extraData, extraSize ));
  b = b && endChunk();
  return b;
}

bool WaveFile1::beginData()
{
  assert( !curChunkOffset_ );
  assert( file_ );
  assert( fmtHeader_.format != 0 );
  assert( frameCount_ == 0 );
  assert( mode_ == WaveOpenCreate );
  bool b = newChunk( DATA_TAG );
  dataOffset_ = curChunkOffset_;
  return b;
}

bool WaveFile1::writeFrames( float const *const* channels, unsigned int frameCount )
{
  assert( dataOffset_ );
  assert( curChunkOffset_ );
  unsigned int size = calcFileSize( frameCount );
  void * block = tempBlock( size );
  if( !block ) {
    return false;
  }
  unsigned int pos = calcFilePos( frameCount_ );
  fseek( file_, pos, 0 );
  translateDataOut( channels, frameCount, block );
  if( size != fwrite( block, 1, size, file_ ) ) {
    assert( !"file write error" );
    return false;
  }
  frameCount_ += frameCount;
  return true;
}

bool WaveFile1::endData()
{
  return endChunk();
}



bool WaveFile1::slurpChunks()
{
  assert( file_ );
  assert( !headers_.size() );
  assert( mode_ != WaveOpenCreate );
  WAVFileHeader fhdr;
  fseek( file_, 0, 0 );
  if( fread( &fhdr, 1, sizeof( fhdr ), file_ ) != sizeof( fhdr ) ) {
    //  not enough for header -- empty file?
    return false;
  }
  if( fhdr.riff != RIFF_TAG || fhdr.wave != WAVE_TAG ) {
    //  not a wave file
    return false;
  }
  while( 1 ) {
    WAVChunkHeader chdr;
    unsigned int pos;
    if( fread( &chdr, 1, sizeof( chdr ), file_ ) != sizeof( chdr ) ) {
      //  done reading
      break;
    }
    pos = ftell( file_ );
    headers_.push_back( pair< unsigned int, WAVChunkHeader >( pos, chdr ) );
    fseek( file_, ((chdr.size + 3) & -4) + pos, 0 );
  }
  bool gotFmt = false;
  bool gotData = false;
  unsigned int dataSize = 0;
  unsigned int cc = countChunks();
  for( unsigned int ix = 0; ix < cc; ++ix ) {
    if( chunkType( ix ) == FMT__TAG ) {
      unsigned int sz = chunkSize( ix );
      unsigned int rdSize = sizeof( fmtHeader_ );
      if( !readChunkData( ix, &fmtHeader_, 0, &rdSize ) || rdSize < sizeof( fmtHeader_ ) ) {
        //  format chunk too small?
        assert( sz >= sizeof( fmtHeader_ ) );
        return false;
      }
      if( !parseFmtHeader() ) {
        assert( !"fmt header not recognized" );
        return false;
      }
      gotFmt = true;
      if( sz > rdSize ) {
        fmtExtraSize_ = sz - rdSize;
        fmtExtraData_ = malloc( fmtExtraSize_ );
        rdSize = fmtExtraSize_;
        if( !readChunkData( ix, fmtExtraData_, sizeof( fmtHeader_ ), &rdSize ) || rdSize != fmtExtraSize_ ) {
          assert( !"read error" );
          free( fmtExtraData_ );
          fmtExtraData_ = 0;
          fmtExtraSize_ = 0;
        }
      }
    }
    else if( chunkType( ix ) == DATA_TAG ) {
      gotData = true;
      dataOffset_ = headers_[ ix ].first;
      dataSize = chunkSize( ix );
    }
    else {
      //  skip it
      continue;
    }
  }
  if( !gotFmt || !gotData ) {
    //  not enough header data to make sense!
    return false;
  }
  frameCount_ = dataSize / fmtHeader_.frameSize;
  return true;
}

bool WaveFile1::writeRiff()
{
  assert( mode_ == WaveOpenCreate );
  assert( file_ != 0 );
  WAVFileHeader fhdr;
  fhdr.riff = RIFF_TAG;
  fhdr.size = 0;
  fhdr.wave = WAVE_TAG;
  fseek( file_, 0, 0 );
  if( 12 != fwrite( &fhdr, 1, sizeof( fhdr ), file_ ) ) {
    assert( !"write error" );
    return false;
  }
  return true;
}

unsigned int WaveFile1::calcFileSize( unsigned int frameCount )
{
  assert( fmtHeader_.frameSize > 0 );
  return frameCount * fmtHeader_.frameSize;
}

unsigned int WaveFile1::calcFilePos( unsigned int frameOffset )
{
  assert( fmtHeader_.frameSize > 0 );
  assert( dataOffset_ > 0 );
  return frameOffset * fmtHeader_.frameSize + dataOffset_;
}

void * WaveFile1::tempBlock( unsigned int size )
{
  if( tempBlockSize_ < size ) {
    void * tb = realloc( tempBlock_, size );
    if( !tb ) {
      assert( !"out of memory!" );
      return 0;
    }
    tempBlock_ = tb;
    tempBlockSize_ = size;
  }
  return tempBlock_;
}

bool WaveFile1::waveHeaderIsSane( WAVFmtHeader const & hdr )
{
  if( hdr.format != INT_PCM_FORMAT && hdr.format != FLOAT_PCM_FORMAT ) {
    return false;
  }
  //fixme: maybe we want to support more channels at some point?
  if( hdr.samplesPerFrame < 1 || hdr.samplesPerFrame > 2 ) {
    return false;
  }
  if( hdr.framesPerSecond < 8000 || hdr.framesPerSecond > 200000 ) {
    return false;
  }
  if( hdr.bytesPerSecond != hdr.framesPerSecond * hdr.frameSize ) {
    //return false;   // we can probably live with this if all else checks out OK
  }
  if( hdr.frameSize < hdr.bitsPerSample * hdr.samplesPerFrame / 8 ||
      hdr.frameSize > (hdr.bitsPerSample + 8) * hdr.samplesPerFrame / 8 ) {
    return false;
  }
  if( (hdr.samplesPerFrame > 1) && (hdr.frameSize & 1) ) {
    return false;
  }
  //fixme: is unsigned char necessary? Probably not.
  if( hdr.bitsPerSample < 16 || hdr.bitsPerSample > 32 ) {
    return false;
  }
  return true;
}

bool WaveFile1::parseFmtHeader()
{
  if( !waveHeaderIsSane( fmtHeader_ ) ) {
    return false;
  }
  if( fmtHeader_.format == INT_PCM_FORMAT ) {
    if( fmtHeader_.samplesPerFrame == 1 ) {
      if( fmtHeader_.bitsPerSample == 16 ) {
        if( fmtHeader_.frameSize != 2 ) {
          return false;
        }
        packFunc_ = sep_flt_to_int_16_1;
        unpackFunc_ = int_16_to_sep_flt_1;
      }
      else if( fmtHeader_.frameSize == 3 ) {
        packFunc_ = sep_flt_to_int_24_1;
        unpackFunc_ = int_24_to_sep_flt_1;
      }
      else if( fmtHeader_.frameSize == 4 ) {
        packFunc_ = sep_flt_to_int_32_1;
        unpackFunc_ = int_32_to_sep_flt_1;
      }
      else {
        return false;
      }
    }
    else if( fmtHeader_.samplesPerFrame == 2 ) {
      if( fmtHeader_.bitsPerSample == 16 ) {
        if( fmtHeader_.frameSize != 4 ) {
          return false;
        }
        packFunc_ = sep_flt_to_int_16_2;
        unpackFunc_ = int_16_to_sep_flt_2;
      }
      else if( fmtHeader_.frameSize == 6 ) {
        packFunc_ = sep_flt_to_int_24_2;
        unpackFunc_ = int_24_to_sep_flt_2;
      }
      else if( fmtHeader_.frameSize == 8 ) {
        packFunc_ = sep_flt_to_int_32_2;
        unpackFunc_ = int_32_to_sep_flt_2;
      }
      else {
        return false;
      }
    }
    else {
      return false;
    }
  }
  else if( fmtHeader_.format == FLOAT_PCM_FORMAT ) {
    if( fmtHeader_.bitsPerSample != 32 ) {
      return false;
    }
    if( fmtHeader_.frameSize != 4 * fmtHeader_.samplesPerFrame ) {
      return false;
    }
    if( fmtHeader_.samplesPerFrame == 1 ) {
      packFunc_ = sep_flt_to_int_flt_1;
      unpackFunc_ = int_flt_to_sep_flt_1;
    }
    else {
      packFunc_ = sep_flt_to_int_flt_2;
      unpackFunc_ = int_flt_to_sep_flt_2;
    }
  }
  else {
    return false;
  }
  return true;
}


void WaveFile1::translateDataIn( void const * data, unsigned int frameCount, float ** buffers )
{
  (this->*unpackFunc_)( data, frameCount, buffers );
}

void WaveFile1::translateDataOut( float const *const* buffers, unsigned int frameCount, void * data )
{
  (this->*packFunc_)( buffers, frameCount, data );
}



//  fixme: optimize these conversion functions if necessary

void WaveFile1::sep_flt_to_int_16_1( float const *const* fbuf, unsigned int cnt, void * obuf )
{
  float const * in1 = fbuf[ 0 ];
  short * out1 = (short *)obuf;
  float d1 = dither1_;
  while( cnt > 0 ) {
    --cnt;
    float v = *in1 + d1;
    ++in1;
    short o;
    if( v > 32767.0f/32768.0f ) {
      o = 32767;
    }
    else if( v < -1.0f ) {
      o = -32768;
    }
    else {
      o = (short)floorf( v * 32768 );
    }
    d1 = v - o/32768.f;
    *out1 = o;
    ++out1;
  }
  dither1_ = d1;
}

void WaveFile1::sep_flt_to_int_16_2( float const *const* fbuf, unsigned int cnt, void * obuf )
{
  float const * in1 = fbuf[ 0 ];
  float const * in2 = fbuf[ 1 ];
  short * out12 = (short *)obuf;
  float d1 = dither1_;
  float d2 = dither2_;
  while( cnt > 0 ) {
    --cnt;
    float v1 = *in1 + d1;
    float v2 = *in2 + d2;
    ++in1;
    ++in2;
    short o1;
    short o2;
    if( v1 > 32767.0f/32768.0f ) {
      o1 = 32767;
    }
    else if( v1 < -1.0f ) {
      o1 = -32768;
    }
    else {
      o1 = (short)floorf( v1 * 32768 );
    }
    if( v2 > 32767.0f/32768.0f ) {
      o2 = 32767;
    }
    else if( v2 < -1.0f ) {
      o2 = -32768;
    }
    else {
      o2 = (short)floorf( v2 * 32768 );
    }
    d1 = v1 - o1/32768.f;
    d2 = v2 - o2/32768.f;
    *out12 = o1;
    ++out12;
    *out12 = o2;
    ++out12;
  }
  dither1_ = d1;
  dither2_ = d2;
}

void WaveFile1::sep_flt_to_int_24_1( float const *const* fbuf, unsigned int cnt, void * obuf )
{
  float const * in1 = fbuf[ 0 ];
  unsigned char * out1 = (unsigned char *)obuf;
  float d1 = dither1_;
  while( cnt > 0 ) {
    --cnt;
    float v = *in1 + d1;
    ++in1;
    int o;
    if( v > 8388607.0f/8388608.0f ) {
      o = 8388607;
    }
    else if( v < -1.0f ) {
      o = -8388608;
    }
    else {
      o = (int)floorf( v * 8388608 );
    }
    d1 = v - o/8388608.f;
    out1[0] = unsigned char( o & 0xff );
    out1[1] = unsigned char( (o>>8) & 0xff );
    out1[2] = unsigned char( (o>>16) & 0xff );
    out1 += 3;
  }
  dither1_ = d1;
}

void WaveFile1::sep_flt_to_int_24_2( float const *const* fbuf, unsigned int cnt, void * obuf )
{
  float const * in1 = fbuf[ 0 ];
  float const * in2 = fbuf[ 1 ];
  unsigned char * out12 = (unsigned char *)obuf;
  float d1 = dither1_;
  float d2 = dither2_;
  while( cnt > 0 ) {
    --cnt;
    float v1 = *in1 + d1;
    float v2 = *in2 + d2;
    ++in1;
    ++in2;
    int o1;
    int o2;
    if( v1 > 8388607.0f/8388608.0f ) {
      o1 = 8388607;
    }
    else if( v1 < -1.0f ) {
      o1 = -8388608;
    }
    else {
      o1 = (int)floorf( v1 * 8388608 );
    }
    if( v2 > 8388607.0f/8388608.0f ) {
      o2 = 8388607;
    }
    else if( v2 < -1.0f ) {
      o2 = -8388608;
    }
    else {
      o2 = (int)floorf( v2 * 8388608 );
    }
    d1 = v1 - o1/8388608.f;
    d2 = v2 - o2/8388608.f;
    out12[0] = unsigned char( o1 & 0xff );
    out12[1] = unsigned char( (o1>>8) & 0xff );
    out12[2] = unsigned char( (o1>>16) & 0xff );
    out12 += 3;
    out12[0] = unsigned char( o2 & 0xff );
    out12[1] = unsigned char( (o2>>8) & 0xff );
    out12[2] = unsigned char( (o2>>16) & 0xff );
    out12 += 3;
  }
  dither1_ = d1;
  dither2_ = d2;
}

void WaveFile1::sep_flt_to_int_32_1( float const *const* fbuf, unsigned int cnt, void * obuf )
{
  float const * in1 = fbuf[ 0 ];
  int * out1 = (int *)obuf;
  float d1 = dither1_;
  while( cnt > 0 ) {
    --cnt;
    float v = *in1 + d1;
    ++in1;
    int o;
    if( v > 2147483647.0f/2147483648.0f ) {
      o = 2147483647;
    }
    else if( v < -1.0f ) {
      o = -1-2147483647;
    }
    else {
      o = (int)floorf( v * 2147483648 );
    }
    d1 = v - o/2147483648.f;
    *out1 = o;
    ++out1;
  }
  dither1_ = d1;
}

void WaveFile1::sep_flt_to_int_32_2( float const *const* fbuf, unsigned int cnt, void * obuf )
{
  float const * in1 = fbuf[ 0 ];
  float const * in2 = fbuf[ 1 ];
  int * out12 = (int *)obuf;
  float d1 = dither1_;
  float d2 = dither2_;
  while( cnt > 0 ) {
    --cnt;
    float v1 = *in1 + d1;
    float v2 = *in2 + d2;
    ++in1;
    ++in2;
    int o1;
    int o2;
    if( v1 > 2147483647.0f/2147483648.0f ) {
      o1 = 2147483647;
    }
    else if( v1 < -1.0f ) {
      o1 = -1-2147483647;
    }
    else {
      o1 = (int)floorf( v1 * 2147483648 );
    }
    if( v2 > 2147483647.0f/2147483648.0f ) {
      o2 = 2147483647;
    }
    else if( v2 < -1.0f ) {
      o2 = -1-2147483647;
    }
    else {
      o2 = (int)floorf( v2 * 2147483648 );
    }
    d1 = v1 - o1/2147483648.f;
    d2 = v2 - o2/2147483648.f;
    *out12 = o1;
    ++out12;
    *out12 = o2;
    ++out12;
  }
  dither1_ = d1;
  dither2_ = d2;
}

void WaveFile1::sep_flt_to_int_flt_1( float const *const* fbuf, unsigned int cnt, void * obuf )
{
  memcpy( obuf, fbuf[0], cnt*4 );
}

void WaveFile1::sep_flt_to_int_flt_2( float const *const* fbuf, unsigned int cnt, void * obuf )
{
  float const * in1 = fbuf[ 0 ];
  float const * in2 = fbuf[ 1 ];
  float * out12 = (float *)obuf;
  while( cnt > 0 ) {
    --cnt;
    *out12 = *in1;
    ++out12;
    ++in1;
    *out12 = *in2;
    ++out12;
    ++in2;
  }
}


void WaveFile1::int_16_to_sep_flt_1( void const * ibuf, unsigned int cnt, float ** ofbuf )
{
  short const * in12 = (short const *)ibuf;
  float * out1 = ofbuf[0];
  while( cnt > 0 ) {
    --cnt;
    float f = *in12;
    ++in12;
    f *= 1.f/32768.f;
    *out1 = f;
    ++out1;
  }
}

void WaveFile1::int_16_to_sep_flt_2( void const * ibuf, unsigned int cnt, float ** ofbuf )
{
  short const * in12 = (short const *)ibuf;
  float * out1 = ofbuf[0];
  float * out2 = ofbuf[1];
  while( cnt > 0 ) {
    --cnt;
    float f = *in12;
    ++in12;
    f *= 1.f/32768.f;
    *out1 = f;
    ++out1;
    f = *in12;
    ++in12;
    f *= 1.f/32768.f;
    *out2 = f;
    ++out2;
  }
}

void WaveFile1::int_24_to_sep_flt_1( void const * ibuf, unsigned int cnt, float ** ofbuf )
{
  unsigned char const * in12 = (unsigned char const *)ibuf;
  float * out1 = ofbuf[0];
  while( cnt > 0 ) {
    --cnt;
    int i = (int)(((unsigned int)in12[0]<<8) | ((unsigned int)in12[1]<<16) | ((unsigned int)in12[2]<<24));
    in12 += 3;
    *out1 = i * 1.f/2147483648.f;
    ++out1;
  }
}

void WaveFile1::int_24_to_sep_flt_2( void const * ibuf, unsigned int cnt, float ** ofbuf )
{
  unsigned char const * in12 = (unsigned char const *)ibuf;
  float * out1 = ofbuf[0];
  float * out2 = ofbuf[1];
  while( cnt > 0 ) {
    --cnt;
    int i = (int)(((unsigned int)in12[0]<<8) | ((unsigned int)in12[1]<<16) | ((unsigned int)in12[2]<<24));
    in12 += 3;
    *out1 = i * 1.f/2147483648.f;
    ++out1;
    i = (int)(((unsigned int)in12[0]<<8) | ((unsigned int)in12[1]<<16) | ((unsigned int)in12[2]<<24));
    in12 += 3;
    *out2 = i * 1.f/2147483648.f;
    ++out2;
  }
}

void WaveFile1::int_32_to_sep_flt_1( void const * ibuf, unsigned int cnt, float ** ofbuf )
{
  int const * in12 = (int const *)ibuf;
  float * out1 = ofbuf[0];
  while( cnt > 0 ) {
    --cnt;
    float f = float( *in12 );
    ++in12;
    f *= 1.f/2147483648.f;
    *out1 = f;
    ++out1;
  }
}

void WaveFile1::int_32_to_sep_flt_2( void const * ibuf, unsigned int cnt, float ** ofbuf )
{
  short const * in12 = (short const *)ibuf;
  float * out1 = ofbuf[0];
  float * out2 = ofbuf[1];
  while( cnt > 0 ) {
    --cnt;
    float f = *in12;
    ++in12;
    f *= 1.f/2147483648.f;
    *out1 = f;
    ++out1;
    f = *in12;
    ++in12;
    f *= 1.f/2147483648.f;
    *out2 = f;
    ++out2;
  }
}

void WaveFile1::int_flt_to_sep_flt_1( void const * ibuf, unsigned int cnt, float ** ofbuf )
{
  memcpy( ofbuf[0], ibuf, 4*cnt );
}

void WaveFile1::int_flt_to_sep_flt_2( void const * ibuf, unsigned int cnt, float ** ofbuf )
{
  float const * in12 = (float const *)ibuf;
  float * out1 = ofbuf[0];
  float * out2 = ofbuf[1];
  while( cnt > 0 ) {
    --cnt;
    *out1 = *in12;
    ++in12;
    ++out1;
    *out2 = *in12;
    ++in12;
    ++out2;
  }
}




OWaveFile * NewWaveFile()
{
  return new WaveFile1;
}



#if WAVCONVERT_MAIN

#include <windows.h>
#include <setjmp.h>

#include "oresampler.h"
#include "msamplesource.h"
#include "ofloatblock.h"
#include "owaveloadersource.h"

#include "nsp.h"
#include "nsperror.h"

char errorStr[ 600 ];
jmp_buf jmpBuf;

static int __STDCALL myNspErrorFunc(
  NSPStatus status,
  const char *funcname, const char *context,
  const char *file, int line )
{
  sprintf( errorStr, "%s\nfunc: %s\ncontext: %s\nfile: %s:%d",
      nspErrorStr( status ),
      funcname, context, file, line );
  longjmp( jmpBuf, 1 );
  return 0;
}


class WavePump
{
  MSampleSource * input_;
  OWaveFile * output_;
public:
  WavePump()
    {
      input_ = 0;
      output_ = 0;
    }
  ~WavePump()
    {
    }
  void setInput( MSampleSource * i )
    {
      input_ = i;
    }
  void setOutput( OWaveFile * o )
    {
      output_ = o;
    }
  void work( unsigned long count )
    {
      float * buf[ 2 ] = { alignedFloatBlock( 1024 ), alignedFloatBlock( 1024 ) };
      unsigned int total = 0;
      while( total < count ) {
        unsigned int m = 1024;
        if( m > count ) {
          m = count;
        }
#if !defined( NDEBUG )
        memset( buf[0], 0xff, sizeof(float)*1024 );
        memset( buf[1], 0xff, sizeof(float)*1024 );
#endif
        unsigned int n = input_->getData( buf, m );
#if !defined( NDEBUG )
        for( unsigned int ix = 0; ix < 1024; ++ix ) {
          if( ((unsigned int *)buf[0])[ ix ] == 0xffffffff ) {
            assert( 0 );    //  didn't write correctly
          }
        }
#endif
        checkFloatBlock( buf[ 0 ] );
        checkFloatBlock( buf[ 1 ] );
        if( !n ) {
          break;
        }
        total += n;
        output_->writeFrames( buf, n );
        checkFloatBlock( buf[ 0 ] );
        checkFloatBlock( buf[ 1 ] );
        fprintf( stderr, "%6d/%6d\r", total, count );
        fflush( stderr );
      }
      freeFloatBlock( buf[ 0 ] );
      freeFloatBlock( buf[ 1 ] );
    }
};


int main( int argc, char * argv[] ) {
  if( (argc != 4) && (argc != 5) && (argc != 6) ) {
usage:
    fprintf( stderr, "usage: convert input.wav output.wav out-fmt [@out-sr] [out-bits]\n" );
    fprintf( stderr, "out-fmt can be 1 (int) or 3 (float)\n" );
    fprintf( stderr, "out-bits can be 16, 24 or 32 for int, and 32 for float\n" );
    fprintf( stderr, "default for out-bits is 32\n" );
    fprintf( stderr, "if you specify @out-sr, the file will be resampled\n" );
    return 1;
  }
  if( setjmp( jmpBuf ) ) {
    fprintf( stderr, "error: %s\n", errorStr );
    return 5;
  }
  nspRedirectError( myNspErrorFunc );
  char const * inputName = argv[1];
  char const * outputName = argv[2];
  unsigned int resampleTarget = 0;
  int fmt = atoi( argv[3] );
  if( argv[4] && argv[4][0] == '@' ) {
    resampleTarget = atoi( &argv[4][1] );
    if( resampleTarget < 8000 || resampleTarget > 96000 ) {
      fprintf( stderr, "resampling must be to sampling rate between 8000 and 96000 Hz\n" );
      goto usage;
    }
    ++argv;
    --argc;
  }
  if( argv[5] ) {
    goto usage;
  }
  int bits = argv[4] ? atoi( argv[4] ) : 32;
  OWaveFile * input = NewWaveFile();
  OWaveFile * output = NewWaveFile();
  input->setPath( inputName );
  input->setOpenMode( WaveOpenRead );
  if( !input->open() ) {
    fprintf( stderr, "%s: cannot open\n", inputName );
    return 2;
  }
  WAVFmtHeader hdr;
  char extraData[ 100 ];
  unsigned int extraSize = 100;
  if( !input->fmtHeader( &hdr, extraData, &extraSize ) ) {
    fprintf( stderr, "%s: format is bad\n", inputName );
    return 2;
  }
  fprintf( stderr, "%s: format %d bits %d framesize %d rate %d\n", inputName, 
      hdr.format, hdr.bitsPerSample, hdr.frameSize, hdr.framesPerSecond );
  hdr.format = fmt;
  hdr.bitsPerSample = bits;
  hdr.frameSize = hdr.samplesPerFrame * bits / 8;
  unsigned int inputRate = hdr.framesPerSecond;
  if( resampleTarget ) {
    fprintf( stderr, "resampling to %d Hz\n", resampleTarget );
    hdr.framesPerSecond = resampleTarget;
  }
  hdr.bytesPerSecond = hdr.frameSize * hdr.framesPerSecond;
  output->setPath( outputName );
  output->setOpenMode( WaveOpenCreate );
  if( !output->open() ) {
    fprintf( stderr, "%s: cannot create\n", outputName );
    return 3;
  }
  unsigned int cnt = input->countChunks();
  fprintf( stderr, "%s contains %d chunks\n", inputName, cnt );
  float bufs[ 2 ][ 2048 ];
  float * bufPtrs[2] = { bufs[0], bufs[1] };
  char data[ 4096 ];
  for( unsigned int ix = 0; ix < cnt; ++ix ) {
    if( input->chunkType( ix ) == FMT__TAG ) {
      if( !output->setFmtHeader( hdr, extraData, extraSize ) ) {
        fprintf( stderr, "%s: cannot set output format\n", outputName );
        return 3;
      }
    }
    else if( input->chunkType( ix ) == DATA_TAG ) {
      if( !output->beginData() ) {
        fprintf( stderr, "%s: cannot begin data chunk\n", outputName );
        return 3;
      }
      if( resampleTarget != 0 ) {
        OWaveLoaderSource * wls = NewWaveLoaderSource();
        wls->setWaveFile( input );
        wls->start();
        OResampler * r = NewResampler();
        r->setChannelCount( hdr.samplesPerFrame );
        r->setInputFrameRate( float( inputRate ) );
        r->setOutputFrameRate( float( resampleTarget ) );
        r->setMinBlockSize( 2048 );
        WavePump p;
        p.setOutput( output );
        p.setInput( r->create( wls->sampleSource() ) );
//        p.setInput( wls->sampleSource() );
        r->destroy();
        p.work( unsigned long( ceil( input->countFrames() * double( resampleTarget / double( inputRate ) ) ) ) );
        wls->reset();
        wls->setWaveFile( 0 );
        wls->deleteMe();
      }
      else {
        unsigned int fc = input->countFrames();
        unsigned int cur = 0;
        fprintf( stderr, "%s: %d frames\n", inputName, fc );
        while( cur < fc ) {
          fflush( stderr );
          unsigned int tr = 2048;
          if( tr > fc-cur ) {
            tr = fc-cur;
          }
          memset( bufs, 0xfd, sizeof( bufs ) );
          if( !(tr = input->readFrames( bufPtrs, cur, tr )) ) {
            fprintf( stderr, "%s: error reading at frame %d\n", inputName, cur );
            return 2;
          }
          cur += tr;
          fprintf( stderr, "%s: %6d/%6d  [%5.1f%%]\r", outputName, cur, fc, float(cur)/float(fc)*100.f );
          if( !output->writeFrames( bufPtrs, tr ) ) {
            fprintf( stderr, "%s: error writing at frame %d\n", outputName, cur );
            return 3;
          }
        }
      }
      fprintf( stderr, "\n" );
      if( !output->endData() ) {
        fprintf( stderr, "%s: error closing data chunk\n", outputName );
        return 3;
      }
    }
    else {
      unsigned int size = input->chunkSize( ix );
      fprintf( stderr, "%s: chunk %d has size %d\n", inputName, ix, size );
      if( !output->newChunk( input->chunkType( ix ) ) ) {
        fprintf( stderr, "%s: cannot create chunk %d\n", outputName, ix );
        return 3;
      }
      unsigned int off = 0;
      while( off < size ) {
        fprintf( stderr, "%6d/%6d\r", off, size );
        fflush( stderr );
        unsigned int sz = 4096;
        if( sz > size-off ) {
          sz = size-off;
        }
        if( !input->readChunkData( ix, data, off, &sz ) ) {
          fprintf( stderr, "\n%s: error reading\n", inputName );
          return 2;
        }
        if( !output->writeNewData( data, sz ) ) {
          fprintf( stderr, "\n%s: error writing\n", outputName );
          return 3;
        }
        off += sz;
      }
      fprintf( stderr, "\n" );
      if( !output->endChunk() ) {
        fprintf( stderr, "%s: error completing chunk %d\n", outputName, ix );
        return 3;
      }
    }
  }
  fprintf( stderr, "Conversion complete\n" ) ;
  input->destroy();
  output->destroy();
  return 0;
}

#endif
