
#include <windows.h>

#include "owaveplayer.h"
#include "log.h"

#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#define REPORT_UNDERFLOW_ERRORS 1

#pragma warning( disable: 4201 )

#pragma warning( push, 1 )
#include "mmsystem.h"
#include <vector>
using namespace std;
#pragma warning( pop )


class WaveBuffer1
  : public MWaveBuffer
{
public:

  WAVEHDR hdr_;
  HWAVEOUT h_;
  short * shortData_;
  float * floatData_;
  unsigned int channelCount_;
  unsigned int frameCount_;
  unsigned int written_;
  unsigned int played_;
  float mainVolume_;

  WaveBuffer1()
    {
      floatData_ = 0;
      shortData_ = 0;
      channelCount_ = 0;
      frameCount_ = 0;
      written_ = 0;
      played_ = false;
      mainVolume_ = 1;
    }
  ~WaveBuffer1()
    {
      if( floatData_ != 0 || shortData_ != 0 ) {
        waveOutUnprepareHeader( h_, &hdr_, sizeof( hdr_ ) );
        delete[] floatData_;
        delete[] shortData_;
      }
    }
  bool init( unsigned int channels, unsigned int size, unsigned int format, HWAVEOUT h )
    {
      assert( channels == 1 || channels == 2 );
      assert( size >= 128 && size <= 16384 && !(size & (size-1)) );
      assert( format == 1 || format == 3 );
      assert( h );
      h_ = h;
      channelCount_ = channels;
      frameCount_ = size;
      memset( &hdr_, 0, sizeof( hdr_ ) );
      if( format == 1 ) {
        shortData_ = new short[ channels * size ];
        hdr_.lpData = (LPSTR)shortData_;
        hdr_.dwBufferLength = channels * size * 2;
      }
      else {
        floatData_ = new float[ channels * size ];
        hdr_.lpData = (LPSTR)floatData_;
        hdr_.dwBufferLength = channels * size * 4;
      }
      MMRESULT r = waveOutPrepareHeader( h_, &hdr_, sizeof( hdr_ ) );
      if( r != MMSYSERR_NOERROR ) {
        delete[] shortData_;
        shortData_ = 0;
        delete[] floatData_;
        floatData_ = 0;
        return false;
      }
      return true;
    }

  virtual unsigned int size()
    {
      return frameCount_;
    }
  virtual unsigned int setData( float const * const * buffers, unsigned int frames )
    {
      assert( frames == frameCount_ );

      float const * in1 = buffers[ 0 ];
      float const * in2 = buffers[ 1 ];
      float cvt = 32768.0f * mainVolume_;
      if( shortData_ ) {
        if( channelCount_ == 1 ) {
          for( unsigned int ix = 0; ix < frameCount_; ++ix ) {
            float i = in1[ix] * cvt;
            if( i <= -32768.f ) {
              shortData_[ ix ] = -32768;
            }
            else if( i >= 32767.f ) {
              shortData_[ ix ] = 32767;
            }
            else {
              //short * sp = &shortData_[ix];
              short s;
              __asm {
                fld i
                fistp s     //  don't call ftol()!
              }
              shortData_[ix] = s;
            }
          }
        }
        else {
          for( unsigned int ix = 0; ix < frameCount_; ++ix ) {
            float i = in1[ix] * cvt;
            if( i <= -32768.f ) {
              shortData_[ ix*2 ] = -32768;
            }
            else if( i >= 32767.f ) {
              shortData_[ ix*2 ] = 32767;
            }
            else {
              //short * sp = &shortData_[ix*2];
              short s;
              __asm {
                fld i
                fistp s     //  don't call ftol()!
              }
              shortData_[ix*2] = s;
            }
            i = in2[ix] * cvt;
            if( i <= -32768.f ) {
              shortData_[ ix*2+1 ] = -32768;
            }
            else if( i >= 32767.f ) {
              shortData_[ ix*2+1 ] = 32767;
            }
            else {
              //short * sp = &shortData_[ix*2+1];
              short s;
              __asm {
                fld i
                fistp s     //  don't call ftol()!
              }
              shortData_[ix*2+1] = s;
            }
          }
        }
      }
      else {
        if( channelCount_ == 1 ) {
          float cvt = mainVolume_;
          if( mainVolume_ == 1 ) {
            memcpy( floatData_, in1, sizeof( float ) * frameCount_ );
          }
          else {
            for( uint i = 0; i < frameCount_; ++i ) {
              floatData_[i] = in1[i] * cvt;
            }
          }
        }
        else {
          float cvt = mainVolume_;
          for( unsigned int ix = 0; ix < frameCount_; ++ix ) {
            floatData_[ ix*2 ] = in1[ ix ] * cvt;
            floatData_[ ix*2+1 ] = in2[ ix ] * cvt;
          }
        }
      }
      written_ = frames;
      played_ = false;
      return written_;
    }

  void setMainVolume( float f )
    {
      mainVolume_ = f;
    }

private:
  WaveBuffer1( WaveBuffer1 const & );                   //  no go
  WaveBuffer1 & operator=( WaveBuffer1 const & );       //  no go
};

class WavePlayer1
  : public OWavePlayer
  , public MWavePlayer
{
public:

  WavePlayer1();
  ~WavePlayer1();

  //  OWavePlayer
  virtual void destroy();
  virtual void setFormat( unsigned int channelCount, float framesPerSecond );
  virtual void setBuffers( unsigned int count, unsigned int size, bool tryFp );
  virtual void setVolume( float mainVolume );
  virtual bool open();
  virtual void close();
  virtual MWavePlayer * wavePlayer();
  virtual bool start( MWaveDataNotify * provider );
  virtual void stop();

  unsigned int channelCount_;
  float framesPerSecond_;
  unsigned int bufferCount_;
  unsigned int bufferSize_;
  bool tryFp_;
  float mainVolume_;

  //  thread, buffer and device management
  HANDLE event_;
  WAVEFORMATEX wfex_;
  HWAVEOUT waveOut_;
  vector< WaveBuffer1 * > buffers_;
  HANDLE thread_;
  DWORD threadId_;
  MWaveDataNotify * provider_;
  bool running_;
  unsigned int curWrite_;
  unsigned int curReap_;

  static DWORD WINAPI threadEntry( void * );
  void threadFunc();
  unsigned int writeQueuedBuffers();
  void reapAndNotifyBuffers();
};

OWavePlayer * NewWavePlayer()
{
  WavePlayer1 * wp1 = new WavePlayer1;
  return static_cast< OWavePlayer * >( wp1 );
}



WavePlayer1::WavePlayer1()
{
  debug( "WavePlayer1 created 0x%x", this );
  channelCount_ = 0;
  framesPerSecond_ = 0;
  bufferCount_ = 0;
  bufferSize_ = 0;
  tryFp_ = false;
  mainVolume_ = 1;
  event_ = CreateEvent( 0, false, false, 0 );
  assert( event_ != 0 );
  memset( &wfex_, 0, sizeof( wfex_ ) );
  waveOut_ = 0;
  thread_ = 0;
  threadId_ = 0;
  provider_ = 0;
  running_ = false;
}

WavePlayer1::~WavePlayer1()
{
  CloseHandle( event_ );
  debug( "WavePlayer1 deleted 0x%x", this );
}

void WavePlayer1::destroy()
{
  if( waveOut_ != 0 ) {
    close();
  }
  delete this;
}

void WavePlayer1::setFormat( unsigned int channelCount, float framesPerSecond )
{
  assert( waveOut_ == 0 );
  assert( channelCount == 1 || channelCount == 2 );
//  assert( framesPerSecond == 44100.f ); //  oof!
  channelCount_ = channelCount;
  framesPerSecond_ = framesPerSecond;
}

void WavePlayer1::setBuffers( unsigned int count, unsigned int size, bool tryFp )
{
  assert( waveOut_ == 0 );
  assert( count > 1 && count < 10 );
  assert( size >= 32 && size <= 16384 && !(size & (size-1)) );
  bufferCount_ = count;
  bufferSize_ = size;
  tryFp_ = tryFp;
}

void WavePlayer1::setVolume( float mainVolume )
{
  mainVolume_ = mainVolume;
}

bool WavePlayer1::open()
{
  assert( waveOut_ == 0 );
  MMRESULT r = MMSYSERR_NOTSUPPORTED;
  if( tryFp_ ) {
    //  try floating point
    wfex_.wFormatTag = 3;
    wfex_.nChannels = WORD( channelCount_ );
    wfex_.nSamplesPerSec = DWORD( framesPerSecond_ );
    wfex_.nBlockAlign = WORD( 4 * channelCount_ );
    wfex_.wBitsPerSample = 32;
    wfex_.cbSize = 0;
    wfex_.nAvgBytesPerSec = wfex_.nBlockAlign * wfex_.nSamplesPerSec;
#pragma warning( push )
#pragma warning( disable: 4311 )
    r = waveOutOpen( &waveOut_, WAVE_MAPPER, &wfex_, (DWORD)event_, 0, CALLBACK_EVENT );
#pragma warning( pop )
  }
  if( r != MMSYSERR_NOERROR ) {
    //  try short
    wfex_.wFormatTag = 1;
    wfex_.nChannels = WORD( channelCount_ );
    wfex_.nSamplesPerSec = DWORD( framesPerSecond_ );
    wfex_.nBlockAlign = WORD( 2 * channelCount_ );
    wfex_.wBitsPerSample = 16;
    wfex_.cbSize = 0;
    wfex_.nAvgBytesPerSec = wfex_.nBlockAlign * wfex_.nSamplesPerSec;
#pragma warning( push )
#pragma warning( disable: 4311 )
    r = waveOutOpen( &waveOut_, WAVE_MAPPER, &wfex_, (DWORD)event_, 0, CALLBACK_EVENT );
#pragma warning( pop )
  }
  if( r != MMSYSERR_NOERROR ) {
    warning( "Open error: 0x%lx\n", (long)r );
    waveOut_ = 0;
    return false;
  }
  buffers_.resize( bufferCount_ );
  for( unsigned int ix = 0; ix < bufferCount_; ++ix ) {
    buffers_[ ix ] = new WaveBuffer1;
  }
  for( unsigned int ix = 0; ix < bufferCount_; ++ix ) {
    if( !buffers_[ ix ]->init( channelCount_, bufferSize_, wfex_.wFormatTag, waveOut_ ) ) {
      close();
      return false;
    }
  }
  return true;
}

void WavePlayer1::close()
{
  assert( GetCurrentThreadId() != threadId_ );
  assert( waveOut_ );
  if( thread_ ) {
    stop();
    WaitForSingleObject( thread_, INFINITE );
  }
  if( waveOut_ ) {
    waveOutClose( waveOut_ );
    waveOut_ = 0;
  }
  for( unsigned int ix = 0; ix < buffers_.size(); ++ix ) {
    delete buffers_[ ix ];
  }
  buffers_.resize( 0 );
}

MWavePlayer * WavePlayer1::wavePlayer()
{
  return static_cast< MWavePlayer * >( this );
}

bool WavePlayer1::start( MWaveDataNotify * provider )
{
  assert( thread_ == 0 );
  assert( waveOut_ != 0 );
  assert( provider );
  thread_ = CreateThread( 0, 22*4096, threadEntry, (void *)this, CREATE_SUSPENDED, &threadId_ );
  if( thread_ == 0 ) {
    error( "Could not start playing back your sound file. The computer appears to be critically low on system resources." );
    return false;
  }
  SetThreadPriority( thread_, THREAD_PRIORITY_TIME_CRITICAL );
  provider_ = provider;
  curWrite_ = 0;
  curReap_ = 0;
  running_ = true;
  provider_->playbackStarted( this );
  for( unsigned int ix = 0; ix < buffers_.size(); ++ix ) {
    if( running_ ) {
      provider_->bufferCallback( this, buffers_[ ix ] );
    }
  }
  ResumeThread( thread_ );
  return true;
}

void WavePlayer1::stop()
{
  running_ = false;
}

DWORD WINAPI WavePlayer1::threadEntry( void * data )
{
  WavePlayer1 * wp1 = (WavePlayer1 *)data;
  wp1->threadFunc();
  return 0;
}

void WavePlayer1::threadFunc()
{
  unsigned int errCount = 0;
  bool startupImmunity_ = true;
  while( 1 ) {
    if( !writeQueuedBuffers() ) {
      if( curReap_ == curWrite_ ) {
        if( !running_ ) {
          break;
        }
        //  I used to have code here that bombed out, but if we 
        //  circle around without writing any buffers when the 
        //  queue is already full, we would die even though there 
        //  isn't a problem.
      }
    }
    if( WaitForSingleObject( event_, INFINITE ) != WAIT_OBJECT_0 ) {
      ++errCount;
      if( errCount > 5 ) { //  prevent runaway problems
        warning( "errCount is %d after WaitForSingleObject", errCount );
        assert( !"this is a bad case" );  //  although this might happen with some signaling conditions (?)
        break;
      }
    }
    else {
      if( errCount > 0 ) {
        --errCount;
      }
    }
    reapAndNotifyBuffers();
  }
  provider_->playbackStopped( this );
}

unsigned int WavePlayer1::writeQueuedBuffers()
{
  unsigned int cnt = 0;

  //  run through buffers with data and queue them
  while( buffers_[ curWrite_ ]->written_ ) {
    WAVEHDR * phdr = &buffers_[ curWrite_ ]->hdr_;
    MMRESULT r = waveOutWrite( waveOut_, phdr, sizeof( *phdr ) );
    //assert( buffers_[ curReap_ ]->hdr_.dwFlags & WHDR_INQUEUE );
    buffers_[ curWrite_ ]->played_ = true;
    buffers_[ curWrite_ ]->written_ = 0;
    if( r != MMSYSERR_NOERROR ) {
      assert( !"this is a bad error case" );
      error( "Something went wrong with the sound card -- suddenly, I can't play sound through it! (code 0x%x)", r );
      stop();
      return cnt;
    }
    ++cnt;
    ++curWrite_;
    if( curWrite_ == bufferCount_ ) {
      curWrite_ = 0;
    }
  }
  return cnt;
}

void WavePlayer1::reapAndNotifyBuffers()
{
  //  poll for buffers completed
  while( buffers_[ curReap_ ]->played_ && !(buffers_[ curReap_ ]->hdr_.dwFlags & WHDR_INQUEUE) ) {
//    buffers_[ curReap_ ]->hdr_.dwFlags &= ~WHDR_INQUEUE;
    buffers_[ curWrite_ ]->played_ = false;
    if( running_ ) {
      buffers_[ curReap_ ]->setMainVolume( mainVolume_ );
      provider_->bufferCallback( this, buffers_[ curReap_ ] );
    }
    ++curReap_;
    if( curReap_ == bufferCount_ ) {
      curReap_ = 0;
    }
  }
}


#if WAVEPLAYER_MAIN

#include "owavefile.h"

class WavePlayah 
  : public MWaveDataNotify
{
public:
  OWaveFile * file_;
  unsigned int offset_;
  HANDLE event_;

  WavePlayah( OWaveFile * file )
    {
      file_ = file;
      offset_ = 0;
      event_ = CreateEvent( 0, false, false, 0 );
    }
  ~WavePlayah()
    {
      CloseHandle( event_ );
    }

  virtual void playbackStarted( MWavePlayer * wp )
    {
    }
  virtual void bufferCallback( MWavePlayer * wp, MWaveBuffer * buf )
    {
      float flts[ 2048 ];
      float * bufs[ 2 ] = { &flts[ 0 ], &flts[ 1024 ] };
      unsigned int rd = file_->readFrames( bufs, offset_, 1024 );
      if( rd < 1024 ) {
        memset( &flts[ rd ], 0, sizeof( float )*(1024-rd) );
        memset( &flts[ rd+1024 ], 0, sizeof( float )*(1024-rd) );
      }
      buf->setData( bufs, 1024 );
      offset_ += rd;
      fprintf( stderr, "%d\r", offset_ );
      if( rd < 1024 ) {
        wp->stop();
        fprintf( stderr, "\nStopped.\n" );
      }
    }
  virtual void playbackStopped( MWavePlayer * wp )
    {
      SetEvent( event_ );
    }
};

int main( int argc, char * argv[] )
{
  if( (argc != 2 && argc != 3) || argv[1][0] == '-' ) {
    fprintf( stderr, "usage: waveplayer file.wav [2-9]\n" );
    return 1;
  }
  unsigned int n = 3;
  if( argc == 3 ) {
    n = atoi( argv[2] );
    if( n < 2 || n > 9 ) {
      fprintf( stderr, "usage: waveplayer file.wav [2-9]\n" );
      return 1;
    }
  }
  OWaveFile * wf = NewWaveFile();
  wf->setOpenMode( WaveOpenRead );
  wf->setPath( argv[1] );
  if( !wf->open() ) {
    fprintf( stderr, "%s: cannot open\n", argv[1] );
    return 2;
  }
  WAVFmtHeader fmt;
  char data[ 500 ];
  unsigned int dataLen;
  if( !wf->fmtHeader( &fmt, data, &dataLen ) ) {
    fprintf( stderr, "%s: cannot get format\n", argv[1] );
    return 2;
  }
  fprintf( stderr, "%s: %s %d bit %s %.1f Hz\n", argv[1], (fmt.format == 1) ? "integer" : "float", 
      fmt.bitsPerSample, (fmt.samplesPerFrame == 1) ? "mono" : "stereo", float(fmt.framesPerSecond) );
  OWavePlayer * wp = NewWavePlayer();
  wp->setFormat( fmt.samplesPerFrame, float( fmt.framesPerSecond ) );
  wp->setBuffers( n, 1024, true );
  if( !wp->open() ) {
    fprintf( stderr, "%s: cannot open device\n", argv[1] );
    return 3;
  }
  WavePlayah playah( wf );
  if( !wp->start( &playah ) ) {
    fprintf( stderr, "%s: cannot start device\n", argv[1] );
    return 3;
  }
  fprintf( stderr, "%s: playing...\n", argv[1] );
  WaitForSingleObject( playah.event_, INFINITE );
  wp->destroy();
  wf->destroy();
  return 0;
}

#endif
