
#include "scene.h"
#include "glut-start.h"

#include <string.h>
#include <assert.h>

#include <list>


namespace {
  enum ChildCodes {
    lxlylz,
    lxlyhz,
    lxhylz,
    lzhyhz,
    hxlylz,
    hxlyhz,
    hxhylz,
    hzhyhz,
  };
  class _Node {
    public:
      _Node * next;
      _Node * prev;
  };
  class Node : public _Node {
    public:
      Node() { memset( this, 0, sizeof( *this ) ); }
      I_SceneNode * node;
      float radius;
      Bounds bounds;
      OBB obb;
      Matrix4 posOri;
  };
  class Cell : public _Node {
    public:
      Cell() { memset( this, 0, sizeof( *this ) ); next = prev = this; }
      Cell( Cell const & o ) {
        memcpy( this, &o, sizeof( *this ) );
        if( o.next == &o ) { next = this; prev = this; }
      }
      Cell & operator=( Cell const & o ) {
        memcpy( this, &o, sizeof( *this ) );
        if( o.next == &o ) { next = this; prev = this; }
      }
      Cell * children[8];
  };
  enum Constants {
    CellLevels = 12,
    MinLevel = 4,
    ChildSizeAtTopLevel = (1<<(CellLevels)),
    MinChildSize = (1<<(MinLevel)),
  };
  class Scene : public I_Scene {
    public:
      Scene() {
        low_ = Vector3::Null;
        high_ = Vector3::Null;
        numNodes_ = 0;
        root_ = 0;
        cellsFree_ = 0;
        nodesFree_ = 0;
        worldLow_.x = worldLow_.y = worldLow_.z = -ChildSizeAtTopLevel;
        worldHigh_.x = worldHigh_.y = worldHigh_.z = ChildSizeAtTopLevel;
        root_ = newCell();
      }

      virtual void dispose();
      virtual SceneRef addNode( I_SceneNode * node, Matrix4 const & posOri );
      virtual void removeNode( SceneRef ref );
      virtual void moveNode( SceneRef ref, Matrix4 const & posOri );
      virtual void getNodePos( SceneRef ref, Matrix4 * outPosOri );
      virtual void visitNodes( I_SceneVisitor * visitor );
      virtual void visitNodesAABB( Vector3 const & low, Vector3 const & high, I_SceneVisitor * visitor );
      virtual void visitNodesPlanes( Vector4 const * planes, uint numPlanes, I_SceneVisitor * visitor );

      Vector3 low_;
      Vector3 high_;
      Vector3 worldLow_;
      Vector3 worldHigh_;
      uint numNodes_;
      Cell * root_;
      std::list< Cell > cells_;
      Cell * cellsFree_;
      std::list< Node > nodes_;
      Node * nodesFree_;

      Cell * newCell() {
        if( cellsFree_ ) {
          Cell * ret = cellsFree_;
          cellsFree_ = cellsFree_->children[0];
          new( ret ) Cell();
          return ret;
        }
        static Cell c;
        cells_.push_back(c);
        return &cells_.back();
      }
      void deleteCell( Cell * c ) {
        c->~Cell();
        c->children[0] = cellsFree_;
        cellsFree_ = c;
      }
      Node * newNode() {
        if( nodesFree_ ) {
          Node * ret = nodesFree_;
          nodesFree_ = static_cast< Node * >( nodesFree_->next );
          new( ret ) Node();
          return ret;
        }
        static Node n;
        nodes_.push_back(n);
        return &nodes_.back();
      }
      void deleteNode( Node * n ) {
        n->~Node();
        n->next = nodesFree_;
        nodesFree_ = n;
      }
      static SceneRef refFromNode( Node * n ) {
        return reinterpret_cast< SceneRef >( n );
      }
      static Node * nodeFromRef( SceneRef r ) {
        return reinterpret_cast< Node * >( r );
      }

      void insertNode( Node * n );
      void removeNode( Node * n );
      Cell * findCell( Cell * c, Vector3 low, Vector3 const & pos, uint r, uint size );
      template< class Culler > static void visit( Cell * c, Vector3 const & lo, uint size, Culler & cull, bool knownVisible = false );

  };
  class CullNone {
    public:
      CullNone( I_SceneVisitor * v, I_Scene * scene ) : visitor_( v ), scene_( scene ) {}
      I_SceneVisitor * visitor_;
      I_Scene * scene_;
      void considerNode( Node * n, bool knownVisible = false ) const {
        visitor_->onSceneVisit( scene_, Scene::refFromNode( n ), n->node, n->posOri );
      }
      bool considerCell( Vector3 const & v, uint s, bool * outKnownVisible ) const {
        *outKnownVisible = true;
        return true;
      }
  };
  class CullAABB {
    public:
      CullAABB( Vector3 const & lo, Vector3 const & hi, I_SceneVisitor * v, I_Scene * scene ) : 
          low_( lo ), high_( hi ), visitor_( v ), scene_( scene ) {}
      Vector3 low_;
      Vector3 high_;
      I_SceneVisitor * visitor_;
      I_Scene * scene_;
      void considerNode( Node * n, bool knownVisible = false ) const {
        if( knownVisible || 
            (PointRadiusIntersectsAABB( n->posOri.pos(), n->radius, low_, high_ ) &&
            OBBIntersectsAABB( n->obb, low_, high_ )) ) {
          visitor_->onSceneVisit( scene_, Scene::refFromNode( n ), n->node, n->posOri );
        }
      }
      bool considerCell( Vector3 const & v, uint s, bool * outKnownVisible ) const {
        Vector3 w( v.x+s, v.y+s, v.z+s );
        return AABBIntersectsAABB( v, w, low_, high_, outKnownVisible );
      }
  };
  class CullPlanes {
    public:
      CullPlanes( Vector4 const * p, uint n, I_SceneVisitor * v, I_Scene * scene ) : 
          planes_(p), numPlanes_(n), visitor_( v ), scene_( scene ) {}
      I_SceneVisitor * visitor_;
      I_Scene * scene_;
      Vector4 const * planes_;
      uint numPlanes_;
      void considerNode( Node * n, bool knownVisible = false ) const {
        if( knownVisible || 
            (PointRadiusIntersectsPlanes( n->posOri.pos(), n->radius, planes_, numPlanes_ ) &&
            OBBIntersectsPlanes( n->obb, planes_, numPlanes_ )) ) {
          visitor_->onSceneVisit( scene_, Scene::refFromNode( n ), n->node, n->posOri );
        }
      }
      bool considerCell( Vector3 const & v, uint s, bool * outKnownVisible ) const {
      return true;
        Vector3 w( v.x+s, v.y+s, v.z+s );
        return AABBIntersectsPlanes( v, w, planes_, numPlanes_, outKnownVisible );
      }
  };
}

void Scene::dispose()
{
  assert( numNodes_ == 0 );
  //  don't need to dispose anything, becase the lists do the memory management
  delete this;
}

SceneRef Scene::addNode( I_SceneNode * node, Matrix4 const & posOri )
{
  Node * n = newNode();
  n->posOri = posOri;
  n->node = node;
  node->properties( &n->bounds, &n->radius );
  insertNode( n );
  return refFromNode( n );
}

void Scene::removeNode( SceneRef ref )
{
  Node * n = nodeFromRef( ref );
  removeNode( n );
}

void Scene::moveNode( SceneRef ref, Matrix4 const & posOri )
{
  Node * n = nodeFromRef( ref );
  //fixme: check whether the center has actually moved cells
  removeNode( n );
  n->posOri = posOri;
  insertNode( n );
}

void Scene::getNodePos( SceneRef ref, Matrix4 * outPosOri )
{
  Node * n = nodeFromRef( ref );
  *outPosOri = n->posOri;
}

void Scene::visitNodes( I_SceneVisitor * visitor )
{
  visit( root_, worldLow_, ChildSizeAtTopLevel, CullNone( visitor, this ) );
}

void Scene::visitNodesAABB( Vector3 const & low, Vector3 const & high, I_SceneVisitor * visitor )
{
  visit( root_, worldLow_, ChildSizeAtTopLevel, CullAABB( low, high, visitor, this ) );
}

void Scene::visitNodesPlanes( Vector4 const * planes, uint numPlanes, I_SceneVisitor * visitor )
{
  visit( root_, worldLow_, ChildSizeAtTopLevel, CullPlanes( planes, numPlanes, visitor, this ) );
}

template< class Culler >
void Scene::visit( Cell * c, Vector3 const & lo, uint size, Culler & cull, bool knownVisible )
{
  _Node * n = c->next;
  while( n != c ) {
    cull.considerNode( static_cast< Node * >( n ) );
    n = n->next;
  }
  if( knownVisible ) {
    for( uint i = 0; i < 8; ++i ) {
      if( c->children[i] ) {
        visit( c->children[i], lo, size, cull, true );
      }
    }
    return;
  }
  for( uint i = 0; i < 8; ++i ) {
    if( !c->children[i] ) {
      continue;
    }
    Vector3 low = lo;
    if( i & 4 ) {
      low.x += size;
    }
    if( i & 2 ) {
      low.y += size;
    }
    if( i & 1 ) {
      low.z += size;
    }
    assert( size >= 2 );
    uint s2 = size >> 1;
    //  This is a loose octree; each child is inflated by half its size 
    //  each way for purposes of visit testing.
    Vector3 low2( low.x - s2, low.y - s2, low.z - s2 );
    if( !cull.considerCell( low2, size + size, &knownVisible ) ) {
      continue;
    }
    visit( c->children[i], low, s2, cull, knownVisible );
  }
}

void Scene::insertNode( Node * n )
{
  assert( !n->next && !n->prev );
  n->obb.set( n->bounds, n->posOri );
  Cell * c = findCell( root_, worldLow_, (Vector3 const &)n->posOri.xw, (uint)ceil(n->radius), ChildSizeAtTopLevel );
  n->next = c->next;
  n->prev = c;
  c->next = n;
  n->next->prev = n;
  ++numNodes_;
}

void Scene::removeNode( Node * n )
{
  //fixme: if I run low on memory, removing and recycling empty cells here would help.
  assert( numNodes_ > 0 );
  assert( n->prev && n->next );
  n->prev->next = n->next;
  n->next->prev = n->prev;
  n->next = n->prev = 0;
  --numNodes_;
}

Cell * Scene::findCell( Cell * c, Vector3 low, Vector3 const & pos, uint r, uint size )
{
  if( r < MinChildSize ) {
    r = MinChildSize;
  }
again:
  if( r >= size ) {
    return c;
  }
  uint ix = 0;
  if( pos.x >= low.x+size ) {
    ix += 4;
    low.x += size;
  }
  if( pos.y >= low.y+size ) {
    ix += 2;
    low.y += size;
  }
  if( pos.z >= low.z+size ) {
    ix += 1;
    low.z += size;
  }
  if( !c->children[ix] ) {
    c->children[ix] = newCell();
  }
  c = c->children[ix];
  size >>= 1;
  goto again;
}


I_Scene * MakeScene()
{
  return new Scene();
}

class FakeNode : public I_SceneNode, public I_SceneVisitor {
  public:
    FakeNode() { obb_.set(Vector3(1,1,1),Vector3(3,4,5)); r_ = 3; vCount_ = 0; }
    Bounds obb_;
    uint vCount_;
    float r_;
    SceneRef vRef_;
    Matrix4 vPosOri_;
    void properties( Bounds * bounds, float * radius ) { *bounds = obb_; *radius = r_; }
    void onSceneVisit( I_Scene * scene, SceneRef ref, I_SceneNode * node, Matrix4 const & posOri ) {
      ++vCount_;
      vRef_ = ref;
      vPosOri_ = posOri;
    }
};
void UnittestScene()
{
  FakeNode fn1;
  FakeNode fn2;
  Scene * s = new Scene();
  Matrix4 posOri = Matrix4::Null;
  fn1.r_ = 1;
  SceneRef sr1 = s->addNode( &fn1, posOri );
  assert( sr1 != 0 );
  fn2.r_ = 1000;
  SceneRef sr2 = s->addNode( &fn2, posOri );
  assert( sr2 != 0 );
  posOri.pos() = Vector3( 2000, 0, -500 );
  s->moveNode( sr1, posOri );
  posOri.pos() = Vector3( -3000, 500, 0 );
  s->moveNode( sr2, posOri );
  Matrix4 m4;
  s->getNodePos( sr1, &m4 );
  assert( m4.pos().x == 2000 );
  assert( m4.pos().y == 0 );
  assert( m4.pos().z == -500 );
  s->visitNodes( &fn1 );
  assert( fn1.vCount_ == 2 );
  Vector3 p( 0, 0, 0 );
  s->visitNodesAABB( p, p, &fn2 );
  assert( fn2.vCount_ == 0 );
  Vector3 q( -5000, -5000, -5000 );
  p = Vector3( 5000, 5000, 5000 );
  s->visitNodesAABB( q, p, &fn2 );
  assert( fn2.vCount_ == 2 );
  fn1.vCount_ = 0;
  s->visitNodesPlanes( 0, 0, &fn1 );
  assert( fn1.vCount_ == 2 );
  s->removeNode( sr2 );
  s->removeNode( sr1 );
  s->dispose();
}
