
#include <windows.h>
#include <math.h>

#include "soundmanager.h"

#include "owavefile.h"
#include "owaveplayer.h"

#include "log.h"


//  Make this smaller (say, 1024 or 512) to play with 
//  lower latency -- but some systems will start crackling 
//  with too small buffers.
#define WAV_BUFFER_SIZE 1024
#define WAV_NUM_BUFFERS 3
//  Block size for mixing.
#define BLOCK_SIZE 256
//  Maximum value of "pitch"
#define MAX_PITCH 4


struct SoundInstance {
  //  Must have next, then prev, at top
  SoundInstance * next;
  SoundInstance ** prev;
  int handleRefCount;
  int playingRefCount;
  char * path;
  float * data;
  size_t sampleCount;
  PlayingSound * completed;
};


//  fixme: low-pass filter for anti-alias
//  Should keep a lowpass filter around for pitch bending
//  to avoid aliasing artifacts. Filter should probably 
//  always be on for pitch > 1.0 OR not auto-stop (i e, 
//  may be modulated).
//  Use lowpassfilter1D.h, perhaps (though that's not yet tested)
struct PlayingSound {
  //  Must have next, then prev, at top
  PlayingSound * next;
  PlayingSound ** prev;
  SoundInstance * sound;
  PlayParameters params;
  float curPos;
};

namespace {
  class Lock {
      CRITICAL_SECTION & c_;
    public:
      Lock( CRITICAL_SECTION & c ) : c_( c ) {
        EnterCriticalSection( &c_ );
      }
      ~Lock() {
        LeaveCriticalSection( &c_ );
      }
  };
}


class SoundManager1 : public SoundManager, public MWaveDataNotify {
  public:
    virtual void dispose();
    virtual SoundHandle loadSound( char const * path );
    virtual void unloadSound( SoundHandle h );
    virtual PlayingHandle playInstance( SoundHandle h, float volume );
    virtual PlayingHandle playInstance( SoundHandle h, PlayParameters const * p );
    virtual void stopPlaying( PlayingHandle h );
    virtual void setVolume( PlayingHandle h, float volL, float volR );
    virtual void setVolume( float mainVolume );
    virtual void setPitch( PlayingHandle h, float relative );
    //virtual void setMuffle( PlayingHandle h, float muffle );

    //  called from within start()
    virtual void playbackStarted( MWavePlayer * wp );
    //  called from another thread (!)
    virtual void bufferCallback( MWavePlayer * wp, MWaveBuffer * buf );
    //  called from another thread (!)
    virtual void playbackStopped( MWavePlayer * wp );

    SoundManager1();
    ~SoundManager1();

    void stopAllPlaying( SoundInstance * s );
    void decrementHandle( SoundInstance * s );
    void decrementPlaying( SoundInstance * s );
    void checkForZeroRefs( SoundInstance * s );
    void donePlaying( PlayingSound * p );
    void stopInternal( PlayingSound * p );
    void auditLists();
    void auditList( void ** head );

    void mix( unsigned int cnt, float * left, float * right );
    void mixInto( unsigned int cnt, float * left, float * right, PlayingSound * ps );

    void open();
    void close();

    SoundInstance * newSoundInstance();
    PlayingSound * newPlayingSound( SoundInstance * i );

    SoundInstance * sounds_;  //  THREAD STATE
    PlayingSound * playing_;  //  THREAD STATE
    
    SoundInstance * freeInstance_;
    PlayingSound * freePlaying_;

    CRITICAL_SECTION lock_;
    HANDLE stopEvent_;
    OWavePlayer * owp_;
    float mainVolume_;
};


void SoundManager1::dispose()
{
  delete this;
}

SoundHandle SoundManager1::loadSound( char const * path )
{
  if( !owp_ ) {
    return 0;
  }

  {
    Lock lock( lock_ );
    SoundInstance * si = sounds_;
    while( si != 0 ) {
      if( !stricmp( path, si->path ) ) {
        ++si->handleRefCount;
        return si;
      }
      si = si->next;
    }
  }
  OWaveFile * wf = NewWaveFile();
  wf->setPath( path );
  wf->setOpenMode( WaveOpenRead );
  if( !wf->open() ) {
    warning( "Could not find WAV file %s\n", path );
    wf->destroy();
    return 0;
  }
  SoundHandle ret = newSoundInstance();
  ret->path = strdup( path );
  strlwr( ret->path );

  //  Pad up to 4 * WAVE_BUFFER_SIZE so that we can play 
  //  with cubic interpolation at 4x the pitch.
  //  Add 1 for the interpolator (silent first sample).
  size_t frames = wf->countFrames();
  if( frames < WAV_BUFFER_SIZE ) {
    warning( "warning! sounds %s must be at least %d samples (is %d)\n",
        path, WAV_BUFFER_SIZE, frames );
    frames = WAV_BUFFER_SIZE;
  }
  ret->data = new float[ frames + MAX_PITCH * (3 + BLOCK_SIZE) + 1 ];
  ret->sampleCount = frames;
  memset( ret->data, 0, sizeof(float) * (frames + MAX_PITCH * (3 + BLOCK_SIZE) + 1) );

  //  read the data, down-mix from stereo to mono if it's stereo
  float temp[2][WAV_BUFFER_SIZE];
  float * ptrs[2] = { temp[0], temp[1] };
  size_t offset = 0;
  WAVFmtHeader hdr;
  unsigned int tsize = sizeof( temp );
  if( !wf->fmtHeader( &hdr, temp, &tsize ) || (hdr.samplesPerFrame > 2) ) {
    error( "Could not get format from WAV file %s (or not mono or stereo).\n", path );
    goto error_silence;
  }
  while( offset < frames ) {
    unsigned int toRead;
    if( frames > offset + WAV_BUFFER_SIZE ) {
      toRead = WAV_BUFFER_SIZE;
    }
    else {
      toRead = (unsigned int)(frames - offset);
    }
    unsigned int cnt = wf->readFrames( ptrs, (unsigned int)offset, toRead );
    if( cnt == 0 || cnt > 0x7fffffff ) {
      error( "Could not read data from WAV file %s. (0x%x read)\n", path, cnt );
      break;
    }
    if( hdr.samplesPerFrame == 1 ) {
      memcpy( ret->data+offset+1, temp[0], sizeof(float)*cnt );
    }
    else {
      for( unsigned int i = 0; i < cnt; ++i ) {
        ret->data[offset+i+1] = (temp[0][i] + temp[1][i]) * 0.5f;
      }
    }
    offset += cnt;
  }
error_silence:
  //  we're done with this wave file object
  wf->destroy();
  return ret;
}

void SoundManager1::unloadSound( SoundHandle h )
{
  if( !owp_ ) {
    return;
  }

  Lock lock( lock_ );
  decrementHandle( h );
}

PlayingHandle SoundManager1::playInstance( SoundHandle h, float volume )
{
  if( !owp_ ) {
    return 0;
  }

  PlayParameters pp;
  pp.volL = volume;
  pp.volR = volume;
  pp.pitch = 1.0f;
  pp.looping = false;
  pp.autoStop = true;
  return playInstance( h, &pp );
}

PlayingHandle SoundManager1::playInstance( SoundHandle h, PlayParameters const * p )
{
  assert( !(p->looping && p->autoStop) );

  if( !owp_ ) {
    return 0;
  }
  info( "play %s vol %.1f,%.1f %s pitch %.1f\n",
      h->path, p->volL, p->volR, p->looping ? "looping" : p->autoStop ? "autoStop" : "", p->pitch );
  Lock lock( lock_ );
  PlayingHandle ret = newPlayingSound( h );
  ret->curPos = 0;
  ret->params = *p;
  auditLists();
  return ret;
}

void SoundManager1::stopPlaying( PlayingHandle h )
{
  if( !owp_ ) {
    return;
  }

  Lock lock( lock_ );
  auditLists();
#if !defined( NDEBUG )
  //  first, check the waiting list of the sound
  bool found = false;
  PlayingSound * ps;
  ps = h->sound->completed;
  while( ps != 0 ) {
    if( ps == h ) {
      found = true;
      break;
    }
    ps = ps->next;
  }
  //  then, check the active list
  if( !found ) {
    ps = playing_;
    while( ps != 0 ) {
      if( ps == h ) {
        found = true;
        break;
      }
      ps = ps->next;
    }
  }
  //  if this assert pops, you're cancelling a sound twice!
  assert( found );
#endif
  stopInternal( h );
}

void SoundManager1::stopInternal( PlayingSound * p )
{
  auditLists();
  //  unlink from whatever list it's on
  *p->prev = p->next;
  if( p->next ) {
    p->next->prev = p->prev;
  }

  //  link into free list
  p->next = freePlaying_;
  p->prev = &freePlaying_;
  if( freePlaying_ ) {
    freePlaying_->prev = &p->next;
  }
  freePlaying_ = p;

  auditLists();
  //  maybe we're done now?
  decrementPlaying( p->sound );
}


void SoundManager1::auditLists()
{
  auditList( (void **)&sounds_ );
  auditList( (void **)&playing_ );
  auditList( (void **)&freeInstance_ );
  auditList( (void **)&freePlaying_ );
  SoundInstance * si = sounds_;
  while( si ) {
    auditList( (void **)&si->completed );
    si = si->next;
  }
}

namespace {
  struct LH {
    LH * next;
    LH ** prev;
  };
};
void SoundManager1::auditList( void ** head )
{
  LH ** ptr = (LH **)head;
  while( *ptr ) {
    assert( (*ptr)->prev == ptr );
    ptr = &(*ptr)->next;
  }
}

void SoundManager1::setVolume( PlayingHandle h, float volL, float volR )
{
  if( !owp_ ) {
    return;
  }
  assert( !h->params.autoStop );

  Lock lock( lock_ );
  h->params.volL = volL;
  h->params.volR = volR;
}

void SoundManager1::setVolume( float mainVolume )
{
  mainVolume_ = mainVolume;
  if( !owp_ ) {
    return;
  }
  owp_->setVolume( mainVolume_ );
}

void SoundManager1::setPitch( PlayingHandle h, float relative )
{
  if( !owp_ ) {
    return;
  }
  if( relative < 0.01f ) {
    relative = 0.01f;
  }
  else if( relative > MAX_PITCH ) {
    relative = MAX_PITCH;
  }
  assert( !h->params.autoStop );
  Lock lock( lock_ );
  h->params.pitch = relative;
}

//void SoundManager1::setMuffle( PlayingHandle h, float muffle )
//{
//}

SoundManager1::SoundManager1()
{
  sounds_ = 0;
  playing_ = 0;
  freePlaying_ = 0;
  freeInstance_ = 0;
  owp_ = 0;
  mainVolume_ = 1;
  stopEvent_ = CreateEvent( 0, false, false, 0 );
  InitializeCriticalSection( &lock_ );

  open();
}

SoundManager1::~SoundManager1()
{
  close();

  CloseHandle( stopEvent_ );
  DeleteCriticalSection( &lock_ );
}

void SoundManager1::stopAllPlaying( SoundInstance * s )   //  THREAD FUNCTION
{
  //  assert that the lock is held!
assert( !"this function isn't really tested" );

  auditLists();

  //  check playing sounds
  PlayingSound * ps = playing_;
  while( ps != 0 ) {
    PlayingSound * toStop = ps;
    ps = ps->next;
    if( toStop->sound == s ) {
      stopInternal( toStop );
    }
  }
  //  check completed sounds
  ps = s->completed;
  while( ps ) {
    PlayingSound * toStop = ps;
    ps = ps->next;
    stopInternal( toStop );
  }
}

void SoundManager1::decrementHandle( SoundInstance * s )
{
  //  assert lock is held
  assert( s->handleRefCount > 0 );
  s->handleRefCount--;
  if( !s->handleRefCount ) {
    //  I don't necessarily want to stop, just because I've 
    //  released the handle. I'll release playing sound 
    //  handles when I'm done.
    //stopAllPlaying( s );
    checkForZeroRefs( s );
  }
}

void SoundManager1::decrementPlaying( SoundInstance * s )
{
  //  there's one less voice playing this sound
  --s->playingRefCount;
  if( !s->playingRefCount ) {
    checkForZeroRefs( s );
  }
}

void SoundManager1::checkForZeroRefs( SoundInstance * s )   //  THREAD FUNCTION
{
  // assert that the lock is held!
  if( s->handleRefCount == 0 && s->playingRefCount == 0 ) {
    //  the sound is done!
    debug( "sound %s has 0 handle and 0 playing refs\n", s->path );
    //  un-load it
    free( s->path ); s->path = 0;
    delete[] s->data; s->data = 0;

    //  un-link it
    *s->prev = s->next;
    if( s->next ) {
      s->next->prev = s->prev;
    }

    //  put it in the free list
    s->next = freeInstance_;
    s->prev = &freeInstance_;
    if( freeInstance_ ) {
      freeInstance_->prev = &s->next;
    }
    freeInstance_ = s;
    //  done
  }
}

void SoundManager1::donePlaying( PlayingSound * p )        //  THREAD FUNCTION
{
  // assert that the lock is held!
  assert( p->sound->playingRefCount > 0 );

  //  un-link from playing list
  *p->prev = p->next;
  if( p->next ) {
    p->next->prev = &p->next;
  }

  //  link to waiting list
  SoundInstance * s = p->sound;
  p->next = s->completed;
  p->prev = &s->completed;
  if( s->completed ) {
    s->completed->prev = &p->next;
  }
  s->completed = p;
}

void SoundManager1::open()
{
  owp_ = NewWavePlayer();
  owp_->setFormat( 2, 44100 );
  owp_->setBuffers( WAV_NUM_BUFFERS, WAV_BUFFER_SIZE, true );
  if( !owp_->open() ) {
    error( "Could not create wave player -- bailing!\n" );
    assert( !"how did this happen?" );
    owp_->close();
    owp_->destroy();
    owp_ = 0;
  }
  owp_->setVolume( mainVolume_ );
  owp_->wavePlayer()->start( this );
}

void SoundManager1::close()
{
  if( owp_ ) {
    owp_->close();
    WaitForSingleObject( stopEvent_, INFINITE );
  }
}

PlayingSound * SoundManager1::newPlayingSound( SoundInstance * instance )
{
  //  Must have a handle to actually play a sound
  assert( instance->handleRefCount > 0 );
  if( freePlaying_ == 0 ) {
    freePlaying_ = new PlayingSound;
    freePlaying_->next = 0;
    freePlaying_->prev = &freePlaying_;
  }

  Lock lock( lock_ );

  freePlaying_->sound = instance;
  ++instance->playingRefCount;
  freePlaying_->params.volL = freePlaying_->params.volR = 0.5f;
  //freePlaying_->params.muffle = 0;
  freePlaying_->params.pitch = 1;
  freePlaying_->params.looping = false;
  freePlaying_->params.autoStop = false;
  freePlaying_->curPos = 0;

  PlayingSound * ps = freePlaying_->next;
  
  freePlaying_->next = playing_;
  freePlaying_->prev = &playing_;
  if( playing_ ) {
    playing_->prev = &freePlaying_->next;
  }
  playing_ = freePlaying_;

  freePlaying_ = ps;
  if( ps ) {
    ps->prev = &freePlaying_;
  }

  auditLists();

  return playing_;
}

SoundInstance * SoundManager1::newSoundInstance()
{
  if( freeInstance_ == 0 ) {
    freeInstance_ = new SoundInstance;
    freeInstance_->next = 0;
    freeInstance_->prev = &freeInstance_;
  }

  Lock lock( lock_ );

  freeInstance_->handleRefCount = 1;
  freeInstance_->playingRefCount = 0;
  freeInstance_->path = 0;
  freeInstance_->data = 0;
  freeInstance_->sampleCount = 0;
  freeInstance_->completed = 0;

  SoundInstance * si = freeInstance_->next;

  freeInstance_->next = sounds_;
  freeInstance_->prev = &sounds_;
  if( sounds_ ) {
    sounds_->prev = &freeInstance_->next;
  }
  sounds_ = freeInstance_;

  freeInstance_ = si;
  if( si ) {
    si->prev = &freeInstance_;
  }

  auditLists();

  return sounds_;
}

void SoundManager1::playbackStarted( MWavePlayer * wp )
{
}

void SoundManager1::mix( unsigned int cnt, float * left, float * right )
{
  //  Set the rounding control the way we want it.
  //  Because it might change, I'm setting it ahead of 
  //  each larger buffer mix.
  unsigned short rc;
  __asm {
    fstcw word ptr [rc]
    and word ptr [rc], 0xf3ff
    or word ptr [rc], 0x0400
    fldcw word ptr [rc]
  }
  memset( left, 0, sizeof(float)*cnt );
  memset( right, 0, sizeof(float)*cnt );
  //  assert that the lock is held
  PlayingSound * ps = playing_;
  while( ps != 0 ) {
    mixInto( cnt, left, right, ps );
    PlayingSound * maybeDone = ps;
    ps = ps->next;
    if( maybeDone->curPos > maybeDone->sound->sampleCount ) {
      //fixme: this is not the right handling of looping!
      //  That should be done within the mixing loop.
      if( maybeDone->params.looping ) {
        maybeDone->curPos -= maybeDone->sound->sampleCount;
      }
      else if( maybeDone->params.autoStop ) {
        stopInternal( maybeDone );
      }
      else {
        donePlaying( maybeDone );
      }
    }
  }
}

float cubic_hermite( float const * data, float t )
{
  float const y0 = data[-1];
  float const y1 = data[0];
  float const y2 = data[1];
  float const y3 = data[2];

  float const c0 = y1;
  float const c1 = 0.5f * (y2 - y0);
  float const c3 = 1.5f * (y1 - y2) + 0.5f * (y3 - y0);
  float const c2 = y0 - y1 + c1 - c3;

  return ((c3 * t + c2) * t + c1) * t + c0;
}

void SoundManager1::mixInto( unsigned int cnt, float * left, float * right, PlayingSound * ps )
{
  float * data = ps->sound->data + 1;
  float volL = ps->params.volL;
  float volR = ps->params.volR;
  float pitch = ps->params.pitch;
  float curPos = ps->curPos;
  while( cnt > 0 ) {
    unsigned int ipos;
    float delta;
#if 0
    ipos = (unsigned int)floorf( curPos );
    delta = curPos - ipos;
#else
    __asm {
      fld curPos
      fist ipos
      fild ipos
      fsubp st(1), st(0)
      fstp delta
    }
#endif
    float s = cubic_hermite( data + ipos, delta );
    *left++ += s * volL;
    *right++ += s * volR;
    //fixme:  Need to lowpass filter the data when pitch > 1
    curPos += pitch;
    --cnt;
  }
  ps->curPos = curPos;
}

void SoundManager1::bufferCallback( MWavePlayer * wp, MWaveBuffer * buf )
{
  Lock lock( lock_ );

  float temp[2][WAV_BUFFER_SIZE];
  float * ptrs[2] = {
    temp[0], temp[1]
  };
  unsigned int toGo = buf->size();
  //  I require BLOCK_SIZE byte multiple sizes
  assert( !(toGo & (BLOCK_SIZE-1)) );
  unsigned int offset = 0;
  while( offset < toGo ) {
    mix( BLOCK_SIZE, ptrs[0]+offset, ptrs[1]+offset );
    offset += BLOCK_SIZE;
  }
  buf->setData( ptrs, offset );
}

void SoundManager1::playbackStopped( MWavePlayer * wp )
{
  SetEvent( stopEvent_ );
}

SoundManager * NewSoundManager()
{
  SoundManager1 * ret = new SoundManager1;
  return ret;
}
