
#include <windows.h>
#include <GL/GL.h>

#include "extsetup.h"
#include "textures.h"

#include "terrain.h"
#include "misc.h"
#include "image.h"
#include "mymath.h"
#include "extsetup.h"


#define SHOW_LOD 0

//TODO: factor the chunk size, texel scale, and other 
//  parameters to be less hard-coded.

class C_Terrain : public I_Terrain, public I_TerrainODE {

  public:
    class Chunk : public I_TerrainChunk {
      public:
        Chunk() : terrain_( 0 ) {}
        ~Chunk() {}
        virtual Vector3 center() { return center_; }
        virtual Vector3 origin() { return origin_; }
        virtual Vector3 lower() { return lower_; }
        virtual Vector3 upper() { return upper_; }
        virtual void measure( Vector3 & oCenter, Vector3 & oOrigin, Vector3 & oLower, Vector3 & oUpper )
                                { oCenter = center_; oOrigin = origin_; oLower = lower_; oUpper = upper_; }
        virtual void issue( unsigned int lod, unsigned int neighbors );

        C_Terrain * terrain_;
        unsigned int vStart_;
        unsigned int nVertices_;
        Vector3 center_;
        Vector3 origin_;
        Vector3 lower_;
        Vector3 upper_;
    };

    struct Vertex {
      Vector3 pos;
      Vector3 normal;
      float   uv[2];
    };
    typedef int Index;

    enum {
      NegativeUEdge = 1,
      PositiveUEdge = 2,
      NegativeVEdge = 4,
      PositiveVEdge = 8
    };

    C_Terrain();
    ~C_Terrain();

    bool load( Image const & i, float scale, float offset );

    //  class I_TerrainODE
    virtual dTriMeshDataID triMeshDataID();
    virtual dGeomID makeGeom( dSpaceID w );

    //  class I_Terrain
    virtual void dispose();
    virtual void measure( float & west, float & south, float & width, float & height );
    virtual float heightAt( float u, float v );
    virtual unsigned int groundAt( float u, float v );
    virtual void issueAll( Vector3 const & camPos );
    virtual unsigned int numChunksU();
    virtual unsigned int numChunksV();
    virtual unsigned int chunkWidth();
    virtual unsigned int chunkHeight();
    virtual I_TerrainChunk * chunkUV( unsigned int u, unsigned int v );
    virtual I_TerrainODE * terrainODE();

    Chunk ** chunks_;
    unsigned int uChunks_;
    unsigned int vChunks_;
    enum {
      ChunkSize = 32,
      TexelWidth = 4,
    };
    float width_;
    float height_;
    float west_;
    float south_;
    float scale_;
    float offset_;
    float texelCoord_;

    GLuint vertexObject_;
    Vertex * renVerts_;
    Vector3 * collVerts_;
    // LOD, neighbor-lower (u+,u-,v+,v-)
    Index * collIndices_;
    static Index * indices_[3][16];
    static unsigned int nIndices_[3];
    struct RenderInfo {
      unsigned int lod_;
      unsigned int neighbors_;
    };
    RenderInfo * renderChunks_;

    unsigned char * data_;
    unsigned int dataWidth_;
    unsigned int dataHeight_;

    dTriMeshDataID triMeshDataID_;

    unsigned char * roadAlpha_;
    unsigned int roadAlphaPitch_;
    unsigned int roadWidth_;
    unsigned int roadHeight_;
    float roadCoord_;
};


C_Terrain::Index * C_Terrain::indices_[3][16];
unsigned int C_Terrain::nIndices_[3];

void C_Terrain::Chunk::issue( unsigned int lod, unsigned int neighbors )
{
#if SHOW_LOD
  glDisable( GL_FOG );
  assert( lod < 3 && neighbors < 16 );
  assert( !glGetError() );
  switch( lod ) {
  case 0:
    glColor3f( 1, 1, 1 );
    break;
  case 1:
    glColor3f( 0, 0, 1 );
    break;
  case 2:
    glColor3f( 1, 0, 0 );
    break;
  }
#endif

  glVertexPointer( 3, GL_FLOAT, sizeof( C_Terrain::Vertex ), &terrain_->renVerts_[vStart_].pos.x );
  glNormalPointer( GL_FLOAT, sizeof( C_Terrain::Vertex ), &terrain_->renVerts_[vStart_].normal.x );
  glTexCoordPointer( 2, GL_FLOAT, sizeof( C_Terrain::Vertex ), terrain_->renVerts_[vStart_].uv );
  assert( !glGetError() );

  glDrawRangeElements( GL_TRIANGLES, 0, nVertices_-1, terrain_->nIndices_[lod], 
      GL_UNSIGNED_INT, terrain_->indices_[lod][neighbors] );
  assert( !glGetError() );

#if SHOW_LOD
  glColor3f( 1, 1, 1 );
  glEnable( GL_FOG );
#endif
}


C_Terrain::C_Terrain()
{
  chunks_ = 0;
  renderChunks_ = 0;
  uChunks_ = 0;
  vChunks_ = 0;
  width_ = 0;
  height_ = 0;
  west_ = 0;
  south_ = 0;
  scale_ = 0;
  offset_ = 0;
  texelCoord_ = 0;
  vertexObject_ = 0;
  renVerts_ = 0;
  collVerts_ = 0;
  collIndices_ = 0;
  triMeshDataID_ = 0;
  data_ = 0;
  dataWidth_ = 0;
  dataHeight_ = 0;
  roadAlpha_ = 0;
  roadAlphaPitch_ = 0;
  roadWidth_ = 0;
  roadHeight_ = 0;
}

C_Terrain::~C_Terrain()
{
  if( vertexObject_ ) {
    glDeleteBuffersARB( 1, &vertexObject_ );
  }
  for( unsigned int v = 0; v < vChunks_; ++v ) {
    delete[] chunks_[v];
  }
  delete[] chunks_;
  delete[] renderChunks_;
  delete[] renVerts_;
  delete[] collVerts_;
  delete[] collIndices_;
  delete[] data_;
  delete[] roadAlpha_;
}

extern "C" bool _CrtCheckMemory();

#define VERT_COUNT (uChunks_*vChunks_*(ChunkSize*2+1)*(ChunkSize*2+1))
#define INDEX_COUNT (uChunks_*vChunks_*(ChunkSize*2)*(ChunkSize*2)*3*2)

bool C_Terrain::load( Image const & i, float scale, float offset )
{
  assert( chunks_ == 0 );
  if( chunks_ != 0 ) {
    return false;
  }

#if 0   //  test hermite function
{
float blah[5][5] = {
  0, 1, 2, 2, 1,
  1, 2, 3, 3, 2,
  2, 3, 4, 4, 3,
  2, 3, 4, 4, 3,
  1, 2, 3, 3, 2,
};

  float d[9];
  for( int i = 0; i < 9; ++i ) {
    float x = 1+float(i)/4;
    d[i] = hermite4_2D( &blah[0][0], 5, uint(floorf(x)), 2, x-floorf(x), 0 );
  }

  assert( _CrtCheckMemory() );
}
#endif

  info( "Terrain loading image '%s'.\n", i.name() );

  //  Images must be power-of-two-plus-one of data each 
  //  side, plus one texel of fringe/border, which is not 
  //  sampled, but which is used for normals and curvature.
  //  This means that successive terrains can tile!
  assert( (i.width & (ChunkSize-1)) == 3 );
  assert( i.width >= ChunkSize );
  assert( (i.height & (ChunkSize-1)) == 3 );
  assert( i.height >= ChunkSize );

  scale_ = scale;
  offset_ = offset;

  __int64 startLoadingImage = rdtsc();

  //  compact the RGB image into storage
  //  We need 2 (!) lines of padding each side, because 
  //  we will be sampling outside the vertex data area 
  //  to calculate normals.
  dataWidth_ = i.width + 2;
  dataHeight_ = i.height + 2;
  width_ = float( (i.width-3)*TexelWidth );
  height_ = float( (i.height-3)*TexelWidth );
  west_ = -width_/2;
  south_ = -height_/2;
  texelCoord_ = 1.0f / TexelWidth;
  //  Add one byte, because we may walk one off the end 
  //  when reaching exactly the end, for interpolation.
  data_ = new unsigned char[ dataWidth_ * dataHeight_ + 1 ];
  memset( data_, 191, dataWidth_*dataHeight_+1 );
  unsigned char * p = data_ + dataWidth_ + 1;
  unsigned int d = i.bits >> 3;
  for( unsigned int v = 0; v < dataHeight_-2; ++v ) {
    unsigned char * b = (unsigned char *)i.base + v*i.pitch;
    for( unsigned int u = 0; u < dataWidth_-2; ++u ) {
      *p++ = *b;
      b += d;
    }
    p += 2;

    assert( _CrtCheckMemory() );

  }

  memcpy( data_+1, data_+dataWidth_+1, dataWidth_-2 );
  memcpy( data_+1+(dataHeight_-1)*dataWidth_, data_+1+(dataHeight_-2)*dataWidth_, dataWidth_-2 );
  for( unsigned int v = 0; v < dataHeight_; ++v ) {
    data_[ v*dataWidth_ ] = data_[ v*dataWidth_ + 1 ];
    data_[ v*dataWidth_ + dataWidth_-1 ] = data_[ v*dataWidth_ + dataWidth_-2 ];
  }

  assert( _CrtCheckMemory() );

  __int64 startConstructMesh = rdtsc();

  //  construct vertices, indices and chunks
  //  OPCODE collision needs copy of all indices, 
  //  unless we return a geomID per block.
  //  Do one static level of subdivision per block, for now, to generate lod0; 
  //  lod1 is 1:1 sampled, and lod2 is 1:2 sampled.
  //  These chunks won't fit in L1, but they will each fit in L2 (about 160 kB each).
  //  Total terrain size for a 256*256 terrain is 10 MB. Sourcing 16-bit data would 
  //  probably help our cause, but let's save that for the console port :-)
  uint vc = (ChunkSize*2+1)*(ChunkSize*2+1);
  uint ic = (ChunkSize*2)*(ChunkSize*2)*2*3;
  info( "Terrain: %d vertices, %d indices, per chunk using %dx%d samples\n",
      vc, ic, ChunkSize+1, ChunkSize+1 );
  uChunks_ = (dataWidth_-3)/ChunkSize;
  vChunks_ = (dataHeight_-3)/ChunkSize;
  renVerts_ = new Vertex[ uChunks_*vChunks_*vc ];
  collVerts_ = new Vector3[ uChunks_*vChunks_*vc ];
  collIndices_ = new Index[ uChunks_*vChunks_*ic ];
  chunks_ = new Chunk *[ vChunks_ ];
  renderChunks_ = new RenderInfo[ vChunks_*uChunks_ ];
  float deltaWeight = scale_ / TexelWidth;
  for( uint i = 0; i < vChunks_; ++i ) {

    assert( _CrtCheckMemory() );

    chunks_[i] = new Chunk[ uChunks_ ];
    for( uint j = 0; j < uChunks_; ++j ) {
      float minY = 0; float maxY = 0;
      Vertex * vBase = renVerts_ + (i * uChunks_ + j) * vc;
      Vector3 loLeft( west_ + ChunkSize * TexelWidth * j, 0, south_ + ChunkSize * TexelWidth * i );
      //  adjust by 2*(dataWidth_ + 1) for the extra border texels for normal sampling and vertex sampling
      //  but then adjust back again because hermite4 sampler wants pointer to fringe.
      unsigned char * cBase = data_ + i * dataWidth_ * ChunkSize + j * ChunkSize;
      C_Terrain::Chunk * c = &chunks_[i][j];
      c->terrain_ = this;
      c->lower_ = loLeft;
      c->lower_.z *= -1;  //  right-handed
      c->upper_ = Vector3( west_ + ChunkSize * TexelWidth * (j+1), 0, south_ + ChunkSize * TexelWidth * (i+1) );
      c->upper_.z *= -1;  //  right-handed
      c->origin_ = Vector3::Null;
      c->nVertices_ = vc;
      c->vStart_ = unsigned int(vBase - renVerts_);
      //  find heights
      //  calculate normals
      for( uint v = 0; v <= ChunkSize*2; ++v ) {
        for( uint u = 0; u <= ChunkSize*2; ++u ) {
          float x = float(u>>1) + ((u&1)?0.5f : 0);
          float z = float(v>>1) + ((v&1)?0.5f : 0);
          vBase->pos.x = loLeft.x + x * TexelWidth;
          //  generate right-handed geometry
          vBase->pos.z = - loLeft.z - z * TexelWidth;
          float yValue = hermite4_2D( cBase, dataWidth_, 
              uint(floorf(x))+2, uint(floorf(z))+2, x-floorf(x), z-floorf(z) )
              * scale_ + offset_;
          if( !(v|u) ) {
            minY = maxY = yValue;
          }
          if( yValue < minY ) minY = yValue;
          if( yValue > maxY ) maxY = yValue;
          vBase->pos.y = yValue;
          //  sample half a sample step (0.5f/2) in each direction
          vBase->normal.x = (
            hermite4_2D( cBase, dataWidth_, 
              uint(floorf(x+0.75f))+1, uint(floorf(z+1.0f))+1, x+0.75f-floorf(x+0.75f), z+1.0f-floorf(z+1.0f) ) -
            hermite4_2D( cBase, dataWidth_, 
              uint(floorf(x+1.25f))+1, uint(floorf(z+1.0f))+1, x+1.25f-floorf(x+1.25f), z+1.0f-floorf(z+1.0f) )) * deltaWeight;
          //  generate right-handed normals (!) by negating Z
          vBase->normal.z = (
            hermite4_2D( cBase, dataWidth_, 
              uint(floorf(x+1.0f))+1, uint(floorf(z+1.25f))+1, x+1.0f-floorf(x+1.0f), z+1.25f-floorf(z+1.25f) ) -
            hermite4_2D( cBase,  dataWidth_, 
              uint(floorf(x+1.0f))+1, uint(floorf(z+0.75f))+1, x+1.0f-floorf(x+1.0f), z+0.75f-floorf(z+0.75f) )) * deltaWeight;
          vBase->normal.y = 1;
          // normalize
          vBase->normal *= 1.0f / sqrtf( vBase->normal.lengthSquared() );
          ++vBase;
        }
      }
      //  swap Z values, because we invert to right-handed
      float zz = c->lower_.z;
      c->lower_.z = c->upper_.z;
      c->upper_.z = zz;
      c->lower_.y = minY;
      c->upper_.y = maxY;
      c->center_ = (c->lower_ + c->upper_) * 0.5f;

  assert( _CrtCheckMemory() );

      //  calculate uv coordinates
      //  I smear them out a little bit to get a more even texture 
      //  distribution based on triangle side length (a relaxation 
      //  method, separated by axis, basically).
      float const relaxation = 0.3f;    // set it to 1 to stretch maximally, 0 for planar projection
      vBase = renVerts_ + (i * uChunks_ + j) * vc;
      //  for each v, sum up the lengths of all sides
      for( uint v = 0; v <= ChunkSize*2; ++v ) {
        float len = 0;
        Vector3 prev = vBase[v*(ChunkSize*2+1)].pos;
        for( uint u = 1; u <= ChunkSize*2; ++u ) {
          Vector3 nu = vBase[u+v*(ChunkSize*2+1)].pos;
          Vector3 lv = nu - prev;
          prev = nu;
          len += sqrtf( lv.lengthSquared() );
        }
        assert( len > 0 );
        //  divide 1 in that, and spread out
        float vScale = 1.0f / len;
        len = 0;
        prev = vBase[v*(ChunkSize*2+1)].pos;
        for( uint u = 0; u <= ChunkSize*2; ++u ) {
          Vector3 nu = vBase[u+v*(ChunkSize*2+1)].pos;
          Vector3 lv = nu - prev;
          prev = nu;
          len += sqrtf( lv.lengthSquared() );
          vBase[u+v*(ChunkSize*2+1)].uv[1] = len * vScale * relaxation + float(u)/(ChunkSize*2)*(1-relaxation);
        }
      }
      //  for each u, sum up the lengths of all sides
      for( uint u = 0; u <= ChunkSize*2; ++u ) {
        float len = 0;
        Vector3 prev = vBase[u].pos;
        for( uint v = 1; v <= ChunkSize*2; ++v ) {
          Vector3 nu = vBase[u+v*(ChunkSize*2+1)].pos;
          Vector3 lv = nu - prev;
          prev = nu;
          len += sqrtf( lv.lengthSquared() );
        }
        assert( len > 0 );
        //  divide 1 in that, and spread out
        float uScale = 1.0f / len;
        len = 0;
        prev = vBase[u].pos;
        for( uint v = 0; v <= ChunkSize*2; ++v ) {
          Vector3 nu = vBase[u+v*(ChunkSize*2+1)].pos;
          Vector3 lv = nu - prev;
          prev = nu;
          len += sqrtf( lv.lengthSquared() );
          vBase[u+v*(ChunkSize*2+1)].uv[0] = len * uScale * relaxation + float(v)/(ChunkSize*2)*(1-relaxation);
        }
      }
      //TODO: should run one or two passes of neighbor diffusion filtering here, 
      //  with the edges locked (for seamless tiling).

  assert( _CrtCheckMemory() );

      Index * iBase = &collIndices_[(i*uChunks_+j)*ic];
      uint ixBase = (i*uChunks_+j)*vc;
      for( uint v = 0; v < ChunkSize*2; ++v ) {
        for( uint u = 0; u < ChunkSize*2; ++u ) {
          *iBase++ = ixBase;
          *iBase++ = ixBase+(ChunkSize*2+1+1);
          *iBase++ = ixBase+(ChunkSize*2+1);
          *iBase++ = ixBase;
          *iBase++ = ixBase+1;
          *iBase++ = ixBase+(ChunkSize*2+1+1);
          ixBase += 1;
        }
        ixBase += 1;
      }
    }
  }

  //  Now, copy position information into collVerts_
  for( unsigned int i = 0; i < vc * uChunks_ * vChunks_; ++i ) {
    collVerts_[i] = renVerts_[i].pos;
  }

  if( GL_ARB_vertex_buffer_object_flag != GL_FALSE ) {
    info( "GL_ARB_vertex_buffer_object is available.\n" );
    assert( !glGetError() );
    glGenBuffersARB( 1, &vertexObject_ );
    glBindBufferARB( GL_ARRAY_BUFFER_ARB, vertexObject_ );
    glBufferDataARB( GL_ARRAY_BUFFER_ARB, sizeof(Vertex)*VERT_COUNT, renVerts_, GL_STATIC_DRAW_ARB );
    glBindBufferARB( GL_ARRAY_BUFFER_ARB, 0 );
    assert( !glGetError() );
    if( glGetError() ) {
      warning( "glGetError() when calling buffer objects; disabling.\n" );
      glDeleteBuffersARB( 1, &vertexObject_ );
      vertexObject_ = 0;
    }
    else {
      delete[] renVerts_;
      renVerts_ = 0;
    }
  }
  else {
    info( "GL_ARB_vertex_buffer_object is not available.\n" );
  }

  //  calculate global indices
  for( unsigned int edge = 0; edge < 16; ++edge ) {
    indices_[0][edge] = new Index[ ic ];
    nIndices_[0] = ic;
    Index * iBase = indices_[0][edge];
    if( edge > 0 ) {
      memcpy( iBase, indices_[0][0], ic*sizeof(Index) );
      for( uint i = 0; i < ic; ++i ) {
        uint v = iBase[i] / (ChunkSize*2+1);
        uint u = iBase[i] % (ChunkSize*2+1);
        if( edge & NegativeUEdge ) {
          if( (v & 1) && (u == 0) ) {
            iBase[i] = (v&~1)*(ChunkSize*2+1)+u;
          }
        }
        if( edge & PositiveUEdge ) {
          if( (v & 1) && (u == ChunkSize*2) ) {
            iBase[i] = (v&~1)*(ChunkSize*2+1)+u;
          }
        }
        if( edge & NegativeVEdge ) {
          if( (u & 1) && (v == 0) ) {
            iBase[i] = v*(ChunkSize*2+1)+(u&~1);
          }
        }
        if( edge & PositiveVEdge ) {
          if( (u & 1) && (v == (ChunkSize*2)) ) {
            iBase[i] = v*(ChunkSize*2+1)+(u&~1);
          }
        }
      }
    }
    else {
      uint ixBase = 0;
      for( uint v = 0; v < ChunkSize*2; ++v ) {
        for( uint u = 0; u < ChunkSize*2; ++u ) {
          *iBase++ = ixBase;
          *iBase++ = ixBase+(ChunkSize*2+1+1);
          *iBase++ = ixBase+(ChunkSize*2+1);
          *iBase++ = ixBase;
          *iBase++ = ixBase+1;
          *iBase++ = ixBase+(ChunkSize*2+1+1);
          ixBase += 1;
        }
        ixBase += 1;
      }
    }
    unsigned int ic2 = (ChunkSize*ChunkSize)*3*2;
    indices_[1][edge] = new Index[ ic2 ];
    nIndices_[1] = ic2;
    iBase = indices_[1][edge];
    if( edge > 0 ) {
      memcpy( iBase, indices_[1][0], ic2*sizeof(Index) );
      for( uint i = 0; i < ic2; ++i ) {
        uint v = iBase[i] / (ChunkSize*2+1);
        uint u = iBase[i] % (ChunkSize*2+1);
        if( edge & NegativeUEdge ) {
          if( (v & 2) && (u == 0) ) {
            iBase[i] = (v&~3)*(ChunkSize*2+1)+u;
          }
        }
        if( edge & PositiveUEdge ) {
          if( (v & 2) && (u == ChunkSize*2) ) {
            iBase[i] = (v&~3)*(ChunkSize*2+1)+u;
          }
        }
        if( edge & NegativeVEdge ) {
          if( (u & 2) && (v == 0) ) {
            iBase[i] = v*(ChunkSize*2+1)+(u&~3);
          }
        }
        if( edge & PositiveVEdge ) {
          if( (u & 2) && (v == (ChunkSize*2)) ) {
            iBase[i] = v*(ChunkSize*2+1)+(u&~3);
          }
        }
      }
    }
    else {
      uint ixBase = 0;
      for( uint v = 0; v < ChunkSize; ++v ) {
        for( uint u = 0; u < ChunkSize; ++u ) {
          *iBase++ = ixBase;
          *iBase++ = ixBase+(ChunkSize*4+2+2);
          *iBase++ = ixBase+(ChunkSize*4+2);
          *iBase++ = ixBase;
          *iBase++ = ixBase+2;
          *iBase++ = ixBase+(ChunkSize*4+2+2);
          ixBase += 2;
        }
        ixBase += 1 + ChunkSize*2+1;
      }
    }
    unsigned int ic3 = (ChunkSize*ChunkSize/4)*3*2;
    indices_[2][edge] = new Index[ ic3 ];
    nIndices_[2] = ic3;
    iBase = indices_[2][edge];
    if( edge > 0 ) {
      memcpy( iBase, indices_[2][0], ic3*sizeof(Index) );
      for( uint i = 0; i < ic3; ++i ) {
        uint v = iBase[i] / (ChunkSize*2+1);
        uint u = iBase[i] % (ChunkSize*2+1);
        if( edge & NegativeUEdge ) {
          if( (v & 4) && (u == 0) ) {
            iBase[i] = (v&~7)*(ChunkSize*2/2+1)+u;
          }
        }
        if( edge & PositiveUEdge ) {
          if( (v & 4) && (u == ChunkSize*2/2) ) {
            iBase[i] = (v&~7)*(ChunkSize*2/2+1)+u;
          }
        }
        if( edge & NegativeVEdge ) {
          if( (u & 4) && (v == 0) ) {
            iBase[i] = v*(ChunkSize*2/2+1)+(u&~17);
          }
        }
        if( edge & PositiveVEdge ) {
          if( (u & 4) && (v == (ChunkSize*2/2)) ) {
            iBase[i] = v*(ChunkSize*2/2+1)+(u&~7);
          }
        }
      }
    }
    else {
      uint ixBase = 0;
      for( uint v = 0; v < ChunkSize/2; ++v ) {
        for( uint u = 0; u < ChunkSize/2; ++u ) {
          *iBase++ = ixBase;
          *iBase++ = ixBase+(ChunkSize*8+4+4);
          *iBase++ = ixBase+(ChunkSize*8+4);
          *iBase++ = ixBase;
          *iBase++ = ixBase+4;
          *iBase++ = ixBase+(ChunkSize*8+4+4);
          ixBase += 4;
        }
        ixBase += 1 + ChunkSize*6+3;
      }
    }
  }

  assert( _CrtCheckMemory() );

  extern char const * imagePath( char const * );

  char const * roadName = imagePath( "roads" );
  Image roadTga;
  loadImage( roadName, &roadTga );
  //  a bitmask of whether there's a road there or not
  roadAlpha_ = new unsigned char[ roadTga.width*roadTga.height/8 ];
  roadAlphaPitch_ = roadTga.width/8;
  roadWidth_ = roadTga.width;
  roadHeight_ = roadTga.height;
  unsigned char * oa = roadAlpha_;
  assert( !(roadTga.width & 7) );
  for( uint v = 0; v < (uint)roadTga.height; ++v ) {
    unsigned char bits = 0;
    for( uint u = 0; u < (uint)roadTga.width; ++u ) {
      if( ((uint *)roadTga.base)[u + v*roadTga.pitch/4] >= 0x80000000 ) {
        bits |= (1 << (u&7));
      }
      if( (u&7) == 7 ) {
        *oa++ = bits;
        bits = 0;
      }
    }
  }
  roadCoord_ = float(roadTga.width) / (chunkWidth() * numChunksU());

  __int64 doneConstructMesh = rdtsc();

  info( "Terrain construction: %.3f Mc image, %.3f Mc mesh\n",
      double(startConstructMesh-startLoadingImage) * 1e-6,
      double(doneConstructMesh-startConstructMesh) * 1e-6 );

  return true;
}


//  class I_TerrainODE
dTriMeshDataID C_Terrain::triMeshDataID()
{
  assert( chunks_ != 0 );
  if( chunks_ == 0 ) {
    return 0;
  }
  if( triMeshDataID_ == 0 ) {
    triMeshDataID_ = dGeomTriMeshDataCreate();
    __int64 a = rdtsc();
    dGeomTriMeshDataBuildSingle( triMeshDataID_, collVerts_, sizeof(*collVerts_), VERT_COUNT, collIndices_, INDEX_COUNT, sizeof(Index)*3 );
    __int64 b = rdtsc();
    info( "CTerrain: dGeomTriMeshDataBuildSingle() on %d verts %d indices took %.3f Mc\n", 
        VERT_COUNT, INDEX_COUNT, double(b-a)*1e-6 );
  }
  return triMeshDataID_;
}

dGeomID C_Terrain::makeGeom( dSpaceID w )
{
  return dCreateTriMesh( w, triMeshDataID(), 0, 0, 0 );
}


//  class I_Terrain
void C_Terrain::dispose()
{
  delete this;
}

void C_Terrain::measure( float & west, float & south, float & width, float & height )
{
  width = width_;
  height = height_;
  west = west_;
  south = south_;
}

float C_Terrain::heightAt( float u, float v )
{
  assert( chunks_ );
  if( !chunks_ ) {
    return 0;
  }
  u = clamp( u, west_, west_+width_ );
  v = clamp( v, south_, south_+height_ );
  float xf = (u - west_) * texelCoord_;
  float zf = (v - south_) * texelCoord_;
  assert( xf >= 0 );
  assert( zf >= 0 );
  unsigned int x = unsigned int(floorf(xf));
  unsigned int z = unsigned int(floorf(zf));
  float xp = xf - x;
  float zp = xf - z;
  return hermite4_2D( data_ + dataWidth_ + 1, dataWidth_, x, z, xp, zp ) * scale_ + offset_;
}

unsigned int C_Terrain::groundAt( float u, float v )
{
  assert( chunks_ );
  if( !chunks_ ) {
    return 0;
  }
  float xf = (u - west_) * roadCoord_;
  float zf = (v - south_) * roadCoord_;
  unsigned int x = unsigned int(floorf(xf));
  unsigned int z = roadHeight_ - unsigned int(floorf(zf)) - 1;
  if( x >= roadWidth_ ) {
   x = roadWidth_-1;
  }
  if( z >= roadHeight_ ) {
   z = roadHeight_-1;
  }
  if( (roadAlpha_[(x>>3)+z*roadAlphaPitch_] & (1<<(x&7))) != 0 ) {
    debugVar( 20, 2, "pos %5.1f,%5.1f xz %4d,%4d ROAD  ", u, v, x, z );
    return 1;
  }
  debugVar( 20, 2, "pos %5.1f,%5.1f xz %4d,%4d GROUND", u, v, x, z );
  return 0;
}

int gLod = 1; //===

static bool PointInInflatedBox( Vector3 const & pt, Vector3 const & lo, Vector3 const & hi, Vector3 const & i )
{
  return pt.x > lo.x-i.x && pt.y > lo.y-i.y && pt.z > lo.z-i.z &&
      pt.x < hi.x+i.x && pt.y < hi.y+i.y && pt.z < hi.z+i.z;
}

void C_Terrain::issueAll( Vector3 const & camPos )
{
  memset( renderChunks_, 0xff, sizeof( RenderInfo ) * uChunks_ * vChunks_ );
  assert( chunks_ );
  for( uint v = 0; v < vChunks_; ++v ) {
    for( uint u = 0; u < uChunks_; ++u ) {
      //  todo: frustum and distance culling
      if( PointInInflatedBox( camPos, chunks_[v][u].lower_, chunks_[v][u].upper_, Vector3( 5, 15, 2 ) ) ) {
        renderChunks_[u+v*uChunks_].lod_ = 0;
      }
      else {
        float d = sqrtf( (camPos - chunks_[v][u].center_).lengthSquared() );
        //=== todo: horrible hard-coded LOD
        if( d > (ChunkSize*TexelWidth*0.5f*(gLod+1)) ) {
          if( d > (ChunkSize*TexelWidth*(gLod+2)) ) {
            renderChunks_[u+v*uChunks_].lod_ = 2;
          }
          else {
            renderChunks_[u+v*uChunks_].lod_ = 1;
          }
        }
        else {
          renderChunks_[u+v*uChunks_].lod_ = 0;
        }
      }
    }
  }

  if( vertexObject_ ) {
    glBindBufferARB( GL_ARRAY_BUFFER_ARB, vertexObject_ );
  }
  glEnableClientState( GL_VERTEX_ARRAY );
  glEnableClientState( GL_NORMAL_ARRAY );
  glEnableClientState( GL_TEXTURE_COORD_ARRAY );
  assert( !glGetError() );

  for( uint v = 0; v < vChunks_; ++v ) {
    for( uint u = 0; u < uChunks_; ++u ) {
      uint rix = u+v*uChunks_;
      uint lod = renderChunks_[rix].lod_;
      if( lod > 2 ) {
        continue; //  culled
      }
      uint edge = 0;
      if( (u > 0) && (renderChunks_[rix-1].lod_ > lod) ) {
        edge |= NegativeUEdge;
      }
      if( (u < uChunks_-1) && (renderChunks_[rix+1].lod_ > lod) ) {
        edge |= PositiveUEdge;
      }
      if( (v > 0) && (renderChunks_[rix-uChunks_].lod_ > lod ) ) {
        edge |= NegativeVEdge;
      }
      if( (v < vChunks_-1) && (renderChunks_[rix+uChunks_].lod_ > lod) ) {
        edge |= PositiveVEdge;
      }
      chunks_[v][u].issue( lod, edge );
    }
  }

  glDisableClientState( GL_VERTEX_ARRAY );
  glDisableClientState( GL_NORMAL_ARRAY );
  glDisableClientState( GL_TEXTURE_COORD_ARRAY );
  assert( !glGetError() );

  if( vertexObject_ ) {
    glBindBufferARB( GL_ARRAY_BUFFER_ARB, 0 );
  }
}

unsigned int C_Terrain::numChunksU()
{
  return uChunks_;
}

unsigned int C_Terrain::numChunksV()
{
  return vChunks_;
}

unsigned int C_Terrain::chunkWidth()
{
  return ChunkSize * TexelWidth;
}

unsigned int C_Terrain::chunkHeight()
{
  return ChunkSize * TexelWidth;
}

I_TerrainChunk * C_Terrain::chunkUV( unsigned int u, unsigned int v )
{
  assert( u < uChunks_ );
  assert( v < vChunks_ );
  return &chunks_[v][u];
}

I_TerrainODE * C_Terrain::terrainODE()
{
  return this;
}



I_Terrain * NewTerrainFromImage( Image const & i, float scale, float offset )
{
  if( ((i.width & (C_Terrain::ChunkSize-1)) != 3) || ((i.height & (C_Terrain::ChunkSize-1)) != 3) ) {
    error( "Terrain image %s must be multiple of %d plus 3 in each dimension.\n", i.name(), C_Terrain::ChunkSize );
    return NULL;
  }
  if( i.bits != 32 && i.bits != 24 && i.bits != 8 ) {
    error( "Terrain image %s must be 32, 24 or 8 bits depth.\n", i.name() );
    return NULL;
  }
  C_Terrain * t = new C_Terrain();
  t->load( i, scale, offset );
  return t;
}

