
#include "glut-start.h"

#include <GL/GL.h>
#include <new>

#include "textures.h"
#include "myext.h"
#include "internaltexture.h"
#include "image.h"
#include "glshaderstate.h"

#include "myfilesystem.h"
#include "mydds.h"
#include "mydxt.h"



#if !defined( NDEBUG )
#define TexRefCountMax 20
#else
#define TexRefCountMax 1000
#endif


static InternalTexture sNullTexture;



class TextureManager
{
public:
  TextureManager( );
  ~TextureManager( );

  InternalTexture *   makeTextureFromId( int id, GLenum target );
  InternalTexture *   getTexture( char const * path, bool mipmaps );
  InternalTexture *   getTexture( InternalTexture * t );
  void                putTexture( InternalTexture * t );

private:
  InternalTexture *   findOrMakeEntry( char const * path, bool mipmaps );
  InternalTexture *   recursiveFindOrMakeEntry( InternalTexture ** root, char const * path );
  void                readAndUploadEntry( InternalTexture * t, bool mipmaps );
  void                removeEntry( InternalTexture * t );
  void                recursiveDelete( InternalTexture * t );

  InternalTexture *   mRoot;
};


TextureManager::TextureManager( )
{
  mRoot = 0;
}

TextureManager::~TextureManager( )
{
  recursiveDelete( mRoot );
  mRoot = 0;
}

InternalTexture *
TextureManager::makeTextureFromId( int id, GLenum target )
{
  GLuint t = (GLuint)id;
  InternalTexture * ret = new InternalTexture;
  ret->mTexId = t;
  ret->mTarget = target;
  ret->mRefCount = 1;

  if( id == 0 ) {
    //  This happens during start-up, initializing the 
    //  cube null texture.
    assert( target == GL_TEXTURE_CUBE_MAP_ARB );
    ret->mWidth = 0;
    ret->mHeight = 0;
    ret->mBits = 0;
    return ret;
  }

  assert( !glGetError() );

  //  read back appropriate parameters
  glBindTexture( target, t );
  GLint w;
  glGetTexLevelParameteriv( ret->mTarget, 0, GL_TEXTURE_WIDTH, &w ); ret->mWidth = w;
  glGetTexLevelParameteriv( ret->mTarget, 0, GL_TEXTURE_HEIGHT, &w ); ret->mHeight = w;

  //  work out the bit depth (if we can)
  GLint i;
  glGetTexLevelParameteriv( ret->mTarget, 0, GL_TEXTURE_INTERNAL_FORMAT, &i );
  switch( i ) {
  case GL_RGB16:
  case GL_RGB8:
  case GL_RGB:
    glGetTexLevelParameteriv( ret->mTarget, 0, GL_TEXTURE_RED_SIZE, &i );
    ret->mBits = 3 * i;
    break;
  case GL_RGBA16:
  case GL_RGBA8:
  case GL_RGBA:
    glGetTexLevelParameteriv( ret->mTarget, 0, GL_TEXTURE_RED_SIZE, &i );
    ret->mBits = 3 * i;
    glGetTexLevelParameteriv( ret->mTarget, 0, GL_TEXTURE_ALPHA_SIZE, &i );
    ret->mBits += i;
    break;
  case GL_ALPHA8:
  case GL_ALPHA:
    glGetTexLevelParameteriv( ret->mTarget, 0, GL_TEXTURE_ALPHA_SIZE, &i );
    ret->mBits = i;
    break;
  case GL_COMPRESSED_RGBA_S3TC_DXT1_EXT:
    ret->mBits = -4;
    break;
  case GL_COMPRESSED_RGBA_S3TC_DXT5_EXT:
    ret->mBits = -8;
    break;
  default:
    ret->mBits = 0;
    break;
  }

  GLShader::flushState();

  return ret;
}

InternalTexture *   
TextureManager::getTexture( char const * path, bool mipmaps )
{
  if( !path || !path[0] ) return &sNullTexture;
  assert( strlen( path ) < PathLengthMax );
  if( strlen( path ) > PathLengthMax ) return 0;
  InternalTexture * t = findOrMakeEntry( path, mipmaps );
  assert( t->mRefCount < TexRefCountMax );
  t->mRefCount += 1;
  return t;
}

InternalTexture *   
TextureManager::getTexture( InternalTexture * t )
{
  t->mRefCount += 1;
  return t;
}

void   
TextureManager::putTexture( InternalTexture * t )
{
  if( t == &sNullTexture || !mRoot ) {    //  we allow NULL mRoot because of shutdown case
    return;
  }
  assert( t->mRefCount > 0 );
  t->mRefCount -= 1;
  if( t->mRefCount == 0 ) {
    if( t->mTexId ) {
      //  I might hold on to the texture image here, 
      //  as the driver should do a decent job of managing 
      //  the active working set. Oh, well.
      glDeleteTextures( 1, &t->mTexId );
      t->mTexId = 0;
    }
    //  I don't ever delete/remove file name records,
    //  as they're likely to be needed again some day.
    //  However, by-id textures are deleted.
    if( !t->mPath[0] ) {
      delete t;
    }
  }
}

InternalTexture *   
TextureManager::findOrMakeEntry( char const * path, bool mipmaps )
{
  assert( path[0] );
  if( !path[0] ) {
    return 0;
  }
  InternalTexture * t = recursiveFindOrMakeEntry( &mRoot, path );
  if( !t->mTexId ) {
    readAndUploadEntry( t, mipmaps );
  }
  return t;
}

InternalTexture *
TextureManager::recursiveFindOrMakeEntry( InternalTexture ** root, char const * path )
{
  //  This is a very simple binary tree, which makes searching for 
  //  textures more efficient than scanning a linear list, while 
  //  not including all the cruft of a std::map< std::string, >.
  //  It fails miserably if you load textures in alphabetical order ;-)
  if( !*root ) {
    InternalTexture * ret = new InternalTexture;
    strncpy( ret->mPath, path, PathLengthMax );
    ret->mPath[ PathLengthMax-1 ] = 0;
    *root = ret;
    return ret;
  }
  int cmp = stricmp( (*root)->mPath, path );
  if( !cmp ) {
    return *root;
  }
  if( cmp < 0 ) {
    return recursiveFindOrMakeEntry( &(*root)->mLeft, path );
  }
  return recursiveFindOrMakeEntry( &(*root)->mRight, path );
}

void                
TextureManager::readAndUploadEntry( InternalTexture * t, bool mipmaps )
{
  int width = 0, height = 0, bits = 0;
  t->mTarget = GL_TEXTURE_2D;
  static GLenum target2d[] = { GL_TEXTURE_2D };
  static GLenum targetCube[] = {
    GL_TEXTURE_CUBE_MAP_POSITIVE_X, GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
    GL_TEXTURE_CUBE_MAP_POSITIVE_Y, GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,
    GL_TEXTURE_CUBE_MAP_POSITIVE_Z, GL_TEXTURE_CUBE_MAP_NEGATIVE_Z,
  };
  if( strstr( t->mPath, ".dds" ) ) {
    Data d;
    if( !FS::readFileByPath( t->mPath, &d ) ) {
      warning( "texture %s wasn't found\n", t->mPath );
      return;
    }
    DDS_header h;
    if( d.read( &h, sizeof(h) ) != sizeof( h ) ) {
      warning( "texture file %s is too small\n", t->mPath );
      return;
    }
    if( h.dwMagic != DDS_MAGIC ) {
      warning( "texture %s is not a real DDS file\n", t->mPath );
      return;
    }
    GLenum compressedFormat;
    if( PF_IS_DXT1( h.sPixelFormat ) ) {
      compressedFormat = GL_COMPRESSED_RGBA_S3TC_DXT1_EXT;
      bits = -4;
    }
    else if( PF_IS_DXT5( h.sPixelFormat ) ) {
      compressedFormat = GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
      bits = -8;
    }
    else {
      warning( "texture %s is of an unsupported DDS format\n", t->mPath );
      return;
    }
    if( bits < 0 ) {
      if( !(GL_ARB_texture_compression_Flag && GL_EXT_texture_compression_s3tc_Flag) ) {
        warning( "the graphics card does not support DXTC compressed textures (%s).\n", t->mPath );
        return;
      }
    }
    mipmaps = ((h.sCaps.dwCaps1 & DDSCAPS_MIPMAP) != 0);
    if( mipmaps ) {
      mipmaps = (h.dwMipMapCount > 1);
    }
    bool cubemap = ((h.sCaps.dwCaps2 & DDSCAPS2_CUBEMAP) != 0);
    size_t nImages = (cubemap ? 6 : 1);
    GLenum const * imageTarget = target2d;
    if( cubemap ) {
      t->mTarget = GL_TEXTURE_CUBE_MAP;
      imageTarget = targetCube;
    }
    width = h.dwWidth;
    height = h.dwHeight;
    glGenTextures( 1, &t->mTexId );
    assert( !glGetError() );
    glBindTexture( t->mTarget, t->mTexId );
    assert( !glGetError() );
    glPixelStorei( GL_UNPACK_ROW_LENGTH, 0 );
    assert( !glGetError() );
    size_t offset = sizeof( h );
    for( size_t cf = 0; cf < nImages; ++cf ) {
      for( size_t i = 0; i < h.dwMipMapCount; ++i ) {
        int w = width >> i;
        if( !w ) w = 1;
        int h = height >> i;
        if( !h ) h = 1;
        size_t toRead = ((w < 4) ? 4 : w) * ((h < 4) ? 4 : h) * -bits / 8;
        if( d.size() < offset + toRead ) {
          warning( "texture %s is too short (corrupt!)\n", t->mPath );
        }
        else {
          dxt_flip_data( (unsigned char *)&d.writeBase()[offset], 
              (w < 4) ? 4 : w, (h < 4) ? 4 : h, (bits == -4) );
          glCompressedTexImage2DARB( imageTarget[cf], (GLint)i, compressedFormat, w, h, 0, 
              (GLsizei)toRead, &d.readBase()[offset] );
          assert( !glGetError() );
          offset += toRead;
        }
      }
    }
  }
  else {
    Image tga;
    if( !loadTgaImage( t->mPath, &tga ) ) {
      warning( "texture %s could not be loaded\n", t->mPath );
      return;
    }
    width = tga.width;
    height = tga.height;
    bits = tga.bits;
    bool cubemap = false;
    if( (width & (width-1)) || (height & (height-1)) || (bits != 32 && bits != 24) ) {
      if( !(height & (height-1)) && (width == 6*height) ) {
        cubemap = true;
        width /= 6;
        t->mTarget = GL_TEXTURE_CUBE_MAP_ARB;
      }
      else {
        warning( "texture %s is non-power-of-two size %dx%dx%d\n", t->mPath, width, height, bits );
        return;
      }
    }
    else {
      debug( "texture %s is sized %dx%dx%d\n", t->mPath, width, height, bits );
    }
    glGenTextures( 1, &t->mTexId );
    glBindTexture( t->mTarget, t->mTexId );
    assert( bits == 32 || bits == 24 || bits == 8 ); //fixme:  for now
    glPixelStorei( GL_UNPACK_ROW_LENGTH, tga.pitch / (tga.bits/8) );
    assert( !glGetError() );
    if( mipmaps ) {
      glTexParameteri( t->mTarget, GL_GENERATE_MIPMAP, GL_TRUE );
      glHint( GL_GENERATE_MIPMAP_HINT, GL_NICEST );
      assert( !glGetError() );
    }
    size_t nImages = (cubemap ? 6 : 1);
    GLenum const * imageTarget = target2d;
    if( cubemap ) {
      t->mTarget = GL_TEXTURE_CUBE_MAP;
      imageTarget = targetCube;
    }
    for( size_t cf = 0; cf < nImages; ++cf ) {
      glTexImage2D( imageTarget[cf], 0, (bits == 8) ? GL_ALPHA : GL_RGBA, width, height, 0, 
          (bits == 32) ? GL_BGRA : (bits == 24) ? GL_BGR : GL_ALPHA, GL_UNSIGNED_BYTE, 
          (unsigned char const *)tga.base + (width * cf * bits/8) );
    }
    assert( !glGetError() );
  }
  if( mipmaps ) {
    glTexParameteri( t->mTarget, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
  }
  else {
    glTexParameteri( t->mTarget, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
  }
  glTexParameteri( t->mTarget, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
  GLShader::flushState();

  t->mWidth = width;
  t->mHeight = height;
  t->mBits = bits;
  t->mMipmaps = mipmaps;
  t->mClamp = false;

  assert( !glGetError( ) );
}

void
TextureManager::recursiveDelete( InternalTexture * t )
{
  if( !t ) return;
  recursiveDelete( t->mLeft );
  recursiveDelete( t->mRight );
  glDeleteTextures( 1, &t->mTexId );
  delete t;
}


static TextureManager sTexManager;



TextureRef TextureRef::null;
TextureRef TextureRef::nullCube( 0, GL_TEXTURE_CUBE_MAP );

TextureRef::TextureRef()
{
  mTexture = sTexManager.getTexture( &sNullTexture );
  mTexture->mTarget = GL_TEXTURE_2D;
}

TextureRef::TextureRef( int texture, GLenum target )
{
  mTexture = sTexManager.makeTextureFromId( texture, target );
}

TextureRef::TextureRef( char const * path, bool mipmaps )
{
  mTexture = sTexManager.getTexture( path, mipmaps );
}

TextureRef::TextureRef( TextureRef const & other )
{
  mTexture = sTexManager.getTexture( other.mTexture );
}

TextureRef::~TextureRef( )
{
  sTexManager.putTexture( mTexture );
  mTexture = 0;
}

TextureRef &
TextureRef::operator=( TextureRef const & other )
{
  InternalTexture * temp = sTexManager.getTexture( other.mTexture );
  sTexManager.putTexture( mTexture );
  mTexture = temp;
  return *this;
}

bool 
TextureRef::operator<( TextureRef const & other ) const
{
  return stricmp( mTexture->mPath, other.mTexture->mPath ) < 0;
}

bool 
TextureRef::consistent( ) const
{
  return mTexture->consistent( );
}

void 
TextureRef::getSize( int * xSize, int * ySize )
{
  *xSize = mTexture->mWidth;
  *ySize = mTexture->mHeight;
}

void 
TextureRef::getInset( float * xInset, float * yInset )
{
  *xInset = 1.f/(2*mTexture->mWidth);
  *yInset = 1.f/(2*mTexture->mHeight);
}

void TextureRef::setClamp( bool clamp )
{
  mTexture->mClamp = clamp;
  glActiveTexture( GL_TEXTURE0_ARB );
  glBindTexture( mTexture->mTarget, mTexture->mTexId );
  glTexParameteri( mTexture->mTarget, GL_TEXTURE_WRAP_S, clamp ? GL_CLAMP_TO_EDGE : GL_REPEAT );
  glTexParameteri( mTexture->mTarget, GL_TEXTURE_WRAP_T, clamp ? GL_CLAMP_TO_EDGE : GL_REPEAT );
  glTexParameteri( mTexture->mTarget, GL_TEXTURE_WRAP_R_EXT, clamp ? GL_CLAMP_TO_EDGE : GL_REPEAT );
  GLShader::flushState();
}

bool TextureRef::clamp() const
{
  return mTexture->mClamp;
}

GLuint TextureRef::textureIdHack() const
{
  return mTexture->mTexId;
}

GLenum TextureRef::textureTargetHack() const
{
  return mTexture->mTarget;
}

